/*____________________________________________________________________________
    Copyright (C) 2002 PGP Corporation
    All rights reserved.

    $Id: PGPAppController.m,v 1.115 2002/11/22 08:44:36 jeffc Exp $
____________________________________________________________________________*/

#import <Cocoa/Cocoa.h>
#import <PGP/PGP.h>
#import <PGPclient/pgpClientLibOSX.h>
#import <PGPclient/pgpClientLibCocoa.h>
#import <pgpFileSpec.h>

#import <pwd.h>
#import <unistd.h>
#import <sys/stat.h>
#import <syslog.h>

#import "PGPAppController.h"
#import "PGPDocument.h"
#import "CPGPDockMenu.h"
#import "CPGPKeyringSelection.h"
#import "CPGPLicenseAuth.h"
#import "CPGPLicenseDisplay.h"

#define kPGPAppString_BetaExpiration			NSLocalizedString(@"Beta Expiration", @"")
#define kPGPAppString_ExpirationMessage			NSLocalizedString(@"This beta version of PGP has expired. \
Please contact your software vendor to obtain a licensed version.", @"")
#define kPGPAppString_ReadOnlyFS				NSLocalizedString(@"Read-Only File System", @"")
#define kPGPAppString_ReadOnlyFSMessage			NSLocalizedString(@"PGP requires write access to the file system in which it resides. It is recommended that you copy PGP to the Applications folder of your main volume.", @"")
#define kPGPAppString_Quit						NSLocalizedString(@"Quit", @"")
#define kPGPAppString_Uninstall					NSLocalizedString(@"Uninstall", @"")
#define kPGPAppString_LimitedFunctionality		NSLocalizedString(@"Features Disabled", @"")
#define kPGPAppString_LimitedFunctionalityMsg	NSLocalizedString(@"Without a valid license, PGP will run in a limited functionality mode. PGP Mail Plugins, PGP Disk, and Enterprise Configuration functions have been disabled. Select License from the PGP menu at any time to enter your license information.", @"")
#define kPGPAppString_LicenseChanged			NSLocalizedString(@"License Successfully Updated", @"")
#define kPGPAppString_LicenseChangedMsg			NSLocalizedString(@"PGP must be restarted in order for your new license changes to take effect.", @"")

static PGPAppController*		sAppController = nil;
static BOOL						sUpdateKeyDBNeeded = NO;
	
typedef struct internal_PGPLicenseAuthorization {
	PGPByte		type;
	PGPUInt16	issuanceDate; // days since 9/1/2002
	PGPSize		signatureSize;	
	PGPByte		signature[1];
	// + signatureSize-1 bytes	// Exported: [2 bytes total size] [1 byte type] [2 bytes issuanceDate] [reserved=0] [signature]
} internal_PGPLicenseAuthorization;

	static void
sSDKNotificationCallback(
	PGPUserValue 			userValue,
	PGPNotificationReason	reason,
	PGPUInt32				param1,
	PGPUInt32				param2)
{
	PGPAppController*	appController = [PGPAppController appController];
	NSArray*			bundles = nil;
	
	pgpAssert(appController != nil);
	
	bundles = [[appController bundleManager] bundleControllers];
	
	switch (reason)
	{
		case kPGPNotification_KeyDBChanged:
			sUpdateKeyDBNeeded = YES;
			break;
		case kPGPNotification_PassphraseCacheChanged:
			PGPLog((@"Notifying bundles of passphrase cache change"));
			[bundles makeObjectsPerformSelector:@selector(passphraseCacheChanged)];
			break;
		default:
			break;
	}
}

@implementation PGPAppController

+ (id)appController
{
	return sAppController;
}

- (id)init
{
    if (!inited)
    {
        [super init];

        [[NSDocumentController sharedDocumentController] setShouldCreateUI:NO];
		
		[self registerNotificationHandlers];
        [self installDocumentFlusher];

        inited = YES;
    }

    if (!pgpAttributes)
		pgpAttributes = [[NSMutableDictionary alloc] initWithCapacity:kPGPAppInitialDictionarySize];
    
    if (!pgpInterface)
        [self initWithInterface:[[PGPInterfaceController alloc] initWithController:self]];
 
    if (!pgpBundleManager)
        [self initWithBundleManager:[[PGPBundleManager alloc] initWithController:self]];
    
    return self;
}

- (id)initWithInterface:(PGPInterfaceController*)interface
{
    [self setInterface:interface];
    
    return [self init];
}

- (id)initWithBundleManager:(PGPBundleManager*)manager
{
    [self setBundleManager:manager];
    
    return [self init];
}

- (id)initWithPrefsManager:(PGPPrefsManager*)manager
{
    [self setPrefsManager:manager];
    
    return [self init];
}

- (void)initSDK
{
	PGPError		err;

	if (!mSDKInited)
	{
		mSDKInited = YES;
		
		err = PGPsdkInit(0);
		pgpAssertNoErr(err);
		err = PGPsdkNetworkLibInit(0);
		pgpAssertNoErr(err);
		err = PGPsdkUILibInit(0);
		pgpAssertNoErr(err);
		err = PGPNewContext(kPGPsdkAPIVersion, &pgpContext);
		pgpAssertNoErr(err);
		err = PGPKeyServerInit();
		pgpAssertNoErr(err);
		err = PGPNewTLSContext(pgpContext, &tlsContext);
		pgpAssertNoErr(err);
		mMainKeyDB = kInvalidPGPKeyDBRef;
		err = PGPSetNotificationCallback(sSDKNotificationCallback, (PGPUserValue) 0x0);
		pgpAssertNoErr(err);
		
		[self installEntropyCollectionTimer];
		[self installKeyDBCheckerTimer];
	}
}

- (void)initPrefs
{
    if (!pgpPrefsManager)
        [self initWithPrefsManager:[[PGPPrefsManager alloc] initWithController:self]];
}

- (void)loadBundles
{
	if (!mBundlesLoaded)
	{
		if (pgpBundleManager)
		{
			[pgpBundleManager loadBundles];
			
			mBundlesLoaded = YES;
		}
	}
}

- (BOOL)finishedInitialization
{
	return mFinishedInitialization;
}

- (BOOL)isRunningOnReadOnlyFS
{
	NSScanner*			scanner;
	NSString*			bundlePath = [[NSBundle mainBundle] bundlePath];
	NSString*			tempString;
	NSMutableString*	volumeName = nil;
	
	/*
	Get the volume name. If we're running on the default volume, /Volumes won't
	appear in the path, and we just assume that the volume is writable.
	*/
	scanner = [[NSScanner alloc] initWithString:bundlePath];
	if ([scanner scanString:NSLocalizedString(@"/Volumes/", @"") intoString:&tempString])
	{
		volumeName = [[[NSMutableString alloc] init] autorelease];
		[volumeName appendString:tempString];
		
		if ([scanner scanUpToString:@"/" intoString:&tempString])
			[volumeName appendString:tempString];
		else
			volumeName = nil;
	}
	
	if (volumeName)
	{
		NSWorkspace*	workspace = [NSWorkspace sharedWorkspace];
		BOOL			removable;
		BOOL			writable;
		BOOL			unmountable;
		NSString*		description;
		NSString*		type;
		
		if ([workspace getFileSystemInfoForPath:volumeName
				isRemovable:&removable
				isWritable:&writable
				isUnmountable:&unmountable
				description:&description
				type:&type])
		{
			if (!writable)
				return YES;
		}
	}
	
	return NO;
}

- (BOOL)checkLicense
{
	CPGPLicenseAuth*			licenseAuth;
	NSUserDefaults*				defaults = [NSUserDefaults standardUserDefaults];
	NSDictionary*				dict = [defaults persistentDomainForName:kPGPPrefsDomain];
	NSNumber*					evalMode = nil;
	PGPLicenseNumberRef			lnRef = kInvalidPGPLicenseNumberRef;
	PGPLicenseAuthorizationRef	laRef = kInvalidPGPLicenseAuthorizationRef;
	PGPInt32					days = 0;
	PGPError					err;
	int							retVal;
	
	PGPclLicenseCheck([self pgpContext]);
	
	if (mLicenseAuthDisplayed)
		return NO;
	
	/*
	If not authorized, we check to see if an LN and/or LA is present. If so, the auth
	check probably failed because of the expiration date (handled further down).
	*/
	if (!PGPclIsAuthorized())
	{
		licenseAuth = [CPGPLicenseAuth sharedLicenseAuth];
		
		/*
		Failure to import the LN/LA indicates that none exists. Possible actions include:
			a) Bring up license auth dialog
				1) User enters a valid LN (and LA, or authorizes over the net)
				2) User hits "Later", putting us in freeware mode.
			b) Continue without license auth dialog if we're in freeware mode.
		*/
		if (IsPGPError(err = [CPGPLicenseAuth prefLNRef:&lnRef]) || (IsPGPError(err = [CPGPLicenseAuth prefLARef:&laRef])))
		{
			if (PGPLicenseNumberRefIsValid(lnRef))
			{
				PGPLNFree(lnRef);
				lnRef = kInvalidPGPLicenseNumberRef; 
			}
			
			if (PGPLicenseAuthorizationRefIsValid(laRef))
			{
				PGPLAFree(laRef);
				laRef = kInvalidPGPLicenseAuthorizationRef;
			}
			
			/* Check evalMode (freeware) pref value */
			evalMode = [dict objectForKey:kPGPAppDefault_EvalMode];
			
			if (!evalMode || (evalMode && ![evalMode boolValue]))
			{
				[licenseAuth displayLicenseEntryDialog:NO];
				return YES;
			}
		}
		/*
		We're not authorized, and importing the LN/LA was successful. Therefore, our
		license has expired.
		*/
		else
		{
			days = PGPclLicenseDaysLeft();
			
			if (days > 0)
				PGPLog((@"Something is wrong here... Treating as if the build has expired"));
			
			/*
			If the eval flag is turned on, there is no grace period. An alert panel is
			brought forward. Possible actions include:
				a) User presses "Purchase Now". Web page is opened, LA dialog displayed.
				b) User presses "Quit".
				c) User presses "Uninstall".
			*/
			if (PGPclEvaluation())
			{
				if (PGPLicenseNumberRefIsValid(lnRef))
				{
					PGPLNFree(lnRef);
					lnRef = kInvalidPGPLicenseNumberRef; 
				}
				
				if (PGPLicenseAuthorizationRefIsValid(laRef))
				{
					PGPLAFree(laRef);
					laRef = kInvalidPGPLicenseAuthorizationRef;
				}
				
				retVal = NSRunAlertPanel(kPGPString_ExpiredEval,
										kPGPString_ExpiredEvalMsg,
										kPGPString_PurchaseNow,
										kPGPString_Quit,
										kPGPString_Uninstall);
				
				if (retVal == NSAlertDefaultReturn)				/* Purchase Now */
				{
					[[NSWorkspace sharedWorkspace] openURL:[NSURL URLWithString:kPGPLicenseDisplayPurchaseURL]];
					[licenseAuth displayLicenseEntryDialog:YES];
					
					return YES;
				}
				else if (retVal == NSAlertAlternateReturn)		/* Quit */
				{
					[licenseAuth release];
					licenseAuth = nil;
					
					[self cleanupAndExit];
				}
				else											/* Uninstall */
				{
					[licenseAuth release];
					licenseAuth = nil;
					
					[self uninstall:YES];
				}
			}
			/*
			Non-eval license expiration. A grace period is extended to the user. Once
			the grace period has ended, the user must purchase a new license to continue
			usage.
			*/
			else
			{
				NSCalendarDate*		date = [NSCalendarDate dateWithTimeIntervalSince1970:0x3d71c901];
				PGPInt16			licenseDays = PGPclLicenseEvalPeriod();
				PGPInt32			daysLeft = PGPclLicenseDaysLeft();
				PGPUInt16			issuance = ((internal_PGPLicenseAuthorization*) laRef)->issuanceDate;
				
				date = [date dateByAddingYears:0
							 months:0
							 days:(issuance + licenseDays + PGP_LICENSEGRACEPERIOD)
							 hours:0
							 minutes:0
							 seconds:0];
				
				if (PGPLicenseNumberRefIsValid(lnRef))
				{
					PGPLNFree(lnRef);
					lnRef = kInvalidPGPLicenseNumberRef; 
				}
				
				if (PGPLicenseAuthorizationRefIsValid(laRef))
				{
					PGPLAFree(laRef);
					laRef = kInvalidPGPLicenseAuthorizationRef;
				}
				
				/* Grace period has been met/exceeded */
				if ((daysLeft < 0) && ((daysLeft + PGP_LICENSEGRACEPERIOD) < 0))
				{
					retVal = NSRunAlertPanel(kPGPString_ExpiredLicense,
												kPGPString_ExpiredLicenseMsg,
												kPGPString_PurchaseNow,
												kPGPString_Quit,
												kPGPString_Uninstall);
					
					if (retVal == NSAlertDefaultReturn)				/* Purchase Now */
					{
						[[NSWorkspace sharedWorkspace] openURL:[NSURL URLWithString:kPGPLicenseDisplayPurchaseURL]];
						[licenseAuth displayLicenseEntryDialog:YES];
						
						return YES;
					}
					else if (retVal == NSAlertAlternateReturn)		/* Quit */
					{
						[licenseAuth release];
						licenseAuth = nil;
						
						[self cleanupAndExit];
					}
					else											/* Uninstall */
					{
						[licenseAuth release];
						licenseAuth = nil;
						
						[self uninstall:YES];
					}
				}
				/* License expired, but within the grace period, so we nag. */
				else if (daysLeft < 0)
				{
					retVal = NSRunAlertPanel(kPGPString_ContinueUsing,
												kPGPString_ContinueUsingMsg,
												kPGPString_OK,
												kPGPString_PurchaseNow,
												nil,
												[date descriptionWithCalendarFormat:@"%b. %d, %Y"]);
					
					if (retVal == NSAlertAlternateReturn)
					{
						[[NSWorkspace sharedWorkspace] openURL:[NSURL URLWithString:kPGPLicenseDisplayPurchaseURL]];
						[licenseAuth displayLicenseEntryDialog:NO];
						return YES;
					}
				}
			}
		}
	}
	
	return NO;
}

- (void)keyringSelectionFinished:(CPGPKeyringSelection*)keyringSelection
{
	if (keyringSelection)
		[keyringSelection release];
	
	[self finishInitialization];
}

- (void)licenseAuthFinished:(CPGPLicenseAuth*)licenseAuth
{
	NSUserDefaults*			defaults;
	NSMutableDictionary*	dict;
	
	mLicenseAuthDisplayed = YES;
	
	defaults = [NSUserDefaults standardUserDefaults];
	dict = [[[defaults persistentDomainForName:kPGPPrefsDomain] mutableCopy] autorelease];
	
	if (!dict)
		dict = [[[NSMutableDictionary alloc] init] autorelease];
	
	[dict setObject:[NSNumber numberWithBool:NO] forKey:kPGPAppDefault_EvalMode];
	
	[defaults removePersistentDomainForName:kPGPPrefsDomain];
	[defaults setPersistentDomain:dict forName:kPGPPrefsDomain];
	[defaults synchronize];
	
	PGPclLicenseFree([self pgpContext]);
	PGPclLicenseCheck([self pgpContext]);
	
	if (!mFinishedInitialization)
		[self finishInitialization];
	else
	{
		NSRunInformationalAlertPanel(kPGPAppString_LicenseChanged,
									 kPGPAppString_LicenseChangedMsg,
									 kPGPString_OK,
									 nil,
									 nil);
	}
	
	[self displayLicense];
}

- (void)licenseAuthDeferred:(CPGPLicenseAuth*)licenseAuth
{
	NSUserDefaults*			defaults;
	NSMutableDictionary*	dict;
	PGPLicenseNumberRef		lnRef = kInvalidPGPLicenseNumberRef;
	PGPError				err;
	
	mLicenseAuthDisplayed = YES;
	
	defaults = [NSUserDefaults standardUserDefaults];
	dict = [[[defaults persistentDomainForName:kPGPPrefsDomain] mutableCopy] autorelease];
	
	if (!dict)
		dict = [[[NSMutableDictionary alloc] init] autorelease];
	
	[dict setObject:[NSNumber numberWithBool:YES] forKey:kPGPAppDefault_EvalMode];
	
	[defaults removePersistentDomainForName:kPGPPrefsDomain];
	[defaults setPersistentDomain:dict forName:kPGPPrefsDomain];
	[defaults synchronize];
	
	PGPclLicenseFree([self pgpContext]);
	PGPclLicenseCheck([self pgpContext]);
	
	if (!PGPclIsAuthorized() && !mFinishedInitialization)
	{
		if (IsPGPError(err = [CPGPLicenseAuth prefLNRef:&lnRef]) || !PGPLicenseNumberRefIsValid(lnRef))
		{
			NSRunAlertPanel(kPGPAppString_LimitedFunctionality,
							kPGPAppString_LimitedFunctionalityMsg,
							kPGPString_OK,
							nil,
							nil);
		}
		else
		{
			if (PGPLicenseNumberRefIsValid(lnRef))
			{
				PGPLNFree(lnRef);
				lnRef = kInvalidPGPLicenseNumberRef; 
			}
		}
		
		[self finishInitialization];
	}
	else if (!mFinishedInitialization)
		[self finishInitialization];
	
	[licenseAuth release];
}

- (void)finishedHandlingDocument:(NSString*)file
{
	if (mLaunchedDocuments && [mLaunchedDocuments count])
	{
		if ([mLaunchedDocuments containsObject:file])
		{
			[mLaunchedDocuments removeObject:file];
			
			if (![mLaunchedDocuments count] && mExitAfterDocumentsHandled)
				[self cleanupAndExit];
		}
	}
}

- (void)uninstall:(BOOL)quitOnClose
{
	PGPError			err;
	CPGPUninstaller*	uninstaller;
	NSArray*			windows = [NSApp windows];
	NSMutableArray*		hiddenWindows = [[NSMutableArray alloc] init];
	NSMutableArray*		installItems = [[NSMutableArray alloc] init];
	NSEnumerator*		enumerator;
    NSWindow*			keyWindow;
	NSWindow*			window;
	BOOL				shouldQuit = NO;
	
    keyWindow = [NSApp keyWindow];
	
	enumerator = [windows objectEnumerator];
	
	while (window = [enumerator nextObject])
	{
		if ([window isVisible])
		{
			[hiddenWindows addObject:window];
			[window orderOut:nil];
		}
	}
	
	if ([[pgpBundleManager bundleNames] count])
		[installItems setArray:[pgpBundleManager bundleNames]];
	else
	{
		[installItems addObject:@"PGPdisk.bundle"];
		[installItems addObject:@"PGPkeys.bundle"];
		[installItems addObject:@"PGPmail.bundle"];
		//[installItems addObject:@"PGPnet.bundle"];
	}
	
	[installItems addObject:@"PGPapp"];
	uninstaller = [[CPGPUninstaller alloc] init];
	[uninstaller setInfo:installItems];
	
	err = [uninstaller displayUninstaller];

	if (IsPGPError(err))
	{
		shouldQuit = quitOnClose;
		
		enumerator = [hiddenWindows objectEnumerator];
		
		while (window = [enumerator nextObject])
			[window orderFront:nil];
        
		if (!shouldQuit)
			[keyWindow makeKeyAndOrderFront:nil];
	}
	else
		shouldQuit = YES;
	
	[uninstaller release];
	[hiddenWindows release];
	[installItems release];
	
	if (shouldQuit)
		[self cleanupAndExit];
}

- (void)displayLicense
{
	CPGPLicenseDisplay*		license = [CPGPLicenseDisplay sharedLicenseDisplay];
	CPGPLicenseAuth*		licenseAuth = [CPGPLicenseAuth sharedLicenseAuth];
	
	if (!PGPclIsAuthorized())
	{
		if ([[licenseAuth window] isVisible])
			[[licenseAuth window] makeKeyAndOrderFront:nil];
		else
			[licenseAuth displayLicenseEntryDialog:NO];
	}
	else
	{
		if ([[license window] isVisible])
			[[license window] makeKeyAndOrderFront:nil];
		else
			[license display];
	}
}

- (void)changeLicense
{
	CPGPLicenseAuth*		licenseAuth = [CPGPLicenseAuth sharedLicenseAuth];
	
	if ([[licenseAuth window] isVisible])
		[[licenseAuth window] makeKeyAndOrderFront:nil];
	else
		[licenseAuth displayLicenseEntryDialog:NO];
}

- (void)displayExpirationAlertAndExit
{
	int		retval;
	
	retval = NSRunAlertPanel(kPGPAppString_BetaExpiration,
							 kPGPAppString_ExpirationMessage,
							 kPGPAppString_Quit,
							 kPGPAppString_Uninstall,
							 nil);
	
	if (retval == NSAlertDefaultReturn)
		[self cleanupAndExit];
	else
		[self uninstall:YES];
}

- (void)displayReadOnlyAlertAndExit
{
	NSRunAlertPanel(kPGPAppString_ReadOnlyFS, kPGPAppString_ReadOnlyFSMessage, kPGPAppString_Quit, nil, nil);
	
	[self cleanupAndExit];
}

- (void)finishInitialization
{
	NSArray*					bundleControllers;
	NSArray*					activeBundles;
	NSUserDefaults*				defaults = [NSUserDefaults standardUserDefaults];
	NSDictionary*				dict = [defaults persistentDomainForName:kPGPPrefsDomain];
	NSDictionary*				attributes;
    NSEnumerator*				enumerator;
    NSString*					filename;
    NSString*					currentBundleIdentifier;
	NSString*					bundleName;
	NSString*					extension;
	id <PGPBundle>				bundleController;
	id <PGPBundleNet>			netController;
	CPGPKeyringSelection*		keyringSelection;
	
	[self initSDK];
	[self initPrefs];
	
	if ([self checkLicense])
		return;
	
	/* Check for keyrings */
	keyringSelection = [[CPGPKeyringSelection alloc] init];
	
	if (![keyringSelection haveKeyrings])
	{
		[keyringSelection displayKeyringSelection];
		return;
	}
	else
		[keyringSelection release];
	
	[self loadBundles];
	[pgpInterface initializeInterface];
	
	if (!mLaunchedDocuments)
	{
		if (bundleControllers = [pgpBundleManager bundleControllers])
		{	
			activeBundles = [dict objectForKey:kPGPAppDefault_ActiveBundles];
			
			if (activeBundles)
			{
				enumerator = [activeBundles objectEnumerator];
				
				while (bundleName = [enumerator nextObject])
				{
					if (bundleController = [pgpBundleManager bundleByName:bundleName])
						[pgpInterface setCurrentBundle:bundleController passive:NO];
				}
			}
			
			currentBundleIdentifier = [dict objectForKey:kPGPAppDefault_CurrentBundle];
			
			if (!currentBundleIdentifier)
				currentBundleIdentifier = @"PGPkeys.bundle";
			
			if (!(bundleController = [[self bundleManager] bundleByName:currentBundleIdentifier]))
				bundleController = [bundleControllers objectAtIndex:0];
			
			[pgpInterface setCurrentBundle:bundleController passive:NO];
		}
	}
	
    [pgpAttributes setObject:[pgpBundleManager bundles] forKey:kPGPAppBundlesKey];
	
	if (netController = [pgpBundleManager bundleByName:@"PGPnet.bundle"])
		[pgpPrefsManager setNetPrefRef:[netController netPrefRef]];
	
    [NSApp setServicesProvider:self];
	
	if (mLaunchedDocuments && [mLaunchedDocuments count])
    {
		mExitAfterDocumentsHandled = YES;
		
        enumerator = [mLaunchedDocuments objectEnumerator];
        
        while (filename = [enumerator nextObject])
		{
			extension = [filename pathExtension];
			attributes = [[NSFileManager defaultManager] fileAttributesAtPath:filename traverseLink:YES];
			
			if (!([extension isEqualToString:@"pgp"] || [extension isEqualToString:@"pgd"] ||
				([attributes fileHFSTypeCode] == 'Dsk1') || ([attributes fileHFSTypeCode] == 'pgEF')))
			{
				mExitAfterDocumentsHandled = NO;
			}
			
			[[NSDocumentController sharedDocumentController] openDocumentWithContentsOfFile:filename display:YES];
		}
    }
	
	mFinishedInitialization = YES;
}

- (void)installEntropyCollectionTimer
{
	if (!entropyCollectionTimer)
    {
        lastMouseLocation.v = 0;
        lastMouseLocation.h = 0;
        
		entropyCollectionTimer = [NSTimer scheduledTimerWithTimeInterval:kPGPAppEntropyCollectionInterval
                                          target:self
                                          selector:@selector(addEntropy:)
                                          userInfo:nil
                                          repeats:YES];
    }
}

- (void)installDocumentFlusher
{
    if (!mDocumentFlusher)
    {
        mDocumentFlusher = [NSTimer scheduledTimerWithTimeInterval:kPGPAppDocumentFlusherInterval
            target:self
            selector:@selector(documentFlusher:)
            userInfo:nil
            repeats:YES];
    }
}

- (void)installKeyDBCheckerTimer
{
	if (!mKeyDBChecker)
	{
		mKeyDBChecker = [NSTimer scheduledTimerWithTimeInterval:kPGPAppKeyDBCheckerInterval
			target:self
			selector:@selector(keyDBUpdateChecker:)
			userInfo:nil
			repeats:YES];
	}
}

- (void)registerNotificationHandlers;
{
	NSNotificationCenter*				center = [NSNotificationCenter defaultCenter];
	
	[center addObserver:self
			selector:@selector(passphraseCacheTypeDidChange:)
			name:kPGPclPrefNotification_CacheType
			object:nil];
	
	[center addObserver:self
			selector:@selector(keyringsDidChange:)
			name:kPGPclPrefNotification_UpdateKeyrings
			object:nil];
}

- (void)addEntropy:(NSTimer*)timer
{
    Point		mouseLocation;
    
    GetMouse(&mouseLocation);

    if ((mouseLocation.v != lastMouseLocation.v) ||
        (mouseLocation.h != lastMouseLocation.h))
    {
        PGPGlobalRandomPoolAddMouse(mouseLocation.h, mouseLocation.v);
        lastMouseLocation = mouseLocation;
    }
}

- (void)documentFlusher:(NSTimer*)timer
{
    /*
    Apparently, NSDocumentController keeps track of which files have been opened, or
    those which the user has attempted to open. If we don't flush this list of recently
    accessed documents, we won't get subsequent attempts to open a document we'd
    otherwise be interested in.
    */
    [[NSDocumentController sharedDocumentController] closeAllDocumentsWithDelegate:nil
                                                     didCloseAllSelector:nil
                                                     contextInfo:nil];
}

- (void)keyDBUpdateChecker:(NSTimer*)timer
{
	NSArray*			bundles = nil;
	id <PGPBundleMail>	mailController = [pgpBundleManager bundleByName:@"PGPmail.bundle"];
	
	if ([NSApp modalWindow] || !mMainKeyDB)
		return;
	
	if (mailController && [mailController jobsPending])
		return;
	
	if (sUpdateKeyDBNeeded)
	{
		bundles = [[self bundleManager] bundleControllers];	
		PGPLog((@"Notifying bundles of keyDB change"));
		[bundles makeObjectsPerformSelector:@selector(keyDBChanged)];
		
		sUpdateKeyDBNeeded = NO;
	}
}

- (void)purgePassphraseCache
{
	id <PGPBundleNet>	netController = [pgpBundleManager bundleByName:@"PGPnet.bundle"];
	
	PGPPurgePassphraseCache([self pgpContext]);
    
    if (netController)
        [netController purgePassphrases];
}

- (NSMutableDictionary*)attributes
{
    return pgpAttributes;
}

- (void)setAttributes:(NSMutableDictionary*)attributes
{
    if (attributes)
        pgpAttributes = attributes;
}

- (PGPInterfaceController*)interface
{
    return pgpInterface;
}

- (void)setInterface:(PGPInterfaceController*)interface
{
    pgpInterface = interface;
}

- (PGPBundleManager*)bundleManager
{
    return pgpBundleManager;
}

- (void)setBundleManager:(PGPBundleManager*)manager
{
    pgpBundleManager = manager;
}

- (PGPPrefsManager*)prefsManager
{
    return pgpPrefsManager;
}

- (void)setPrefsManager:(PGPPrefsManager*)manager
{
    if (manager)
        pgpPrefsManager = manager;
}

- (PGPContextRef)pgpContext
{
	return pgpContext;
}

- (PGPtlsContextRef)tlsContext
{
	return tlsContext;
}

- (PGPPrefRef)pgpPrefRef
{
    return [pgpPrefsManager pgpPrefRef];
}

- (PGPKeyDBRef)mainKeyDB
{
	if( !PGPKeyDBRefIsValid( mMainKeyDB ) )
	{
		PGPError	err = kPGPError_NoErr;

		err = PGPclGetPath([self pgpPrefRef], kInvalidPGPPrefRef, kPGPclPublicKeyringFile, mPubKeyringPath, PATH_MAX);
		err = PGPclGetPath([self pgpPrefRef], kInvalidPGPPrefRef, kPGPclPrivateKeyringFile, mSecKeyringPath, PATH_MAX);

		err = PGPclOpenDefaultKeyrings( pgpContext, [self pgpPrefRef], kPGPOpenKeyDBFileOptions_Mutable, &mMainKeyDB );
		if( !PGPKeyDBRefIsValid( mMainKeyDB ) )
			PGPLog((@"%s:%d: Warning: could not load main KeyDB(PGPerror=%d).", __FILE__, __LINE__,err));
	}
	return mMainKeyDB;
}

- (void)setView:(NSView*)view
{
}

- (void)setCurrentBundle:(id <PGPBundle>)controller passive:(BOOL)passive
{
	if (controller)
		[pgpInterface setCurrentBundle:controller passive:passive];
}

- (BOOL)canDisplayMainWindow
{
	return !mHideMainWindow;
}

- (NSWindow*)mainWindow
{
	return [NSApp mainWindow];
}

- (NSMenu*)mainMenu
{
	return [pgpInterface mainMenu];
}

- (NSArray*)menuItems
{
    return [pgpInterface menuItems];
}

- (NSArray*)launchedDocuments
{
	return mLaunchedDocuments;
}

- (void)addMenu:(NSMenu*)menu withIdentifier:(NSString*)identifier;
{
	[pgpInterface addMenu:menu withIdentifier:identifier];
}

- (void)addMenu:(NSMenu*)menu withIdentifier:(NSString*)identifier atIndex:(unsigned)index;
{
	[pgpInterface addMenu:menu withIdentifier:identifier atIndex:index];
}

- (void)addMenuItem:(NSMenuItem*)menuItem withIdentifier:(NSString*)identifier;
{
	[pgpInterface addMenuItem:menuItem withIdentifier:identifier];
}

- (void)addMenuItem:(NSMenuItem*)menuItem withIdentifier:(NSString*)identifier atIndex:(unsigned)index;
{
	[pgpInterface addMenuItem:menuItem withIdentifier:identifier atIndex:index];
}

- (void)awakeFromNib
{
	NSAutoreleasePool*	pool = [[NSAutoreleasePool alloc] init];
	NSString*			bundlePath = [[NSBundle mainBundle] bundlePath];
	NSArray*			pgpBundlePaths = [[NSArray alloc] initWithObjects:[bundlePath stringByAppendingString:@"/Contents/PlugIns"], nil];
	NSArray*			tempBundlePaths = [PGPBundleManager bundlesAtPaths:pgpBundlePaths];
	PGPError			err;

	sAppController = self;	/* init has just been called. */
	
    [pgpBundleManager setBundlePaths:tempBundlePaths];
	
    [tempBundlePaths release];
	[pgpBundlePaths release];
	
	if ((err = PGPclIsExpired()) == kPGPError_FeatureNotAvailable)
		[self displayExpirationAlertAndExit];
	
	if ((err = PGPclIsOSSupported()) == kPGPError_FeatureNotAvailable)
		[self cleanupAndExit];
	if ([self isRunningOnReadOnlyFS])
		[self displayReadOnlyAlertAndExit];
	
	[pool release];
}

- (void)applicationDidFinishLaunching:(NSNotification*)notification
{
	mAppFinishedLaunching = YES;

	[self finishInitialization];
}

- (BOOL)application:(NSApplication*)app openFile:(NSString*)filename
{
	if (mAppFinishedLaunching)
	{
		if (filename)
			[[NSDocumentController sharedDocumentController] openDocumentWithContentsOfFile:filename display:YES];
	}
	else
	{
		mHideMainWindow = YES;
        
        if (!mLaunchedDocuments)
            mLaunchedDocuments = [[NSMutableArray alloc] init];
        
        [mLaunchedDocuments addObject:filename];
	}
	
	return YES;
}

- (NSApplicationTerminateReply)applicationShouldTerminate:(NSApplication*)app
{
	[self cleanupAndExit];
	
	return NSTerminateNow;
}

- (void)cleanupAndExit
{
	BOOL					allowExit = YES;
    NSUserDefaults*			defaults;
	NSMutableDictionary*	dict;
	id <PGPBundle>			bundle;
    id <PGPBundle>			currentBundle = [[self interface] currentBundle];
	NSArray*				windows = [NSApp windows];
	NSArray*				bundles = [pgpBundleManager bundleControllers];
	NSNumber*				cleanUninstall = nil;
	NSMutableArray*			activeBundles = [NSMutableArray array];
	NSEnumerator*			enumerator;
	NSWindow*				window;
	
	defaults = [NSUserDefaults standardUserDefaults];
	dict = [[defaults persistentDomainForName:kPGPPrefsDomain] mutableCopy];
	
	if (!dict)
		dict = [[NSMutableDictionary alloc] init];
	else
		cleanUninstall = [dict objectForKey:kPGPAppDefault_CleanUninstall];
	
	enumerator = [bundles objectEnumerator];
	while (bundle = [enumerator nextObject])
	{
		if ([[bundle window] isVisible])
			[activeBundles addObject:[[(NSObject*) bundle class] identifier]];
	}
	
	[dict setObject:activeBundles forKey:kPGPAppDefault_ActiveBundles];
	
	[defaults removePersistentDomainForName:kPGPPrefsDomain];
	[defaults setPersistentDomain:dict forName:kPGPPrefsDomain];
	[defaults synchronize];
	[dict release];
	
	enumerator = [windows objectEnumerator];
	while (window = [enumerator nextObject])
	{
		if ([window isVisible])
			[window orderOut:nil];
	}
	[enumerator release];

	if (pgpBundleManager)
		allowExit = [pgpBundleManager releaseBundlesIfReadyToShutdown];
	
	if (allowExit)
	{
		defaults = [NSUserDefaults standardUserDefaults];
		
		dict = [[defaults persistentDomainForName:kPGPPrefsDomain] mutableCopy];
		
		if (!dict)
			dict = [[NSMutableDictionary alloc] init];
		
        if (currentBundle)
            [dict setObject:[[(NSObject*) currentBundle class] identifier] forKey:kPGPAppDefault_CurrentBundle];
		
		[defaults removePersistentDomainForName:kPGPPrefsDomain];
		[defaults setPersistentDomain:dict forName:kPGPPrefsDomain];
        [dict release];
        
		if (pgpInterface)
			[pgpInterface release];
			
		if (pgpBundleManager)
			[pgpBundleManager release];
		
		if (pgpPrefsManager)
			[pgpPrefsManager release];
		
		if (pgpAttributes)
			[pgpAttributes release];
		
		PGPclLicenseFree(pgpContext);
		
		if (pgpContext)
		{
			if (PGPKeyDBRefIsValid(mMainKeyDB))
			{
				PGPFreeKeyDB(mMainKeyDB);
				mMainKeyDB = kInvalidPGPKeyDBRef;
			}
			if (PGPtlsContextRefIsValid(tlsContext))
			{
				PGPFreeTLSContext(tlsContext);
				tlsContext = kInvalidPGPtlsContextRef;
			}
			if (PGPContextRefIsValid(pgpContext))
			{
				PGPFreeContext(pgpContext);
				pgpContext = kInvalidPGPContextRef;
			}

			PGPKeyServerCleanup();
			PGPsdkUILibCleanup();
			PGPsdkNetworkLibCleanup();
			PGPsdkCleanup();
		}
		
		if (cleanUninstall && [cleanUninstall boolValue])
		{
			unlink([[NSHomeDirectory() stringByAppendingPathComponent:@"Library/Preferences/com.pgp.PGP.plist"] UTF8String]);
			unlink([[NSHomeDirectory() stringByAppendingPathComponent:@"Library/Preferences/com.pgp.app.PGP.plist"] UTF8String]);
		}
		
		exit(EXIT_SUCCESS);
	}
}

- (void)signSelection:(NSPasteboard *)pboard userData:(NSString *)data error:(NSString **)error
{
	[self signSelection:pboard];
}

- (void)signSelection:(NSPasteboard*)pboard
{
	NSAutoreleasePool*	pool = [[NSAutoreleasePool alloc] init];
	NSArray*			pBoardTypes = [pboard types];
	NSArray* 			types = [NSArray arrayWithObject:NSStringPboardType];
	NSString* 			data = [pboard stringForType:NSStringPboardType];
	NSString*			signedData = nil;
	id<PGPBundleMail>	mailController = [[self bundleManager] bundleByName:@"PGPmail.bundle"];
	ProcessSerialNumber	frontPSN;
	BOOL				switchProcess = YES;
	
	GetFrontProcess(&frontPSN);

	if ([pBoardTypes containsObject:NSFilenamesPboardType])
    {
		NSArray*	fileNames = nil;
		NSString*	plist = nil;
		
		[NSApp activateIgnoringOtherApps:YES];
		
		plist = [[[NSString alloc] initWithData:[pboard dataForType:NSFilenamesPboardType]
								   encoding:NSUTF8StringEncoding] autorelease];
		
		fileNames = [plist propertyList];
		
		/* No need to work in the main thread for file operations originating from the Services menu */
		[NSThread detachNewThreadSelector:@selector(signSelectedFiles:) toTarget:mailController withObject:fileNames];
		
		switchProcess = NO;
    }
    else if ([pBoardTypes containsObject:NSStringPboardType])
	{
		if ([mailController jobsPending])
			goto done;
		
		[NSApp activateIgnoringOtherApps:YES];
		
		signedData = [mailController signText:data error:nil];
		
		if (signedData)
		{
			[pboard declareTypes:types owner:nil];
			[pboard setString:signedData forType:NSStringPboardType];
		}
	}
	
done:
	if (switchProcess)
		SetFrontProcess(&frontPSN);
	
	[pool release];
}

- (void)encryptSelection:(NSPasteboard *)pboard userData:(NSString *)data error:(NSString **)error
{
	[self encryptSelection:pboard];
}

- (void)encryptSelection:(NSPasteboard*)pboard
{
	NSAutoreleasePool*	pool = [[NSAutoreleasePool alloc] init];
	NSArray*			pBoardTypes = [pboard types];
	NSArray* 			types = [NSArray arrayWithObject:NSStringPboardType];
	NSString* 			data = [pboard stringForType:NSStringPboardType];
	NSString*			encryptedData = nil;
	id<PGPBundleMail>	mailController = [[self bundleManager] bundleByName:@"PGPmail.bundle"];
	ProcessSerialNumber	frontPSN;
	BOOL				switchProcess = YES;
	
	GetFrontProcess(&frontPSN);
	
	if ([pBoardTypes containsObject:NSFilenamesPboardType])
    {
        NSArray			*fileNames = nil;
        NSString		*plist = nil;
        
		[NSApp activateIgnoringOtherApps:YES];
		
		plist = [[[NSString alloc] initWithData:[pboard dataForType:NSFilenamesPboardType]
                                   encoding:NSUTF8StringEncoding] autorelease];
		
        fileNames = [plist propertyList];
		
		/* No need to work in the main thread for file operations originating from the Services menu */
		[NSThread detachNewThreadSelector:@selector(encryptSelectedFiles:) toTarget:mailController withObject:fileNames];
		
		switchProcess = NO;
    }
    else if ([pBoardTypes containsObject:NSStringPboardType])
	{
		if ([mailController jobsPending])
			goto done;
		
		[NSApp activateIgnoringOtherApps:YES];
		
		encryptedData = [mailController encryptText:data toRecipients:nil error:nil];
		
		if (encryptedData)
		{
			[pboard declareTypes:types owner:nil];
			[pboard setString:encryptedData forType:NSStringPboardType];
		}
	}
	
done:
	if (switchProcess)
		SetFrontProcess(&frontPSN);
	
	[pool release];
}

- (void)signEncryptSelection:(NSPasteboard *)pboard userData:(NSString *)data error:(NSString **)error
{
	[self signEncryptSelection:pboard];
}

- (void)signEncryptSelection:(NSPasteboard*)pboard
{
	NSAutoreleasePool*	pool = [[NSAutoreleasePool alloc] init];
	NSArray*			pBoardTypes = [pboard types];
	NSArray* 			types = [NSArray arrayWithObject:NSStringPboardType];
	NSString* 			data = [pboard stringForType:NSStringPboardType];
	NSString*			encryptedData = nil;
	id<PGPBundleMail>	mailController = [[self bundleManager] bundleByName:@"PGPmail.bundle"];
	ProcessSerialNumber	frontPSN;
	BOOL				switchProcess = YES;
	
	GetFrontProcess(&frontPSN);
	
	if ([pBoardTypes containsObject:NSFilenamesPboardType])
    {
        NSArray			*fileNames = nil;
        NSString		*plist = nil;
        
		[NSApp activateIgnoringOtherApps:YES];
		
		plist = [[[NSString alloc] initWithData:[pboard dataForType:NSFilenamesPboardType]
                                   encoding:NSUTF8StringEncoding] autorelease];
		
        fileNames = [plist propertyList];
        
		/* No need to work in the main thread for file operations originating from the Services menu */
		[NSThread detachNewThreadSelector:@selector(encryptSignSelectedFiles:) toTarget:mailController withObject:fileNames];
		
		switchProcess = NO;
    }
    else if ([pBoardTypes containsObject:NSStringPboardType])
	{
		if ([mailController jobsPending])
			goto done;
		
		[NSApp activateIgnoringOtherApps:YES];
		
		encryptedData = [mailController signEncryptText:data toRecipients:nil error:nil];
		
		if (encryptedData)
		{
			[pboard declareTypes:types owner:nil];
			[pboard setString:encryptedData forType:NSStringPboardType];
		}
	}
	
done:
	if (switchProcess)
		SetFrontProcess(&frontPSN);
	
	[pool release];
}

- (void)decryptVerifySelection:(NSPasteboard *)pboard userData:(NSString *)data error:(NSString **)error
{
	[self decryptVerifySelection:pboard];
}

- (void)decryptVerifySelection:(NSPasteboard*)pboard
{
	NSAutoreleasePool*	pool = [[NSAutoreleasePool alloc] init];
	NSArray*			pBoardTypes = [pboard types];
	NSArray* 			types = [NSArray arrayWithObject:NSStringPboardType];
	NSString* 			data = [pboard stringForType:NSStringPboardType];
	NSString*			decryptedData = nil;
	id<PGPBundleMail>	mailController = [[self bundleManager] bundleByName:@"PGPmail.bundle"];
	ProcessSerialNumber	frontPSN;
	BOOL				switchProcess = YES;
	
	GetFrontProcess(&frontPSN);
	
	if ([pBoardTypes containsObject:NSFilenamesPboardType])
    {
        NSArray			*fileNames = nil;
        NSString		*plist = nil;
        
		[NSApp activateIgnoringOtherApps:YES];
		
		plist = [[[NSString alloc] initWithData:[pboard dataForType:NSFilenamesPboardType]
                                   encoding:NSUTF8StringEncoding] autorelease];
		
        fileNames = [plist propertyList];
        
		/* No need to work in the main thread for file operations originating from the Services menu */
		[NSThread detachNewThreadSelector:@selector(decryptVerifySelectedFiles:) toTarget:mailController withObject:fileNames];
		
		switchProcess = NO;
    }
    else if ([pBoardTypes containsObject:NSStringPboardType])
	{
		if ([mailController jobsPending])
			goto done;
		
		[NSApp activateIgnoringOtherApps:YES];
		
		decryptedData = [mailController decryptVerifyText:data error:nil showDialog:YES];
		
		if (decryptedData)
		{
			[pboard declareTypes:types owner:nil];
			[pboard setString:decryptedData forType:NSStringPboardType];
		}
	}
	
done:
	if (switchProcess)
		SetFrontProcess(&frontPSN);
	
	[pool release];
}

- (void)wipeFile:(NSPasteboard *)pboard userData:(NSString *)data error:(NSString **)error
{
	[self wipeFile:pboard];
}

- (void)wipeFile:(NSPasteboard*)pboard
{
	NSAutoreleasePool*	pool = [[NSAutoreleasePool alloc] init];
	NSArray*			pBoardTypes = [pboard types];
	id<PGPBundleMail>	mailController = [[self bundleManager] bundleByName:@"PGPmail.bundle"];
	ProcessSerialNumber	frontPSN;
	BOOL				switchProcess = YES;
	
	GetFrontProcess(&frontPSN);
	
    if ([pBoardTypes containsObject:NSFilenamesPboardType])
    {
        NSArray*	fileNames = nil;
        NSString*	plist = nil;
        
		[NSApp activateIgnoringOtherApps:YES];
		
		plist = [[[NSString alloc] initWithData:[pboard dataForType:NSFilenamesPboardType]
                                   encoding:NSUTF8StringEncoding] autorelease];
                                   
        fileNames = [plist propertyList];
		
		/* No need to work in the main thread for file operations originating from the Services menu */
		[NSThread detachNewThreadSelector:@selector(wipeSelectedFiles:) toTarget:mailController withObject:fileNames];
		
		switchProcess = NO;
    }
	
	if (switchProcess)
		SetFrontProcess(&frontPSN);
	
	[pool release];
}

- (void)passphraseCacheTypeDidChange:(NSNotification*)notification
{
	PGPError		err;
	PGPUInt32		cacheType = kPGPPrefCacheTypeNone;
	
	err = PGPGetPrefNumber([self pgpPrefRef], kPGPPrefCacheType, &cacheType);
	
	switch (cacheType)
	{
		case kPGPPrefCacheTypeNone:
			[self purgePassphraseCache];
			break;
		default:
			break;
	}
	
	if (IsPGPError(err))
		PGPLog((@"Failed to retrieve cache type"));
}

- (void)keyringsDidChange:(NSNotification*)notification
{
	id <PGPBundleKeys>		keysController;
	id <PGPBundleNet>		netController;
	id <PGPBundleDisk>		diskController;
	char					pubKeyringPath[PATH_MAX];
	char					secKeyringPath[PATH_MAX];
	char					netPubKeyringPath[PATH_MAX];
	char					netSecKeyringPath[PATH_MAX];
	BOOL					didChange = NO;
	BOOL					netDidChange = NO;

	netController = [pgpBundleManager bundleByName:@"PGPnet.bundle"];
	keysController = [pgpBundleManager bundleByName:@"PGPkeys.bundle"];
	diskController = [pgpBundleManager bundleByName:@"PGPdisk.bundle"];
	
	pubKeyringPath[0] = '\0';
	secKeyringPath[0] = '\0';
	netPubKeyringPath[0] = '\0';
	netSecKeyringPath[0] = '\0';
	
	PGPclGetPath([self pgpPrefRef], kInvalidPGPPrefRef, kPGPclPublicKeyringFile, pubKeyringPath, PATH_MAX);
	PGPclGetPath([self pgpPrefRef], kInvalidPGPPrefRef, kPGPclPrivateKeyringFile, secKeyringPath, PATH_MAX);

	if (netController)
	{
		PGPclGetPath(kInvalidPGPPrefRef, [netController netPrefRef], kPGPclNetPublicKeyringFile, netPubKeyringPath, PATH_MAX);
		PGPclGetPath(kInvalidPGPPrefRef, [netController netPrefRef], kPGPclNetPrivateKeyringFile, netSecKeyringPath, PATH_MAX);
				
		if (strncmp(pubKeyringPath, netPubKeyringPath, PATH_MAX) ||
			strncmp(secKeyringPath, netSecKeyringPath, PATH_MAX))
		{
			netDidChange = YES;
		}
	}
	
	if (strncmp(pubKeyringPath, mPubKeyringPath, PATH_MAX) ||
		strncmp(secKeyringPath, mSecKeyringPath, PATH_MAX))
	{
		didChange = YES;
		
		strncpy(mPubKeyringPath, pubKeyringPath, PATH_MAX);
		strncpy(mSecKeyringPath, secKeyringPath, PATH_MAX);
	}
	
	if (didChange)
	{
		if (keysController)
			[keysController mainDBInvalidated];

		if (PGPKeyDBRefIsValid(mMainKeyDB))
		{
			PGPFreeKeyDB(mMainKeyDB);
			mMainKeyDB = kInvalidPGPKeyDBRef;
		}
		
		if (keysController)
			[keysController updateKeyrings];
		
		if (netController)
			[netController updateNetKeyrings];
		
		if (diskController)
			[diskController updateKeyrings];
	}
	else if (netController && netDidChange)
		[netController updateNetKeyrings];
}

- (BOOL)applicationShouldHandleReopen:(NSApplication*)app hasVisibleWindows:(BOOL)flag
{
	NSEnumerator*		enumerator;
	NSArray*			windows;
	NSWindow*			window;
	BOOL				visible = NO;
	
	if (mFinishedInitialization)
	{
		windows = [NSApp windows];
		enumerator = [windows objectEnumerator];
		
		while (!visible && (window = [enumerator nextObject]))
			visible = [window isVisible];
		
		if (!visible)
			[pgpInterface setCurrentBundle:[pgpInterface currentBundle] passive:NO];
	}
	
	return NO;
}

- (void)dealloc
{
	if (sAppController == self)
		sAppController = nil;
	
	[[NSNotificationCenter defaultCenter] removeObserver:self];
	
	[super dealloc];
}

@end




#define kSignTextString				@"SignText"
#define kEncryptTextString			@"EncryptText"
#define kEncryptSignTextString		@"EncryptSignText"
#define	kDecryptVerifyTextString	@"DecryptVerifyText"
#define kSignFilesString			@"SignFiles"
#define kEncryptFilesString			@"EncryptFiles"
#define kEncryptSignFilesString		@"EncryptSignFiles"
#define kDecryptVerifyFilesString	@"DecryptVerifyFiles"
#define kWipeFilesString			@"WipeFiles"

#define kRecipientsKey				@"recipients"


@implementation PGPScriptCommand

- (id) performDefaultImplementation
{
	NSString*			result = nil;
	NSString*			commandName = [[self commandDescription] commandName];
	id <PGPBundleMail>	mailController = [[[PGPAppController appController] bundleManager] bundleByName:@"PGPmail.bundle"];
	NSString*			error = nil;
	
	[NSApp activateIgnoringOtherApps:YES];
	
	if ([commandName isEqualToString:kSignTextString])
		result = [mailController signText:[self directParameter] error:&error];
	else if ([commandName isEqualToString:kEncryptTextString])
	{
		result = [mailController encryptText:[self directParameter]
					toRecipients:[[self evaluatedArguments] objectForKey:kRecipientsKey] error:&error];
	}
	else if ([commandName isEqualToString:kEncryptSignTextString])
	{
		result = [mailController signEncryptText:[self directParameter]
					toRecipients:[[self evaluatedArguments] objectForKey:kRecipientsKey] error:&error];
	}
	else if ([commandName isEqualToString:kDecryptVerifyTextString])
		result = [mailController decryptVerifyText:[self directParameter] error:&error showDialog:NO];
	else if ([commandName isEqualToString:kSignFilesString])
		[NSThread detachNewThreadSelector:@selector(signSelectedFiles:) toTarget:mailController withObject:[self directParameter]];
	else if ([commandName isEqualToString:kEncryptFilesString])
		[NSThread detachNewThreadSelector:@selector(encryptSelectedFiles:) toTarget:mailController withObject:[self directParameter]];
	else if ([commandName isEqualToString:kEncryptSignFilesString])
		[NSThread detachNewThreadSelector:@selector(encryptSignSelectedFiles:) toTarget:mailController withObject:[self directParameter]];
	else if ([commandName isEqualToString:kDecryptVerifyFilesString])
		[NSThread detachNewThreadSelector:@selector(decryptVerifySelectedFiles:) toTarget:mailController withObject:[self directParameter]];
	else if ([commandName isEqualToString:kWipeFilesString])
		[NSThread detachNewThreadSelector:@selector(wipeSelectedFiles:) toTarget:mailController withObject:[self directParameter]];
	else
	{
		result = [NSString stringWithFormat:@"PGPScriptCommand description:%@\n commandName:%@",
					[self description], [[self commandDescription] commandName]];
	}
	
	if (result == nil)
		result = error;
	
	return result;
}

@end