/*____________________________________________________________________________
    Copyright (C) 2002 PGP Corporation
    All rights reserved.

    $Id: CPGPdiskObject.m,v 1.43 2002/10/03 20:43:44 jeffc Exp $
____________________________________________________________________________*/

#import "CPGPdiskObject.h"
#import "CPGPdiskUsers.h"
#import <sys/param.h>
#import <sys/ucred.h>
#import <sys/mount.h>

#define kPGPdiskKilobytesThreshold	2048
#define kPGPdiskMegabytesThreshold	2097152

@implementation CPGPdiskObject

+ (id)diskAtPath:(NSString*)path
{
	return [[[CPGPdiskObject alloc] initWithPath:path] autorelease];
}

- (id)init
{
	if (!mInited)
	{
		[super init];
		
		mInited = YES;
	}
	
	if (!mDiskUsers)
		mDiskUsers = [[NSMutableArray alloc] init];
	
	return self;
}

- (id)initWithPath:(NSString*)path
{
	[self setDiskPath:path];
	
	return [self init];
}

- (PGPdiskController*)diskController
{
	return mDiskController;
}

- (void)setController:(PGPdiskController*)controller
{
	mDiskController = controller;
}

- (PGPdiskDiskRef)diskRef
{
	return mDiskRef;
}

- (void)setDiskRef:(PGPdiskDiskRef)diskRef
{
	mDiskRef = diskRef;
}

- (PGPdiskUserSetRef)userSet
{
	return mDiskUserSet;
}

- (void)setUserSet:(PGPdiskUserSetRef)userSet
{
	mDiskUserSet = userSet;
}

- (PGPError)importUserSet:(PGPdiskUserSetRef)userSet adminPassphrase:(const char*)passphrase adminUser:(PGPdiskUserRef)adminUser
{
	PGPError			err = kPGPError_NoErr;
	
	if (!userSet || !passphrase)
		return kPGPError_BadParams;
	
	if (!adminUser)
		adminUser = [CPGPdiskUser findAdminInUserSet:userSet];
	
	err = PGPdiskDiskImportUserSet(mDiskRef, passphrase, userSet, adminUser);
	
	return err;
}

- (void)updateUserSet
{
	PGPError			err;
	PGPdiskUserSetRef	userSet;

	if (PGPdiskDiskRefIsValid(mDiskRef))
	{
		if (PGPdiskUserSetRefIsValid(mDiskUserSet))
			PGPdiskFreeUserSet(mDiskUserSet);
			
		mDiskUserSet = NULL;
		
		err = PGPdiskNewUserSet(PGPdiskPeekDiskContext(mDiskRef), &userSet);
		
		if (IsPGPError(err))
			PGPLog((@"%s:%d: PGPdiskNewUserSet failed with PGPError (err=%d)", __FILE__, __LINE__, err));
		else
		{
			err = PGPdiskDiskExportUserSet(mDiskRef, userSet);
			
			if (IsPGPError(err))
				PGPLog((@"%s:%d: PGPdiskDiskExportUserSet failed with PGPError (err=%d)", __FILE__, __LINE__, err));
			else
				mDiskUserSet = userSet;
		}
	}
	else
		mDiskUserSet = NULL;
}

- (NSString*)diskPath
{
	return mDiskPath;
}

- (NSString*)diskRoot
{
	return [self getStringProperty:kPGPdiskDiskProp_Root];
}

- (NSString*)defaultRoot
{
	return [self getStringProperty:kPGPdiskDiskProp_DefaultRoot];
}

- (NSString*)devicePath
{
	return [self getStringProperty:kPGPdiskDiskProp_DeviceName];
}

- (void)setDiskRoot:(NSString*)root
{
	[self setStringProperty:kPGPdiskDiskProp_DeviceName value:root];
}

- (NSString*)getStringProperty:(PGPdiskDiskPropName)property
{
	PGPError		err;
	char			string[PATH_MAX];
	PGPSize			actualSize = 0;
	BOOL			openedDisk = NO;
	
	if (!mDiskRef)
	{
		[self openDiskWithContext:[[PGPdiskController sharedController] pgpDiskContext]];
		openedDisk = YES;
	}
	
	err = PGPdiskGetDiskDataProp(mDiskRef, property, string, PATH_MAX, &actualSize);
	string[actualSize] = '\0';

	if (IsPGPError(err))
	{
		PGPLog((@"Failed to get string property:%d (err=%d)", property, err));
		
		if (openedDisk)
			[self closeDisk];
		
		return nil;
	}
	
	if (openedDisk)
		[self closeDisk];
	
	return [NSString stringWithCString:string];
}

- (PGPError)setStringProperty:(PGPdiskDiskPropName)property value:(NSString*)string
{
	PGPError		err = kPGPError_NoErr;
	BOOL			openedDisk = NO;
	
	if (!mDiskRef)
	{
		[self openDiskWithContext:[[PGPdiskController sharedController] pgpDiskContext]];
		openedDisk = YES;
	}

	if (string)
	{
		err = PGPdiskSetDiskDataProp(mDiskRef, property, [string cString], [string length] + 1);

		if (IsPGPError(err))
			PGPLog((@"Failed to set string property:%d (err=%d)", property, err));
	}
	else
		PGPLog((@"Invalid string value for property %d", property));
	
	if (openedDisk)
		[self closeDisk];
	
	return err;
}

- (NSString*)shortName
{
	if (mDiskPath)
		return [[mDiskPath lastPathComponent] stringByDeletingPathExtension];
	
	return nil;
}

- (BOOL)isWritable
{
	return [[NSFileManager defaultManager] isWritableFileAtPath:mDiskPath];
}

- (BOOL)hasPubKeyUser
{
	PGPError			err;
	PGPBoolean			isPublicKeyUser = FALSE;
	PGPdiskUserIterRef	userIter;
	PGPdiskUserRef		user;
	
	if (!mDiskUserSet)
		return NO;
		
	if (IsPGPError(err = PGPdiskNewUserIter(mDiskUserSet, &userIter)))
		return NO;
		
	while (!isPublicKeyUser && IsntPGPError(err = PGPdiskUserIterNext(userIter, &user)))
		err = PGPdiskGetUserBooleanProp(user, kPGPdiskUserProp_HasPublicKey, &isPublicKeyUser);
			
	if (userIter)
		PGPdiskFreeUserIter(userIter);	
	
	return isPublicKeyUser;
}

- (BOOL)hideFileExtension
{
	return mHideExtension;
}

- (BOOL)diskNeedsReEncryption
{
	PGPError	err;
	BOOL		flag = [self usesCustomTimeout];
	BOOL		openedDisk = NO;
	
	if (!mDiskRef)
	{
		[self openDiskWithContext:[[PGPdiskController sharedController] pgpDiskContext]];
		openedDisk = YES;
	}

	err = PGPdiskSetDiskBooleanProp(mDiskRef, kPGPdiskDiskProp_UsesCustomTimeout, flag);
	
	if (openedDisk)
		[self closeDisk];
	
	return (err == kPGPdiskError_DiskNeedsReEncryption) ? YES : NO;
}

- (BOOL)diskHasMoved
{
	return ![[NSFileManager defaultManager] fileExistsAtPath:[self diskPath]];
}

- (void)setDiskPath:(NSString*)path
{
	NSFileManager*		fileManager = [NSFileManager defaultManager];
	NSDictionary*		attributes;
	
	if (path)
	{
		if (mDiskPath)
			[mDiskPath release];
		
		mDiskPath = [path retain];
		
		attributes = [fileManager fileAttributesAtPath:mDiskPath traverseLink:NO];
		
		if (attributes)
			mHideExtension = [attributes fileExtensionHidden];
	}
	else
		mHideExtension = NO;
}

- (NSString*)diskSize
{
	PGPError		err = kPGPError_NoErr;
	PGPUInt32		blocksHigh = 0;
	PGPUInt32		blocksLow = 0;
	PGPUInt64		blocks = 0;
	PGPUInt64		diskSize;
	NSString*		sizeString;
	
	if (mDiskRef)
	{
		err = PGPdiskGetDiskNumericProp(mDiskRef, kPGPdiskDiskProp_BlocksDiskHigh, &blocksHigh);
		
		if (IsPGPError(err))
			return nil;
		
		err = PGPdiskGetDiskNumericProp(mDiskRef, kPGPdiskDiskProp_BlocksDiskLow, &blocksLow);
		
		if (IsPGPError(err))
			return nil;
			
		blocks = (PGPUInt64) (blocksLow | ((PGPUInt64) blocksHigh << 32));
	
		if (blocks < kPGPdiskKilobytesThreshold)
		{
			diskSize = blocks * 2;
			
			sizeString = [[[NSNumber numberWithLongLong:diskSize] stringValue] stringByAppendingString:LSPD(@" KB", "Kilobytes")];
		}
		else if (blocks < kPGPdiskMegabytesThreshold)
		{
			diskSize = blocks / kPGPdiskKilobytesThreshold;

			sizeString = [[[NSNumber numberWithLongLong:diskSize] stringValue] stringByAppendingString:LSPD(@" MB", "Megabytes")];
		}
		else
			sizeString = @"";
			
		return sizeString;
	}
	
	return nil;
}

- (NSMutableArray*)diskUsers
{
	return mDiskUsers;
}

- (void)setDiskUsers:(NSMutableArray*)users
{
	if (users)
	{
		if (mDiskUsers)
			[mDiskUsers release];
		
		mDiskUsers = users;
		[mDiskUsers retain];
	}
}

- (BOOL)isMounted
{
	return [self getBooleanProperty:kPGPdiskDiskProp_IsMounted];
}

- (BOOL)isReadOnly
{
	return [self getBooleanProperty:kPGPdiskDiskProp_IsReadOnly];
}

- (BOOL)isBeingReEncrypted
{
	return [self getBooleanProperty:kPGPdiskDiskProp_IsBeingReEncrypted];
}

- (BOOL)hasTimedOut
{
	return [self getBooleanProperty:kPGPdiskDiskProp_HasTimedOut];
}

- (BOOL)usesCustomTimeout
{
	return [self getBooleanProperty:kPGPdiskDiskProp_UsesCustomTimeout];
}

- (BOOL)usesCustomEjectOnLogout
{
	return [self getBooleanProperty:kPGPdiskDiskProp_UsesCustomEjectOnLogout];
}

- (BOOL)ejectOnLogout
{
	return [self getBooleanProperty:kPGPdiskDiskProp_EjectOnLogout];
}

- (void)setUsesCustomTimeout:(PGPBoolean)flag
{
	[self setBooleanProperty:kPGPdiskDiskProp_UsesCustomTimeout value:flag];
}

- (void)setUsesCustomEjectOnLogout:(PGPBoolean)flag
{
	[self setBooleanProperty:kPGPdiskDiskProp_UsesCustomEjectOnLogout value:flag];
}

- (void)setEjectOnLogout:(PGPBoolean)flag
{
	[self setBooleanProperty:kPGPdiskDiskProp_EjectOnLogout value:flag];
}

- (BOOL)getBooleanProperty:(PGPdiskDiskPropName)property
{
	PGPError		err;
	PGPBoolean		value;
	BOOL			openedDisk = NO;
	
	if (!mDiskRef)
	{
		[self openDiskWithContext:[[PGPdiskController sharedController] pgpDiskContext]];
		openedDisk = YES;
	}
	
	err = PGPdiskGetDiskBooleanProp(mDiskRef, property, &value);

	if (IsPGPError(err))
	{
		PGPLog((@"Failed to get boolean property:%d (err=%d)", property, err));
		
		if (openedDisk)
			[self closeDisk];
		
		return FALSE;
	}
	
	if (openedDisk)
		[self closeDisk];
	
	return value;
}

- (PGPError)setBooleanProperty:(PGPdiskDiskPropName)property value:(PGPBoolean)flag
{
	PGPError		err;
	BOOL			openedDisk = NO;
	
	if (!mDiskRef)
	{
		[self openDiskWithContext:[[PGPdiskController sharedController] pgpDiskContext]];
		openedDisk = YES;
	}
	
	err = PGPdiskSetDiskBooleanProp(mDiskRef, property, flag);

	if (IsPGPError(err))
		PGPLog((@"Failed to set boolean property:%d (err=%d)", property, err));
	
	if (openedDisk)
		[self closeDisk];
	
	return err;
}

- (PGPdiskDiskFormat)diskFormat
{
	if (mDiskRef)
		mDiskFormat = [self getNumericProperty:kPGPdiskDiskProp_DiskFormat];
	
	return mDiskFormat;
}

- (PGPdiskEncryptionAlgorithm)diskAlgorithm
{
	return [self getNumericProperty:kPGPdiskDiskProp_Algorithm];
}

- (PGPUInt32)customTimeout
{
	return [self getNumericProperty:kPGPdiskDiskProp_CustomTimeout];
}

- (PGPUInt32)blocksDiskLow
{
	return [self getNumericProperty:kPGPdiskDiskProp_BlocksDiskLow];
}

- (PGPUInt32)blocksDiskHigh
{
	return [self getNumericProperty:kPGPdiskDiskProp_BlocksDiskHigh];
}

- (void)setCustomTimeout:(PGPUInt32)timeout
{
	[self setNumericProperty:kPGPdiskDiskProp_CustomTimeout value:timeout];
}

- (PGPUInt32)getNumericProperty:(PGPdiskDiskPropName)property
{
	PGPError		err;
	PGPUInt32		value;
	BOOL			openedDisk = NO;
	
	if (!mDiskRef)
	{
		[self openDiskWithContext:[[PGPdiskController sharedController] pgpDiskContext]];
		openedDisk = YES;
	}
			
	err = PGPdiskGetDiskNumericProp(mDiskRef, property, &value);

	if (IsPGPError(err))
	{
		PGPLog((@"Failed to get numeric property:%d (err=%d)", property, err));
		
		if (openedDisk)
			[self closeDisk];
		
		return 0;
	}
	
	if (openedDisk)
		[self closeDisk];
	
	return value;
}

- (PGPError)setNumericProperty:(PGPdiskDiskPropName)property value:(PGPUInt32)value
{
	PGPError		err;
	BOOL			openedDisk = NO;
	
	if (!mDiskRef)
	{
		[self openDiskWithContext:[[PGPdiskController sharedController] pgpDiskContext]];
		openedDisk = YES;
	}
	
	err = PGPdiskSetDiskNumericProp(mDiskRef, property, value);

	if (IsPGPError(err))
		PGPLog((@"Failed to set numeric property:%d (err=%d)", property, err));
	
	if (openedDisk)
		[self closeDisk];
	
	return err;
}

- (PGPError)mountWithPassphrase:(const char*)passphrase
{
	PGPError		err = kPGPError_NoErr;

	if (!mDiskRef)
	{
		if (IsPGPError(err = [self openDiskWithContext:[mDiskController pgpDiskContext]]))
			return err;
	}
	
	err = PGPdiskMountDisk(mDiskRef, "ROOT", passphrase, FALSE);
	
	if (IsPGPError(err))
		return err;
    
	/* If the disk controller doesn't know about this disk... */
	if (![mDiskController diskWithPath:mDiskPath])
	{
		[mDiskController addDisk:self];
		[mDiskController updateDiskPopup];
	}
	
    [[mDiskController mountedDisks] addObject:self];
	
	return err;
}

- (PGPError)unmount
{
	PGPError		err = kPGPError_NoErr;
	NSMutableArray*	mountPaths = [[[NSMutableArray alloc] init] autorelease];
	NSString*		devicePath;
	struct statfs*	mntBuf = NULL;
	struct statfs*	mntInfo = NULL;
	int				numElements;
	int				i;
	BOOL			unmounted = YES;
	BOOL			openedDisk = NO;
	
	if (!mDiskRef)
	{
		[self openDiskWithContext:[[PGPdiskController sharedController] pgpDiskContext]];
		openedDisk = YES;
	}
	
	devicePath = [[NSString stringWithString:@"/dev"] stringByAppendingPathComponent:[self devicePath]];
	
	if (numElements = getmntinfo(&mntBuf, MNT_WAIT))
	{
		for (i = 0; i < numElements; ++i)
		{
			NSString*	mountPath;
			NSScanner*	pathScanner;
			
			mntInfo = &mntBuf[i];
			mountPath = [NSString stringWithUTF8String:mntInfo->f_mntfromname];
			pathScanner = [[[NSScanner alloc] initWithString:mountPath] autorelease];
			
			if ([pathScanner scanString:devicePath intoString:nil])
			{
				if (![pathScanner isAtEnd] && [pathScanner scanString:@"s" intoString:nil])
					[mountPaths addObject:[NSString stringWithUTF8String:mntInfo->f_mntonname]];
                else if ([pathScanner isAtEnd])
                    [mountPaths addObject:[NSString stringWithUTF8String:mntInfo->f_mntonname]];
			}
		}
	}

	if ([mountPaths count])
	{
		NSEnumerator*	enumerator = [mountPaths objectEnumerator];
		NSString*		mountPath;
		
		while (mountPath = [enumerator nextObject])
		{
			if (![[NSWorkspace sharedWorkspace] unmountAndEjectDeviceAtPath:mountPath])
			{
				unmounted = NO;
				err = kPGPError_FileOpFailed;
			}
		}
	}
	else
		unmounted = NO;
	
	if (unmounted)
		[[mDiskController mountedDisks] removeObject:self];
	
	[mDiskController updateUsersForDisk:self];
	
	if (openedDisk)
		[self closeDisk];
	
	return err;
}

- (BOOL)diskIsBusy
{
	return mDiskIsBusy;
}

- (void)setDiskIsBusy:(BOOL)flag
{
	mDiskIsBusy = flag;
}

- (PGPError)getAdminPassphrase:(char**)passphrasePtr
{
	PGPdiskUserRef		adminUser;
	PGPdiskUserSetRef	userSet;
	
	if (!(userSet = [self userSet]))
	{
		[self updateUserSet];
		userSet = [self userSet];
	}
	
	adminUser = [CPGPdiskUser findAdminInUserSet:userSet];
	
	return [self getPassphrase:passphrasePtr forUser:adminUser];
}

- (PGPError)getPassphrase:(char**)passphrasePtr forUser:(PGPdiskUserRef)user
{
	PGPError			err = kPGPError_NoErr;
	PGPBoolean			correctPassphrase = FALSE;
	PGPdiskUserSetRef	userSet = NULL;
	PGPdiskUserRef		userRef = NULL;
	PGPdiskContextRef	diskContext;
	PGPContextRef		context;
	NSString*			prompt;
	BOOL				isAdmin = NO;
	BOOL				openedDisk = NO;
	
	if (!passphrasePtr || !user)
		return kPGPError_BadParams;
	
	if (!mDiskRef)
	{
		[self openDiskWithContext:[[PGPdiskController sharedController] pgpDiskContext]];
		openedDisk = YES;
	}
	
	*passphrasePtr = NULL;
	diskContext = PGPdiskPeekDiskContext(mDiskRef);
	userRef = user;
	
	if (userRef)
	{
		PGPdiskGetUserBooleanProp(user, kPGPdiskUserProp_IsAdminUser, &isAdmin);
	
		context = PGPdiskPeekContextPGPContext(diskContext);

retry:
		if (err == kPGPError_BadPassphrase)
			prompt = kPGPdiskString_IncorrectPassphrase;
		else
		{
			if (isAdmin)
			{
				prompt = [NSString stringWithFormat:
								kPGPdiskString_AdminPassphrasePrompt,
								[self shortName]];
			}
			else
			{
				PGPSize		actualSize;
				char		name[kPGPdiskMaxUserNameSize + 1];

				err = PGPdiskGetUserDataProp(userRef, kPGPdiskUserProp_UserName, name, kPGPdiskMaxUserNameSize, &actualSize);
				if (IsPGPError(err))
				{
					PGPLog((@"%s:%d: PGPdiskGetUserDataProp failed with PGPError (err=%d)", __FILE__, __LINE__, err));
					
					if (openedDisk)
						[self closeDisk];
					
					return nil;
				}
				
				if (actualSize)
					name[actualSize] = '\0';

				prompt = [NSString stringWithFormat:
								kPGPdiskString_UserPassphrasePrompt,
								[NSString stringWithCString:name]];
			}
		}
		
		if (IsPGPError(err = PGPdiskIsUserPassphraseCorrect(userRef, "", &correctPassphrase)) || !correctPassphrase)
		{
			err = PGPPassphraseDialog(context,
						PGPOUIDialogPrompt(context, [prompt cString]),
						PGPOUIOutputPassphrase(context, passphrasePtr), PGPOLastOption(context));
		}
		else
			*passphrasePtr = PGPNewData(PGPPeekContextMemoryMgr(context), 1, kPGPMemoryMgrFlags_Clear);
		
		if (IsntPGPError(err))
		{
			err = PGPdiskIsUserPassphraseCorrect(userRef, *passphrasePtr, &correctPassphrase);
			
			if (IsntPGPError(err) && correctPassphrase)
				goto done;
			else if (IsPGPError(err) && (err != kPGPError_BadPassphrase))
				goto done;
			else if (!correctPassphrase)
			{
				PGPFreeData(*passphrasePtr);
				*passphrasePtr = NULL;
				err = kPGPError_BadPassphrase;
				goto retry;
			}
		}
	}
	else
		err = kPGPError_ItemNotFound;
		
done:
	if (userSet)
		PGPdiskFreeUserSet(userSet);
	
	if (openedDisk)
		[self closeDisk];
	
	return err;
}

- (PGPError)openDiskWithContext:(PGPdiskContextRef)contextRef
{
	return [self openDiskWithContext:contextRef diskRef:&mDiskRef ignoreReEncryption:NO];
}

- (PGPError)openDiskWithContext:(PGPdiskContextRef)contextRef diskRef:(PGPdiskDiskRef*)diskRef ignoreReEncryption:(BOOL)flag
{
	PGPError		err = kPGPError_NoErr;
	NSMutableArray*	mountedDisks;
	
	if (!contextRef || !diskRef)
		return kPGPError_BadParams;
	
	if (IsPGPError(err = PGPdiskOpenDisk(contextRef, [mDiskPath UTF8String], FALSE, diskRef)))
		PGPLog((@"%s:%d: PGPdiskOpenDisk failed with PGPError(err=%d)", __FILE__, __LINE__, err));
	else
	{
		mDiskFormat = [self diskFormat];
        
        if ([self isMounted])
		{
			mountedDisks = [mDiskController mountedDisks];
			
			if ([mountedDisks indexOfObject:self] == NSNotFound)
				[[mDiskController mountedDisks] addObject:self];
		}
	}
	
	return err;
}

- (void)closeDisk
{	
	PGPError		err;
	
	if (mDiskRef)
	{
		if (IsPGPError(err = PGPdiskCloseDisk(mDiskRef)))
			PGPLog((@"%s:%d: Failed to close disk \"%@\" (err=%d)", __FILE__, __LINE__, mDiskPath, err));
		
		mDiskRef = NULL;
	}
}

- (NSComparisonResult)compareName:(CPGPdiskObject*)disk
{
	NSComparisonResult	result = NSOrderedAscending;
	
	result = [[self shortName] compare:[disk shortName] options:NSCaseInsensitiveSearch];
	
	return result;
}

- (void)dealloc
{
	if (mDiskPath)
	{
		[mDiskPath release];
		mDiskPath = nil;
	}
		
	if (mDiskUsers)
	{
		[mDiskUsers release];
		mDiskUsers = nil;
	}

	if (mDiskUserSet)
	{
		PGPdiskFreeUserSet(mDiskUserSet);
		mDiskUserSet = NULL;
	}
		
	[self closeDisk];
	
	[super dealloc];
}

@end
