/*____________________________________________________________________________
	Copyright (C) 2002 PGP Corporation
	All rights reserved.

	$Id: PGPdiskDriver.cpp,v 1.21 2002/10/17 06:59:30 wprice Exp $
____________________________________________________________________________*/
#include <IOKit/IOLib.h>
#include <IOKit/system.h>

#include "PGPdiskDriver.h"
#include "PGPdiskDevice.h"
#include "PGPdiskDriverUserClient.h"

#define super IOService
OSDefineMetaClassAndStructors(PGPdiskDriver, IOService)

#define kPGPdiskDriver_IOBufferSize			(64 * 1024)

	bool
PGPdiskDriver::initPGPdisk(
	const PGPdiskCreateDiskParams	*diskParams,
	UInt32							diskID,
	OSDictionary					*dictionary )
{
	bool res = true;
	
	if( !super::init( dictionary ) )
		return false;
	STATUS_LOG(( "PGPdiskDriver::initPGPdisk(size=0x%08lX%08lX,ro=%d)\n",
		(UInt32)(diskParams->diskSizeBytes >> 32),
		(UInt32)(diskParams->diskSizeBytes & 0xFFFFFFFF),
		diskParams->mountReadOnly ));

	fDiskID			= diskID;
	fDiskSize		= diskParams->diskSizeBytes;
	fReadOnly		= diskParams->mountReadOnly;
	fLogoutEject	= diskParams->ejectOnLogout;
	fAutoDiskMount	= diskParams->autoDiskMount;
	fIconPath		= (OSData *)OSData::withBytes( diskParams->iconPath, ::strlen( diskParams->iconPath ) );
	fDiskWasLastPresent = false;
	fLocked = false;
	fActivated = false;
	fIOBuffer = NULL;
	fInDiskControlLock = NULL;
	fWorkQueue = NULL;
	fWorkQueueLock = NULL;
	fWorkQueueSemaphore = NULL;
	fThrottleSemaphore = NULL;
	
	fIOBuffer = IOMalloc( kPGPdiskDriver_IOBufferSize );
	if( !fIOBuffer )
	{
		STATUS_LOG(( "PGPdiskDriver::initPGPdisk could not allocate fIOBuffer\n" ));
		res = false;
	}
	fInDiskControlLock = IOLockAlloc();
	if( !fInDiskControlLock )
	{
		STATUS_LOG(( "PGPdiskDriver::initPGPdisk could not allocate IOLock1\n" ));
		res = false;
	}
	fWorkQueueLock = IOLockAlloc();
	if( !fWorkQueueLock )
	{
		STATUS_LOG(( "PGPdiskDriver::initPGPdisk could not allocate IOLock2\n" ));
		res = false;
	}
	semaphore_create( current_task(), &fWorkQueueSemaphore, SYNC_POLICY_FIFO, 0 );
	if( !fWorkQueueSemaphore )
	{
		STATUS_LOG(( "PGPdiskDriver::initPGPdisk could not allocate fWorkQueueSemaphore\n" ));
		res = false;
	}
	semaphore_create( current_task(), &fThrottleSemaphore, SYNC_POLICY_FIFO, kPGPdiskDriver_ThrottleMaxPending );
	if( !fThrottleSemaphore )
	{
		STATUS_LOG(( "PGPdiskDriver::initPGPdisk could not allocate fThrottleSemaphore\n" ));
		res = false;
	}
	if( !res )
	{
		if( fInDiskControlLock )
		{
			IOLockFree( fInDiskControlLock );
			fInDiskControlLock = NULL;
		}
		if( fWorkQueueLock )
		{
			IOLockFree( fWorkQueueLock );
			fWorkQueueLock = NULL;
		}
		if( fWorkQueueSemaphore )
			semaphore_destroy( current_task(), fWorkQueueSemaphore );
		if( fThrottleSemaphore )
			semaphore_destroy( current_task(), fThrottleSemaphore );
		if( fIOBuffer )
			IOFree( fIOBuffer, kPGPdiskDriver_IOBufferSize );
		if( fIconPath )
		{
			fIconPath->release();
			fIconPath = NULL;
		}
	}
	return res;
}

	void
PGPdiskDriver::free( void )
{
	PGPdiskWorkQueueEntry	*next,
							*walk;
	
	STATUS_LOG(( "PGPdiskDriver::free\n" ));
	
	fActivated = false;
	if( fWorkQueueSemaphore )
		semaphore_signal( fWorkQueueSemaphore );
	if( fThrottleSemaphore )
		semaphore_signal_all( fThrottleSemaphore );
	if( fInDiskControlLock )
	{
		IOLockLock( fInDiskControlLock );
		IOLockUnlock( fInDiskControlLock );
		IOLockFree( fInDiskControlLock );
		fInDiskControlLock = NULL;
	}
	if( fWorkQueueLock )
	{
		IOLockFree( fWorkQueueLock );
		fWorkQueueLock = NULL;
	}
	for( walk = fWorkQueue; walk; walk = next )
	{
		next = walk->next;
		IOFree( walk, sizeof( PGPdiskWorkQueueEntry ) );
	}
	if( fThrottleSemaphore )
		semaphore_destroy( current_task(), fThrottleSemaphore );
	if( fWorkQueueSemaphore )
		semaphore_destroy( current_task(), fWorkQueueSemaphore );
	if( fIOBuffer )
		IOFree( fIOBuffer, kPGPdiskDriver_IOBufferSize );
	if( fIconPath )
	{
		fIconPath->release();
		fIconPath = NULL;
	}
	
	super::free();
}

	IOReturn
PGPdiskDriver::doEjectMedia( void )
{
	if( !fLocked )
	{
		STATUS_LOG(( "PGPdiskDriver::ejected - terminating self\n" ));
		fActivated = false;
		semaphore_signal( fWorkQueueSemaphore );
		terminate(kIOServiceRequired);
		return kIOReturnSuccess;
	}
	else
	{
		STATUS_LOG(( "PGPdiskDriver::eject aborted -- disk locked\n" ));
		return kIOReturnNotPermitted;
	}
}

	bool
PGPdiskDriver::start(
	IOService	*provider )
{
	bool		res = false;
	
	if ( !super::start( provider ) )
		return false;
	STATUS_LOG(( "PGPdiskDriver::start\n" ));

	const OSSymbol *userClient = OSSymbol::withCStringNoCopy( kPGPdiskDriverUserClient_ClassName );
	
	if( !userClient )
		goto abortStart;
	setProperty( gIOUserClientClassKey, (OSObject *) userClient );
	userClient->release();
	registerService();
	setProperty( kPGPdiskDriver_DiskIDProperty, (unsigned long long)fDiskID, 32 );
	setProperty( kPGPdiskDriver_EjectOnLogoutProperty, fLogoutEject );
	setProperty( kPGPdiskDriver_AutoDiskMountProperty, fAutoDiskMount );
	if( fIconPath )
	{
		setProperty( kPGPdiskDriver_IconPathProperty, fIconPath );
		fIconPath->release();
		fIconPath = NULL;
	}
	res = true;
abortStart:
	return res;
}

	bool
PGPdiskDriver::createNub()
{
	IOService	*device;
	
	STATUS_LOG(( "PGPdiskDriver::createNub\n" ));
	device = new PGPdiskDevice;
	if( device == NULL )
	{
		STATUS_LOG(( "PGPdiskDriver::createNub: could not instantiate\n" ));
		return false;
	}
	if( !device->init() )
	{
		STATUS_LOG(( "PGPdiskDriver::createNub: could not init\n" ));
		device->release();
		return false;
	}
	if( !device->attach( this ) )
	{
		STATUS_LOG(( "PGPdiskDriver::createNub: could not attach\n" ));
		device->release();
		return false;
	}
	device->setProperty( kIOBlockStorageDeviceTypeKey, kIOBlockStorageDeviceTypeGeneric );
	device->registerService();
	device->release();
	fActivated = true;
	STATUS_LOG(( "PGPdiskDriver::createNub: exiting true\n" ));
	return true;
}

	IOReturn
PGPdiskDriver::doAsyncReadWrite(
	IOMemoryDescriptor		*buffer,
	UInt32					block,
	UInt32					nblks,
	IOStorageCompletion		completion )
{	
	vm_size_t				byteCount = (UInt64) nblks * (UInt64) kPGPdiskDriver_BlockSize;
	vm_size_t				byteStart = (UInt64) block * (UInt64) kPGPdiskDriver_BlockSize;
	
	if( !fActivated )
		return kIOReturnNoMedia;
	if( ( buffer->getLength() < byteCount ) || ( byteStart + byteCount > fDiskSize ) )
		return kIOReturnInternalError;
	if( fReadOnly && ( buffer->getDirection() == kIODirectionOut ) )
		return kIOReturnLockedWrite;
	{
		PGPdiskWorkQueueEntry	*newWork;
		
		semaphore_wait( fThrottleSemaphore );
		if( !fActivated )
			return kIOReturnNoMedia;
		newWork = (PGPdiskWorkQueueEntry *)IOMalloc( sizeof( PGPdiskWorkQueueEntry ) );
		if( !newWork )
		{
			semaphore_signal( fThrottleSemaphore );
			STATUS_LOG(( "PGPdiskDriver::doAsyncReadWrite: failed to allocate work entry\n" ));
			return kIOReturnNoMemory;
		}
		newWork->next			= NULL;
		newWork->sent			= false;
		newWork->ioBuffer		= buffer;
		newWork->startBlock		= block;
		newWork->numBlocks		= nblks;
		newWork->blocksDone		= 0;
		newWork->completion		= completion;
		//STATUS_LOG(( "PGPdiskDriver::doAsyncReadWrite: queuing work request(%ld/%ld)\n", block, nblks ));
		IOLockLock( fWorkQueueLock );
			if( fWorkQueue )
			{
				PGPdiskWorkQueueEntry *walk;
				
				for( walk = fWorkQueue; walk->next; walk = walk->next )
					;
				walk->next = newWork;
			}
			else
				fWorkQueue = newWork;
		IOLockUnlock( fWorkQueueLock );
		semaphore_signal( fWorkQueueSemaphore );
	}
	return kIOReturnSuccess;
}

	IOReturn
PGPdiskDriver::doSyncReadWrite(
	IOMemoryDescriptor		*buffer,
	UInt32					block,
	UInt32					nblks )
{
	STATUS_LOG(( "PGPdiskDriver::doSyncReadWrite unsupported\n" ));
	return kIOReturnUnsupported;
}

	IOReturn
PGPdiskDriver::doFormatMedia(
	UInt64		byteCapacity )
{
	STATUS_LOG(( "PGPdiskDriver::doFormatMedia (%ld)\n", (UInt32)byteCapacity ));
	if( byteCapacity != fDiskSize )
		return kIOReturnBadArgument;
	STATUS_LOG(( "PGPdiskDriver::doFormatMedia unsupported\n" ));
	return kIOReturnUnsupported;
}

	UInt32
PGPdiskDriver::doGetFormatCapacities(
	UInt64		*capacities,
	UInt32		capacitiesMaxCount ) const
{
	if( capacities == 0 || capacitiesMaxCount == 0 )
		return 1;
	capacities[0] = fDiskSize;
	return 1;	
}

	IOReturn
PGPdiskDriver::doLockUnlockMedia(
	bool		doLock )
{
	fLocked = doLock;
	if( fLocked )
	{
		STATUS_LOG(( "PGPdiskDriver::Lock\n" ));
	}
	else
	{
		STATUS_LOG(( "PGPdiskDriver::Unlock\n" ));
	}
	return kIOReturnSuccess;
}

	IOReturn
PGPdiskDriver::doSynchronizeCache( void )
{
	return kIOReturnSuccess;
}

	char *
PGPdiskDriver::getVendorString( void )
{
	return kPGPdiskDriver_VendorString;
}

	char *
PGPdiskDriver::getProductString( void )
{
	return kPGPdiskDriver_ProductString;
}

	char *
PGPdiskDriver::getRevisionString( void )
{
	return kPGPdiskDriver_RevisionString;
}

	char *
PGPdiskDriver::getAdditionalDeviceInfoString( void )
{
	return kPGPdiskDriver_AdditionalString;
}

	IOReturn
PGPdiskDriver::reportBlockSize(
	UInt64		*blockSize )
{
	*blockSize = kPGPdiskDriver_BlockSize;
	return kIOReturnSuccess;
}

	IOReturn
PGPdiskDriver::reportEjectability(
	bool		*isEjectable )
{
	if( fLocked )
		*isEjectable = false;
	else
		*isEjectable = true;
	return kIOReturnSuccess;
}

	IOReturn
PGPdiskDriver::reportLockability(
	bool		*isLockable )
{
	*isLockable = true;
	return kIOReturnSuccess;
}

	IOReturn
PGPdiskDriver::reportMaxReadTransfer(
	UInt64		blockSize,
	UInt64		*max)
{
	*max = kPGPdiskDriver_MaxTransferSize;
	return kIOReturnSuccess;
}

	IOReturn
PGPdiskDriver::reportMaxWriteTransfer(
	UInt64		blockSize,
	UInt64		*max)
{
	*max = kPGPdiskDriver_MaxTransferSize;
	return kIOReturnSuccess;
}

	IOReturn
PGPdiskDriver::reportMaxValidBlock(
	UInt64		*maxBlock )
{
	*maxBlock = ( fDiskSize / kPGPdiskDriver_BlockSize ) - 1;
	return kIOReturnSuccess;
}

	IOReturn
PGPdiskDriver::reportMediaState(
	bool		*isMediaPresent,
	bool		*hasChanged )
{
	*isMediaPresent = ( fActivated ) ? true : false;
	*hasChanged = ( fDiskWasLastPresent != *isMediaPresent ) ? true : false;
	fDiskWasLastPresent = *isMediaPresent;
	
	return kIOReturnSuccess;
}

	IOReturn
PGPdiskDriver::reportPollRequirements(
	bool		*pollRequired,
	bool		*pollIsExpensive )
{
	*pollRequired = false;
	*pollIsExpensive = false;
	return kIOReturnSuccess;
}

	IOReturn
PGPdiskDriver::reportRemovability(
	bool		*isRemovable )
{
	*isRemovable = true;
	return kIOReturnSuccess;
}

	IOReturn
PGPdiskDriver::reportWriteProtection(
	bool		*isWriteProtected )
{
	*isWriteProtected = fReadOnly;
	return kIOReturnSuccess;
}

	void
PGPdiskDriver::stop(
	IOService	*provider )
{
	STATUS_LOG(( "PGPdiskDriver::stop\n" ));
	super::stop( provider );
}

	IOReturn 
PGPdiskDriver::message(
	UInt32		type,
	IOService	*provider,
	void		*argument)
{
	IOReturn	res = kIOReturnSuccess;
	
	switch (type)
	{
		case kIOMessageServiceIsTerminated:
			if( provider )
			{
				STATUS_LOG(( "PGPdiskDriver::message: kIOMessageServiceIsTerminated from %s\n", provider->getName() ));
				provider->close(this);
			}
			break;
		default:
			res = super::message(type, provider, argument);
			break;
	}
	return res;
}

// PGPdiskDriverUserClient calls follow

	IOReturn 
PGPdiskDriver::activateDrive( UInt32 *diskID )
{
	STATUS_LOG(( "PGPdiskController::activateDisk\n" ));
	if( createNub() )
	{
		*diskID = fDiskID;
		return kIOReturnSuccess;
	}
	else
		return kIOReturnError;
}

	IOReturn
PGPdiskDriver::diskControl(
	const PGPdiskIOParamBlock		*diskIOIn,
	PGPdiskIOParamBlock				*diskIOOut,
	IOMemoryDescriptor				*dataMD )
{
	IOReturn						res			= kIOReturnSuccess;
	PGPdiskWorkQueueEntry			*walk,
									*request	= NULL,
									*last;
	vm_size_t						bDone		= 0,
									bTotal,
									bCurrent,
									bFullTotal = 0,
									bDoneBefore = 0;
	
	if( !fActivated || !fInDiskControlLock )
	{
		diskIOOut->controlCode		= kPGPdiskDriverControlCode_Inactive;
		goto done;
	}
	IOLockLock( fInDiskControlLock );
	if( !fActivated )
	{
		diskIOOut->controlCode		= kPGPdiskDriverControlCode_Inactive;
		goto done;
	}
	diskIOOut->controlCode			= kPGPdiskDriverControlCode_None;
	switch( diskIOIn->controlCode )
	{
		case kPGPdiskDriverControlCode_GetRequest:
			semaphore_wait( fWorkQueueSemaphore );
			if( !fActivated )
			{
				diskIOOut->controlCode		= kPGPdiskDriverControlCode_Inactive;
				goto done;
			}
			IOLockLock( fWorkQueueLock );
			for( walk = fWorkQueue; walk; walk = walk->next )
			{
				if( !walk->sent )
				{
					walk->sent	= true;
					request		= walk;
					break;
				}
			}
			IOLockUnlock( fWorkQueueLock );
			
			if( request )
			{
				diskIOOut->refCon			= (UInt32)request;
				diskIOOut->startBlock		= request->startBlock + request->blocksDone;
				if( request->numBlocks - request->blocksDone > kPGPdiskDriver_MaxTransferBlocks )
					diskIOOut->numBlocks	= kPGPdiskDriver_MaxTransferBlocks;
				else
					diskIOOut->numBlocks	= request->numBlocks - request->blocksDone;
				diskIOOut->err				= 0;
				if( request->ioBuffer->getDirection() == kIODirectionOut )
				{
					diskIOOut->controlCode	= kPGPdiskDriverControlCode_WriteRequest;								
					bTotal = diskIOOut->numBlocks * kPGPdiskDriver_BlockSize;
					while( bDone < bTotal )
					{
						bCurrent = bTotal - bDone;
						if( bCurrent > kPGPdiskDriver_IOBufferSize )
							bCurrent = kPGPdiskDriver_IOBufferSize;
						bDone += request->ioBuffer->readBytes(
									request->blocksDone * kPGPdiskDriver_BlockSize + bDone,
									fIOBuffer, bCurrent );
						dataMD->writeBytes( bDone - bCurrent, fIOBuffer, bCurrent );
					}
				}
				else
					diskIOOut->controlCode	= kPGPdiskDriverControlCode_ReadRequest;
			}
			break;
		case kPGPdiskDriverControlCode_ReadReply:
		case kPGPdiskDriverControlCode_WriteReply:
		{
			bool					sendCompletion = false;
			IOStorageCompletion		completion;
			
			IOLockLock( fWorkQueueLock );
			for( walk = fWorkQueue; walk; walk = walk->next )
			{
				if( (UInt32)walk == diskIOIn->refCon )
					break;
			}
			if( walk )
			{
				bDoneBefore			= walk->blocksDone;
				walk->blocksDone	+= diskIOIn->numBlocks;
				if( diskIOIn->err || ( walk->blocksDone == walk->numBlocks ) )
				{
					sendCompletion	= true;
					completion		= walk->completion;
					bFullTotal		= walk->numBlocks * kPGPdiskDriver_BlockSize;
				}
				else
					walk->sent = false;
			}
			IOLockUnlock( fWorkQueueLock );
			if( ( diskIOIn->controlCode == kPGPdiskDriverControlCode_ReadReply ) && !diskIOIn->err )
			{
				bTotal = diskIOIn->numBlocks * kPGPdiskDriver_BlockSize;
				while( bDone < bTotal )
				{
					bCurrent = bTotal - bDone;
					if( bCurrent > kPGPdiskDriver_IOBufferSize )
						bCurrent = kPGPdiskDriver_IOBufferSize;
					dataMD->readBytes( bDone, fIOBuffer, bCurrent );
					bDone += walk->ioBuffer->writeBytes(
								bDoneBefore * kPGPdiskDriver_BlockSize + bDone,
								fIOBuffer, bCurrent );
				}
			}
			if( sendCompletion )
			{
				if( completion.action )
				{
					( *completion.action )( completion.target, completion.parameter,
											( !diskIOIn->err ) ? kIOReturnSuccess : kIOReturnIOError,
											bFullTotal );
				}
				
				// We're done, dequeue and free the request
				IOLockLock( fWorkQueueLock );
				for( last = NULL, walk = fWorkQueue; walk; walk = walk->next )
				{
					if( (UInt32)walk == diskIOIn->refCon )
						break;
					last = walk;
				}
				if( last )
					last->next = walk->next;
				else
					fWorkQueue = walk->next;
				IOLockUnlock( fWorkQueueLock );
				IOFree( walk, sizeof( PGPdiskWorkQueueEntry ) );
				semaphore_signal( fThrottleSemaphore );
			}
			break;
		}
	}
done:
	IOLockUnlock( fInDiskControlLock );
	return res;
}

