/*____________________________________________________________________________
	Copyright (C) 2002 PGP Corporation
	All rights reserved.

	$Id: pgpDiskEngineDisk.c,v 1.72 2002/11/21 22:38:55 dallen Exp $
____________________________________________________________________________*/

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pthread.h>
#include <sys/param.h>
#include <sys/mount.h>
#include <syslog.h>

#include <CoreFoundation/CoreFoundation.h>

#include "pgpMem.h"
#include "pgpEndianConversion.h"
#include "pgpRandomPool.h"
#include "pgpKeys.h"
#include "pgpPublicKey.h"
#include "pgpPFLPriv.h"
#include "pgpUtilities.h"
#include "pgpHash.h"
#include "pgpRPCPacker.h"
#include "pgpDiskEngineDisk.h"
#include "pgpDiskEngineKernelWorker.h"
#include "pgpDiskHeaders.h"
#include "pgpDiskOS9Headers.h"

/* Macros */

#define CKERR		if( IsPGPError( err ) ) goto done
#define CKNULL(x)	do { if( IsNull( x ) ) { err = kPGPError_OutOfMemory; goto done; } } while(0)
#define RETERR(x)	do { err = x; goto done; } while(0)

#define BYTES_TO_BLOCKS(x)	(((x) % kPGPdiskBlockSize  == 0 ) ? \
	( (x) / kPGPdiskBlockSize ) : ( (x) / kPGPdiskBlockSize ) + 1 )
	
#define kPGPdiskCASTBlockSizeInBytes	(64 / 8)
#define kPGPdiskCASTKeySizeInBytes		(128 / 8)
#define kPGPdiskCASTSaltSizeInBytes		(64 / 8)
#define kPGPdiskCASTNumIVs				4

#define kPGPdiskAESBlockSizeInBytes		(128 / 8)
#define kPGPdiskAESKeySizeInBytes		(256 / 8)
#define kPGPdiskAESSaltSizeInBytes		(128 / 8)
#define kPGPdiskAESNumIVs				8

/* Globals */

typedef struct PGPdiskIODiskTimerData
{
	PGPdiskIODiskRef				disk;
	PGPBoolean						bDiskValid;
} PGPdiskIODiskTimerData;

typedef struct PGPdiskIODiskListItem
{
	dev_t							device;
	ino_t							inode;
	PGPdiskIODiskRef				disk;
	struct PGPdiskIODiskListItem *	next;
} PGPdiskIODiskListItem;

PGPdiskIODiskListItem *		gDiskList			= NULL;
pthread_mutex_t				gDiskListMutex;

/* Functions */

	static inline PGPUInt16
sPGPUInt16ToLittleEndian(
	PGPUInt16		in )
{
	PGPUInt16		out		= 0;
	
	PGPUInt16ToEndian( in, kPGPLittleEndian, (PGPByte *) &out );
	return out;
}

	static inline PGPUInt16
sLittleEndianToPGPUInt16(
	PGPUInt16		in )
{
	return PGPEndianToUInt16( kPGPLittleEndian, (PGPByte *) &in );
}

	static inline PGPUInt32
sPGPUInt32ToLittleEndian(
	PGPUInt32		in )
{
	PGPUInt32		out		= 0;
	
	PGPUInt32ToEndian( in, kPGPLittleEndian, (PGPByte *) &out );
	return out;
}

	static inline PGPUInt32
sLittleEndianToPGPUInt32(
	PGPUInt32		in )
{
	return PGPEndianToUInt32( kPGPLittleEndian, (PGPByte *) &in );
}

	static inline PGPUInt64
sPGPUInt64ToLittleEndian(
	PGPUInt64		in )
{
	PGPUInt64		out		= 0;
	
	PGPUInt64ToEndian( in, kPGPLittleEndian, (PGPByte *) &out );
	return out;
}

	static inline PGPUInt64
sLittleEndianToPGPUInt64(
	PGPUInt64		in )
{
	return PGPEndianToUInt64( kPGPLittleEndian, (PGPByte *) &in );
}

	static inline PGPError
sRead(
	PGPInt32		fd,
	void *			data,
	PGPUInt32		size )
{
	PGPError		err			= kPGPError_NoErr;
	PGPInt32		retval		= 0;
	PGPUInt32		bytesRead	= 0;
	do
	{
		retval = read( fd, data + bytesRead, size - bytesRead );
		if( retval < 0 )
			RETERR( kPGPError_ReadFailed );
		else if( retval == 0 )
			RETERR( kPGPError_EOF );
		bytesRead += retval;
	} while( bytesRead < size );

done:
	return err;
}

	static inline PGPError
sWrite(
	PGPInt32		fd,
	void *			data,
	PGPUInt32		size )
{
	PGPError		err				= kPGPError_NoErr;
	PGPInt32		retval			= 0;
	PGPUInt32		bytesWritten	= 0;
	do
	{
		retval = write( fd, data + bytesWritten, size - bytesWritten );
		if( retval < 0 )
			RETERR( kPGPError_WriteFailed );
		bytesWritten += retval;
	} while( bytesWritten < size );

done:
	return err;
}

	static inline PGPError
sSeek(
	PGPInt32		fd,
	PGPUInt64		offset,
	PGPInt32		whence )
{
	PGPInt32		retval		= 0;

	retval = lseek( fd, offset, whence );
	if( retval == -1 )
		return kPGPError_FileOpFailed;
	return kPGPError_NoErr;
}

	PGPError
PGPdiskEngineInit( void )
{
	pthread_mutex_init( &gDiskListMutex, 0 );
	return kPGPError_NoErr;
}

	PGPError
PGPdiskEngineCleanup( void )
{
	pthread_mutex_destroy( &gDiskListMutex );
	return kPGPError_NoErr;
}

	static PGPUInt32
sComputeCRC(
	const PGPUInt32 *	theDWords,
	PGPUInt32			nDWords )
{
    PGPUInt32       k4C11DB7    = 0x04C11DB7;
    PGPUInt32 const *curDWord   = theDWords;
    PGPUInt32 const *lastDWord  = theDWords + (nDWords - 1);
	PGPUInt32		crc			= 0;
	PGPUInt32		i			= 0;
	PGPInt32		check		= 0;
	
    while (curDWord <= lastDWord )
    {
        PGPUInt8        bytes[4];
        const PGPUInt8  *curByte;
        PGPInt32        remaining;
        PGPUInt32       theDWord;

		/*
		 * Looks like the failed attempt at making an endian-independent CRC 
		 * had the unfortunate sideeffect of making us reverse the bytes,
		 * regardless of the endianness of our platform.
		 */
        theDWord = (* curDWord++);
		
#if BYTE_ORDER == LITTLE_ENDIAN
		PGPUInt32ToEndian( theDWord, kPGPBigEndian, bytes );
#else
		PGPUInt32ToEndian( theDWord, kPGPLittleEndian, bytes );
#endif

        remaining = 4;
        curByte = bytes;

        while (remaining-- != 0)
        {
            crc ^= ((PGPInt32 ) (* curByte)) << 8;
            ++curByte;

            for ( i = 0; i < 8; i++ )
            {
                check = crc;
                crc <<= 1;

                if (check < 0)
                    crc ^= k4C11DB7;
            }
        }
    }
	return crc;
}

	static PGPError
sGetCipherSizes(
	PGPdiskEncryptionAlgorithm		alg,
	PGPUInt32 *						outBlockSize,
	PGPUInt32 *						outKeySize,
	PGPUInt32 *						outSaltSize )
{
	PGPError						err			= kPGPError_NoErr;
	PGPUInt32						keySize		= 0;
	PGPUInt32						blockSize	= 0;
	PGPUInt32						saltSize	= 0;

	switch( alg )
	{
		case kPGPdiskCAST5Algorithm:
			blockSize = kPGPdiskCASTBlockSizeInBytes;
			keySize = kPGPdiskCASTKeySizeInBytes;
			saltSize = kPGPdiskCASTSaltSizeInBytes;
			break;
		case kPGPdiskAESAlgorithm:
			blockSize = kPGPdiskAESBlockSizeInBytes;
			keySize = kPGPdiskAESKeySizeInBytes;
			saltSize = kPGPdiskAESSaltSizeInBytes;
			break;
		default:
			RETERR( kPGPError_FeatureNotAvailable );
			break;
	}

	if( IsntNull( outBlockSize ) )
		*outBlockSize = blockSize;
	if( IsntNull( outKeySize ) )
		*outKeySize = keySize;
	if( IsntNull( outSaltSize ) )
		*outSaltSize = saltSize;
		
done:
	return err;
}

	static void
sEncryptCFBdbl(
	PGPdiskEncryptionAlgorithm	alg,
	void *						cipherContext,
	PGPUInt32					ivs[],
	void *						in,
	void *						out,
	PGPUInt32					len )
{
	switch( alg )
	{
		case kPGPdiskCAST5Algorithm:
			CAST5encryptCFBdbl( cipherContext,
				ivs[0], ivs[1], ivs[2], ivs[3],
				in, out, len );
			break;

		case kPGPdiskAESAlgorithm:
			aesEncryptCFBdbl( cipherContext, ivs, in, out, len );
			break;
		
		default:
			break;
	}
	return;
}

	static void
sDecryptCFBdbl(
	PGPdiskEncryptionAlgorithm	alg,
	void *						cipherContext,
	PGPUInt32					ivs[],
	void *						in,
	void *						out,
	PGPUInt32					len )
{
	switch( alg )
	{
		case kPGPdiskCAST5Algorithm:
			CAST5decryptCFBdbl( cipherContext,
				ivs[0], ivs[1], ivs[2], ivs[3],
				in, out, len );
			break;

		case kPGPdiskAESAlgorithm:
			aesDecryptCFBdbl( cipherContext, ivs, in, out, len );
			break;
		
		default:
			break;
	}
	return;
}

	static PGPError
sEncryptBlock(
	PGPdiskEncryptionAlgorithm	alg,
	void *						cipherContext,
	PGPdiskPassphraseSalt *		salt,
	PGPUInt64					blockNumber,
	PGPByte *					plaintext,
	PGPByte *					ciphertext,
	PGPBoolean					littleEndian )
{
	PGPError				err					= kPGPError_NoErr;
	PGPUInt32				i					= 0;
	PGPInt32				j					= 0;
	UInt32 *				inBlock				= (UInt32 *) plaintext;
	UInt32 *				outBlock			= (UInt32 *) ciphertext;
	const PGPUInt32			kBlockSizeInUInt32	= kPGPdiskBlockSize / sizeof(PGPUInt32);
	PGPInt32				halfIVs				= 0;
	PGPUInt32				cipherBlockSize		= 0;
	PGPUInt32				ivs[8];
	PGPUInt32				sums[4];
	
	PGPValidatePtr( cipherContext );
	PGPValidatePtr( salt );
	PGPValidatePtr( plaintext );
	PGPValidatePtr( ciphertext );
	
	switch( alg )
	{
		case kPGPdiskCAST5Algorithm:	halfIVs = kPGPdiskCASTNumIVs / 2;			break;
		case kPGPdiskAESAlgorithm:		halfIVs = kPGPdiskAESNumIVs / 2;			break;
		default:						RETERR( kPGPError_FeatureNotAvailable );	break;
	}
	
	err = sGetCipherSizes( alg, &cipherBlockSize, NULL, NULL ); CKERR;

	pgpAssert( halfIVs <= 4 );	// if not so, bump up size of above arrays

	for (i = 0; i < halfIVs; i++)
	{
		sums[i] = ( littleEndian ? sLittleEndianToPGPUInt32( salt->words[i] ) : salt->words[i] ) +
			( ( blockNumber & 0xFFFFFFFF00000000 ) >> 32 ) +
			( blockNumber & 0xFFFFFFFF );
	}

	for (i = 0; i < kBlockSizeInUInt32 - halfIVs; i++)
	{
		/* Simple Fletcher checksum of block */
		sums[halfIVs - 1] += ( littleEndian ? sLittleEndianToPGPUInt32( inBlock[i] ) : inBlock[i] );

		for (j = halfIVs - 2; j >= 0; j--)
			sums[j] += sums[j+1];
	}

	for (i = 0; i < halfIVs; i++)
	{
		ivs[i] = ( littleEndian ? sPGPUInt32ToLittleEndian( sums[i] ) : sums[i] )
			^ inBlock[kBlockSizeInUInt32 - halfIVs + i];
	}

	for (i = 0; i < halfIVs; i++)
		ivs[i + halfIVs] = ~ivs[i];

	sEncryptCFBdbl( alg, cipherContext, ivs, inBlock, outBlock, kPGPdiskBlockSize / ( cipherBlockSize * 2 ) );

	for (i = 0; i < halfIVs; i++)
	{
		outBlock[kBlockSizeInUInt32 - halfIVs + i] ^=
			( littleEndian ? sPGPUInt32ToLittleEndian( sums[i] ) : sums[i] );
	}

done:
	return err;
}

	static PGPError
sDecryptBlock(
	PGPdiskEncryptionAlgorithm	alg,
	void *						cipherContext,
	PGPdiskPassphraseSalt *		salt,
	PGPUInt64					blockNumber,
	PGPByte *					ciphertext,
	PGPByte *					plaintext,
	PGPBoolean					littleEndian )
{
	PGPError				err					= kPGPError_NoErr;
	PGPUInt32				ivs[8];
	UInt32 *				inBlock				= (UInt32 *) ciphertext;
	UInt32 *				outBlock			= (UInt32 *) plaintext;
	PGPUInt32				i					= 0;
	PGPInt32				j					= 0;
	PGPInt32				halfIVs				= 0;
	PGPUInt32				cipherBlockSize		= 0;
	const PGPUInt32			kBlockSizeInUInt32	= kPGPdiskBlockSize / sizeof(PGPUInt32);
	const PGPUInt32			kSplitPointInUInt32 = kBlockSizeInUInt32/2;
	const size_t			kBytesBeforeSplit	= kSplitPointInUInt32 * sizeof(PGPUInt32);
	const size_t			kBytesAfterSplit	= (kBlockSizeInUInt32 - kSplitPointInUInt32) * sizeof(PGPUInt32);

	PGPValidatePtr( cipherContext );
	PGPValidatePtr( salt );
	PGPValidatePtr( plaintext );
	PGPValidatePtr( ciphertext );
	
	err = sGetCipherSizes( alg, &cipherBlockSize, NULL, NULL ); CKERR;

	switch( alg )
	{
		case kPGPdiskCAST5Algorithm:	halfIVs = kPGPdiskCASTNumIVs / 2;			break;
		case kPGPdiskAESAlgorithm:		halfIVs = kPGPdiskAESNumIVs / 2;			break;
		default:						RETERR( kPGPError_FeatureNotAvailable );	break;
	}

	pgpAssert( halfIVs <= 4 );	/* if not so, bump up size of above arrays */

	/* This must be done in two halves to recover the IV. */
	for (i = 0; i < halfIVs*2; i++)
		ivs[i] = inBlock[kSplitPointInUInt32 - halfIVs*2 + i];

	sDecryptCFBdbl( alg, cipherContext, ivs,
		inBlock + kSplitPointInUInt32, outBlock + kSplitPointInUInt32,
		kBytesAfterSplit/(cipherBlockSize * 2));

	for (i = 0; i < halfIVs; i++)
		ivs[i] = outBlock[kBlockSizeInUInt32 - halfIVs + i];

	for (i = 0; i < halfIVs; i++)
		ivs[i + halfIVs] = ~ivs[i];

	sDecryptCFBdbl( alg, cipherContext, ivs,
		inBlock, outBlock,
		kBytesBeforeSplit/(cipherBlockSize * 2) );

	/* Undo the checksum */
	for (i = 0; i < halfIVs; i++)
		ivs[i] = ( littleEndian ? sLittleEndianToPGPUInt32( salt->words[i] ) : salt->words[i] ) +
			( ( blockNumber & 0xFFFFFFFF00000000 ) >> 32 ) +
			( blockNumber & 0xFFFFFFFF );

	for (i = 0; i < kBlockSizeInUInt32 - halfIVs; i++)
	{
		/* simple Fletcher checksum of block */
		ivs[halfIVs - 1] += ( littleEndian ? sLittleEndianToPGPUInt32( outBlock[i] ) : outBlock[i] );

		for (j = halfIVs - 2; j >= 0; j--)
			ivs[j] += ivs[j+1];
	}
	
	for (i = 0; i < halfIVs; i++)
		outBlock[kBlockSizeInUInt32 - halfIVs + i] ^= ( littleEndian ? sPGPUInt32ToLittleEndian( ivs[i] ) : ivs[i] );

done:
	return err;
}

	static PGPError
sEncryptDataCAST(
	PGPMemoryMgrRef			mgr,
	PGPdiskSymmetricKey *	hashedKey,
	PGPByte *				in,
	PGPByte *				out,
	PGPUInt32				size )
{
	PGPError				err				= kPGPError_NoErr;
	CASTCFBContext *		cast			= NULL;
	PGPUInt32				i				= 0;
	
	cast = PGPNewSecureData( mgr,
		sizeof( CASTCFBContext ),
		kPGPMemoryMgrFlags_Clear );
	CKNULL( cast );

	CASTCFB_Init( cast, hashedKey->bytes );

	pgpClearMemory( out, size );
	
	for( i = 0; i < size / kPGPdiskCASTBlockSizeInBytes; i++ )
	{
		CAST5encrypt( (const byte *)in + ( i * kPGPdiskCASTBlockSizeInBytes ),
			(byte *)out + ( i * kPGPdiskCASTBlockSizeInBytes ),
			cast->expandedKey.keyWords );
	}

done:
	if( IsntNull( cast ) )
	{
		CASTCFB_Destroy( cast );
		(void) PGPFreeData( cast );
	}
	return err;
}

	static PGPError
sEncryptDataAES(
	PGPMemoryMgrRef			mgr,
	PGPdiskSymmetricKey *	hashedKey,
	PGPByte *				in,
	PGPByte *				out,
	PGPUInt32				size )
{
	PGPError				err				= kPGPError_NoErr;
	PGPByte *				aes				= NULL;
	PGPUInt32				i				= 0;
	
	aes = PGPNewSecureData( mgr,
		kMaxAESContextSize,
		kPGPMemoryMgrFlags_Clear );
	CKNULL( aes );
	
	aes256Key( aes, hashedKey->bytes );

	pgpClearMemory( out, size );
	
	for( i = 0; i < size / kPGPdiskAESBlockSizeInBytes; i++ )
	{
		aesEncrypt( aes,
			in + ( i * kPGPdiskAESBlockSizeInBytes ),
			out + ( i * kPGPdiskAESBlockSizeInBytes ) );
	}

done:
	if( IsntNull( aes ) )
	{
		pgpClearMemory( aes, kMaxAESContextSize );
		PGPFreeData( aes );
	}
	return err;
}

	static PGPError
sEncryptData(
	PGPMemoryMgrRef				mgr,
	PGPdiskEncryptionAlgorithm	alg,
	PGPdiskSymmetricKey *		hashedKey,
	PGPByte *					in,
	PGPByte *					out,
	PGPUInt32					size )
{
	PGPError					err			= kPGPError_NoErr;

	if( alg == kPGPdiskCAST5Algorithm )
		err = sEncryptDataCAST( mgr, hashedKey, in, out, size );
	else if( alg == kPGPdiskAESAlgorithm )
		err = sEncryptDataAES( mgr, hashedKey, in, out, size );
	else
		err = kPGPError_BadParams;

	return err;
}

	static PGPError
sDecryptDataCAST(
	PGPMemoryMgrRef			mgr,
	PGPdiskSymmetricKey *	hashedKey,
	PGPByte *				in,
	PGPByte *				out,
	PGPUInt32				size )
{
	PGPError				err				= kPGPError_NoErr;
	CASTCFBContext *		cast			= NULL;
	PGPUInt32				i				= 0;
	
	cast = PGPNewSecureData( mgr,
		sizeof( CASTCFBContext ),
		kPGPMemoryMgrFlags_Clear );
	CKNULL( cast );

	CASTCFB_Init( cast, hashedKey->bytes );
	pgpClearMemory( out, size );
	
	for( i = 0; i < size / kPGPdiskCASTBlockSizeInBytes; i++ )
	{
		CAST5decrypt( (const byte *)in + ( i * kPGPdiskCASTBlockSizeInBytes ),
			(byte *)out + ( i * kPGPdiskCASTBlockSizeInBytes ),
			cast->expandedKey.keyWords );
	}

done:
	if( IsntNull( cast ) )
	{
		CASTCFB_Destroy( cast );
		(void) PGPFreeData( cast );
	}
	return err;
}

	static PGPError
sDecryptDataAES(
	PGPMemoryMgrRef			mgr,
	PGPdiskSymmetricKey *	hashedKey,
	PGPByte *				in,
	PGPByte *				out,
	PGPUInt32				size )
{
	PGPError				err				= kPGPError_NoErr;
	PGPByte *				aes				= NULL;
	PGPUInt32				i				= 0;

	aes = PGPNewSecureData( mgr,
		kMaxAESContextSize,
		kPGPMemoryMgrFlags_Clear );
	CKNULL( aes );
	
	aes256Key( aes, hashedKey->bytes );

	pgpClearMemory( out, size );
	
	for( i = 0; i < size / kPGPdiskAESBlockSizeInBytes; i++ )
	{
		aesDecrypt( aes,
			in + ( i * kPGPdiskAESBlockSizeInBytes ),
			out + ( i * kPGPdiskAESBlockSizeInBytes ) );
	}

done:
	if( IsntNull( aes ) )
	{
		pgpClearMemory( aes, kMaxAESContextSize );
		PGPFreeData( aes );
	}
	return err;
}

	static PGPError
sDecryptData(
	PGPMemoryMgrRef				mgr,
	PGPdiskEncryptionAlgorithm	alg,
	PGPdiskSymmetricKey *		hashedKey,
	PGPByte *					in,
	PGPByte *					out,
	PGPUInt32					size )
{
	PGPError					err			= kPGPError_NoErr;
	if( alg == kPGPdiskCAST5Algorithm )
		err = sDecryptDataCAST( mgr, hashedKey, in, out, size );
	else if( alg == kPGPdiskAESAlgorithm )
		err = sDecryptDataAES( mgr, hashedKey, in, out, size );
	else
		err = kPGPError_BadParams;

	return kPGPError_NoErr;
}

	static PGPError
sHashPassphraseWithSalt(
	PGPContextRef			context,
	PGPdiskEncryptionAlgorithm		alg,
	PGPdiskPassphraseSalt *	pSalt,
	const char *			passphrase,
	PGPUInt16				hashReps,
	PGPdiskSymmetricKey *	pHashedKey )
{
	PGPError					err				= kPGPError_NoErr;
	PGPHashContextRef			hashCon			= kInvalidPGPHashContextRef;
	PGPUInt32					i				= 0;
	PGPUInt8					pad				= 0;
	PGPByte						hashedPassphrase[20];
	PGPUInt32					keySize			= 0;
	PGPUInt32					saltSize		= 0;
	PGPUInt32					bytesThisTime	= 0;
	PGPUInt32					bytesNeeded		= 0;
	PGPUInt32					offset			= 0;
	
	err = sGetCipherSizes( alg, NULL, &keySize, &saltSize ); CKERR;
	
	bytesNeeded = keySize;
	while( bytesNeeded > 0 )
	{
		bytesThisTime = pgpMin( sizeof( hashedPassphrase ), bytesNeeded );
		
		err = PGPNewHashContext( context, kPGPHashAlgorithm_SHA, &hashCon ); CKERR;
		if( offset > 0 )
		{
			err = PGPContinueHash( hashCon, pHashedKey->bytes, 20 );
			CKERR;
		}
		err = PGPContinueHash( hashCon, passphrase, strlen( passphrase ) ); CKERR;
		err = PGPFinalizeHash( hashCon, hashedPassphrase ); CKERR;
		(void) PGPFreeHashContext( hashCon );
	
		err = PGPNewHashContext( context, kPGPHashAlgorithm_SHA, &hashCon ); CKERR;
		err = PGPContinueHash( hashCon, pSalt->bytes, saltSize ); CKERR;
		for( i = 0, pad = 0; i < hashReps; i++, pad++ )
		{
			err = PGPContinueHash( hashCon, hashedPassphrase, bytesThisTime ); CKERR;
			err = PGPContinueHash( hashCon, &pad, 1 ); CKERR;
		}
		err = PGPFinalizeHash( hashCon, pHashedKey->bytes + offset ); CKERR;
		(void) PGPFreeHashContext( hashCon );

		bytesNeeded -= bytesThisTime;
		offset += bytesThisTime;
	}
	
done:
	return err;
}

	static PGPUInt32
sComputeMainHeaderCRC(
	PGPdiskOnDiskMainHeader *	pMainHdr )
{
	PGPUInt32		oldCRC		= 0;
	PGPUInt32		crc			= 0;
	
	oldCRC = pMainHdr->hdrInfo.headerCRC;

	pMainHdr->hdrInfo.headerCRC = 0;
	crc = sComputeCRC( (PGPUInt32 *) pMainHdr,
		sLittleEndianToPGPUInt32( pMainHdr->hdrInfo.headerSize ) / sizeof( PGPUInt32 ) );
	pMainHdr->hdrInfo.headerCRC = oldCRC;
		
	return crc;
}

/*
 * This is slightly more generic.  We pass in a pointer, which we assume
 * to be the UserHeader plus the UserWithPub or UserWithSym headers
 * after it.  We compute the CRC over the entire buffer.
 */
	static PGPUInt32
sComputeUserHeaderCRC(
	void *		pData )
{
	PGPUInt32					oldCRC		= 0;
	PGPUInt32					crc			= 0;
	PGPdiskOnDiskUserHeader *	pUserHdr	= pData;
	
	oldCRC = pUserHdr->hdrInfo.headerCRC;

	pUserHdr->hdrInfo.headerCRC = 0;
	crc = sComputeCRC( (PGPUInt32 *) pUserHdr,
		sLittleEndianToPGPUInt32( pUserHdr->hdrInfo.headerSize ) / sizeof( PGPUInt32 ) );
	pUserHdr->hdrInfo.headerCRC = oldCRC;
		
	return crc;
}

	static PGPError
sCreateMainHeader(
	PGPContextRef				context,
	PGPdiskOnDiskMainHeader *	pMainHdr,
	PGPUInt64					ullSize,		/* Number of data blocks * kPGPdiskBlockSize */
	const char *				pszDefaultMountPoint,
	PGPdiskEncryptionAlgorithm	alg )
{
	PGPError					err			= kPGPError_NoErr;
	
	PGPValidatePtr( pMainHdr );
	PGPValidateParam( ullSize % kPGPdiskBlockSize == 0 );
	
	pgpClearMemory( pMainHdr, sizeof( PGPdiskOnDiskMainHeader ) );

	pMainHdr->hdrInfo.headerMagic		= sPGPUInt32ToLittleEndian( kPGPdiskHeaderMagic );
	pMainHdr->hdrInfo.headerType		= sPGPUInt32ToLittleEndian( kPGPdiskMainType );
	pMainHdr->hdrInfo.headerSize		= sPGPUInt32ToLittleEndian( sizeof( PGPdiskOnDiskMainHeader ) );
	pMainHdr->hdrInfo.headerCRC			= 0;	/* We'll set this when we write the header to disk */
	pMainHdr->hdrInfo.nextHeaderOffset	= sPGPUInt64ToLittleEndian( 0 );
	
	pMainHdr->majorVersion			= kPGPdiskMainHeader_MajorVersion;
	pMainHdr->minorVersion			= kPGPdiskMainHeader_MinorVersion;

	pMainHdr->numHeaderBlocks		= sPGPUInt64ToLittleEndian( kPGPdiskReservedHeaderBlocks );
	pMainHdr->numDataBlocks			= sPGPUInt64ToLittleEndian( ullSize / kPGPdiskBlockSize );
	/* We assume there are no users yet.  We'll have this include the data and the main header
	 * for now, and add the  number of user blocks later, after we have added some users */
	pMainHdr->numFileBlocks			= sPGPUInt64ToLittleEndian(
										sLittleEndianToPGPUInt64( pMainHdr->numDataBlocks ) +
										sLittleEndianToPGPUInt64( pMainHdr->numHeaderBlocks ) );
	
	pMainHdr->algorithm				= sPGPUInt32ToLittleEndian( alg );
	err = PGPContextGetRandomBytes( context, &pMainHdr->salt, sizeof( pMainHdr->salt ) ); CKERR;
	
	pMainHdr->isBeingReEncrypted	= 0;
	pMainHdr->isWiped				= 0;
	pMainHdr->usesCustomTimeout		= 0;
	
	pMainHdr->customTimeout			= 0;
	pMainHdr->numBlocksReEncrypted	= 0;
	
	if( IsntNull( pszDefaultMountPoint ) )
	{
		strncpy( pMainHdr->defaultRoot, pszDefaultMountPoint,
			pgpMin( sizeof( pMainHdr->defaultRoot ), strlen( pszDefaultMountPoint ) ) );
	}

done:
	return err;
}

	static PGPError
sWriteMainHeaderAtOffset(
	PGPInt32					fd,
	PGPdiskOnDiskMainHeader *	pMainHdr,
	PGPUInt64					offset )
{
	PGPError		err				= kPGPError_NoErr;
	PGPUInt64		blocksWritten	= 0;
	PGPByte			block[kPGPdiskBlockSize];
	
	/* Recompute CRC */
	pMainHdr->hdrInfo.headerCRC = sPGPUInt32ToLittleEndian( sComputeMainHeaderCRC( pMainHdr ) );
	
	err = sSeek( fd, offset, SEEK_SET ); CKERR;
	
	for( blocksWritten = 0;
		blocksWritten < sLittleEndianToPGPUInt64( pMainHdr->numHeaderBlocks );
		blocksWritten++ )
	{
		pgpClearMemory( block, sizeof( block ) );
		if( blocksWritten == 0 )
			pgpCopyMemory( pMainHdr, block, sizeof( PGPdiskOnDiskMainHeader ) );
			
		err = sWrite( fd, block, sizeof( block ) ); CKERR;
	}
done:
	return err;
}

	static PGPError
sWriteMainHeader(
	PGPdiskIODiskRef			disk )
{
	PGPError		err				= kPGPError_NoErr;
	
	PGPdiskValidateIODiskRef( disk );
	
	if( disk->diskFormat != kPGPdiskDiskFormat_V6 )
		RETERR( kPGPError_FeatureNotAvailable );
	
	disk->mainHdr.usesCustomTimeout = disk->bCustomTimeout;
	if( disk->bCustomTimeout )
		disk->mainHdr.customTimeout = sPGPUInt32ToLittleEndian( disk->timeout / 60 );	/* Convert to minutes */
	else
		disk->mainHdr.customTimeout = 0;
		
	err = sWriteMainHeaderAtOffset( disk->fd, &disk->mainHdr, 0 ); CKERR;
	
done:

	return err;
}

	static PGPError
sReadMainHeaderAtOffset(
	PGPInt32					fd,
	PGPdiskOnDiskMainHeader *	pMainHdr,
	PGPUInt64					offset )
{
	PGPError					err			= kPGPError_NoErr;
	PGPUInt32					crc			= 0;
	PGPdiskOS9FileHeader *		pFileHdr	= NULL;
	
	err = sSeek( fd, offset, SEEK_SET );
		
	err = sRead( fd, (PGPByte *) pMainHdr, sizeof( PGPdiskOnDiskMainHeader ) ); CKERR;
	
	if( sLittleEndianToPGPUInt32( pMainHdr->hdrInfo.headerMagic ) != kPGPdiskHeaderMagic )
	{
		pFileHdr = (PGPdiskOS9FileHeader *) pMainHdr;
		if( pFileHdr->headerInfo.headerMagic == kPGPdiskOS9HeaderMagic )
		{
			/* Could be an OS9 disk - work with V2.anything */
			if( ( pFileHdr->headerInfo.headerType == kPGPdiskOS9PrimaryHeaderType )
				&& ( pFileHdr->majorVersion == kPGPdiskOS9PrimaryHeaderMajorVersion ) )
				RETERR( kPGPdiskError_OS9DiskNeedsConversion );
			else
				RETERR( kPGPError_CorruptData );
		}
		else
			RETERR( kPGPError_CorruptData );
	}

	/* Make sure it's a version we understand - ignore the minor version */
	if( pMainHdr->majorVersion != kPGPdiskMainHeader_MajorVersion )
		RETERR( kPGPError_CorruptData );

	if( sLittleEndianToPGPUInt32( pMainHdr->hdrInfo.headerType ) != kPGPdiskMainType )
		RETERR( kPGPError_CorruptData );
	
	crc = sLittleEndianToPGPUInt32( pMainHdr->hdrInfo.headerCRC );
	if( crc != sComputeMainHeaderCRC( pMainHdr ) )
		RETERR( kPGPError_CorruptData );
	
done:
	return err;
}

	static PGPError
sReadMainHeader(
	PGPdiskIODiskRef			disk )
{
	return sReadMainHeaderAtOffset( disk->fd, &disk->mainHdr, 0 );
}

	static PGPError
sFlattenUserIntoBuffer(
	PGPMemoryMgrRef			mgr,
	PGPdiskIODiskUser *		pUser,
	PGPByte **				ppBuf,
	PGPSize *				pSize )
{
	PGPError				err			= kPGPError_NoErr;
	PGPByte *				buf			= NULL;
	PGPSize					size		= 0;
	
	PGPValidatePtr( pUser );
	PGPValidatePtr( ppBuf );

	size = sLittleEndianToPGPUInt32( pUser->userHdr.exportedUserOffset ) +
		sLittleEndianToPGPUInt32( pUser->userHdr.exportedUserSize );

	buf = PGPNewData( mgr, size, kPGPMemoryMgrFlags_Clear );
	CKNULL( buf );
	pgpClearMemory( buf, size );
	
	if( pUser->bSymUser )
	{
		pgpCopyMemory( &pUser->userHdr, buf, sizeof( PGPdiskOnDiskUserHeader ) );
		pgpCopyMemory( &pUser->user, buf + sizeof( PGPdiskOnDiskUserHeader ),
			sLittleEndianToPGPUInt32( pUser->userHdr.exportedUserSize ) );
	}
	else
	{
		pgpCopyMemory( &pUser->userHdr, buf, sizeof( PGPdiskOnDiskUserHeader ) );
		pgpCopyMemory( &pUser->user, buf + sizeof( PGPdiskOnDiskUserHeader ),
			sizeof( PGPdiskOnDiskUserWithPub ) );
		pgpCopyMemory( pUser->encryptedKey,
			buf + sizeof( PGPdiskOnDiskUserHeader ) + sizeof( PGPdiskOnDiskUserWithPub ),
			pUser->encryptedKeySize );
	}
	
	*ppBuf = buf;
	*pSize = size;
		
done:
	if( IsPGPError( err ) )
	{
		if( IsntNull( buf ) )
			(void) PGPFreeData( buf );
	}
	return err;
}

	static PGPError
sWriteUserHeader(
	PGPMemoryMgrRef				mgr,
	PGPInt32					fd,
	PGPUInt64					offset,
	PGPdiskIODiskUser *			pUser )
{
	PGPError					err				= kPGPError_NoErr;
	PGPByte *					data			= NULL;
	PGPSize						size			= 0;
	PGPByte *					paddedData		= NULL;
	PGPdiskOnDiskUserHeader *	pTempUserHdr	= NULL;
	PGPUInt32					numBlocks		= 0;
	
	err = sFlattenUserIntoBuffer( mgr, pUser, &data, &size ); CKERR;

	/* Pad to 512 bytes */
	numBlocks = BYTES_TO_BLOCKS( size );
		
	paddedData = PGPNewData( mgr,
		numBlocks * kPGPdiskBlockSize,
		kPGPMemoryMgrFlags_Clear );
	CKNULL( paddedData );
	
	pgpClearMemory( paddedData, numBlocks * kPGPdiskBlockSize );
	pgpCopyMemory( data, paddedData, size );
	
	pTempUserHdr = (PGPdiskOnDiskUserHeader *) paddedData;
		
	pTempUserHdr->hdrInfo.headerCRC = sPGPUInt32ToLittleEndian( sComputeUserHeaderCRC( paddedData ) );

	err = sSeek( fd, offset, SEEK_SET ); CKERR;
	err = sWrite( fd, paddedData, numBlocks * kPGPdiskBlockSize ); CKERR;

done:
	if( IsntNull( data ) )
		(void) PGPFreeData( data );
	if( IsntNull( paddedData ) )
		(void) PGPFreeData( paddedData );
	return err;
}

	static PGPError
sValidateUserHeaderCRC(
	PGPMemoryMgrRef			mgr,
	PGPdiskIODiskUser *		pUser )
{
	PGPError				err			= kPGPError_NoErr;
	PGPUInt32				crc			= 0;
	PGPByte *				buf			= NULL;
	PGPSize					size		= 0;
	
	/* Check user header CRC */
	err = sFlattenUserIntoBuffer( mgr, pUser, &buf, &size ); CKERR;
	
	crc = sLittleEndianToPGPUInt32( pUser->userHdr.hdrInfo.headerCRC );
	if( crc != sComputeUserHeaderCRC( buf ) )
		RETERR( kPGPError_CorruptData );
		
done:
	if( IsntNull( buf ) )
		(void) PGPFreeData( buf );
	return err;
}

	static PGPError
sReadUserHeader(
	PGPMemoryMgrRef			mgr,
	PGPInt32				fd,
	PGPUInt64				headerOffset,
	PGPdiskIODiskUser *		pUser )
{
	PGPError				err			= kPGPError_NoErr;
	PGPdiskOnDiskUserInfo	userInfo;
	
	/* Read UserHeader */
	err = sSeek( fd, headerOffset, SEEK_SET ); CKERR;
	err = sRead( fd, (PGPByte *) &pUser->userHdr, sizeof( PGPdiskOnDiskUserHeader ) ); CKERR;
	
	if( sLittleEndianToPGPUInt32( pUser->userHdr.hdrInfo.headerMagic ) != kPGPdiskHeaderMagic )
		RETERR( kPGPError_CorruptData );
		
	if( sLittleEndianToPGPUInt32( pUser->userHdr.hdrInfo.headerType ) != kPGPdiskUserType )
		RETERR( kPGPError_CorruptData );
	
	/* Read UserWithPub or UserWithSym */
	err = sSeek( fd,
		headerOffset + sLittleEndianToPGPUInt32( pUser->userHdr.exportedUserOffset ),
		SEEK_SET ); CKERR;
	err = sRead( fd, &userInfo, sizeof( userInfo ) ); CKERR;
	
	if( sLittleEndianToPGPUInt32( userInfo.userMagic ) != kPGPdiskUserMagic )
		RETERR( kPGPError_CorruptData );

	if( sLittleEndianToPGPUInt32( userInfo.userType ) == kPGPdiskUserWithSymType )
	{
		/* Sanity check */
		if( sLittleEndianToPGPUInt32( pUser->userHdr.exportedUserSize ) != sizeof( PGPdiskOnDiskUserWithSym ) )
			RETERR( kPGPError_CorruptData );

		pUser->bSymUser = TRUE;
		pgpCopyMemory( &userInfo, &pUser->user.symUser, sizeof( PGPdiskOnDiskUserInfo ) );
		err = sRead( fd,
			(PGPByte *) &pUser->user.symUser + sizeof( PGPdiskOnDiskUserInfo ),
			sizeof( PGPdiskOnDiskUserWithSym ) - sizeof( PGPdiskOnDiskUserInfo ) );
		CKERR;
	}
	else if( sLittleEndianToPGPUInt32( userInfo.userType ) == kPGPdiskUserWithPubType )
	{
		pUser->bSymUser = FALSE;
		pgpCopyMemory( &userInfo, &pUser->user.pubUser, sizeof( PGPdiskOnDiskUserInfo ) );
		err = sRead( fd,
			(PGPByte *) &pUser->user.pubUser + sizeof( PGPdiskOnDiskUserInfo ),
			sizeof( PGPdiskOnDiskUserWithPub ) - sizeof( PGPdiskOnDiskUserInfo ) );
		CKERR;

		/* Sanity check */
		if( sLittleEndianToPGPUInt32( pUser->userHdr.exportedUserSize ) !=
			sizeof( PGPdiskOnDiskUserWithPub ) + sLittleEndianToPGPUInt32( pUser->user.pubUser.encryptedKeySize ) )
			RETERR( kPGPError_CorruptData );
		
		/* Read public key portion */
		pUser->encryptedKeySize = sLittleEndianToPGPUInt32( pUser->user.pubUser.encryptedKeySize );
		pUser->encryptedKey = PGPNewData( mgr,
			pUser->encryptedKeySize,
			kPGPMemoryMgrFlags_Clear );
		CKNULL( pUser->encryptedKey );
		
		err = sRead( fd, pUser->encryptedKey, pUser->encryptedKeySize ); CKERR;
	}
	else
		RETERR( kPGPError_CorruptData );
		
	err = sValidateUserHeaderCRC( mgr, pUser ); CKERR;
	
done:
	return err;
}

	static PGPError
sWriteDiskUsers(
	PGPdiskIODiskRef	disk )
{
	PGPError					err					= kPGPError_NoErr;
	PGPdiskIODiskUser *			ioDiskUser;
	PGPUInt64					offset				= 0;
	PGPUInt64					firstUserOffset		= 0;
	PGPUInt64					nextHeaderOffset	= 0;
	PGPUInt32					numUserBlocks		= 0;
	
	if( disk->diskFormat != kPGPdiskDiskFormat_V6 )
		RETERR( kPGPError_FeatureNotAvailable );
	
	firstUserOffset = ( sLittleEndianToPGPUInt64( disk->mainHdr.numHeaderBlocks ) +
		sLittleEndianToPGPUInt64( disk->mainHdr.numDataBlocks ) ) * kPGPdiskBlockSize;
	offset = firstUserOffset;
	
	/* If disk->reencryptingMainHdr and disk->reencryptingAdminUser, write these first */
	if( IsntNull( disk->reencryptingMainHdr ) && IsntNull( disk->reencryptingAdminUser ) )
	{
		nextHeaderOffset = offset + BYTES_TO_BLOCKS( sizeof( PGPdiskOnDiskMainHeader ) ) * kPGPdiskBlockSize;
		numUserBlocks += BYTES_TO_BLOCKS( sizeof( PGPdiskOnDiskMainHeader ) );
		
		disk->reencryptingMainHdr->hdrInfo.nextHeaderOffset = sPGPUInt64ToLittleEndian( nextHeaderOffset );
		err = sWriteMainHeaderAtOffset( disk->fd, disk->reencryptingMainHdr, offset ); CKERR;
		
		offset = nextHeaderOffset;
		
		nextHeaderOffset = 
			offset + BYTES_TO_BLOCKS( sizeof( PGPdiskOnDiskUserHeader ) +
			sLittleEndianToPGPUInt32( disk->reencryptingAdminUser->userHdr.exportedUserSize ) ) * kPGPdiskBlockSize;
		numUserBlocks += BYTES_TO_BLOCKS( sizeof( PGPdiskOnDiskUserHeader ) +
			sLittleEndianToPGPUInt32( disk->reencryptingAdminUser->userHdr.exportedUserSize ) );
			
		disk->reencryptingAdminUser->userHdr.hdrInfo.nextHeaderOffset = sPGPUInt64ToLittleEndian( nextHeaderOffset );
		err = sWriteUserHeader( disk->mgr, disk->fd, offset, disk->reencryptingAdminUser ); CKERR;
		
		offset = nextHeaderOffset;
	}
	
	/* Write disk users */
	ioDiskUser = disk->userList;
	while( IsntNull( ioDiskUser ) )
	{
		if( IsNull( ioDiskUser->next ) )
		{
			/* Last user.  nextHeaderOffset should be 0 */
			nextHeaderOffset = 0;
		}
		else
		{
			nextHeaderOffset = 
				offset + BYTES_TO_BLOCKS( sizeof( PGPdiskOnDiskUserHeader ) +
				sLittleEndianToPGPUInt32( ioDiskUser->userHdr.exportedUserSize ) ) * kPGPdiskBlockSize;
		}
		ioDiskUser->userHdr.hdrInfo.nextHeaderOffset = sPGPUInt64ToLittleEndian( nextHeaderOffset );
		
		numUserBlocks += BYTES_TO_BLOCKS( sizeof( PGPdiskOnDiskUserHeader ) +
				sLittleEndianToPGPUInt32( ioDiskUser->userHdr.exportedUserSize ) );
				
		err = sWriteUserHeader( disk->mgr, disk->fd, offset, ioDiskUser ); CKERR;
		
		offset = nextHeaderOffset;
		ioDiskUser = ioDiskUser->next;
	}
	
	/* Update main header */
	disk->mainHdr.hdrInfo.nextHeaderOffset = sPGPUInt64ToLittleEndian( firstUserOffset );
	disk->mainHdr.numFileBlocks = sPGPUInt64ToLittleEndian(
		sLittleEndianToPGPUInt64( disk->mainHdr.numHeaderBlocks ) + 
		sLittleEndianToPGPUInt64( disk->mainHdr.numDataBlocks ) +
		numUserBlocks );

	(void) ftruncate( disk->fd, sLittleEndianToPGPUInt64( disk->mainHdr.numFileBlocks ) * kPGPdiskBlockSize );

	err = sWriteMainHeader( disk ); CKERR;

done:
	return err;
}

	static PGPError
sGetRandomHashReps(
	PGPContextRef		context,
	PGPUInt16 *			pHashReps )
{
	PGPError			err			= kPGPError_NoErr;
	PGPUInt16			hashReps	= 0;
	
	do
	{
		err = PGPContextGetRandomBytes( context, &hashReps, sizeof( hashReps ) );
		CKERR;
	} while( hashReps < 11000 || hashReps > 20000 );
	
	*pHashReps = hashReps;
	
done:
	return err;
}

	static PGPError
sMakeEncryptedKeyForSymUser(
	PGPContextRef				context,
	PGPMemoryMgrRef				mgr,
	PGPdiskIODiskUser *			pUser,
	const char *				pszPassphrase,
	PGPdiskOnDiskMainHeader *	pMainHdr,
	PGPdiskSymmetricKey *		pDiskKey )
{
	PGPError					err				= kPGPError_NoErr;
	PGPUInt16					hashReps		= 0;
	PGPdiskSymmetricKey *		pHashedKey		= NULL;
	PGPUInt32					blockSize		= 0;
	PGPUInt32					keySize			= 0;
	
	PGPValidatePtr( pUser );
	PGPValidatePtr( pszPassphrase );
	PGPValidatePtr( pMainHdr );
	PGPValidatePtr( pDiskKey );
	
	err = sGetRandomHashReps( context, &hashReps ); CKERR;
	pUser->user.symUser.keyInfo.hashReps = sPGPUInt16ToLittleEndian( hashReps );
	
	err = sGetCipherSizes( sLittleEndianToPGPUInt32( pMainHdr->algorithm ), &blockSize, &keySize, NULL ); CKERR;
	
	pHashedKey = PGPNewSecureData( mgr,
		sizeof( PGPdiskSymmetricKey ),
		kPGPMemoryMgrFlags_Clear );
	CKNULL( pHashedKey );

	/* Encrypt disk key to H(passphrase) */
	err = sHashPassphraseWithSalt( context,
		sLittleEndianToPGPUInt32( pMainHdr->algorithm ),
		&pMainHdr->salt,
		pszPassphrase,
		hashReps,
		pHashedKey ); CKERR;

	/* Make encryptedKey */
	err = sEncryptData( mgr, 
		sLittleEndianToPGPUInt32( pMainHdr->algorithm ),
		pHashedKey,
		pDiskKey->bytes,
		pUser->user.symUser.keyInfo.encryptedKey.bytes,
		keySize ); CKERR;
		
	/* Make checkBytes */
	err = sEncryptData( mgr, 
		sLittleEndianToPGPUInt32( pMainHdr->algorithm ),
		pHashedKey,
		pHashedKey->bytes,
		pUser->user.symUser.keyInfo.checkBytes.bytes,
		blockSize ); CKERR;

done:
	if( IsntNull( pHashedKey ) )
	{
		pgpClearMemory( pHashedKey, sizeof( PGPdiskSymmetricKey ) );
		(void) PGPFreeData( pHashedKey );
	}
	return err;
}

	static PGPError
sCreateSymmetricKeyUser(
	PGPContextRef				context,
	PGPMemoryMgrRef				mgr,
	PGPdiskOnDiskMainHeader *	pMainHdr,
	PGPdiskIODiskUser *			pUser,
	PGPdiskSymmetricKey *		pDiskKey,
	const char *				pszUsername,
	const char *				pszPassphrase,
	PGPBoolean					bReadOnly,
	PGPBoolean					bAdminUser,
	PGPBoolean					bDisabled,
	PGPBoolean					bLocked )
{
	PGPError					err				= kPGPError_NoErr;
	
	pUser->bSymUser = TRUE;
	
	pgpClearMemory( &pUser->userHdr, sizeof( PGPdiskOnDiskUserHeader ) );
	pgpClearMemory( &pUser->user.symUser, sizeof( PGPdiskOnDiskUserWithSym ) );
	
	pUser->userHdr.hdrInfo.headerMagic		= sPGPUInt32ToLittleEndian( kPGPdiskHeaderMagic );
	pUser->userHdr.hdrInfo.headerType		= sPGPUInt32ToLittleEndian( kPGPdiskUserType );
	pUser->userHdr.hdrInfo.headerSize		= sPGPUInt32ToLittleEndian(
		sizeof( PGPdiskOnDiskUserHeader ) + sizeof( PGPdiskOnDiskUserWithSym ) );
	pUser->userHdr.hdrInfo.headerCRC		= 0;
	pUser->userHdr.hdrInfo.nextHeaderOffset	= 0;
	
	pUser->userHdr.notExported				= 0;
	pUser->userHdr.exportedUserOffset		= sPGPUInt32ToLittleEndian( sizeof( PGPdiskOnDiskUserHeader ) );
	pUser->userHdr.exportedUserSize			= sPGPUInt32ToLittleEndian( sizeof( PGPdiskOnDiskUserWithSym ) );
	
	pUser->user.symUser.header.userMagic	= sPGPUInt32ToLittleEndian( kPGPdiskUserMagic );
	pUser->user.symUser.header.userType		= sPGPUInt32ToLittleEndian( kPGPdiskUserWithSymType );
	pUser->user.symUser.header.userSize		= sPGPUInt32ToLittleEndian( sizeof( PGPdiskOnDiskUserWithSym ) );
	pUser->user.symUser.header.readOnly		= bReadOnly;
	pUser->user.symUser.header.locked		= bLocked;
	pUser->user.symUser.header.adminUser	= bAdminUser;
	pUser->user.symUser.header.disabled		= bDisabled;
	
	strncpy( pUser->user.symUser.userName, pszUsername, kPGPdiskMaxUserNameSize );
	
	err = sMakeEncryptedKeyForSymUser( context, mgr, pUser, pszPassphrase, pMainHdr, pDiskKey ); CKERR;
	
done:
	return err;
}

	static PGPError
sGeneratePubKeyCheckBytes(
	PGPMemoryMgrRef				mgr,
	PGPdiskEncryptionAlgorithm	alg,
	PGPKeyDBObjRef				key,
	PGPdiskSymmetricKey *		pDiskKey,
	PGPdiskCheckBytes *			pCheckBytes )
{
	PGPError					err					= kPGPError_NoErr;
	PGPByte						fingerprint[128];
	PGPSize						fingerprintSize		= 0;
	PGPUInt32					blockSize			= 0;
	
	PGPValidatePtr( pDiskKey );
	PGPValidatePtr( pCheckBytes );

	err = PGPGetKeyDBObjDataProperty( key,
		kPGPKeyProperty_Fingerprint,
		fingerprint,
		sizeof( fingerprint ),
		&fingerprintSize ); CKERR;
		
	err = sGetCipherSizes( alg, &blockSize, NULL, NULL ); CKERR;
		
	pgpClearMemory( pCheckBytes->bytes, sizeof( PGPdiskCheckBytes ) );
	err = sEncryptData( mgr,
		alg,
		pDiskKey,
		fingerprint,
		pCheckBytes->bytes,
		blockSize ); CKERR;

done:
	return err;
}

	static PGPError
sMakeEncryptedKeyForPubUser(
	PGPMemoryMgrRef				mgr,
	PGPdiskIODiskUser *			pUser,
	PGPKeyDBRef					keydb,
	PGPdiskOnDiskMainHeader *	pMainHdr,
	PGPdiskSymmetricKey *		pDiskKey )
{
	PGPError					err					= kPGPError_NoErr;
	PGPKeyDBObjRef				key					= kInvalidPGPKeyDBObjRef;
	PGPKeyDBObjRef				encKey				= kInvalidPGPKeyDBObjRef;
	PGPPublicKeyContextRef		pubContext			= kInvalidPGPPublicKeyContextRef;
	PGPSize						maxDecryptedSize	= 0;
	PGPSize						maxEncryptedSize	= 0;
	PGPSize						maxSignatureSize	= 0;
	PGPUInt32					keySize				= 0;

	err = PGPFindKeyByKeyID( keydb, &pUser->user.pubUser.keyID, &key ); CKERR;
	
	/* Get encryption subkey */
	err = PGPGetKeyForUsage( key, kPGPKeyPropertyFlags_UsageEncryptStorage, &encKey ); CKERR;
	err = PGPGetKeyID( encKey, &pUser->user.pubUser.subKeyID ); CKERR;

	/* Encrypt disk key to subkey */
	err = PGPNewPublicKeyContext( encKey,
		kPGPPublicKeyMessageFormat_PGP,
		&pubContext ); CKERR;
	
	err = PGPGetPublicKeyOperationSizes( pubContext,
		&maxDecryptedSize,
		&maxEncryptedSize,
		&maxSignatureSize ); CKERR;
	
	if( IsntNull( pUser->encryptedKey ) )
	{
		(void) PGPFreeData( pUser->encryptedKey );
		pUser->encryptedKey = NULL;
	}

	pUser->encryptedKey = PGPNewData( mgr,
		maxEncryptedSize,
		kPGPMemoryMgrFlags_Clear );
	CKNULL( pUser->encryptedKey );
	
	err = sGetCipherSizes( sLittleEndianToPGPUInt32( pMainHdr->algorithm ), NULL, &keySize, NULL ); CKERR;
		
	err = PGPPublicKeyEncrypt( pubContext,
		pDiskKey,
		keySize,
		pUser->encryptedKey,
		&pUser->encryptedKeySize ); CKERR;
		
	pUser->user.pubUser.encryptedKeyOffset	= sPGPUInt32ToLittleEndian( sizeof( PGPdiskOnDiskUserWithPub ) );
	pUser->user.pubUser.encryptedKeySize	= sPGPUInt32ToLittleEndian( maxEncryptedSize );
	
	pUser->userHdr.exportedUserSize = sPGPUInt32ToLittleEndian(
		sizeof( PGPdiskOnDiskUserWithPub ) +
		sLittleEndianToPGPUInt32( pUser->user.pubUser.encryptedKeySize ) );
		
	pUser->userHdr.hdrInfo.headerSize = sPGPUInt32ToLittleEndian(
		sLittleEndianToPGPUInt32( pUser->userHdr.exportedUserOffset ) +
		sLittleEndianToPGPUInt32( pUser->userHdr.exportedUserSize ) );
	pUser->user.pubUser.header.userSize = sPGPUInt32ToLittleEndian(
		sLittleEndianToPGPUInt32( pUser->user.pubUser.encryptedKeyOffset ) +
		sLittleEndianToPGPUInt32( pUser->user.pubUser.encryptedKeySize ) );

	/* Generate checkBytes */
	err = sGeneratePubKeyCheckBytes( mgr,
		sLittleEndianToPGPUInt32( pMainHdr->algorithm ),
		encKey,
		pDiskKey,
		&pUser->user.pubUser.checkBytes ); CKERR;

done:
	if( PGPPublicKeyContextRefIsValid( pubContext ) )
		(void) PGPFreePublicKeyContext( pubContext );
	return err;
}

	static PGPError
sCreatePublicKeyUser(
	PGPMemoryMgrRef				mgr,
	PGPdiskOnDiskMainHeader *	pMainHdr,
	PGPdiskIODiskUser *			pUser,
	PGPdiskSymmetricKey *		pDiskKey,
	PGPKeyDBObjRef				key,
	PGPBoolean					bReadOnly,
	PGPBoolean					bAdminUser,
	PGPBoolean					bDisabled,
	PGPBoolean					bLocked )
{
	PGPError					err					= kPGPError_NoErr;
	
	pUser->bSymUser = FALSE;
	
	pgpClearMemory( &pUser->userHdr, sizeof( PGPdiskOnDiskUserHeader ) );
	pgpClearMemory( &pUser->user.pubUser, sizeof( PGPdiskOnDiskUserWithPub ) );
	
	pUser->userHdr.hdrInfo.headerMagic		= sPGPUInt32ToLittleEndian( kPGPdiskHeaderMagic );
	pUser->userHdr.hdrInfo.headerType		= sPGPUInt32ToLittleEndian( kPGPdiskUserType );
	pUser->userHdr.hdrInfo.headerSize		= 0;			/* Set this later */
	pUser->userHdr.hdrInfo.headerCRC		= 0;
	pUser->userHdr.hdrInfo.nextHeaderOffset	= 0;
	
	pUser->userHdr.notExported				= 0;
	pUser->userHdr.exportedUserOffset		= sPGPUInt32ToLittleEndian( sizeof( PGPdiskOnDiskUserHeader ) );
	pUser->userHdr.exportedUserSize			= sPGPUInt32ToLittleEndian( sizeof( PGPdiskOnDiskUserWithPub ) );
	
	pUser->user.pubUser.header.userMagic	= sPGPUInt32ToLittleEndian( kPGPdiskUserMagic );
	pUser->user.pubUser.header.userType		= sPGPUInt32ToLittleEndian( kPGPdiskUserWithPubType );
	pUser->user.pubUser.header.userSize		= 0;			/* Set this later */
	pUser->user.pubUser.header.readOnly		= bReadOnly;
	pUser->user.pubUser.header.locked		= bLocked;
	pUser->user.pubUser.header.adminUser	= bAdminUser;
	pUser->user.pubUser.header.disabled		= bDisabled;
	
	err = PGPGetKeyID( key, &pUser->user.pubUser.keyID ); CKERR;

	err = sMakeEncryptedKeyForPubUser( mgr, pUser, PGPPeekKeyDBObjKeyDB( key ), pMainHdr, pDiskKey ); CKERR;
	
done:
	return err;
}

	static PGPError
sFreeIODiskUser(
	PGPdiskIODiskUser *	user )
{
	PGPValidatePtr( user );
	
	if( IsntNull( user->encryptedKey ) )
		(void) PGPFreeData( user->encryptedKey );
	(void) PGPFreeData( user );
	
	return kPGPError_NoErr;
}

	static PGPError
sAddUserToUserList(
	PGPdiskIODiskRef		disk,
	PGPdiskIODiskUser *		pNewUser )
{
	PGPdiskIODiskUser *		walk			= NULL;
	
	if( IsNull( disk->userList ) )
		disk->userList = pNewUser;
	else
	{
		for( walk = disk->userList; IsntNull( walk ); walk = walk->next )
		{
			if( IsNull( walk->next ) )
			{
				walk->next = pNewUser;
				break;
			}
		}
	}
	disk->numUsers++;
	
	return kPGPError_NoErr;
}

	static PGPError
sRemoveUserFromUserList(
	PGPdiskIODiskRef		disk,
	PGPdiskIODiskUser *		user )
{
	PGPdiskIODiskUser *		walk			= NULL;
	
	if( user == disk->userList )
	{
		disk->userList = disk->userList->next;
		disk->numUsers--;
	}
	else
	{
		for( walk = disk->userList; IsntNull( walk ); walk = walk->next )
		{
			if( walk->next == user )
			{
				walk->next = walk->next->next;
				disk->numUsers--;
				break;
			}
		}
	}
	
	return kPGPError_NoErr;
}

	static PGPdiskIODiskUser *
sGetAdminUser(
	PGPdiskIODiskRef		disk )
{
	PGPdiskIODiskUser *		user	= NULL;

	for( user = disk->userList; IsntNull( user ); user = user->next )
	{
		if( user->bSymUser )
		{
			if( user->user.symUser.header.adminUser )
				return user;
		}
		else
		{
			if( user->user.pubUser.header.adminUser )
				return user;
		}
	}
	return NULL;
}

	static PGPError
sSetAdminUser(
	PGPdiskIODiskRef		disk,
	PGPdiskIODiskUser *		newAdminUser )
{
	PGPdiskIODiskUser *		user		= NULL;

	for( user = disk->userList; IsntNull( user ); user = user->next )
	{
		if( user->bSymUser )
			user->user.symUser.header.adminUser = ( user == newAdminUser ? TRUE : FALSE );
		else
			user->user.pubUser.header.adminUser = ( user == newAdminUser ? TRUE : FALSE );
	}

	return kPGPError_NoErr;
}

	static PGPError
sAddSymmetricKeyUserToDisk(
	PGPContextRef				context,
	PGPdiskIODiskRef			disk,
	PGPdiskSymmetricKey *		pDiskKey,
	const char *				pszUsername,
	const char *				pszPassphrase,
	PGPBoolean					bReadOnly,
	PGPBoolean					bAdminUser,
	PGPBoolean					bDisabled,
	PGPBoolean					bLocked,
	PGPdiskIODiskUser **		ppNewUser )
{
	PGPError					err				= kPGPError_NoErr;
	PGPdiskIODiskUser *			ioDiskUser		= NULL;
	
	PGPValidatePtr( ppNewUser );
	
	ioDiskUser = PGPNewData( disk->mgr,
		sizeof( PGPdiskIODiskUser ),
		kPGPMemoryMgrFlags_Clear );
	CKNULL( ioDiskUser );
	pgpClearMemory( ioDiskUser, sizeof( PGPdiskIODiskUser ) );

	err = sCreateSymmetricKeyUser( context,
		disk->mgr,
		&disk->mainHdr,
		ioDiskUser,
		pDiskKey,
		pszUsername,
		pszPassphrase,
		bReadOnly,
		bAdminUser,
		bDisabled,
		bLocked ); CKERR;

	ioDiskUser->disk = disk;

	/* Append ioDiskUser to disk->userList */
	err = sAddUserToUserList( disk, ioDiskUser ); CKERR;
	
	*ppNewUser = ioDiskUser;
	
done:

	return err;
}

	static PGPError
sAddPublicKeyUserToDisk(
	PGPdiskIODiskRef		disk,
	PGPdiskSymmetricKey *	pDiskKey,
	PGPKeyDBObjRef			key,
	PGPBoolean				bReadOnly,
	PGPBoolean				bAdminUser,
	PGPBoolean				bDisabled,
	PGPBoolean				bLocked,
	PGPdiskIODiskUser **	ppNewUser )
{
	PGPError					err				= kPGPError_NoErr;
	PGPdiskIODiskUser *			ioDiskUser		= NULL;
	
	PGPValidatePtr( ppNewUser );
	
	ioDiskUser = PGPNewData( disk->mgr,
		sizeof( PGPdiskIODiskUser ),
		kPGPMemoryMgrFlags_Clear );
	CKNULL( ioDiskUser );
	pgpClearMemory( ioDiskUser, sizeof( PGPdiskIODiskUser ) );

	err = sCreatePublicKeyUser( disk->mgr,
		&disk->mainHdr,
		ioDiskUser,
		pDiskKey,
		key,
		bReadOnly,
		bAdminUser,
		bDisabled,
		bLocked ); CKERR;

	ioDiskUser->disk = disk;

	/* Append ioDiskUser to disk->userList */
	err = sAddUserToUserList( disk, ioDiskUser ); CKERR;
	
	*ppNewUser = ioDiskUser;
	
done:

	return err;
}

	static PGPError
sWriteEmptyData(
	PGPContextRef				context,
	PGPInt32					fd,
	PGPdiskSymmetricKey *		pDiskKey,
	PGPInt32					sock,
	PGPdiskOnDiskMainHeader *	pMainHdr )
{
	PGPError			err				= kPGPError_NoErr;
	PGPUInt64			blocksWritten	= 0;
	PGPByte				zeroBlock[kPGPdiskBlockSize];
	PGPByte				block[kPGPdiskBlockSize];
	PGPUInt64			numDataBlocks;
	PGPUInt32			percentDone		= 0;
	time_t				lastMessage		= 0;
	PGPError			clienterr		= kPGPError_NoErr;
	CASTCFBContext *	cast			= NULL;
	PGPByte *			aes				= NULL;
	void *				cipher			= NULL;
	
	PGPValidatePtr( pMainHdr );
	
	numDataBlocks = sLittleEndianToPGPUInt64( pMainHdr->numDataBlocks );
	
	err = sSeek( fd,
		sLittleEndianToPGPUInt64( pMainHdr->numHeaderBlocks ) * kPGPdiskBlockSize,
		SEEK_SET ); CKERR;
		
	pgpClearMemory( block, sizeof( block ) );

	cast = PGPNewSecureData( PGPPeekContextMemoryMgr( context ),
		sizeof( CASTCFBContext ),
		kPGPMemoryMgrFlags_Clear );
	CKNULL( cast );
	aes = PGPNewSecureData( PGPPeekContextMemoryMgr( context ),
		kMaxAESContextSize,
		kPGPMemoryMgrFlags_Clear );
	CKNULL( aes );
	
	switch( sLittleEndianToPGPUInt32( pMainHdr->algorithm ) )
	{
		case kPGPdiskCAST5Algorithm:
			CASTCFB_Init( cast, pDiskKey->bytes );
			cipher = cast;
			break;
		case kPGPdiskAESAlgorithm:
			aes256Key( aes, pDiskKey->bytes );
			cipher = aes;
			break;
	}

	memset( zeroBlock, 0, sizeof( zeroBlock ) );
	for( blocksWritten = 0; blocksWritten < numDataBlocks; blocksWritten++ )
	{
		err = sEncryptBlock( sLittleEndianToPGPUInt32( pMainHdr->algorithm ),
			cipher,
			&pMainHdr->salt,
			blocksWritten,
			zeroBlock,
			block,
			TRUE ); CKERR;

		err = sWrite( fd, block, sizeof( block ) );
		CKERR;

		/* Send an update to the app every 1% or every second to allow cancellation */
		if( ( ( blocksWritten * 100 ) / numDataBlocks > percentDone ) || ( time( NULL ) > lastMessage ) )
		{
			lastMessage = time( NULL );
			percentDone = ( blocksWritten * 100 ) / numDataBlocks;
			if( percentDone != 100 )
			{
				/* Callback */
				err = PGPrpcPackAndWrite( context, sock, kPGPdiskEngineMsg_CreateDiskCallback,
					"u", percentDone ); CKERR;
				err = PGPrpcReadAndUnpack( context, sock, kPGPdiskEngineMsg_CreateDiskCallbackFinished,
					"u", &clienterr ); CKERR;
				if( clienterr == kPGPError_UserAbort )
					RETERR( kPGPError_UserAbort );
			}
		}
	}

	err = PGPrpcPackAndWrite( context, sock, kPGPdiskEngineMsg_CreateDiskCallback,
		"u", 100 ); CKERR;
	err = PGPrpcReadAndUnpack( context, sock, kPGPdiskEngineMsg_CreateDiskCallbackFinished,
		"u", &clienterr ); CKERR;
	
done:
	if( IsntNull( cast ) )
	{
		pgpClearMemory( cast, sizeof( CASTCFBContext ) );
		(void) PGPFreeData( cast );
	}
	if( IsntNull( aes ) )
	{
		pgpClearMemory( aes, kMaxAESContextSize );
		(void) PGPFreeData( aes );
	}
	
	return err;
}

	static PGPError
sReadDiskUsers(
	PGPdiskIODiskRef		disk )
{
	PGPError				err					= kPGPError_NoErr;
	PGPUInt64				nextHeaderOffset	= 0;
	PGPdiskIODiskUser *		user				= NULL;
	PGPdiskIODiskUser *		userJustAdded		= NULL;
	
	PGPdiskValidateIODiskRef( disk );
	
	disk->numUsers = 0;

	nextHeaderOffset = sLittleEndianToPGPUInt64( disk->mainHdr.hdrInfo.nextHeaderOffset );
	
	/* Check if being reencrypted */
	if( disk->mainHdr.isBeingReEncrypted )
	{
		/* Read disk->reencryptingMainHdr and disk->reencryptingAdminUser */
		disk->reencryptingMainHdr = PGPNewData( disk->mgr,
			sizeof( PGPdiskOnDiskMainHeader ),
			kPGPMemoryMgrFlags_Clear );
		CKNULL( disk->reencryptingMainHdr );
		pgpClearMemory( disk->reencryptingMainHdr, sizeof( PGPdiskOnDiskMainHeader ) );

		err = sReadMainHeaderAtOffset( disk->fd, disk->reencryptingMainHdr, nextHeaderOffset ); CKERR;
		
		nextHeaderOffset = sLittleEndianToPGPUInt64( disk->reencryptingMainHdr->hdrInfo.nextHeaderOffset );

		disk->reencryptingAdminUser = PGPNewData( disk->mgr,
			sizeof( PGPdiskIODiskUser ),
			kPGPMemoryMgrFlags_Clear );
		CKNULL( disk->reencryptingAdminUser );
		pgpClearMemory( disk->reencryptingAdminUser, sizeof( PGPdiskIODiskUser ) );
		
		err = sReadUserHeader( disk->mgr, disk->fd, nextHeaderOffset, disk->reencryptingAdminUser ); CKERR;
		disk->reencryptingAdminUser->disk = disk;

		nextHeaderOffset = sLittleEndianToPGPUInt64( disk->reencryptingAdminUser->userHdr.hdrInfo.nextHeaderOffset );
	}

	while( nextHeaderOffset )
	{
		user = PGPNewData( disk->mgr,
			sizeof( PGPdiskIODiskUser ),
			kPGPMemoryMgrFlags_Clear );
		CKNULL( user );
		pgpClearMemory( user, sizeof( PGPdiskIODiskUser ) );
		
		err = sReadUserHeader( disk->mgr, disk->fd, nextHeaderOffset, user ); CKERR;

		user->disk = disk;
		
		if( IsNull( disk->userList ) )
		{
			disk->userList = user;
			userJustAdded = user;
		}
		else
		{
			userJustAdded->next = user;
			userJustAdded = user;
		}
		disk->numUsers++;

		nextHeaderOffset = sLittleEndianToPGPUInt64( user->userHdr.hdrInfo.nextHeaderOffset );
	}
	
done:
	return err;
}

	static PGPError
sNewIODisk(
	PGPMemoryMgrRef		mgr,
	const char *		pszPath,
	PGPdiskIODiskRef *	outDisk )
{
	PGPError				err			= kPGPError_NoErr;
	PGPdiskIODiskRef		disk		= kInvalidPGPdiskIODiskRef;
	PGPdiskIODiskListItem *	newListItem	= NULL;
	PGPdiskIODiskListItem *	walk		= NULL;
	PGPInt32				retval		= 0;
	struct stat				sb;
	PGPInt32				fd			= 0;
	
	PGPValidatePtr( pszPath );
	PGPValidatePtr( outDisk );
	
	/*
	 * First, run through our list of open disks to see if we've already opened
	 * this disk before.  If we have, return the same PGPdiskIODiskRef we use
	 * in the other thread (it'll be mutexed, so this is supposed to be safe),
	 * and increment the refcount so we don't accidentally free it before everybody's
	 * finished with it.
	 */
	
	retval = stat( pszPath, &sb );
	if( ( retval < 0 ) && ( errno == ENOENT ) )
	{
		/* File doesn't exist.  Create it */
		fd = open( pszPath, O_RDWR | O_CREAT, 0644 );
		if( fd < 0 )
			RETERR( kPGPError_CantOpenFile );
		(void) close( fd );
		
		retval = stat( pszPath, &sb );
	}
	if( retval < 0 )
		RETERR( kPGPError_CantOpenFile );
	
	for( walk = gDiskList; IsntNull( walk ); walk = walk->next )
	{
		if( ( walk->device == sb.st_dev ) && ( walk->inode == sb.st_ino ) )
		{
			walk->disk->refCount++;
			*outDisk = walk->disk;
			RETERR( kPGPError_NoErr );
		}
	}

	disk = PGPNewData( mgr,
		sizeof( PGPdiskIODisk ),
		kPGPMemoryMgrFlags_Clear );
	CKNULL( disk );
	pgpClearMemory( disk, sizeof( PGPdiskIODisk ) );
	
	disk->mgr = mgr;
	disk->refCount = 1;
	strcpy( disk->filename, pszPath );
	pthread_mutex_init( &disk->mutex, 0 );
	
	newListItem = PGPNewData( mgr,
		sizeof( PGPdiskIODiskListItem ),
		kPGPMemoryMgrFlags_Clear );
	CKNULL( newListItem );
	pgpClearMemory( newListItem, sizeof( PGPdiskIODiskListItem ) );
	
	newListItem->disk = disk;

	newListItem->device	= sb.st_dev;
	newListItem->inode	= sb.st_ino;
	
	/* Add new disk to the global disk list */
	pthread_mutex_lock( &gDiskListMutex );
	
	if( gDiskList == NULL )
		gDiskList = newListItem;
	else
	{
		for( walk = gDiskList; IsntNull( walk ); walk = walk->next )
		{
			if( IsNull( walk->next ) )
			{
				walk->next = newListItem;
				break;
			}
		}
	}
	
	pthread_mutex_unlock( &gDiskListMutex );

	*outDisk = disk;

done:
	return err;
}

	static PGPError
sFreeIODisk(
	PGPdiskIODiskRef	disk )
{
	PGPdiskIODiskListItem *		diskWalk	= NULL;
	PGPdiskIODiskListItem *		diskToFree	= NULL;
	PGPdiskIODiskUser *			userWalk	= NULL;
	PGPdiskIODiskUser *			userNext	= NULL;
	PGPUInt32					i			= 0;
	
	PGPdiskValidateIODiskRef( disk );
	
	pthread_mutex_lock( &disk->mutex );
	
	disk->refCount--;
	
	if( disk->refCount == 0 )
	{
		/* Stop the bit flipper */
		if( IsntNull( disk->pbTimerStillRunning ) )
			*disk->pbTimerStillRunning = FALSE;
		
		pthread_mutex_unlock( &disk->mutex );

		(void) close( disk->fd );
		if( IsntNull( disk->castCFB[0] ) )
		{
			for( i = 0; i < 2; i++ )
			{
				CASTCFB_Destroy( disk->castCFB[i] );
				(void) PGPFreeData( disk->castCFB[i] );
			}
		}
		if( IsntNull( disk->aesCFB[0] ) )
		{
			for( i = 0; i < 2; i++ )
			{
				pgpClearMemory( disk->aesCFB[i], kMaxAESContextSize );
				(void) PGPFreeData( disk->aesCFB[i] );
			}
		}
		
		if( IsntNull( disk->reencryptingMainHdr ) )
			(void) PGPFreeData( disk->reencryptingMainHdr );
		if( IsntNull( disk->reencryptingAdminUser ) )
			(void) sFreeIODiskUser( disk->reencryptingAdminUser );

		/* Empty user list */
		for( userWalk = disk->userList; IsntNull( userWalk ); userWalk = userNext )
		{
			userNext = userWalk->next;
			(void) sFreeIODiskUser( userWalk );
		}
		
		/* Remove disk from global disk list */
		pthread_mutex_lock( &gDiskListMutex );
		if( gDiskList->disk == disk )
		{
			diskToFree = gDiskList;
			gDiskList = gDiskList->next;
		}
		else
		{
			for( diskWalk = gDiskList; IsntNull( diskWalk ) && IsntNull( diskWalk->next ); diskWalk = diskWalk->next )
			{
				if( diskWalk->next->disk == disk )
				{
					diskToFree = diskWalk->next;
					diskWalk->next = diskWalk->next->next;
					break;
				}
			}
		}
		pthread_mutex_unlock( &gDiskListMutex );

		(void) PGPFreeData( diskToFree );

		pthread_mutex_destroy( &disk->mutex );
		(void) PGPFreeData( disk );
	}
	else
		pthread_mutex_unlock( &disk->mutex );

	return kPGPError_NoErr;
}


	static PGPError
sDecryptDiskKeyWithSym(
	PGPContextRef				context,
	PGPMemoryMgrRef				mgr,
	PGPdiskOnDiskMainHeader *	pMainHdr,
	PGPdiskIODiskUser *			user,
	const char *				passphrase,
	PGPdiskSymmetricKey *		pDiskKey )
{
	PGPError					err			= kPGPError_NoErr;
	PGPdiskSymmetricKey *		pHashedKey	= NULL;
	PGPdiskCheckBytes			checkBytes;
	PGPUInt32					keySize		= 0;
	PGPUInt32					blockSize	= 0;
	
	PGPValidatePtr( pMainHdr );
	PGPValidatePtr( user );
	PGPValidatePtr( passphrase );
	
	pHashedKey = PGPNewSecureData( mgr,
		sizeof( PGPdiskSymmetricKey ),
		kPGPMemoryMgrFlags_Clear );
	CKNULL( pHashedKey );

	/* Hash passphrase */
	err = sHashPassphraseWithSalt( context,
		sLittleEndianToPGPUInt32( pMainHdr->algorithm ),
		&pMainHdr->salt,
		passphrase,
		sLittleEndianToPGPUInt16( user->user.symUser.keyInfo.hashReps ),
		pHashedKey ); CKERR;
	
	err = sGetCipherSizes( sLittleEndianToPGPUInt32( pMainHdr->algorithm ), &blockSize, &keySize, NULL ); CKERR;

	/*
	 * Passphrase is correct iff:
	 *   E(H(passphrase),encryptedKey) = checkBytes
	 */
	err = sEncryptData( user->disk->mgr,
		sLittleEndianToPGPUInt32( pMainHdr->algorithm ),
		pHashedKey,
		pHashedKey->bytes,
		checkBytes.bytes,
		blockSize ); CKERR;

	if( !pgpMemoryEqual(
		&checkBytes.bytes, 
		&user->user.symUser.keyInfo.checkBytes.bytes, blockSize ) )
		RETERR( kPGPError_BadPassphrase );
		
	/*
	 * Passphrase is correct.  We can get the real decryption key by:
	 *   k = D(H(passphrase),encryptedKey)
	 */

	/* Passphrase is correct */
	err = sDecryptData( user->disk->mgr,
		sLittleEndianToPGPUInt32( pMainHdr->algorithm ),
		pHashedKey,
		user->user.symUser.keyInfo.encryptedKey.bytes,
		pDiskKey->bytes,
		keySize ); CKERR;
		
done:
	if( IsntNull( pHashedKey ) )
	{
		pgpClearMemory( pHashedKey, sizeof( PGPdiskSymmetricKey ) );
		(void) PGPFreeData( pHashedKey );
	}
	return err;
}

	static PGPError
sDecryptDiskKeyWithPub(
	PGPContextRef				context,
	PGPMemoryMgrRef				mgr,
	PGPdiskOnDiskMainHeader *	pMainHdr,
	PGPKeyDBRef					keydb,
	PGPdiskIODiskUser *			user,
	const char *				passphrase,
	PGPBoolean					cachePassphrase,
	PGPUInt32					cacheTimeout,
	PGPBoolean					shareCache,
	PGPdiskSymmetricKey *		pDiskKey )
{
	PGPError					err					= kPGPError_NoErr;
	PGPdiskCheckBytes			checkBytes;
	PGPPrivateKeyContextRef		privContext			= kInvalidPGPPrivateKeyContextRef;
	PGPKeyDBObjRef				key					= kInvalidPGPKeyDBObjRef;
	PGPKeyDBObjRef				encKey				= kInvalidPGPKeyDBObjRef;
	PGPSize						maxDecryptedSize	= 0;
	PGPSize						maxEncryptedSize	= 0;
	PGPSize						maxSignatureSize	= 0;
	PGPByte *					decryptedKey		= NULL;
	PGPSize						decryptedKeySize	= 0;
	PGPUInt32					blockSize			= 0;
	PGPUInt32					keySize				= 0;

	PGPValidatePtr( pMainHdr );
	PGPValidatePtr( user );
	PGPValidatePtr( passphrase );
	
	err = PGPFindKeyByKeyID( keydb, &user->user.pubUser.keyID, &key ); CKERR;
	err = PGPFindKeyByKeyID( keydb, &user->user.pubUser.subKeyID, &encKey ); CKERR;
	
	/* Be sure to use the correct encryption key with V4 keys */
	err = PGPNewPrivateKeyContext( encKey,
		kPGPPublicKeyMessageFormat_PGP,
		&privContext,
		PGPOPassphrase( PGPPeekKeyDBObjContext( key ), passphrase ),
		( cachePassphrase ? PGPOCachePassphrase( context, cacheTimeout, shareCache ) : PGPONullOption( context ) ),
		PGPOLastOption( PGPPeekKeyDBObjContext( key ) ) ); CKERR;
		
	err = PGPGetPrivateKeyOperationSizes( privContext,
		&maxDecryptedSize,
		&maxEncryptedSize,
		&maxSignatureSize ); CKERR;
	
	decryptedKey = PGPNewSecureData( mgr,
		maxDecryptedSize,
		kPGPMemoryMgrFlags_Clear );
	CKNULL( decryptedKey );
		
	err = PGPPrivateKeyDecrypt( privContext,
		user->encryptedKey,
		user->encryptedKeySize,
		decryptedKey,
		&decryptedKeySize ); CKERR;

	err = sGetCipherSizes( sLittleEndianToPGPUInt32( pMainHdr->algorithm ), &blockSize, &keySize, NULL ); CKERR;

	pgpCopyMemory( decryptedKey, pDiskKey->bytes, keySize );

	/* Verify checkBytes */
	err = sGeneratePubKeyCheckBytes( mgr,
		sLittleEndianToPGPUInt32( pMainHdr->algorithm ),
		encKey,
		pDiskKey,
		&checkBytes ); CKERR;
	
	if( !pgpMemoryEqual( checkBytes.bytes, user->user.pubUser.checkBytes.bytes, blockSize ) )
		RETERR( kPGPError_BadPassphrase );

done:
	if( IsntNull( decryptedKey ) )
		(void) PGPFreeData( decryptedKey );
	if( PGPPrivateKeyContextRefIsValid( privContext ) )
		(void) PGPFreePrivateKeyContext( privContext );
	return err;
}

	static PGPError
sDecryptDiskKey(
	PGPdiskBEContextRef			diskcon,
	PGPdiskOnDiskMainHeader *	pMainHdr,
	PGPdiskIODiskUser *			user,
	const char *				passphrase,
	PGPdiskSymmetricKey **		ppDiskKey )
{
	PGPError				err			= kPGPError_NoErr;
	PGPdiskSymmetricKey	*	pDiskKey	= NULL;
	
	PGPValidatePtr( pMainHdr );
	PGPdiskValidateBEContextRef( diskcon );
	PGPValidatePtr( passphrase );
	PGPValidatePtr( ppDiskKey );

	pDiskKey = PGPNewSecureData( diskcon->mgr,
		sizeof( PGPdiskSymmetricKey ),
		kPGPMemoryMgrFlags_Clear );
	CKNULL( pDiskKey );

	if( user->bSymUser )
		err = sDecryptDiskKeyWithSym( diskcon->context, diskcon->mgr, pMainHdr, user, passphrase, pDiskKey );
	else
	{
		err = sDecryptDiskKeyWithPub( diskcon->context,
			diskcon->mgr,
			pMainHdr,
			diskcon->keydb,
			user,
			passphrase,
			diskcon->cachePassphrases,
			diskcon->cacheTimeout,
			diskcon->shareCache,
			pDiskKey );
	}
	CKERR;
	
	*ppDiskKey = pDiskKey;

done:
	if( IsPGPError( err ) )
	{
		pgpClearMemory( pDiskKey, sizeof( PGPdiskSymmetricKey ) );
		(void) PGPFreeData( pDiskKey );
	}
	return err;
}

	static PGPError
sGenerateNewRandomKey(
	PGPdiskBEContextRef		diskcon,
	PGPdiskEncryptionAlgorithm	alg,
	PGPdiskSymmetricKey **	ppDiskKey )
{
	PGPError				err			= kPGPError_NoErr;
	PGPUInt32				keySize		= 0;
	PGPdiskSymmetricKey *	pTempKey	= NULL;
	PGPdiskSymmetricKey *	pDiskKey	= NULL;
	PGPdiskSymmetricKey *	pMungeKey	= NULL;
	
	err = sGetCipherSizes( alg, NULL, &keySize, NULL ); CKERR;
	
	pTempKey = PGPNewSecureData( diskcon->mgr,
		sizeof( PGPdiskSymmetricKey ),
		kPGPMemoryMgrFlags_Clear );
	CKNULL( pTempKey );

	pDiskKey = PGPNewSecureData( diskcon->mgr,
		sizeof( PGPdiskSymmetricKey ),
		kPGPMemoryMgrFlags_Clear );
	CKNULL( pDiskKey );

	pMungeKey = PGPNewSecureData( diskcon->mgr,
		sizeof( PGPdiskSymmetricKey ),
		kPGPMemoryMgrFlags_Clear );
	CKNULL( pMungeKey );

	err = PGPContextGetRandomBytes( diskcon->context, pTempKey->bytes, keySize ); CKERR;
	err = PGPContextGetRandomBytes( diskcon->context, pMungeKey->bytes, keySize ); CKERR;
	
	/* Encrypt the temp key twice with the munge key to make the real disk key */
	err = sEncryptData( diskcon->mgr, alg, pMungeKey, pTempKey->bytes, pDiskKey->bytes, keySize ); CKERR;
	pgpCopyMemory( pDiskKey->bytes, pTempKey->bytes, keySize );
	err = sEncryptData( diskcon->mgr, alg, pMungeKey, pTempKey->bytes, pDiskKey->bytes, keySize ); CKERR;
	
	*ppDiskKey = pDiskKey;
	
done:
	if( IsntNull( pTempKey ) )
	{
		pgpClearMemory( pTempKey, sizeof( PGPdiskSymmetricKey ) );
		(void) PGPFreeData( pTempKey );
	}
	if( IsntNull( pMungeKey ) )
	{
		pgpClearMemory( pMungeKey, sizeof( PGPdiskSymmetricKey ) );
		(void) PGPFreeData( pMungeKey );
	}
	return err;
}

	PGPUInt32
sComputeOS9CRC(
	const void *theBytes,
	PGPUInt32 numBytes )
{   
	register PGPUInt32  crc			= 0;
	register uchar  *	cur			= (uchar *) theBytes;
	register uchar  *	last		= cur + numBytes - 1;
	register PGPInt32 	k4C11DB7	= 0x04C11DB7;

	while( cur <= last )
	{
		register PGPInt32 check;
		
		crc ^= ((PGPInt32 ) *cur) << 8;
		++cur;

		check = crc; crc <<= 1; if( check < 0 ) crc ^= k4C11DB7;
		check = crc; crc <<= 1; if( check < 0 ) crc ^= k4C11DB7;
		check = crc; crc <<= 1; if( check < 0 ) crc ^= k4C11DB7;
		check = crc; crc <<= 1; if( check < 0 ) crc ^= k4C11DB7;
		check = crc; crc <<= 1; if( check < 0 ) crc ^= k4C11DB7;
		check = crc; crc <<= 1; if( check < 0 ) crc ^= k4C11DB7;
		check = crc; crc <<= 1; if( check < 0 ) crc ^= k4C11DB7;
		check = crc; crc <<= 1; if( check < 0 ) crc ^= k4C11DB7;
	}
	
	return crc;
}

	static PGPError
sDecryptOS9DiskKey(
	PGPContextRef				context,
	PGPMemoryMgrRef				mgr,			/* global memory mgr */
	PGPdiskOS9FileHeader *		pFileHdr,
	const char *				passphrase,
	PGPdiskSymmetricKey **		ppDiskKey )
{
	PGPError					err				= kPGPError_NoErr;
	PGPdiskSymmetricKey *		pHashedKey		= NULL;
	PGPdiskSymmetricKey *		pDiskKey		= NULL;
	PGPdiskOS9CheckBytes		checkBytes;
	PGPUInt32					keySize			= 0;
	PGPUInt32					blockSize		= 0;
	PGPdiskOS9PassphraseKey *	passphraseKey	= NULL;
	
	PGPValidatePtr( pFileHdr );
	PGPValidatePtr( passphrase );
	PGPValidatePtr( ppDiskKey );
	
	passphraseKey = &pFileHdr->passphrases[kPGPdiskOS9MasterPassphraseIndex];
	
	pHashedKey = PGPNewSecureData( mgr,
		sizeof( PGPdiskSymmetricKey ),
		kPGPMemoryMgrFlags_Clear );
	CKNULL( pHashedKey );

	pDiskKey = PGPNewSecureData( mgr,
		sizeof( PGPdiskSymmetricKey ),
		kPGPMemoryMgrFlags_Clear );
	CKNULL( pDiskKey );

	/* Hash passphrase */
	err = sHashPassphraseWithSalt( context,
		kPGPdiskCAST5Algorithm,
		(PGPdiskPassphraseSalt *) &pFileHdr->salt,
		passphrase,
		passphraseKey->hashReps,
		pHashedKey ); CKERR;
	
	err = sGetCipherSizes( kPGPdiskCAST5Algorithm, &blockSize, &keySize, NULL ); CKERR;
	
	/* Decrypt disk key */
	err = sDecryptData( mgr,
		kPGPdiskCAST5Algorithm,
		pHashedKey,
		passphraseKey->encryptedKey.bytes,
		pDiskKey->bytes,
		keySize ); CKERR;
		
	/*
	 * Passphrase is correct iff:
	 *  D(H(passphrase),checkBytes) = H(passphrase)
	 */
	err = sDecryptData( mgr,
		kPGPdiskCAST5Algorithm,
		pHashedKey,
		passphraseKey->checkBytes.bytes,
		checkBytes.bytes,
		sizeof( checkBytes ) ); CKERR;
		
	if( !pgpMemoryEqual( checkBytes.bytes, pHashedKey->bytes, sizeof( checkBytes ) ) )
		RETERR( kPGPError_BadPassphrase );
		
	/* Passphrase is correct and we have the correct disk key in pDiskKey */
	*ppDiskKey = pDiskKey;
		
done:
	if( IsntNull( pHashedKey ) )
	{
		pgpClearMemory( pHashedKey, sizeof( PGPdiskSymmetricKey ) );
		(void) PGPFreeData( pHashedKey );
	}
	if( IsPGPError( err ) )
	{
		if( IsntNull( pDiskKey ) )
		{
			pgpClearMemory( pDiskKey, sizeof( PGPdiskSymmetricKey ) );
			(void) PGPFreeData( pDiskKey );
		}
	}
	return err;
}

	static PGPError
sConvertOS9DiskHeader(
	PGPContextRef			context,
	PGPInt32				os9FD,
	PGPdiskIODiskRef		disk )
{
	PGPError				err				= kPGPError_NoErr;
	PGPdiskOS9FileHeader	fileHdr;
	PGPUInt32				crc				= 0;

	err = sSeek( os9FD, 0, SEEK_SET ); CKERR;
	err = sRead( os9FD, &fileHdr, sizeof( fileHdr ) ); CKERR;
	
	/* Check disk magic and type */
	if( fileHdr.headerInfo.headerMagic != kPGPdiskOS9HeaderMagic )
		RETERR(  kPGPError_CorruptData );
	if( fileHdr.headerInfo.headerType != kPGPdiskOS9PrimaryHeaderType )
		RETERR( kPGPError_CorruptData );
		
	/* Make sure it's a version we understand - ignore the minor version */
	if( ( fileHdr.majorVersion != kPGPdiskOS9PrimaryHeaderMajorVersion ) )
		RETERR( kPGPError_CorruptData );
	
	/* Check the CRC */
	crc = fileHdr.headerInfo.headerCRC;
	fileHdr.headerInfo.headerCRC = 0;
	
	/* Aaarrggh!  Why can't we all just use the same CRC function? */
	if( crc != sComputeOS9CRC( (PGPUInt32 *) &fileHdr.headerInfo, fileHdr.headerInfo.headerSize ) )
		RETERR( kPGPError_CorruptData );
	
	/* Create main header */
	err = sCreateMainHeader( context,
		&disk->mainHdr,
		fileHdr.numDataBlocks * kPGPdiskBlockSize,
		"",		/* Default mount point */
		kPGPdiskCAST5Algorithm ); CKERR;
	pgpCopyMemory( fileHdr.salt.bytes, disk->mainHdr.salt.bytes, sizeof( fileHdr.salt ) );

	disk->dataOffset = sLittleEndianToPGPUInt64( disk->mainHdr.numHeaderBlocks ) * kPGPdiskBlockSize;
	disk->dataSize = sLittleEndianToPGPUInt64( disk->mainHdr.numDataBlocks ) * kPGPdiskBlockSize;
	
	disk->bCustomTimeout = FALSE;
	disk->timeout = 0;
	disk->timeLastUsed = time( NULL );
	
	/* Save the original OS9 file header in case we need it later */
	pgpCopyMemory( &fileHdr, &disk->os9FileHeader, sizeof( PGPdiskOS9FileHeader ) );
	
done:
	return err;
}

	static PGPError
sCheckDiskSanity(
	PGPdiskIODiskRef	disk )
{
	PGPError			err					= kPGPError_NoErr;
	PGPdiskIODiskUser *	walk				= NULL;
	PGPBoolean			bFoundValidUser		= FALSE;
	PGPBoolean			bFoundAdmin			= FALSE;

	/*
	 * We need to do some more sanity checks:
	 * 1. There must be at least one valid user
	 * 2. There must be one and only one admin user
	 * 3. Admin user is not read only
	 */

	/*
	 * We won't be very smart about this right now.  If we find an error, shout
	 * CorruptData and refuse to use the disk
	 */

	for( walk = disk->userList; IsntNull( walk ); walk = walk->next )
	{
		bFoundValidUser = TRUE;
		if( walk->bSymUser ? walk->user.symUser.header.adminUser : walk->user.pubUser.header.adminUser )
		{
			if( bFoundAdmin )
				RETERR( kPGPError_CorruptData );	/* Already have an admin user */
			bFoundAdmin = TRUE;
			
			if( walk->bSymUser ? walk->user.symUser.header.readOnly : walk->user.pubUser.header.readOnly )
				RETERR( kPGPError_CorruptData );	/* Admin is read only */
		}
	}
	
	if( !bFoundValidUser )
		RETERR( kPGPError_CorruptData );			/* No valid users */
	
done:
	
	return err;
}

	PGPError
PGPdiskEngineCreateDisk(
	PGPdiskBEContextRef	diskcon,
	const char *		pszPath,
	PGPUInt64			ullSize,
	const char *		pszDefaultMountPoint,
	PGPdiskEncryptionAlgorithm	alg,
	PGPBoolean			bPubKeyUser,
	PGPKeyDBObjRef		key,
	const char *		pszUsername,
	const char *		pszPassphrase,
	PGPInt32			callbackSock,
	PGPdiskIODiskRef *	outDisk )
{
	PGPError					err			= kPGPError_NoErr;
	PGPdiskIODiskRef			disk		= kInvalidPGPdiskIODiskRef;
	PGPdiskIODiskUser *			ioDiskUser	= NULL;
	PGPdiskIODiskUser *			adkDiskUser	= NULL;
	PGPBoolean					bHaveMutex	= FALSE;
	PGPdiskSymmetricKey *		pDiskKey	= NULL;
	PGPUInt32					keySize		= 0;
	struct statfs				fsInfo;
	PGPInt32					retval		= 0;
	
	PGPValidatePtr( outDisk );

	/* Make ullSize a multiple of 512 bytes */
	ullSize = BYTES_TO_BLOCKS( ullSize ) * kPGPdiskBlockSize;
	
	/* We will need some entropy later */
	err = PGPGlobalRandomPoolAddSystemState(); CKERR;
	
	err = sNewIODisk( diskcon->mgr, pszPath, &disk ); CKERR;

	pthread_mutex_lock( &disk->mutex );
	bHaveMutex = TRUE;
	
	if( disk->refCount != 1 )
	{
		/* Disk exists and is in use.  Can't make another over this file. */
		RETERR( kPGPError_ItemAlreadyExists );
	}
	
	retval = statfs( pszPath, &fsInfo );
	if( retval < 0 )
		RETERR( kPGPError_UnknownError );
		
	/* Check disk free space */
	if( ullSize > ( (PGPUInt64) fsInfo.f_bfree * (PGPUInt64) fsInfo.f_bsize ) )
		RETERR( kPGPError_DiskFull );

	disk->fd = open( pszPath, O_RDWR | O_CREAT | O_TRUNC, 0644 );
	if( disk->fd < 0 )
		RETERR( kPGPError_CantOpenFile );
		
	/* Generate random master disk key */
	err = sGetCipherSizes( alg, NULL, &keySize, NULL ); CKERR;
	err = sGenerateNewRandomKey( diskcon, alg, &pDiskKey ); CKERR;
		
	/* Create Main header */
	err = sCreateMainHeader( diskcon->context, &disk->mainHdr, ullSize, pszDefaultMountPoint, alg ); CKERR;
	
	disk->dataOffset = sLittleEndianToPGPUInt64( disk->mainHdr.numHeaderBlocks ) * kPGPdiskBlockSize;
	disk->dataSize = sLittleEndianToPGPUInt64( disk->mainHdr.numDataBlocks ) * kPGPdiskBlockSize;
	
	disk->kernCon = diskcon->kernCon;
	disk->bCustomTimeout = FALSE;
	disk->timeout = diskcon->globalTimeout * 60;
	disk->forcedUnmount = diskcon->forcedUnmount;
	disk->timeLastUsed = time( NULL );

	/* Create (and encrypt) empty data */
	err = sWriteEmptyData( diskcon->context, disk->fd, pDiskKey, callbackSock, &disk->mainHdr ); CKERR;

	/* Create admin user */
	if( bPubKeyUser )
	{
		err = sAddPublicKeyUserToDisk( disk,
			pDiskKey,
			key,
			FALSE,			/* read only */
			TRUE,			/* admin user */
			FALSE,			/* disabled */
			FALSE,			/* locked */
			&ioDiskUser ); CKERR;
	}
	else
	{
		err = sAddSymmetricKeyUserToDisk( diskcon->context,
			disk,
			pDiskKey,
			pszUsername,
			pszPassphrase,
			FALSE,			/* read only */
			TRUE,			/* admin user */
			FALSE,			/* disabled */
			FALSE,			/* locked */
			&ioDiskUser ); CKERR;
	}
	
	if( PGPKeyDBObjRefIsValid( diskcon->adkKey ) )
	{
		err = sAddPublicKeyUserToDisk( disk,
			pDiskKey,
			diskcon->adkKey,
			FALSE,			/* read only */
			FALSE,			/* admin user */
			FALSE,			/* disabled */
			TRUE,			/* locked */
			&adkDiskUser ); CKERR;
	}
	
	err = sWriteMainHeader( disk ); CKERR;
	err = sWriteDiskUsers( disk ); CKERR;
	
	/*
	 * We've just created the disk, so we don't really need to do any
	 * extra sanity checks, but it might not be a bad idea...
	 */
	err = sCheckDiskSanity( disk ); CKERR;

	*outDisk = disk;
	
done:
	if( bHaveMutex )
		pthread_mutex_unlock( &disk->mutex );
		
	if( IsntNull( pDiskKey ) )
	{
		pgpClearMemory( pDiskKey, sizeof( PGPdiskSymmetricKey ) );
		(void) PGPFreeData( pDiskKey );
	}
		
	if( IsPGPError( err ) )
	{
		if( PGPdiskIODiskRefIsValid( disk ) )
			(void) sFreeIODisk( disk );
	}
	
	if( IsPGPError( err ) )
		syslog( LOG_NOTICE, "Creation of PGPdisk %s failed: %d\n", pszPath, err );
		
	return err;
}

	PGPError
PGPdiskEngineOpenDisk(
	PGPdiskBEContextRef		diskcon,
	const char *			filename,
	PGPBoolean				readonly,
	PGPdiskIODiskRef *		outDisk )
{
	PGPError				err			= kPGPError_NoErr;
	PGPdiskIODiskRef		disk		= kInvalidPGPdiskIODiskRef;
	PGPdiskSymmetricKey		dummyDiskKey;
	PGPBoolean				bHaveMutex	= FALSE;
	PGPdiskIODiskUser *		user		= NULL;
	
	PGPValidatePtr( outDisk );

	/* We might need some entropy later */
	err = PGPGlobalRandomPoolAddSystemState(); CKERR;
	
	err = sNewIODisk( diskcon->mgr, filename, &disk ); CKERR;

	pthread_mutex_lock( &disk->mutex );
	bHaveMutex = TRUE;
	
	if( disk->refCount == 1 )
	{
		/* We just created the PGPdiskIODiskRef.  We need to open the file and read the headers */
		disk->fd = open( filename, readonly ? O_RDONLY : O_RDWR, 0 );
		if( disk->fd < 0 )
			RETERR( kPGPError_CantOpenFile );
			
		err = sReadMainHeader( disk );
		if( err == kPGPdiskError_OS9DiskNeedsConversion )
		{
			/* Convert the disk headers in memory */
			err = sConvertOS9DiskHeader( diskcon->context, disk->fd, disk ); CKERR;

			/*
			 * We don't have the passphrase until we try to mount the disk, so
			 * just substitute "" as a dummy passphrase for now.  We'll verify
			 * the real passphrase when we decrypt the disk key.
			 */
			err = sAddSymmetricKeyUserToDisk( diskcon->context,
				disk,
				&dummyDiskKey,
				"Administrator",
				"",
				FALSE,		/* read only */
				TRUE,		/* admin user */
				FALSE,		/* disabled */
				FALSE,		/* locked */
				&user ); CKERR;

			disk->diskFormat = kPGPdiskDiskFormat_OS9;
			
			disk->dataOffset = disk->os9FileHeader.numHeaderBlocks * kPGPdiskBlockSize;
			disk->dataSize = disk->os9FileHeader.numDataBlocks * kPGPdiskBlockSize;
		}
		else if( IsntPGPError( err ) )
		{
			disk->diskFormat = kPGPdiskDiskFormat_V6;
		
			disk->dataOffset = sLittleEndianToPGPUInt64( disk->mainHdr.numHeaderBlocks ) * kPGPdiskBlockSize;
			disk->dataSize = sLittleEndianToPGPUInt64( disk->mainHdr.numDataBlocks ) * kPGPdiskBlockSize;
	
			err = sReadDiskUsers( disk ); CKERR;
			
			/* Extra sanity checks */
			err = sCheckDiskSanity( disk ); CKERR;
		}
		else
			CKERR;

		disk->kernCon = diskcon->kernCon;
		disk->bCustomTimeout = disk->mainHdr.usesCustomTimeout;
		if( disk->bCustomTimeout )
			disk->timeout = sLittleEndianToPGPUInt32( disk->mainHdr.customTimeout ) * 60;
		else
			disk->timeout = diskcon->globalTimeout * 60;
		disk->timeLastUsed = time( NULL );
	
		disk->forcedUnmount = diskcon->forcedUnmount;
	 }
	
	*outDisk = disk;
	
done:
	if( bHaveMutex )
		pthread_mutex_unlock( &disk->mutex );
		
	if( IsPGPError( err ) )
	{
		if( PGPdiskIODiskRefIsValid( disk ) )
			(void) sFreeIODisk( disk );
	}
	return err;
}

	PGPError
PGPdiskEngineCloseDisk(
	PGPdiskIODiskRef	disk )
{
	PGPdiskValidateIODiskRef( disk );
	
	/* Mutexing the disk is handled inside of sFreeIODisk */
	(void) sFreeIODisk( disk );
	
	return kPGPError_NoErr;
}

	static PGPError
sCopyOS9DiskBlocksToDisk(
	PGPContextRef			context,
	PGPdiskIODiskRef		disk,
	PGPdiskSymmetricKey *	pDiskKey,
	PGPInt32				os9FD,
	PGPUInt64				offset,
	PGPInt32				sock )
{
	PGPError				err				= kPGPError_NoErr;
	PGPError				clienterr		= kPGPError_NoErr;
	CASTCFBContext *		cast1			= NULL;
	CASTCFBContext *		cast2			= NULL;
	PGPByte					block1[kPGPdiskBlockSize];
	PGPByte					block2[kPGPdiskBlockSize];
	PGPUInt64				blocksWritten	= 0;
	PGPUInt64				numDataBlocks	= 0;
	PGPUInt32				percentDone		= 0;

	err = sSeek( disk->fd,
		sLittleEndianToPGPUInt64( disk->mainHdr.numHeaderBlocks ) * kPGPdiskBlockSize,
		SEEK_SET ); CKERR;
	err = sSeek( os9FD, offset, SEEK_SET ); CKERR;

	cast1 = PGPNewSecureData( disk->mgr,
		sizeof( CASTCFBContext ),
		kPGPMemoryMgrFlags_Clear );
	CKNULL( cast1 );
	cast2 = PGPNewSecureData( disk->mgr,
		sizeof( CASTCFBContext ),
		kPGPMemoryMgrFlags_Clear );
	CKNULL( cast2 );

	CASTCFB_Init( cast1, pDiskKey->bytes );
	CASTCFB_Init( cast2, pDiskKey->bytes );

	numDataBlocks = sLittleEndianToPGPUInt64( disk->mainHdr.numDataBlocks );
	for( blocksWritten = 0; blocksWritten < numDataBlocks; blocksWritten++ )
	{
		/* Read into block1 */
		err = sRead( os9FD, block1, sizeof( block1 ) ); CKERR;
		
		/* Decrypt into block2 */
		err = sDecryptBlock( kPGPdiskCAST5Algorithm,
			cast1,
			&disk->mainHdr.salt,
			blocksWritten,
			block1,
			block2,
			FALSE ); CKERR;
			
		/* ReEncrypt back into block1 */
		err = sEncryptBlock( kPGPdiskCAST5Algorithm,
			cast2,
			&disk->mainHdr.salt,
			blocksWritten,
			block2,
			block1,
			TRUE ); CKERR;
		
		/* Write to disk */
		err = sWrite( disk->fd, block1, sizeof( block1 ) ); CKERR;
		
		if( ( blocksWritten * 100 ) / numDataBlocks > percentDone )
		{
			percentDone = ( blocksWritten * 100 ) / numDataBlocks;
			if( percentDone != 100 )
			{
				/* Callback */
				err = PGPrpcPackAndWrite( context, sock, kPGPdiskEngineMsg_ConvertDiskCallback,
					"u", percentDone ); CKERR;
				err = PGPrpcReadAndUnpack( context, sock, kPGPdiskEngineMsg_ConvertDiskCallbackFinished,
					"u", &clienterr ); CKERR;
				if( clienterr == kPGPError_UserAbort )
					RETERR( kPGPError_UserAbort );
			}
		}
	}

	err = PGPrpcPackAndWrite( context, sock, kPGPdiskEngineMsg_ConvertDiskCallback,
		"u", 100 ); CKERR;
	err = PGPrpcReadAndUnpack( context, sock, kPGPdiskEngineMsg_ConvertDiskCallbackFinished,
		"u", &clienterr ); CKERR;

done:
	if( IsntNull( cast1 ) )
	{
		pgpClearMemory( cast1, sizeof( CASTCFBContext ) );
		(void) PGPFreeData( cast1 );
	}
	if( IsntNull( cast2 ) )
	{
		pgpClearMemory( cast2, sizeof( CASTCFBContext ) );
		(void) PGPFreeData( cast2 );
	}
	return err;
}

	PGPError
PGPdiskEngineConvertOS9Disk(
	PGPdiskBEContextRef		diskcon,
	const char *			pszOriginalFilename,
	const char *			pszAdminPassphrase,
	const char *			pszNewFilename,
	PGPInt32				callbackSock )
{
	PGPError				err				= kPGPError_NoErr;
	PGPInt32				origFD			= 0;
	PGPdiskIODiskRef		disk			= kInvalidPGPdiskIODiskRef;
	PGPdiskSymmetricKey *	pDiskKey		= NULL;
	PGPdiskIODiskUser *		user			= NULL;
	PGPBoolean				bHaveMutex		= FALSE;

	/* We will need some entropy later */
	err = PGPGlobalRandomPoolAddSystemState(); CKERR;

	err = sNewIODisk( diskcon->mgr, pszNewFilename, &disk ); CKERR;

	pthread_mutex_lock( &disk->mutex );
	bHaveMutex = TRUE;

	if( disk->refCount != 1 )
	{
		/* Disk exists and is in use.  Can't make another over this file. */
		RETERR( kPGPError_ItemAlreadyExists );
	}

	origFD = open( pszOriginalFilename, O_RDONLY, 0 );
	if( origFD < 0 )
		RETERR( kPGPError_CantOpenFile );

	disk->fd = open( pszNewFilename, O_RDWR | O_CREAT | O_TRUNC, 0644 );
	if( disk->fd < 0 )
		RETERR( kPGPError_CantOpenFile );

	/* Convert the OS9 disk header into a header we can understand */
	err = sConvertOS9DiskHeader( diskcon->context, origFD, disk ); CKERR;

	/* Decrypt disk key */
	err = sDecryptOS9DiskKey( diskcon->context, diskcon->mgr, &disk->os9FileHeader, pszAdminPassphrase, &pDiskKey ); CKERR;

	/* Create lone admin user */
	err = sAddSymmetricKeyUserToDisk( diskcon->context,
		disk,
		pDiskKey,
		"Administrator",
		pszAdminPassphrase,
		FALSE,		/* read only */
		TRUE,		/* admin user */
		FALSE,		/* disabled */
		FALSE,		/* locked */
		&user ); CKERR;

	/* Write main header */
	err = sWriteMainHeader( disk ); CKERR;
	
	/* Write data blocks */
	err = sCopyOS9DiskBlocksToDisk( diskcon->context,
		disk,
		pDiskKey,
		origFD,
		disk->os9FileHeader.numHeaderBlocks * kPGPdiskBlockSize,
		callbackSock ); CKERR;
	
	/* Write admin user */
	err = sWriteDiskUsers( disk ); CKERR;
		
done:
	if( bHaveMutex )
		pthread_mutex_unlock( &disk->mutex );
		
	if( origFD >= 0 )
		(void) close( origFD );
		
	if( IsntNull( pDiskKey ) )
	{
		pgpClearMemory( pDiskKey, sizeof( PGPdiskSymmetricKey ) );
		(void) PGPFreeData( pDiskKey );
	}

	if( PGPdiskIODiskRefIsValid( disk ) )
		(void) sFreeIODisk( disk );

	return err;
}

	static PGPError
sUpdateUsersWithNewKey(
	PGPdiskBEContextRef		diskcon,
	PGPdiskIODiskRef		disk,
	PGPdiskSymmetricKey *	pNewDiskKey )
{
	PGPError				err			= kPGPError_NoErr;
	PGPdiskIODiskUser *		user		= NULL;
	PGPdiskIODiskUser *		next		= NULL;
	PGPKeyDBObjRef			key			= kInvalidPGPKeyDBObjRef;
	
	PGPdiskValidateIODiskRef( disk );
	PGPValidatePtr( pNewDiskKey );
	
	for( user = disk->userList; IsntNull( user ); user = next )
	{
		next = user->next;

		if( user->bSymUser )
		{
			if( !user->user.symUser.header.adminUser )
			{
				/*
				 * Non-admin symmetric key users cannot be converted because we don't have their
				 * passphrase.  Remove them.
				 */
				 err = sRemoveUserFromUserList( disk, user );
				(void) sFreeIODiskUser( user );
			}
		}
		else
		{
			err = PGPFindKeyByKeyID( diskcon->keydb, &user->user.pubUser.keyID, &key );
			if( err == kPGPError_ItemNotFound )
			{
				/*
				 * Public key users for which we don't have the key for cannot be converted.
				 * Remove them.
				 */
				err = sRemoveUserFromUserList( disk, user );
				(void) sFreeIODiskUser( user );
			}
			else
				err = sMakeEncryptedKeyForPubUser( user->disk->mgr, user, diskcon->keydb, disk->reencryptingMainHdr, pNewDiskKey );
		}
		CKERR;
	}
	
done:
	return err;
}

	static PGPError
sReEncryptDataBlocks(
	PGPContextRef			context,
	PGPdiskIODiskRef		disk,
	PGPInt32				sock,
	PGPdiskSymmetricKey *	pOldDiskKey,
	PGPdiskSymmetricKey *	pNewDiskKey )
{
	PGPError				err				= kPGPError_NoErr;
	PGPUInt64				i				= 0;
	PGPUInt64				startBlock		= 0;
	PGPUInt64				endBlock		= 0;
	PGPByte					oldBlock[kPGPdiskBlockSize];
	PGPByte					tempBlock[kPGPdiskBlockSize];
	PGPByte					newBlock[kPGPdiskBlockSize];
	CASTCFBContext *		oldCAST			= NULL;
	CASTCFBContext *		newCAST			= NULL;
	PGPByte *				oldAES			= NULL;
	PGPByte *				newAES			= NULL;
	void *					oldCipher		= NULL;
	void *					newCipher		= NULL;
	PGPUInt32				percentDone		= 0;
	time_t					lastMessage		= 0;
	PGPError				clienterr		= kPGPError_NoErr;

	startBlock = sLittleEndianToPGPUInt64( disk->mainHdr.numHeaderBlocks ) +
		sLittleEndianToPGPUInt64( disk->mainHdr.numBlocksReEncrypted );
	endBlock = sLittleEndianToPGPUInt64( disk->mainHdr.numHeaderBlocks ) +
		sLittleEndianToPGPUInt64( disk->mainHdr.numDataBlocks );
		
	/* Init cipher contexts */
	oldCAST = PGPNewSecureData( disk->mgr,
		sizeof( CASTCFBContext ),
		kPGPMemoryMgrFlags_Clear );
	CKNULL( oldCAST );
	newCAST = PGPNewSecureData( disk->mgr,
		sizeof( CASTCFBContext ),
		kPGPMemoryMgrFlags_Clear );
	CKNULL( newCAST );
	oldAES = PGPNewSecureData( disk->mgr,
		kMaxAESContextSize,
		kPGPMemoryMgrFlags_Clear );
	CKNULL( oldAES );
	newAES = PGPNewSecureData( disk->mgr,
		kMaxAESContextSize,
		kPGPMemoryMgrFlags_Clear );
	CKNULL( newAES );

	switch( sLittleEndianToPGPUInt32( disk->reencryptingMainHdr->algorithm ) )
	{
		case kPGPdiskCAST5Algorithm:
			CASTCFB_Init( oldCAST, pOldDiskKey->bytes );
			oldCipher = oldCAST;
			break;
		case kPGPdiskAESAlgorithm:
			aes256Key( oldAES, pOldDiskKey->bytes );
			oldCipher = oldAES;
			break;
	}
	switch( sLittleEndianToPGPUInt32( disk->mainHdr.algorithm ) )
	{
		case kPGPdiskCAST5Algorithm:
			CASTCFB_Init( newCAST, pNewDiskKey->bytes );
			newCipher = newCAST;
			break;
		case kPGPdiskAESAlgorithm:
			aes256Key( newAES, pNewDiskKey->bytes );
			newCipher = newAES;
			break;
	}

	for( i = startBlock; i < endBlock; i++ )
	{
		/* Read old block */
		err = sSeek( disk->fd,
			( i - sLittleEndianToPGPUInt64( disk->mainHdr.numHeaderBlocks ) ) * kPGPdiskBlockSize + disk->dataOffset,
			SEEK_SET ); CKERR;
		err = sRead( disk->fd, oldBlock, kPGPdiskBlockSize ); CKERR;
		
		/* Decrypt old block */
		err = sDecryptBlock( sLittleEndianToPGPUInt32( disk->reencryptingMainHdr->algorithm ),
			oldCipher,
			&disk->reencryptingMainHdr->salt,
			i - sLittleEndianToPGPUInt64( disk->reencryptingMainHdr->numHeaderBlocks ),
			oldBlock,
			tempBlock,
			TRUE ); CKERR;
			
		/* Encrypt new block */
		err = sEncryptBlock( sLittleEndianToPGPUInt32( disk->mainHdr.algorithm ),
			newCipher,
			&disk->mainHdr.salt,
			i - sLittleEndianToPGPUInt64( disk->mainHdr.numHeaderBlocks ),
			tempBlock,
			newBlock,
			TRUE ); CKERR;
			
		/* Write new block */
		err = sSeek( disk->fd,
			( i - sLittleEndianToPGPUInt64( disk->mainHdr.numHeaderBlocks ) ) * kPGPdiskBlockSize + disk->dataOffset,
			SEEK_SET ); CKERR;
		err = sWrite( disk->fd, newBlock, kPGPdiskBlockSize ); CKERR;
		
		disk->mainHdr.numBlocksReEncrypted = sPGPUInt64ToLittleEndian(
			sLittleEndianToPGPUInt64( disk->mainHdr.numBlocksReEncrypted ) + 1 );
		err = sWriteMainHeader( disk ); CKERR;
		(void) fsync( disk->fd );

		/* Send an update to the app every 1% or every second to allow cancellation */
		if( ( ( i * 100 ) / endBlock > percentDone ) || ( time( NULL ) > lastMessage ) )
		{
			lastMessage = time( NULL );
			percentDone = ( i * 100 ) / endBlock;
			if( percentDone != 100 )
			{
				/* Callback */
				err = PGPrpcPackAndWrite( context, sock, kPGPdiskEngineMsg_ReEncryptDiskCallback,
					"u", percentDone ); CKERR;
				err = PGPrpcReadAndUnpack( context, sock, kPGPdiskEngineMsg_ReEncryptDiskCallbackFinished,
					"u", &clienterr ); CKERR;
				if( clienterr == kPGPError_UserAbort )
					RETERR( kPGPError_UserAbort );
			}
		}
	}

	err = PGPrpcPackAndWrite( context, sock, kPGPdiskEngineMsg_ReEncryptDiskCallback,
		"u", 100 ); CKERR;
	err = PGPrpcReadAndUnpack( context, sock, kPGPdiskEngineMsg_ReEncryptDiskCallbackFinished,
		"u", &clienterr ); CKERR;

done:
	if( IsntNull( oldCAST ) )
	{
		pgpClearMemory( oldCAST, sizeof( CASTCFBContext ) );
		(void) PGPFreeData( oldCAST );
	}
	if( IsntNull( newCAST ) )
	{
		pgpClearMemory( newCAST, sizeof( CASTCFBContext ) );
		(void) PGPFreeData( newCAST );
	}
	if( IsntNull( oldAES ) )
	{
		pgpClearMemory( oldAES, kMaxAESContextSize );
		(void) PGPFreeData( oldAES );
	}
	if( IsntNull( newAES ) )
	{
		pgpClearMemory( newAES, kMaxAESContextSize );
		(void) PGPFreeData( newAES );
	}
	return err;
}

	PGPError
PGPdiskEngineReEncryptDisk(
	PGPdiskBEContextRef	diskcon,
	PGPdiskIODiskRef	disk,
	char *				pszPassphrase,
	PGPdiskEncryptionAlgorithm	newAlg,
	PGPInt32			callbackSock )
{
	PGPError					err				= kPGPError_NoErr;
	PGPdiskSymmetricKey *		pOldDiskKey		= NULL;
	PGPdiskSymmetricKey *		pNewDiskKey		= NULL;
	PGPUInt32					keySize			= 0;
	PGPKeyDBObjRef				key				= kInvalidPGPKeyDBObjRef;
	PGPdiskIODiskUser *			adminUser		= NULL;
	PGPdiskOnDiskMainHeader		tempHdr;

	PGPdiskValidateBEContextRef( diskcon );
	PGPdiskValidateIODiskRef( disk );
	PGPValidatePtr( pszPassphrase );
	
	pthread_mutex_lock( &disk->mutex );
	
	if( disk->diskFormat == kPGPdiskDiskFormat_OS9 )
		RETERR( kPGPError_FeatureNotAvailable );

	/* Make sure not mounted */
	if( disk->bMounted )
		RETERR( kPGPdiskError_DiskAlreadyMounted );
		
	adminUser = sGetAdminUser( disk );
	
	/* Get old disk key */
	err = sGetCipherSizes( newAlg, NULL, &keySize, NULL ); CKERR;

	if( disk->mainHdr.isBeingReEncrypted )
	{
		/* Decrypt new disk key */
		err = sDecryptDiskKey( diskcon, disk->reencryptingMainHdr, disk->reencryptingAdminUser, pszPassphrase, &pOldDiskKey ); CKERR;
		err = sDecryptDiskKey( diskcon, &disk->mainHdr, adminUser, pszPassphrase, &pNewDiskKey ); CKERR;
	}
	else
	{
		/* Generate new disk key */
		err = sDecryptDiskKey( diskcon, &disk->mainHdr, adminUser, pszPassphrase, &pOldDiskKey ); CKERR;
		err = sGenerateNewRandomKey( diskcon, newAlg, &pNewDiskKey );CKERR;

		/* Create temporary MainHeader and admin user to store the old disk key */
		disk->reencryptingMainHdr = PGPNewData( disk->mgr,
			sizeof( PGPdiskOnDiskMainHeader ),
			kPGPMemoryMgrFlags_Clear );
		CKNULL( disk->reencryptingMainHdr );
		pgpClearMemory( disk->reencryptingMainHdr, sizeof( PGPdiskOnDiskMainHeader ) );
		
		err = sCreateMainHeader( diskcon->context,
			disk->reencryptingMainHdr, 
			sLittleEndianToPGPUInt64( disk->mainHdr.numDataBlocks ) * kPGPdiskBlockSize,
			disk->mainHdr.defaultRoot,
			newAlg ); CKERR;
		
		disk->reencryptingAdminUser = PGPNewData( disk->mgr,
			sizeof( PGPdiskIODiskUser ),
			kPGPMemoryMgrFlags_Clear );
		CKNULL( disk->reencryptingAdminUser );
		pgpClearMemory( disk->reencryptingAdminUser, sizeof( PGPdiskIODiskUser ) );
		
		if( adminUser->bSymUser )
		{
			err = sCreateSymmetricKeyUser( diskcon->context,
				diskcon->mgr,
				disk->reencryptingMainHdr,
				disk->reencryptingAdminUser,
				pNewDiskKey,
				adminUser->user.symUser.userName,
				pszPassphrase,
				adminUser->user.symUser.header.readOnly,
				FALSE,		/* admin user */
				adminUser->user.symUser.header.disabled,
				adminUser->user.symUser.header.locked ); CKERR;
		}
		else
		{
			err = PGPFindKeyByKeyID( diskcon->keydb, &adminUser->user.pubUser.keyID, &key ); CKERR;
			err = sCreatePublicKeyUser( disk->mgr,
				disk->reencryptingMainHdr,
				disk->reencryptingAdminUser,
				pNewDiskKey,
				key,
				adminUser->user.pubUser.header.readOnly,
				FALSE,		/* admin user */
				adminUser->user.pubUser.header.disabled,
				adminUser->user.pubUser.header.locked ); CKERR;
		}
		disk->reencryptingAdminUser->disk = disk;

		/* Update all disk users with new key */
		err = sUpdateUsersWithNewKey( diskcon, disk, pNewDiskKey ); CKERR;
		
		/* Swap the admin user and disk->reencryptingAdminUser */
		err = sRemoveUserFromUserList( disk, adminUser ); CKERR;
		err = sAddUserToUserList( disk, disk->reencryptingAdminUser ); CKERR;
		err = sSetAdminUser( disk, disk->reencryptingAdminUser ); CKERR;
		disk->reencryptingAdminUser = adminUser;
		
		/* Swap the main headers */
		pgpCopyMemory( &disk->mainHdr, &tempHdr, sizeof( PGPdiskOnDiskMainHeader ) );
		pgpCopyMemory( disk->reencryptingMainHdr, &disk->mainHdr, sizeof( PGPdiskOnDiskMainHeader ) );
		pgpCopyMemory( &tempHdr, disk->reencryptingMainHdr, sizeof( PGPdiskOnDiskMainHeader ) );
		
		err = sWriteDiskUsers( disk ); CKERR;
	}
	
	/* Re-encrypt data blocks */
	disk->mainHdr.isBeingReEncrypted = TRUE;
	err = sWriteMainHeader( disk ); CKERR;

	err = sReEncryptDataBlocks( diskcon->context, disk, callbackSock, pOldDiskKey, pNewDiskKey ); CKERR;

	disk->mainHdr.isBeingReEncrypted = FALSE;

	(void) PGPFreeData( disk->reencryptingMainHdr );
	disk->reencryptingMainHdr = NULL;
	(void) sFreeIODiskUser( disk->reencryptingAdminUser );
	disk->reencryptingAdminUser = NULL;

	err = sWriteMainHeader( disk ); CKERR;
	err = sWriteDiskUsers( disk ); CKERR;
	(void) fsync( disk->fd );
	
done:

	pthread_mutex_unlock( &disk->mutex );

	if( IsntNull( pNewDiskKey ) )
	{
		pgpClearMemory( pNewDiskKey, sizeof( PGPdiskSymmetricKey ) );
		(void) PGPFreeData( pNewDiskKey );
	}
	if( IsntNull( pOldDiskKey ) )
	{
		pgpClearMemory( pOldDiskKey, sizeof( PGPdiskSymmetricKey ) );
		(void) PGPFreeData( pOldDiskKey );
	}
	return err;
}

	PGPError
PGPdiskEngineGetIODiskFilename(
	PGPdiskIODiskRef	disk,
	char *				filename )
{
	PGPValidatePtr( filename );
	
	pthread_mutex_lock( &disk->mutex );
	strcpy( filename, disk->filename );
	pthread_mutex_unlock( &disk->mutex );
	
	return kPGPError_NoErr;
}

	static void *
sTimer(
	void *					arg )
{
	PGPdiskIODiskTimerData *	tData			= arg;
	PGPdiskIODiskRef			disk			= kInvalidPGPdiskIODiskRef;
	PGPByte *					p				= NULL;
	PGPUInt32					i				= 0;
	PGPUInt32					whichContext	= 0;
	time_t						lastFlipped		= 0;

	lastFlipped = time( NULL );
	while( 1 )
	{
		/* Check for something to do every 5 seconds */
		sleep( 5 );

		if( !tData->bDiskValid )
			break;

		pthread_mutex_lock( &tData->disk->mutex );
		
		disk = tData->disk;
		
		/* Check for disk timeout */
		if( ( disk->timeout != 0 ) && ( time( NULL ) > disk->timeLastUsed + disk->timeout ) )
		{
			/* Timeout expired.  Unmount the disk now */
			disk->bTimedOut = TRUE;
			pthread_mutex_unlock( &tData->disk->mutex );

			/* Allow forced unmount */
			(void) PGPdiskEngineUnmountDisk( disk->kernCon, disk, disk->forcedUnmount );
			break;
		}
		
		/* Flip bits of cipher context */
		if( time( NULL ) > lastFlipped + 30 )
		{
			lastFlipped = time( NULL );
			for( whichContext = 0; whichContext < 2; whichContext++ )
			{
				if( IsntNull( disk->castCFB[whichContext] ) )
				{
					p = (PGPByte *) disk->castCFB[whichContext];
					for( i = 0; i < sizeof( CASTCFBContext ); i++ )
						p[i] = ~p[i];
				}
				if( IsntNull( disk->aesCFB[whichContext] ) )
				{
					p = (PGPByte *) disk->aesCFB[whichContext];
					for( i = 0; i < kMaxAESContextSize; i++ )
						p[i] = ~p[i];
				}
			}
			disk->whichContext = disk->whichContext == 0 ? 1 : 0;
		}
		
		pthread_mutex_unlock( &tData->disk->mutex );
	}
	
	(void) PGPFreeData( tData );
	
	return NULL;
}

	PGPError
PGPdiskEngineDecryptDisk(
	PGPdiskBEContextRef	diskcon,
	PGPdiskIODiskRef	disk,
	PGPdiskIODiskUser *	user,
	const char *		passphrase )
{
	PGPError						err			= kPGPError_NoErr;
	PGPdiskSymmetricKey	*			pDiskKey	= NULL;
	PGPdiskIODiskTimerData *		tData		= NULL;
	PGPByte *						pNormal		= NULL;
	PGPByte *						pFlipped	= NULL;
	PGPUInt32						i			= 0;
	pthread_t						thread;
	
	PGPdiskValidateIODiskRef( disk );
	PGPValidatePtr( user );
	PGPValidatePtr( passphrase );

	pthread_mutex_lock( &disk->mutex );

	if( disk->diskFormat == kPGPdiskDiskFormat_V6 )
	{
		err = sDecryptDiskKey( diskcon, &disk->mainHdr, user, passphrase, &pDiskKey );
		CKERR;
	}
	else if( disk->diskFormat == kPGPdiskDiskFormat_OS9 )
	{
		err = sDecryptOS9DiskKey( diskcon->context, diskcon->mgr, &disk->os9FileHeader, passphrase, &pDiskKey );
		CKERR;
	}
	
	/* We're doing this to mount the disk.  Increment our refCount */
	disk->refCount++;
	
	disk->bTimedOut = FALSE;

	switch( sLittleEndianToPGPUInt32( disk->mainHdr.algorithm ) )
	{
		case kPGPdiskCAST5Algorithm:
			for( i = 0; i < 2; i++ )
			{
				disk->castCFB[i] = PGPNewSecureData( disk->mgr,
					sizeof( CASTCFBContext ),
					kPGPMemoryMgrFlags_Clear );
				CKNULL( disk->castCFB[i] );
			}
			disk->whichContext = 0;
			CASTCFB_Init( disk->castCFB[0], pDiskKey->bytes );

			pNormal = (PGPByte *) disk->castCFB[0];
			pFlipped = (PGPByte *) disk->castCFB[1];
			for( i = 0; i < sizeof( CASTCFBContext ); i++ )
				pFlipped[i] = ~pNormal[i];

			break;
		case kPGPdiskAESAlgorithm:
			for( i = 0; i < 2; i++ )
			{
				disk->aesCFB[i] = PGPNewSecureData( disk->mgr,
					kMaxAESContextSize,
					kPGPMemoryMgrFlags_Clear );
				CKNULL( disk->aesCFB[i] );
			}
			disk->whichContext = 0;
			aes256Key( disk->aesCFB[0], pDiskKey->bytes );

			pNormal = (PGPByte *) disk->aesCFB[0];
			pFlipped = (PGPByte *) disk->aesCFB[1];
			for( i = 0; i < kMaxAESContextSize; i++ )
				pFlipped[i] = ~pNormal[i];

			break;
		default:
			RETERR( kPGPError_FeatureNotAvailable );
			break;
	}
	
	if( IsNull( disk->pbTimerStillRunning ) )
	{
		/* Start bit flipper thread */
		tData = PGPNewData( disk->mgr,
			sizeof( PGPdiskIODiskTimerData ),
			kPGPMemoryMgrFlags_Clear );
		CKNULL( tData );
		
		tData->bDiskValid = TRUE;
		tData->disk = disk;
		disk->pbTimerStillRunning = &tData->bDiskValid;
		
		pthread_create( &thread, NULL, sTimer, tData );
		pthread_detach( thread );
	}
	
done:
	pthread_mutex_unlock( &disk->mutex );
	if( IsntNull( pDiskKey ) )
	{
		pgpClearMemory( pDiskKey, sizeof( PGPdiskSymmetricKey ) );
		(void) PGPFreeData( pDiskKey );
	}
	return err;
}

	PGPUInt64
PGPdiskEngineGetDiskSize(
	PGPdiskIODiskRef	disk )
{
	PGPUInt64			diskSize		= 0;
	
	pthread_mutex_lock( &disk->mutex );
	diskSize = disk->dataSize;
	pthread_mutex_unlock( &disk->mutex );
	
	return diskSize;
}

	PGPError
PGPdiskEngineCountUsers(
	PGPdiskIODiskRef		disk,
	PGPUInt32 *				outNumUsers )
{
	PGPdiskValidateIODiskRef( disk );
	PGPValidatePtr( outNumUsers );

	pthread_mutex_lock( &disk->mutex );
	*outNumUsers = disk->numUsers;
	pthread_mutex_unlock( &disk->mutex );
	
	return kPGPError_NoErr;
}
	
	PGPError
PGPdiskEngineGetIODiskUser(
	PGPdiskBEContextRef		diskcon,
	PGPdiskIODiskRef		disk,
	PGPUInt32				userNumber,
	PGPdiskBEUserRef		user )
{
	PGPError				err			= kPGPError_NoErr;
	PGPdiskIODiskUser *		pUser		= NULL;
	
	PGPdiskValidateIODiskRef( disk );
	PGPdiskValidateBEUserRef( user );
	
	pthread_mutex_lock( &disk->mutex );
	
	for( pUser = disk->userList; IsntNull( pUser ); pUser = pUser->next )
	{
		if( userNumber-- == 0 )
		{
			user->ioDiskUser = pUser;
			break;
		}
	}
	if( IsNull( pUser ) )
		RETERR( kPGPError_ItemNotFound );
		
	/* Copy user->ioDiskUser fields into user */
	if( user->ioDiskUser->bSymUser )
	{
		user->bSymUser = TRUE;
		strcpy( user->username, user->ioDiskUser->user.symUser.userName );
		user->passphrase = NULL;
		user->bReadOnly = user->ioDiskUser->user.symUser.header.readOnly;
		user->bAdminUser = user->ioDiskUser->user.symUser.header.adminUser;
		user->bDisabled = user->ioDiskUser->user.symUser.header.disabled;
	}
	else
	{
		user->bSymUser = FALSE;
		user->bReadOnly = user->ioDiskUser->user.pubUser.header.readOnly;
		user->bAdminUser = user->ioDiskUser->user.pubUser.header.adminUser;
		user->bDisabled = user->ioDiskUser->user.pubUser.header.disabled;
		
		pgpCopyMemory( &user->ioDiskUser->user.pubUser.keyID, &user->keyid, sizeof( PGPKeyID ) );
	}
	
done:
	pthread_mutex_unlock( &disk->mutex );
	return err;
}

	PGPdiskIODiskUser *
PGPdiskEngineGetAdminUser(
	PGPdiskIODiskRef		disk )
{
	PGPdiskIODiskUser *		user	= NULL;

	pthread_mutex_lock( &disk->mutex );
	user = sGetAdminUser( disk );
	pthread_mutex_unlock( &disk->mutex );

	return user;
}

	static PGPBoolean
sIsUserOnDisk(
	PGPdiskIODiskRef		disk,
	PGPdiskBEUserRef		user )
{
	PGPdiskIODiskUser *		ioDiskUser		= NULL;
	
	for( ioDiskUser = disk->userList; IsntNull( ioDiskUser ); ioDiskUser = ioDiskUser->next )
	{
		if( user->ioDiskUser == ioDiskUser )
			return TRUE;

		if( user->bSymUser && ioDiskUser->bSymUser )
		{
			if( strcmp( user->username, ioDiskUser->user.symUser.userName ) == 0 )
				return TRUE;
		}
		else if( !user->bSymUser && !ioDiskUser->bSymUser )
		{
			if( PGPCompareKeyIDs( &user->keyid, &ioDiskUser->user.pubUser.keyID ) == 0 )
				return TRUE;
		}
	}
	return FALSE;
}

	PGPBoolean
PGPdiskEngineIsUserOnDisk(
	PGPdiskIODiskRef		disk,
	PGPdiskBEUserRef		user )
{
	PGPBoolean				bRetVal			= FALSE;
	
	pthread_mutex_lock( &disk->mutex );
	bRetVal = sIsUserOnDisk( disk, user );
	pthread_mutex_unlock( &disk->mutex );
	
	return bRetVal;
}

	PGPError
PGPdiskEngineAddUser(
	PGPdiskBEContextRef		diskcon,
	PGPdiskIODiskRef		disk,
	PGPdiskIODiskUser *		adminUser,
	const char *			adminPassphrase,
	PGPdiskBEUserRef		user )
{
	PGPError				err			= kPGPError_NoErr;
	PGPdiskIODiskUser *		ioDiskUser	= NULL;
	PGPdiskSymmetricKey *	pDiskKey	= NULL;
	PGPKeyDBObjRef			key			= kInvalidPGPKeyDBObjRef;
	
	pthread_mutex_lock( &disk->mutex );
	
	if( disk->diskFormat == kPGPdiskDiskFormat_OS9 )
		RETERR( kPGPError_FeatureNotAvailable );
	
	if( sIsUserOnDisk( disk, user ) )
		RETERR( kPGPError_ItemAlreadyExists );
	
	pDiskKey = PGPNewSecureData( disk->mgr,
		sizeof( PGPdiskSymmetricKey ),
		kPGPMemoryMgrFlags_Clear );
	CKNULL( pDiskKey );

	if( adminUser->bSymUser )
		err = sDecryptDiskKeyWithSym( diskcon->context, diskcon->mgr, &disk->mainHdr, adminUser, adminPassphrase, pDiskKey );
	else
	{
		err = sDecryptDiskKeyWithPub( diskcon->context,
			diskcon->mgr,
			&disk->mainHdr,
			diskcon->keydb,
			adminUser,
			adminPassphrase,
			diskcon->cachePassphrases,
			diskcon->cacheTimeout,
			diskcon->shareCache,
			pDiskKey );
	}
	CKERR;

	if( user->bSymUser )
	{
		err = sAddSymmetricKeyUserToDisk( diskcon->context,
			disk,
			pDiskKey,
			user->username,
			user->passphrase,
			user->bReadOnly,
			user->bAdminUser,
			user->bDisabled,
			FALSE,		/* locked */
			&ioDiskUser ); CKERR;
	}
	else
	{
		err = PGPFindKeyByKeyID( diskcon->keydb, &user->keyid, &key ); CKERR;
		err = sAddPublicKeyUserToDisk( disk,
			pDiskKey,
			key,
			user->bReadOnly,
			user->bAdminUser,
			user->bDisabled,
			user->bLocked,
			&ioDiskUser ); CKERR;
	}

	user->ioDiskUser = ioDiskUser;

	err = sWriteDiskUsers( disk ); CKERR;
	
done:
	pthread_mutex_unlock( &disk->mutex );
	if( IsntNull( pDiskKey ) )
	{
		pgpClearMemory( pDiskKey, sizeof( PGPdiskSymmetricKey ) );
		(void) PGPFreeData( pDiskKey );
	}
	return err;
}

	PGPError
PGPdiskEngineRemoveUser(
	PGPdiskIODiskRef		disk,
	PGPdiskIODiskUser *		user )
{
	PGPError				err			= kPGPError_NoErr;
	PGPBoolean				bAdminUser	= FALSE;
	PGPBoolean				bLocked		= FALSE;
	
	PGPdiskValidateIODiskRef( disk );
	PGPValidatePtr( user );

	pthread_mutex_lock( &disk->mutex );
	
	if( disk->diskFormat == kPGPdiskDiskFormat_OS9 )
		RETERR( kPGPError_FeatureNotAvailable );
	
	bAdminUser = user->bSymUser ? user->user.symUser.header.adminUser : user->user.pubUser.header.adminUser;
	bLocked = user->bSymUser ? user->user.symUser.header.locked : user->user.pubUser.header.locked;
	
	if( bAdminUser )
		RETERR( kPGPdiskError_IsAdminUser );
	if( bLocked )
		RETERR( kPGPdiskError_IsLocked );

	err = sRemoveUserFromUserList( disk, user ); CKERR;
	(void) sFreeIODiskUser( user );

	err = sWriteDiskUsers( disk ); CKERR;

done:
	pthread_mutex_unlock( &disk->mutex );
	return err;
}

	PGPError
PGPdiskEngineIsPassphraseCorrect(
	PGPdiskBEContextRef		diskcon,
	PGPdiskIODiskUser *		user,
	const char *			pszPassphrase,
	PGPBoolean *			bCorrect )
{
	PGPError				err			= kPGPError_NoErr;
	PGPdiskSymmetricKey *	pDiskKey	= NULL;

	if( user->disk->diskFormat == kPGPdiskDiskFormat_V6 )
	{
		err = sDecryptDiskKey( diskcon,
			&user->disk->mainHdr,
			user,
			pszPassphrase,
			&pDiskKey );
	}
	else if( user->disk->diskFormat == kPGPdiskDiskFormat_OS9 )
	{
		err = sDecryptOS9DiskKey( diskcon->context,
			diskcon->mgr,
			&user->disk->os9FileHeader,
			pszPassphrase,
			&pDiskKey );
	}

	if( IsntPGPError( err ) )
		*bCorrect = TRUE;
	else if( err == kPGPError_BadPassphrase )
	{
		*bCorrect = FALSE;
		err = kPGPError_NoErr;
	}

	if( IsntNull( pDiskKey ) )
	{
		pgpClearMemory( pDiskKey, sizeof( PGPdiskSymmetricKey ) );
		(void) PGPFreeData( pDiskKey );
	}
	return err;
}

	PGPError
PGPdiskEngineChangeUserPassphrase(
	PGPdiskBEContextRef		diskcon,
	PGPdiskIODiskUser *		user,
	const char *			oldPassphrase,
	const char *			newPassphrase )
{
	PGPError				err			= kPGPError_NoErr;
	PGPdiskSymmetricKey *	pDiskKey	= NULL;
	
	PGPValidatePtr( user );
	PGPValidatePtr( oldPassphrase );
	PGPValidatePtr( newPassphrase );
	
	pthread_mutex_lock( &user->disk->mutex );

	if( user->disk->diskFormat == kPGPdiskDiskFormat_OS9 )
		RETERR( kPGPError_FeatureNotAvailable );
	
	/* Has to be a symmetric key user */
	if( !user->bSymUser )
		RETERR( kPGPError_FeatureNotAvailable );
	
	err = sDecryptDiskKey( diskcon,
		&user->disk->mainHdr,
		user,
		oldPassphrase,
		&pDiskKey ); CKERR;
		
	err = sMakeEncryptedKeyForSymUser( diskcon->context,
		diskcon->mgr,
		user,
		newPassphrase,
		&user->disk->mainHdr,
		pDiskKey ); CKERR;
		
	err = sWriteDiskUsers( user->disk ); CKERR;
	
done:
	pthread_mutex_unlock( &user->disk->mutex );
	if( IsntNull( pDiskKey ) )
	{
		pgpClearMemory( pDiskKey, sizeof( PGPdiskSymmetricKey ) );
		(void) PGPFreeData( pDiskKey );
	}
	return err;
}

	PGPError
PGPdiskEngineSetAdminUser(
	PGPdiskIODiskRef		disk,
	PGPdiskIODiskUser *		newAdminUser )
{
	PGPError				err			= kPGPError_NoErr;
	
	pthread_mutex_lock( &disk->mutex );
	
	if( disk->diskFormat == kPGPdiskDiskFormat_OS9 )
		RETERR( kPGPError_FeatureNotAvailable );

	err = sSetAdminUser( disk, newAdminUser ); CKERR;
	err = sWriteDiskUsers( disk ); CKERR;

done:
	pthread_mutex_unlock( &disk->mutex );
	return err;
}

	PGPError
PGPdiskEngineBEUserModified(
	PGPdiskBEUserRef		user )
{
	PGPError				err		= kPGPError_NoErr;
	
	PGPdiskValidateBEUserRef( user );
	PGPValidatePtr( user->ioDiskUser );
	
	pthread_mutex_lock( &user->ioDiskUser->disk->mutex );

	if( user->ioDiskUser->disk->diskFormat == kPGPdiskDiskFormat_OS9 )
		RETERR( kPGPError_FeatureNotAvailable );

	if( user->bSymUser )
	{
		strcpy( user->ioDiskUser->user.symUser.userName, user->username );
		user->ioDiskUser->user.symUser.header.readOnly	= user->bReadOnly;
		user->ioDiskUser->user.symUser.header.disabled	= user->bDisabled;
	}
	else
	{
		user->ioDiskUser->user.pubUser.header.readOnly	= user->bReadOnly;
		user->ioDiskUser->user.pubUser.header.disabled	= user->bDisabled;
		user->ioDiskUser->user.pubUser.header.locked	= user->bLocked;
	}
	
	err = sWriteDiskUsers( user->ioDiskUser->disk ); CKERR;

done:
	pthread_mutex_unlock( &user->ioDiskUser->disk->mutex );
	return err;
}

	PGPError
PGPdiskEngineGetDiskBooleanProperty(
	PGPdiskIODiskRef		disk,
	PGPUInt32				prop,
	PGPBoolean *			bProperty )
{
	PGPError				err			= kPGPError_NoErr;
	
	PGPdiskValidateIODiskRef( disk );
	PGPValidatePtr( bProperty );
	
	pthread_mutex_lock( &disk->mutex );
	switch( prop )
	{
		case kPGPdiskDiskProp_IsMounted:
			*bProperty = disk->bMounted;
			break;
		case kPGPdiskDiskProp_IsReadOnly:
			*bProperty = disk->bReadOnly;
			break;
		case kPGPdiskDiskProp_IsBeingReEncrypted:
			*bProperty = disk->mainHdr.isBeingReEncrypted;
			break;
		case kPGPdiskDiskProp_HasTimedOut:
			*bProperty = disk->bTimedOut;
			break;
		case kPGPdiskDiskProp_UsesCustomTimeout:
			*bProperty = disk->mainHdr.usesCustomTimeout;
			break;
		case kPGPdiskDiskProp_UsesCustomEjectOnLogout:
			*bProperty = disk->mainHdr.usesCustomEjectOnLogout;
			break;
		case kPGPdiskDiskProp_EjectOnLogout:
			*bProperty = disk->mainHdr.ejectOnLogout;
			break;
		default:
			RETERR( kPGPError_InvalidProperty );
			break;
	}

done:
	pthread_mutex_unlock( &disk->mutex );
	return err;
}

	PGPError
PGPdiskEngineSetDiskBooleanProperty(
	PGPdiskIODiskRef		disk,
	PGPUInt32				prop,
	PGPBoolean				bProperty )
{
	PGPError				err			= kPGPError_NoErr;
	
	PGPdiskValidateIODiskRef( disk );
	
	pthread_mutex_lock( &disk->mutex );

	if( disk->diskFormat == kPGPdiskDiskFormat_OS9 )
		RETERR( kPGPError_FeatureNotAvailable );
	if( disk->bReadOnly )
		RETERR( kPGPdiskError_DiskIsReadOnly );

	switch( prop )
	{
		case kPGPdiskDiskProp_UsesCustomTimeout:
			disk->bCustomTimeout = bProperty;
			break;
		case kPGPdiskDiskProp_UsesCustomEjectOnLogout:
			disk->mainHdr.usesCustomEjectOnLogout = bProperty;
			break;
		case kPGPdiskDiskProp_EjectOnLogout:
			disk->mainHdr.ejectOnLogout = bProperty;
			break;
		default:
			RETERR( kPGPError_InvalidProperty );
			break;
	}
	err = sWriteMainHeader( disk ); CKERR;

done:
	pthread_mutex_unlock( &disk->mutex );
	return err;
}

	PGPError
PGPdiskEngineGetDiskNumericProperty(
	PGPdiskIODiskRef		disk,
	PGPUInt32				prop,
	PGPInt32 *				iProperty )
{
	PGPError				err			= kPGPError_NoErr;
	
	PGPdiskValidateIODiskRef( disk );
	PGPValidatePtr( iProperty );
	
	pthread_mutex_lock( &disk->mutex );
	switch( prop )
	{
		case kPGPdiskDiskProp_DiskFormat:
			*iProperty = disk->diskFormat;
			break;
		case kPGPdiskDiskProp_Algorithm:
			*iProperty = sLittleEndianToPGPUInt32( disk->mainHdr.algorithm );
			break;
		case kPGPdiskDiskProp_CustomTimeout:
			*iProperty = sLittleEndianToPGPUInt32( disk->mainHdr.customTimeout );
			break;
		case kPGPdiskDiskProp_BlocksDiskLow:
			*iProperty = sLittleEndianToPGPUInt64( disk->mainHdr.numDataBlocks ) & 0xFFFFFFFF;
			break;
		case kPGPdiskDiskProp_BlocksDiskHigh:
			*iProperty = ( sLittleEndianToPGPUInt64( disk->mainHdr.numDataBlocks ) & 0xFFFFFFFF00000000 ) >> 32;
			break;
		default:
			RETERR( kPGPError_InvalidProperty );
			break;
	}

done:
	pthread_mutex_unlock( &disk->mutex );
	return err;
}

	PGPError
PGPdiskEngineSetDiskNumericProperty(
	PGPdiskIODiskRef		disk,
	PGPUInt32				prop,
	PGPInt32 				iProperty )
{
	PGPError				err			= kPGPError_NoErr;
	
	PGPdiskValidateIODiskRef( disk );
	
	pthread_mutex_lock( &disk->mutex );

	if( disk->diskFormat == kPGPdiskDiskFormat_OS9 )
		RETERR( kPGPError_FeatureNotAvailable );
	if( disk->bReadOnly )
		RETERR( kPGPdiskError_DiskIsReadOnly );

	switch( prop )
	{
		case kPGPdiskDiskProp_CustomTimeout:
			disk->timeout = iProperty * 60;		/* minutes -> seconds */
			break;
		default:
			RETERR( kPGPError_InvalidProperty );
			break;
	}
	err = sWriteMainHeader( disk ); CKERR;

done:
	pthread_mutex_unlock( &disk->mutex );
	return err;
}

	PGPError
PGPdiskEngineGetDiskDataProperty(
	PGPdiskIODiskRef		disk,
	PGPUInt32				prop,
	char **					pszProperty )
{
	PGPError				err			= kPGPError_NoErr;
	char *					pDataProp	= NULL;
	
	PGPdiskValidateIODiskRef( disk );
	PGPValidatePtr( pszProperty );
	
	pthread_mutex_lock( &disk->mutex );
	switch( prop )
	{
		case kPGPdiskDiskProp_Path:
			pDataProp = disk->filename;
			break;
		case kPGPdiskDiskProp_Root:
			/* XXX Where it's really mounted */
			break;
		case kPGPdiskDiskProp_DefaultRoot:
			pDataProp = disk->mainHdr.defaultRoot;
			break;
		case kPGPdiskDiskProp_DeviceName:
			pDataProp = disk->szDeviceName;
			break;
		default:
			RETERR( kPGPError_InvalidProperty );
			break;
	}
	
	if( IsntNull( pDataProp ) )
	{
		*pszProperty = PGPNewData( disk->mgr,
			strlen( pDataProp ) + 1,
			kPGPMemoryMgrFlags_Clear );
		CKNULL( *pszProperty );
		strcpy( *pszProperty, pDataProp );
	}

done:	
	pthread_mutex_unlock( &disk->mutex );
	return err;
}

	PGPError
PGPdiskEngineSetDiskDataProperty(
	PGPdiskIODiskRef		disk,
	PGPUInt32				prop,
	char *					pszProperty )
{
	PGPError				err			= kPGPError_NoErr;
	
	PGPdiskValidateIODiskRef( disk );
	PGPValidatePtr( pszProperty );
	
	pthread_mutex_lock( &disk->mutex );

	if( disk->diskFormat == kPGPdiskDiskFormat_OS9 )
		RETERR( kPGPError_FeatureNotAvailable );
	if( disk->bReadOnly )
		RETERR( kPGPdiskError_DiskIsReadOnly );

	switch( prop )
	{
		case kPGPdiskDiskProp_DefaultRoot:
			strncpy( disk->mainHdr.defaultRoot, pszProperty,
				pgpMin( sizeof( disk->mainHdr.defaultRoot ), strlen( pszProperty ) ) );
			break;
		default:
			RETERR( kPGPError_InvalidProperty );
			break;
	}
	err = sWriteMainHeader( disk ); CKERR;

done:	
	pthread_mutex_unlock( &disk->mutex );
	return err;
}

	PGPError
PGPdiskEngineSetDiskMounted(
	PGPdiskIODiskRef		disk,
	PGPBoolean				bReadOnly )
{
	PGPError				err			= kPGPError_NoErr;
	
	PGPdiskValidateIODiskRef( disk );
	pthread_mutex_lock( &disk->mutex );
	
	if( disk->bMounted )
		RETERR( kPGPdiskError_DiskAlreadyMounted );
	disk->bMounted = TRUE;
	disk->bReadOnly = bReadOnly;
done:
	pthread_mutex_unlock( &disk->mutex );
	return err;
}

	PGPError
PGPdiskEngineSetDiskUnmounted(
	PGPdiskIODiskRef		disk )
{
	PGPdiskValidateIODiskRef( disk );

	pthread_mutex_lock( &disk->mutex );
	disk->bMounted = FALSE;
	pthread_mutex_unlock( &disk->mutex );
	
	return kPGPError_NoErr;
}

	PGPError
PGPdiskEngineUpdateGlobalDiskTimeout(
	PGPUInt32				timeout )
{
	PGPdiskIODiskListItem *	walk		= NULL;

	pthread_mutex_lock( &gDiskListMutex );
	
	for( walk = gDiskList; IsntNull( walk ); walk = walk->next )
	{
		if( !walk->disk->bCustomTimeout )
			walk->disk->timeout = timeout * 60;
	}
	
	pthread_mutex_unlock( &gDiskListMutex );

	return kPGPError_NoErr;
}

	PGPUInt32
PGPdiskEngineNumMountedDisks( void )
{
	PGPUInt32					count		= 0;
	PGPdiskIODiskListItem *		walk		= NULL;

	pthread_mutex_lock( &gDiskListMutex );
	
	for( walk = gDiskList; IsntNull( walk ); walk = walk->next )
	{
		if( walk->disk->bMounted )
			count++;
	}
	
	pthread_mutex_unlock( &gDiskListMutex );
	
	return count;
}

	PGPError
PGPdiskEngineUnmountAllDisks( 
	PGPBoolean				bForced )
{
	PGPError					err					= kPGPError_NoErr;
	PGPBoolean					bCouldntUnmount		= FALSE;
	PGPdiskIODiskListItem *		walk				= NULL;
	PGPdiskIODiskListItem *		next				= NULL;
	PGPdiskIODiskRef			disk				= kInvalidPGPdiskIODiskRef;

	pthread_mutex_lock( &gDiskListMutex );
	
	for( walk = gDiskList; IsntNull( walk ); walk = next )
	{
		next = walk->next;
		disk = walk->disk;
		
		if( walk->disk->bMounted )
		{
			pthread_mutex_unlock( &gDiskListMutex );
			
			err = PGPdiskEngineUnmountDisk( disk->kernCon, disk, bForced );
			
			pthread_mutex_lock( &gDiskListMutex );
			if( IsPGPError( err ) )
				bCouldntUnmount = TRUE;
		}
	}
	
	pthread_mutex_unlock( &gDiskListMutex );
	
	/* If we had a problem with any of the disks, return the CantUnmountDisk error */
	if( bCouldntUnmount )
		return kPGPdiskError_CantUnmountDisk;
	
	return kPGPError_NoErr;
}

	PGPError
PGPdiskEngineSetDiskDeviceName(
	PGPdiskIODiskRef		disk,
	char *					pszDeviceName )
{
	PGPdiskValidateIODiskRef( disk );
	
	pthread_mutex_lock( &disk->mutex );
	strcpy( disk->szDeviceName, pszDeviceName );
	pthread_mutex_unlock( &disk->mutex );
	
	return kPGPError_NoErr;
}

	PGPError
PGPdiskEngineSetDiskKernelConnectionContext(
	PGPdiskIODiskRef					disk,
	PGPdiskKernelConnectionContext *	driverConn )
{
	PGPdiskValidateIODiskRef( disk );
	
	pthread_mutex_lock( &disk->mutex );
	disk->driverConn = driverConn;
	pthread_mutex_unlock( &disk->mutex );
	
	return kPGPError_NoErr;
}

	PGPError
PGPdiskEngineGetDiskKernelConnectionContext(
	PGPdiskIODiskRef					disk,
	PGPdiskKernelConnectionContext **	driverConn )
{
	PGPdiskValidateIODiskRef( disk );
	PGPValidatePtr( driverConn );
	
	pthread_mutex_lock( &disk->mutex );
	*driverConn = disk->driverConn;
	pthread_mutex_unlock( &disk->mutex );
	
	return kPGPError_NoErr;
}

	PGPError
PGPdiskEngineReadFromDisk(
	PGPdiskIODiskRef		disk,
	PGPUInt64				startBlock,
	PGPUInt32				numBlocks,
	PGPByte *				data )
{
	PGPError				err				= kPGPError_NoErr;
	PGPUInt32				i				= 0;
	void *					cipherContext	= NULL;
	PGPByte					tempBlock[kPGPdiskBlockSize];
	
	pthread_mutex_lock( &disk->mutex );
	
	switch( sLittleEndianToPGPUInt32( disk->mainHdr.algorithm ) )
	{
		case kPGPdiskCAST5Algorithm:
			cipherContext = disk->castCFB[disk->whichContext]->expandedKey.keyWords;
			break;
		case kPGPdiskAESAlgorithm:
			cipherContext = disk->aesCFB[disk->whichContext];
			break;
		default:
			RETERR( kPGPError_FeatureNotAvailable );
			break;
	}
	
	err = sSeek( disk->fd, startBlock * kPGPdiskBlockSize + disk->dataOffset, SEEK_SET ); CKERR;
	err = sRead( disk->fd, data, numBlocks * kPGPdiskBlockSize ); CKERR;
	
	for( i = 0; i < numBlocks; i++ )
	{
		err = sDecryptBlock( sLittleEndianToPGPUInt32( disk->mainHdr.algorithm ),
			cipherContext,
			&disk->mainHdr.salt,
			startBlock + i,
			data + (i * kPGPdiskBlockSize),
			tempBlock,
			disk->diskFormat == kPGPdiskDiskFormat_OS9 ? FALSE : TRUE ); CKERR;
		pgpCopyMemory( tempBlock, data + (i * kPGPdiskBlockSize), kPGPdiskBlockSize );
	}
	
	disk->timeLastUsed = time( NULL );
	
done:
	pthread_mutex_unlock( &disk->mutex );
	return err;
}

	PGPError
PGPdiskEngineWriteToDisk(
	PGPdiskIODiskRef		disk,
	PGPUInt64				startBlock,
	PGPUInt32				numBlocks,
	PGPByte *				data )
{
	PGPError				err				= kPGPError_NoErr;
	PGPUInt32				i				= 0;
	void *					cipherContext	= NULL;
	PGPByte					tempBlock[kPGPdiskBlockSize];

	pthread_mutex_lock( &disk->mutex );

	switch( sLittleEndianToPGPUInt32( disk->mainHdr.algorithm ) )
	{
		case kPGPdiskCAST5Algorithm:
			cipherContext = disk->castCFB[disk->whichContext]->expandedKey.keyWords;
			break;
		case kPGPdiskAESAlgorithm:
			cipherContext = disk->aesCFB[disk->whichContext];
			break;
		default:
			RETERR( kPGPError_FeatureNotAvailable );
			break;
	}
	
	err = sSeek( disk->fd, startBlock * kPGPdiskBlockSize + disk->dataOffset, SEEK_SET ); CKERR;

	for( i = 0; i < numBlocks; i++ )
	{
		err = sEncryptBlock( sLittleEndianToPGPUInt32( disk->mainHdr.algorithm ),
			cipherContext,
			&disk->mainHdr.salt,
			i + startBlock,
			data + (i * kPGPdiskBlockSize),
			tempBlock,
			disk->diskFormat == kPGPdiskDiskFormat_OS9 ? FALSE : TRUE ); CKERR;
		pgpCopyMemory( tempBlock, data + (i * kPGPdiskBlockSize), kPGPdiskBlockSize );
	}

	err = sWrite( disk->fd, data, numBlocks * kPGPdiskBlockSize ); CKERR;

	disk->timeLastUsed = time( NULL );
	
done:
	pthread_mutex_unlock( &disk->mutex );
	return err;
}


/*__Editor_settings____

	Local Variables:
	tab-width: 4
	End:
	vi: ts=4 sw=4
	vim: si
_____________________*/
