/*____________________________________________________________________________
		Copyright (C) 2002 PGP Corporation
        All rights reserved.

        $Id: CPGPdiskDiskImp.cpp,v 1.22 2002/08/06 20:09:36 dallen Exp $
____________________________________________________________________________*/

#include "pgpClassesConfig.h"

#include "UMath.h"
#include "UTime.h"

#include "pgpClientErrors.h"
#include "CCipherContext.h"
#include "CPGPKey.h"

#include "CDriverComm.h"
#include "CEngineSubsystems.h"
#include "CPGPdiskContext.h"
#include "CPGPdiskDiskImp.h"
#include "CPGPdiskUser.h"
#include "CPGPdiskUserSet.h"

_USING_PGP

// Class CPGPdiskDiskImp member functions

CPGPdiskDiskImp::CPGPdiskDiskImp() :

	mPContext(NULL), mDiskState(kDiskStateNone), mHasTimedOut(FALSE),
		mIsLockedForFormat(FALSE), mIsLockedForIO(FALSE), mLockHandle(0)
{
}

PGPBoolean
CPGPdiskDiskImp::operator==(const CPGPdiskDisk& disk) const
{
	return ComparePaths(disk.Path());
}

PGPBoolean 
CPGPdiskDiskImp::ComparePaths(const char *path) const
{
	CString	cpath(path);
	return cpath.CompareNoCase(Path());
}

PGPBoolean 
CPGPdiskDiskImp::CompareRoots(const char *root) const
{
	CString	croot(root);
	return croot.CompareNoCase(Root());
}

PGPBoolean
CPGPdiskDiskImp::HasOldHeaders() const
{
	MakeSureHeadersRead();
	return mHeaders.HasOldHeaders();
}

PGPdiskEncryptionAlgorithm
CPGPdiskDiskImp::Algorithm() const
{
	MakeSureHeadersRead();
	return mHeaders.Algorithm();
}

PGPBoolean
CPGPdiskDiskImp::IsBeingReEncrypted() const
{
	MakeSureHeadersRead();
	return mHeaders.IsBeingReEncrypted();
}

PGPBoolean
CPGPdiskDiskImp::AreUsersWiped() const
{
	MakeSureHeadersRead();
	return mHeaders.AreUsersWiped();
}

PGPBoolean
CPGPdiskDiskImp::HasInactivityTimedOut() const
{
	MakeSureHeadersRead();
	return mHasTimedOut;
}

PGPBoolean
CPGPdiskDiskImp::UsesCustomTimeout() const
{
	MakeSureHeadersRead();
	return mHeaders.UsesCustomTimeout();
}

const char *
CPGPdiskDiskImp::DefaultRoot() const
{
	MakeSureHeadersRead();
	return mHeaders.DefaultRoot();
}

PGPUInt32
CPGPdiskDiskImp::CustomTimeout() const
{
	MakeSureHeadersRead();
	return mHeaders.CustomTimeout();
}

PGPUInt64
CPGPdiskDiskImp::BlocksDisk() const
{
	MakeSureHeadersRead();
	return mHeaders.BlocksDisk();
}

void
CPGPdiskDiskImp::AttachMounted(
	const CPGPdiskContext	*pContext,
	const char				*path,
	const char				*root,
	const char				*deviceName,
	PGPBoolean				readOnly,
	PGPBoolean				hasTimedOut)
{
	pgpAssert(!IsOpened());
	pgpAssert(!IsMounted());
	pgpAssertAddrValid(pContext, CPGPdiskContext);

	mPContext		= pContext;
	mIsReadOnly		= readOnly;
	mHasTimedOut	= hasTimedOut;
	mPath			= path;
	mRoot			= root;
	mDeviceName		= deviceName;

	mDiskState = kDiskStateMounted;
}

void
CPGPdiskDiskImp::ExportUserSet(CPGPdiskUserSet& userSet) const
{
	MakeSureHeadersRead();
	mHeaders.ExportUserSet(userSet);
}

void
CPGPdiskDiskImp::ImportUserSet(const CPGPdiskUserSet& userSet)
{
	pgpAssert(IsOpened());

	VerifyModifiability();

	mHeaders.ImportUserSet(userSet);
	mHeaders.WriteHeaders(mDiskFile);
}

void
CPGPdiskDiskImp::EnableCipheringOnUser(
	CPGPdiskUser&	user,
	const char		*passphrase,
	int				passkeylen)
{
	MakeSureHeadersRead();
	mHeaders.EnableCipheringOnUser(user, passphrase, passkeylen);
}

void
CPGPdiskDiskImp::WipeUsers()
{
	pgpAssert(IsOpened());

	if (AreUsersWiped())
		THROW_PGPERROR(kPGPError_UserAbort);

	VerifyModifiability();
	mHeaders.WipeUsers(mDiskFile);	// writes out headers for us
}

void
CPGPdiskDiskImp::SetUsesCustomTimeout(PGPBoolean usesCustom)
{
	pgpAssert(IsOpened());

	VerifyModifiability();

	mHeaders.SetUsesCustomTimeout(usesCustom);
	mHeaders.WriteHeaders(mDiskFile);
}

void
CPGPdiskDiskImp::SetDefaultRoot(const char *root)
{
	pgpAssert(IsOpened());

	VerifyModifiability();
	mHeaders.SetDefaultRoot(root);
}

void
CPGPdiskDiskImp::SetCustomTimeout(PGPUInt32 customTimeout)
{
	pgpAssert(IsOpened());

	VerifyModifiability();

	mHeaders.SetCustomTimeout(customTimeout);
	mHeaders.WriteHeaders(mDiskFile);
}

void
CPGPdiskDiskImp::ReEncrypt(
	const char					*adminPassphrase,
	PGPdiskEncryptionAlgorithm	newAlgorithm,
	ReEncryptDiskStatusFuncType	statusFunc,
	void						*userValue)
{
	pgpAssert(IsOpened());
	pgpAssertStrValid(adminPassphrase);
	pgpAssertAddrValid(statusFunc, ReEncryptDiskStatusFuncType);

	if (IsReadOnly())
		THROW_PGPERROR(kPGPClientError_DiskDiskIsWriteProtected);

	if (AreUsersWiped())
		THROW_PGPERROR(kPGPClientError_DiskDiskUsersAreWiped);

	CCipherContext	oldCipher, newCipher;
	PGPUInt64		numBlocksReEncrypted;

	if (!mHeaders.IsBeingReEncrypted())
		mHeaders.BeginReEncryption(mDiskFile, adminPassphrase, newAlgorithm);

	mHeaders.GetReEncryptionInfo(adminPassphrase, oldCipher, newCipher,
		numBlocksReEncrypted);

	ReEncryptBlocks(oldCipher, newCipher, statusFunc, userValue,
		numBlocksReEncrypted);

	mHeaders.FinishReEncryption(mDiskFile);
}

void
CPGPdiskDiskImp::LockVolume(PGPBoolean forFormat)
{
	if (!IsMounted())
		THROW_PGPERROR(kPGPError_FeatureNotAvailable);

	mLockHandle = CEngineSubsystems::DriverComm().LockVolume(Root(),
		forFormat);

	if (forFormat)
		mIsLockedForFormat = TRUE;
	else
		mIsLockedForIO = TRUE;
}

void
CPGPdiskDiskImp::UnlockVolume()
{
	if (!IsLockedForFormat() && !IsLockedForIO())
		THROW_PGPERROR(kPGPError_FeatureNotAvailable);

	CEngineSubsystems::DriverComm().UnlockVolume(mLockHandle);

	mIsLockedForFormat = FALSE;
	mIsLockedForIO = FALSE;
	mLockHandle = 0;
}

void
CPGPdiskDiskImp::ReadVolume(
	void		*buffer,
	PGPUInt64	blockPos,
	PGPUInt32	blocksLength)
{
	pgpAssertAddrValid(buffer, VoidAlign);

	if (!IsLockedForFormat() && !IsLockedForIO())
		THROW_PGPERROR(kPGPError_FeatureNotAvailable);

	CEngineSubsystems::DriverComm().IOVolume(mLockHandle, TRUE, buffer,
		blockPos, blocksLength);
}

void
CPGPdiskDiskImp::WriteVolume(
	const void	*buffer,
	PGPUInt64	blockPos,
	PGPUInt32	blocksLength)
{
	pgpAssertAddrValid(buffer, VoidAlign);

	if (!IsLockedForFormat() && !IsLockedForIO())
		THROW_PGPERROR(kPGPError_FeatureNotAvailable);

	CEngineSubsystems::DriverComm().IOVolume(mLockHandle, FALSE,
		const_cast<void *>(buffer), blockPos, blocksLength);
}

void
CPGPdiskDiskImp::Open(
	const CPGPdiskContext	*pContext,
	const char				*path,
	PGPBoolean				readOnly)
{
	pgpAssertAddrValid(pContext, CPGPdiskContext);

	if (IsMounted() || IsOpened())
		THROW_PGPERROR(kPGPError_FeatureNotAvailable);

	PGPUInt32	openFlags	= CFile::kNoFlags;

	if (readOnly)
		openFlags |= CFile::kReadOnlyFlag | CFile::kShareWriteFlag;

	mDiskFile.Open(path, openFlags);

	try
	{
		mHeaders.ReadHeaders(pContext, mDiskFile);
	}
	catch (CComboError&)
	{
		mDiskFile.Close();
		throw;
	}

	mPContext	= pContext;
	mIsReadOnly	= readOnly;
	mPath		= path;
	mDiskState	= kDiskStateOpened;

	mRoot.Empty();
}

void
CPGPdiskDiskImp::Close()
{
	if (!IsOpened())
		THROW_PGPERROR(kPGPError_FeatureNotAvailable);

	try
	{
		if (!IsReadOnly())
			mHeaders.WriteHeaders(mDiskFile);
	}
	catch (CComboError&) { }

	mHeaders.ClearHeaders();
	mDiskFile.Close();

	mPContext = NULL;
	mPath.Empty();

	mDiskState = kDiskStateNone;
}

void
CPGPdiskDiskImp::MakeSureHeadersRead() const
{
	if (!mHeaders.HeadersRead())
	{
		PGPBoolean	weOpened	= FALSE;

		try
		{
			if (!mDiskFile.IsOpened())
			{
				const_cast<CFile *>(&mDiskFile)->Open(Path(),
					CFile::kReadOnlyFlag | CFile::kShareWriteFlag);
				weOpened = TRUE;
			}

			const_cast<CPGPdiskHeaders *>(&mHeaders)->ReadHeaders(
				Context(), mDiskFile);

			if (weOpened)
				const_cast<CFile *>(&mDiskFile)->Close();
		}
		catch (CComboError&)
		{
			if (weOpened)
				const_cast<CFile *>(&mDiskFile)->Close();
		}
	}
}

void
CPGPdiskDiskImp::VerifyModifiability()
{
	if (IsReadOnly())
		THROW_PGPERROR(kPGPClientError_DiskDiskIsWriteProtected);

	if (IsBeingReEncrypted())
		THROW_PGPERROR(kPGPClientError_DiskDiskIsBeingReEncrypted);

	if (AreUsersWiped())
		THROW_PGPERROR(kPGPClientError_DiskDiskUsersAreWiped);
}

void
CPGPdiskDiskImp::ResetTimedOut()
{
	mHasTimedOut = FALSE;
	CEngineSubsystems::DriverComm().ResetTimedOutDisk(Root());
}

void
CPGPdiskDiskImp::Mount(
	const CPGPdiskContext	*pContext,
	const char				*path,
	const CPGPdiskUser&		diskUser,
	const char				*root,
	PGPBoolean				readOnly)
{
	if (!diskUser.IsCipheringEnabled())
		THROW_PGPERROR(kPGPError_BadParams);

	// Get cipher context of user.
	CCipherContext	cipher;
	diskUser.GetCipherContext(cipher);

	Mount(pContext, path, cipher, root, readOnly || diskUser.IsReadOnly());
}

void
CPGPdiskDiskImp::Mount(
	const CPGPdiskContext	*pContext,
	const char				*path,
	const CCipherContext&	cipher,
	const char				*root,
	PGPBoolean				readOnly)
{
	pgpAssertAddrValid(pContext, CPGPdiskContext);
	pgpAssertStrValid(path);
	pgpAssertStrValid(root);

	if (IsMounted() || IsOpened())
		THROW_PGPERROR(kPGPError_FeatureNotAvailable);

	// Open the PGPdisk file.
	PGPUInt32	openFlags	= CFile::kNoFlags;

	if (readOnly)
		openFlags |= CFile::kReadOnlyFlag;

	mDiskFile.Open(path, openFlags);

	try
	{
		// Read in list of headers.
		mHeaders.ReadHeaders(pContext, mDiskFile);

		if (IsBeingReEncrypted())
			THROW_PGPERROR(kPGPClientError_DiskDiskIsBeingReEncrypted);

		// Perform filesystem compatibility checks (forcing RO if needed).
		CheckFilesystemCompatibility(cipher, readOnly);

		// Save some information.
		PGPUInt64	numHeaderBlocks	= mHeaders.NumHeaderBlocks();
		PGPUInt64	numDataBlocks	= mHeaders.NumDataBlocks();

		// Close the PGPdisk file in preparation for mounting.
		if (!readOnly)
		{
			mHeaders.SetDefaultRoot(root);
			mHeaders.WriteHeaders(mDiskFile);
		}

		mDiskFile.Close();

		// Get exported key data in locked memory.
		CArray<PGPUInt8>	exportedContext(cipher.GetSizeForExport());
		cipher.Export(exportedContext.Get(), exportedContext.Size());

		// Perform the mount.
		CEngineSubsystems::DriverComm().Mount(path, root,
			cipher.Algorithm(), exportedContext.Get(),
			exportedContext.Size(), numHeaderBlocks, numDataBlocks,
			mDeviceName, readOnly);

		// set timeout
		try
		{
			if (UsesCustomTimeout())
			{
				CEngineSubsystems::DriverComm().SetTimeout(root,
					CustomTimeout());
			}
			else
			{
				PGPBoolean	autoUnmount		= pContext->PGPclPrefs().
					GetBoolean(kPGPPrefDiskUnmountOnInactivity);
				PGPUInt32	unmountTimeout	= pContext->PGPclPrefs().
					GetNumber(kPGPPrefDiskAutoUnmountTimeout);

				if (autoUnmount)
				{
					CEngineSubsystems::DriverComm().SetTimeout(root,
						unmountTimeout);
				}
			}
		}
		catch (CComboError&) {}

		mPContext = pContext;
		mIsReadOnly = readOnly;
		mPath = path;
		mRoot = root;

		mDiskState = kDiskStateMounted;
	}
	catch (CComboError&)
	{
		if (IsOpened())
			Close();
		throw;
	}
}

void
CPGPdiskDiskImp::Unmount(PGPBoolean isForced)
{
	if (!IsMounted())
		THROW_PGPERROR(kPGPError_FeatureNotAvailable);

	if (IsLockedForFormat() || IsLockedForIO())
		UnlockVolume();

	// Perform the unmount.
	CString	savedRoot(Root());

	CEngineSubsystems::DriverComm().Unmount(Root(), isForced);

	// Clear headers.
	mHeaders.ClearHeaders();

	mPContext = NULL;
	mPath.Empty();
	mRoot.Empty();

	mDiskState = kDiskStateNone;
}

void
CPGPdiskDiskImp::Create(
	const CPGPdiskContext		*pContext,
	const char					*path,
	PGPUInt64					blocksDisk,
	PGPdiskEncryptionAlgorithm	algorithm,
	const char					*adminPassphrase,
	const char					*userName,
	const char					*defaultRoot,
	NewDiskStatusFuncType		statusFunc,
	void						*userValue)
{
	if (IsMounted() || IsOpened())
		THROW_PGPERROR(kPGPError_FeatureNotAvailable);

	CreateCommon(pContext, path, blocksDisk, algorithm, adminPassphrase,
		userName, NULL, defaultRoot, statusFunc, userValue);
}

void
CPGPdiskDiskImp::Create(
	const CPGPdiskContext		*pContext,
	const char					*path,
	PGPUInt64					blocksDisk,
	PGPdiskEncryptionAlgorithm	algorithm,
	const CPGPKey&				pubKey,
	const char					*defaultRoot,
	NewDiskStatusFuncType		statusFunc,
	void						*userValue)
{
	if (IsMounted() || IsOpened())
		THROW_PGPERROR(kPGPError_FeatureNotAvailable);

	CreateCommon(pContext, path, blocksDisk, algorithm, NULL, NULL, &pubKey,
		defaultRoot, statusFunc, userValue);
}

void
CPGPdiskDiskImp::EncryptNewPGPdiskFile(
	const CCipherContext&	cipher,
	NewDiskStatusFuncType	statusFunc,
	void					*userValue)
{
	pgpAssert(!IsMounted());
	pgpAssert(!IsOpened());
	pgpAssert(mDiskFile.IsOpened());
	pgpAssert(cipher.IsKeyAttached());

	// Get the memory for our data buffer.
	CArray<PGPUInt8>	dataBuf(kBlocksPerWrite * kPGPdiskBlockSize);

	// Calculate the starting and ending blocks.
	PGPUInt64	startBlock	= mHeaders.NumHeaderBlocks();
	PGPUInt64	endBlock	= startBlock + mHeaders.NumDataBlocks();

	// Calculate information used for progress updates.
	PGPUInt32	totalWrites	= static_cast<PGPUInt32>(
		UMath::CeilDiv<PGPUInt64>(endBlock - startBlock, kBlocksPerWrite));

	PGPUInt32	writesDone		= 0;
	PGPUInt32	writesPerUpdate	= static_cast<PGPUInt32>(
		UMath::CeilDiv<PGPUInt64>(totalWrites, 100));

	// Encrypt the PGPdisk.
	for (PGPUInt64 i = startBlock; i < endBlock; i += kBlocksPerWrite)
	{
		PGPUInt32	blocksThisTime	= static_cast<PGPUInt32>(
			pgpMin(kBlocksPerWrite, endBlock - i));

		dataBuf.Wipe();

		// Encrypt the blocks and output them.
		cipher.EncryptCFB(i - startBlock, blocksThisTime, dataBuf.Get(),
			dataBuf.Get());

		mDiskFile.Write(dataBuf.Get(), i * kPGPdiskBlockSize,
			blocksThisTime * kPGPdiskBlockSize);

		// Call the progress function.
		PGPBoolean	continueCreate	= TRUE;

		if (writesDone++ % writesPerUpdate == 0)
		{
			continueCreate = statusFunc(FALSE, writesDone*100/totalWrites,
				userValue);
		}

		if (!continueCreate)
			THROW_PGPERROR(kPGPError_UserAbort);
	}

	statusFunc(FALSE, 100, userValue);
}

void
CPGPdiskDiskImp::CreateCommon(
	const CPGPdiskContext		*pContext,
	const char					*path,
	PGPUInt64					blocksDisk,
	PGPdiskEncryptionAlgorithm	algorithm,
	const char					*adminPass,
	const char					*userName,
	const CPGPKey				*pPubKey,
	const char					*defaultRoot,
	NewDiskStatusFuncType		statusFunc,
	void						*userValue)
{
	pgpAssertAddrValid(pContext, CPGPdiskContext);
	pgpAssert(blocksDisk > 0);
	pgpAssertStrValid(defaultRoot);
	pgpAssertAddrValid(statusFunc, NewDiskStatusFuncType);

	if (IsMounted() || IsOpened())
		THROW_PGPERROR(kPGPError_FeatureNotAvailable);

	mPContext = pContext;

	// Private key of public key must be present and usable
	if (IsntNull(pPubKey))
	{
		pgpAssertAddrValid(pPubKey, CPGPKey);

		if (!pPubKey->GetBooleanProp(kPGPKeyProperty_CanDecrypt))
			THROW_PGPERROR(kPGPClientError_BadDiskPassphrase);
	}

	try
	{
		mDiskFile.Open(path, CFile::kCreateIfFlag);
		mDiskFile.SetLength(0);

		// It's important we don't write out the headers until we're done
		// encrypting all the new disk - else we wait ages for NTFS to zero
		// out the new disk file when we write the admin user header to the
		// end of the disk.

		CCipherContext	cipher;

		if (IsntNull(pPubKey))
		{
			mHeaders.CreateHeaders(Context(), mDiskFile, blocksDisk,
				algorithm, *pPubKey, defaultRoot, cipher);
		}
		else
		{
			mHeaders.CreateHeaders(Context(), mDiskFile, blocksDisk,
				algorithm, adminPass, userName, defaultRoot, cipher);
		}

		mDiskFile.SetLength(mHeaders.NumFileBlocks() * kPGPdiskBlockSize);
		EncryptNewPGPdiskFile(cipher, statusFunc, userValue);
		mHeaders.WriteHeaders(mDiskFile);

		mPContext = pContext;
		mIsReadOnly = FALSE;
		mPath = path;
		mRoot.Empty();

		mDiskFile.Close();

		// Mount the disk.
/*		PGPBoolean	oldBrowsePref;

		try
		{
			// gets around a spurious Win2k warning dialog
			oldBrowsePref = mPContext->PGPclPrefs().GetBoolean(
				kPGPPrefDiskBrowseAfterMount);
			mPContext->PGPclPrefs().SetBoolean(kPGPPrefDiskBrowseAfterMount, 
				FALSE);
		}
		catch (CComboError&) { }*/
		
		Mount(pContext, path, cipher, defaultRoot, FALSE);

/*		try
		{
			mPContext->PGPclPrefs().SetBoolean(kPGPPrefDiskBrowseAfterMount, 
				oldBrowsePref);
		}
		catch (CComboError&) { }*/
	}
	catch (CComboError&)
	{
		mHeaders.ClearHeaders();

		if (mDiskFile.IsOpened())
			mDiskFile.Close();

		// work around warn on delete
		CString temp(path);
		temp += "not";

		CFile	deleter;

		deleter.Move(path, temp);
		deleter.Delete(temp);
		throw;
	}
}

void
CPGPdiskDiskImp::ReEncryptBlocks(
	const CCipherContext&		oldCipher,
	const CCipherContext&		newCipher,
	ReEncryptDiskStatusFuncType	statusFunc,
	void						*userValue,
	PGPUInt64					numBlocksReEncrypted)
{
	pgpAssert(IsOpened());
	pgpAssert(oldCipher.IsInitialized());
	pgpAssert(newCipher.IsInitialized());
	pgpAssertAddrValid(statusFunc, ReEncryptDiskStatusFuncType);

	// Get the memory for our data buffer.
	CArray<PGPUInt8>	dataBuf(kBlocksPerWrite * kPGPdiskBlockSize);

	// Calculate the starting and ending blocks.
	PGPUInt64	firstBlock	= mHeaders.NumHeaderBlocks();
	PGPUInt64	startBlock	= firstBlock + numBlocksReEncrypted;
	PGPUInt64	endBlock	= firstBlock +mHeaders.NumDataBlocks();

	// Calculate information used for progress updates.
	PGPUInt32	totalWrites	= static_cast<PGPUInt32>(
		UMath::CeilDiv<PGPUInt64>(mHeaders.NumDataBlocks(),
		kBlocksPerWrite));

	PGPUInt32	writesDone	= static_cast<PGPUInt32>(
		(totalWrites * numBlocksReEncrypted) / mHeaders.NumDataBlocks());
	PGPUInt32	writesPerUpdate	= UMath::CeilDiv<PGPUInt32>(totalWrites,
		100);

	// Re-encrypt the PGPdisk.
	for (PGPUInt64 i = startBlock; i < endBlock; i += kBlocksPerWrite)
	{
		PGPUInt32	blocksThisTime	= static_cast<PGPUInt32>(
			pgpMin(kBlocksPerWrite, endBlock - i));

		dataBuf.Wipe();

		// Read in the old blocks.
		mDiskFile.Read(dataBuf.Get(), i * kPGPdiskBlockSize,
			blocksThisTime * kPGPdiskBlockSize);

		// Decrypt the blocks using the old algorithm.
		oldCipher.DecryptCFB(i - firstBlock, blocksThisTime, dataBuf.Get(),
			dataBuf.Get());

		// Encrypt the blocks using the new algorithm.
		newCipher.EncryptCFB(i - firstBlock, blocksThisTime, dataBuf.Get(),
			dataBuf.Get());

		// Write out the newly ReEncrypted blocks.
		mDiskFile.Write(dataBuf.Get(), i * kPGPdiskBlockSize,
			blocksThisTime * kPGPdiskBlockSize);

		// Write out now the # of the last ReEncrypted block.
		numBlocksReEncrypted += blocksThisTime;
		mHeaders.WriteReEncryptionProgress(mDiskFile, numBlocksReEncrypted);

		// Call the progress function.
		PGPBoolean	continueReEncrypt	= TRUE;

		if (writesDone++ % writesPerUpdate == 0)
		{
			continueReEncrypt = statusFunc(writesDone*100/totalWrites,
				userValue);
		}

		if (!continueReEncrypt)
			THROW_PGPERROR(kPGPError_UserAbort);
	}

	statusFunc(100, userValue);
}
