/*____________________________________________________________________________
		Copyright (C) 2002 PGP Corporation
        All rights reserved.

        $Id: CPGPdiskHeaders.cpp,v 1.20 2002/11/21 18:42:55 wjb Exp $
____________________________________________________________________________*/

#include "pgpClassesConfig.h"
#include "pgpBuild.h"

#include "CArray.h"
#include "CFile.h"
#include "CString.h"
#include "UMath.h"
#include "UString.h"

#include "pgpClientErrors.h"

#include "CCipherContext.h"
#include "CCRC32.h"
#include "UCommonStrings.h"

#include "CPGPGlobalRandomPool.h"
#include "CPGPKey.h"

#include "CPGPdiskContext.h"
#include "CPGPdiskHeaders.h"
#include "CPGPdiskUser.h"
#include "CPGPdiskUserSet.h"
#include "PGPdiskOldFormat.h"

_USING_PGP

_UNNAMED_BEGIN

// Types

// Align to 1.
#if PGP_WIN32
#pragma pack(push, 1)
#endif

struct OnDiskHeaderInfo
{
	enum
	{
		kReservedHeaderBlocks	= 4, 
		kMaxBlocksHeader		= 128		// arbitrary
	};

#if defined(_M_IX86)
	enum 
	{
		kHeaderMagic	= 'dPGP',	// backwards since little-endian

		kMainType		= 'NIAM', 
		kUserType		= 'RESU', 
		kOldPubKeyType	= 'YEKP', 
	};
#endif	// _M_IX86

	PGPUInt32	headerMagic;	// Always OnDiskHeaderInfo::kHeaderMagic
	PGPUInt32	headerType;		// One of the HeaderType enums
	PGPUInt32	headerSize;		// Total size of this header, in bytes

	PGPUInt32	headerCRC;			// CRC of this header
	PGPUInt64	nextHeaderOffset;	// Offset to next header from file start
									// 0 = no additional headers
	PGPUInt32	reserved[2];
};

struct OnDiskMainHeader
{
	enum
	{
		kMajorVersion		= 6, 
		kMinorVersion		= 0, 

		kCompatMajorVersion		= 1, 
		kPreSubKeyMajorVersion	= 5, 
		kMaxSizeRoot			= 256, 
	};

	OnDiskHeaderInfo	hdrInfo;

	PGPUInt8	majorVersion;
	PGPUInt8	minorVersion;
	PGPUInt16	reserved;

	PGPUInt64	numFileBlocks;			// Size of the file, in blocks
	PGPUInt64	numHeaderBlocks;		// Size of leading header data
	PGPUInt64	numDataBlocks;			// Size of encrypted data

	PGPdiskEncryptionAlgorithm	algorithm;
	Crypto::PassphraseSalt		salt;

	PGPUInt32	isBeingReEncrypted 	: 1;	// is disk being reencrypted?
	PGPUInt32	isWiped				: 1;	// users wiped?
	PGPUInt32	usesCustomTimeout	: 1;	// custom timeout?
	PGPUInt32	unused				: 29;

	PGPUInt32	customTimeout;				// custom unmount timeout
	PGPUInt64	numBlocksReEncrypted;		// number of blocks reencrypted

	char		defaultRoot[kMaxSizeRoot];	// Preferred mount point
};

struct OnDiskUserHeader
{
	OnDiskHeaderInfo	hdrInfo;

	PGPUInt32	notExported		: 1;
	PGPUInt32	unused			: 31;

	PGPUInt32	exportedUserOffset;
	PGPUInt32	exportedUserSize;

	PGPUInt32	reserved[4];

	// Exported PGPdisk user structure goes here...
};

// Restore alignment.
#if PGP_WIN32
#pragma pack(pop)
#endif

_UNNAMED_END

_PGP_BEGIN

// Class CPGPdiskHeader

class CPGPdiskHeader : public CListableObject<CPGPdiskHeader>
{
	NOT_COPYABLE(CPGPdiskHeader);

public:
	CPGPdiskHeader(const CPGPdiskContext *pContext);
	virtual ~CPGPdiskHeader() { }

	const CPGPdiskContext *	Context() const {return mPContext;}

	OnDiskHeaderInfo *	GetAsHeaderInfo() const 
	{
		return reinterpret_cast<OnDiskHeaderInfo *>(mHeaderBytes.Get());
	}
	
	PGPUInt32&	HeaderMagic() const 
	{
		return GetAsHeaderInfo()->headerMagic;
	}

	PGPUInt32&	HeaderType() const 
	{
		return GetAsHeaderInfo()->headerType;
	}

	PGPUInt32&	HeaderSize() const 
	{
		return GetAsHeaderInfo()->headerSize;
	}

	PGPUInt32&	HeaderCRC() const 
	{
		return GetAsHeaderInfo()->headerCRC;
	}

	PGPUInt64&	NextHeaderOffset() const 
	{
		return GetAsHeaderInfo()->nextHeaderOffset;
	}

	PGPBoolean	IsMainType() const
	{
		return (HeaderType() == OnDiskHeaderInfo::kMainType);
	}

	PGPBoolean	IsUserType() const
	{
		return (HeaderType() == OnDiskHeaderInfo::kUserType);
	}

	PGPBoolean	IsOldPubKeyType() const
	{
		return (HeaderType() == OnDiskHeaderInfo::kOldPubKeyType);
	}

	void		UpdateCRC() 
	{
		HeaderCRC() = ComputeCRC();
	}

	virtual void	ReadAtOffset(const CFile& diskFile, PGPUInt64 offset);
	virtual void	WriteAtOffset(CFile& diskFile, PGPUInt64 offset);

	virtual void	Assign(const CPGPdiskHeader *pHeader);

protected:
	const CPGPdiskContext	*mPContext;
	CArray<PGPByte>			mHeaderBytes;

	PGPUInt32		ComputeCRC();

	void			ValidateHeaderSize();
	virtual void	CanonicalizeHeader();
};


// Class CPGPdiskMainHeader

class CPGPdiskMainHeader : public CPGPdiskHeader
{
public:
	CPGPdiskMainHeader(const CPGPdiskContext *pContext);
	~CPGPdiskMainHeader();

	OnDiskMainHeader *	GetAsMainHeader() const
	{
		return reinterpret_cast<OnDiskMainHeader *>(mHeaderBytes.Get());
	}

	PGPBoolean	IsOldHeaderAvailable() const
	{
		return mIsOldHeaderAvailable;
	}

	const PGPdiskOldFormat::MainHeader *	GetOldHeader() const
	{
		return mOldHeader;
	}

	PGPUInt8&	MajorVersion() const 
	{
		return GetAsMainHeader()->majorVersion;
	}

	PGPUInt8&	MinorVersion() const 
	{
		return GetAsMainHeader()->minorVersion;
	}

	PGPUInt64&	NumFileBlocks() const 
	{
		return GetAsMainHeader()->numFileBlocks;
	}

	PGPUInt64&	NumHeaderBlocks() const 
	{
		return GetAsMainHeader()->numHeaderBlocks;
	}

	PGPUInt64&	NumDataBlocks() const 
	{
		return GetAsMainHeader()->numDataBlocks;
	}

	PGPdiskEncryptionAlgorithm&	Algorithm() const
	{
		return GetAsMainHeader()->algorithm;
	}

	Crypto::PassphraseSalt&		Salt() const
	{
		return GetAsMainHeader()->salt;
	}

	PGPBoolean	GetIsBeingReEncrypted() const 
	{
		return GetAsMainHeader()->isBeingReEncrypted;
	}

	PGPBoolean	GetIsWiped() const 
	{
		return GetAsMainHeader()->isWiped;
	}

	PGPBoolean	GetUsesCustomTimeout() const 
	{
		return GetAsMainHeader()->usesCustomTimeout;
	}
	
	const char *	DefaultRoot() const 
	{
		return GetAsMainHeader()->defaultRoot;
	}

	PGPUInt32&	CustomTimeout() const 
	{
		return GetAsMainHeader()->customTimeout;
	}

	PGPUInt64&	NumBlocksReEncrypted() const 
	{
		return GetAsMainHeader()->numBlocksReEncrypted;
	}

	void	SetIsBeingReEncrypted(PGPBoolean isBeingReEncrypted) 
	{
		GetAsMainHeader()->isBeingReEncrypted = isBeingReEncrypted;
	}

	void	SetIsWiped(PGPBoolean isWiped) 
	{
		GetAsMainHeader()->isWiped = isWiped;
	}

	void	SetUsesCustomTimeout(PGPBoolean usesCustom) const 
	{
		GetAsMainHeader()->usesCustomTimeout = usesCustom;
	}

	void	SetDefaultRoot(const char *defaultRoot)
	{
		pgpAssertStrValid(defaultRoot);

		UString::SmartStringCopy(GetAsMainHeader()->defaultRoot, 
			defaultRoot, OnDiskMainHeader::kMaxSizeRoot);
	}

	void	Create(PGPUInt64 headerBlocks, PGPUInt64 dataBlocks, 
		PGPdiskEncryptionAlgorithm algorithm, 
		const Crypto::PassphraseSalt& salt, const char *defaultRoot);

private:
	PGPBoolean						mIsOldHeaderAvailable;
	PGPdiskOldFormat::MainHeader	*mOldHeader;

	void	ConvertOldVersion();
	void	CanonicalizeHeader();

	void	CommonCreate(PGPUInt64 headerBlocks, PGPUInt64 dataBlocks, 
		PGPdiskEncryptionAlgorithm algorithm, 
		const Crypto::PassphraseSalt& salt, const char *defaultRoot);
};


// Class CPGPdiskUserHeader

class CPGPdiskUserHeader : public CPGPdiskHeader
{
public:
	CPGPdiskUserHeader(const CPGPdiskContext *pContext);
	~CPGPdiskUserHeader() { }

	OnDiskUserHeader *	GetAsUserHeader() const
	{
		return reinterpret_cast<OnDiskUserHeader *>(mHeaderBytes.Get());
	}

	PGPUInt32&	ExportedUserOffset() const
	{
		return GetAsUserHeader()->exportedUserOffset;
	}

	PGPUInt32&	ExportedUserSize() const
	{
		return GetAsUserHeader()->exportedUserSize;
	}

	PGPBoolean	GetIsNotExported() const 
	{
		return GetAsUserHeader()->notExported;
	}

	void	SetIsNotExported(PGPBoolean isNotExported) 
	{
		GetAsUserHeader()->notExported = isNotExported;
	}
	
	void	GetUser(CPGPdiskUser& user);

	void	Create(const CPGPdiskUser& user);
	void	CreateFromOldKeyInfo(
		const PGPdiskOldCrypto::PassphraseKeyInfo *pOldKeyInfo, 
		PGPBoolean isAdmin = FALSE);

private:
	void	ConvertFromOldPubKeyHeader();
	void	CanonicalizeHeader();
};

_PGP_END

// Class CPGPdiskHeader member functions

CPGPdiskHeader::CPGPdiskHeader(const CPGPdiskContext *pContext) : 
	mPContext(pContext)
{
	  pgpAssertAddrValid(pContext, CPGPdiskContext);
}

void 
CPGPdiskHeader::ReadAtOffset(const CFile& diskFile, PGPUInt64 offset)
{
	pgpAssert(diskFile.IsOpened());

	// Read in at least one block.
	pgpAssert(sizeof(OnDiskHeaderInfo) <= kPGPdiskBlockSize);

	mHeaderBytes.Resize(kPGPdiskBlockSize);
	diskFile.Read(mHeaderBytes.Get(), offset, kPGPdiskBlockSize);

	// Validate size.
	ValidateHeaderSize();

	// Read in rest of header if necessary.
	if (HeaderSize() > kPGPdiskBlockSize)
	{
		mHeaderBytes.Resize(HeaderSize());
		diskFile.Read(mHeaderBytes.Get(), offset, HeaderSize());
	}

	// Make sure everything is kosher.
	CanonicalizeHeader();
}

void 
CPGPdiskHeader::WriteAtOffset(CFile& diskFile, PGPUInt64 offset)
{
	pgpAssert(diskFile.IsOpened());

	// Update header CRC.
	UpdateCRC();

	// Write it out.
	diskFile.Write(mHeaderBytes.Get(), offset, HeaderSize());

	// Clear bytes up to block boundary.
	if (HeaderSize()%kPGPdiskBlockSize > 0)
	{
		CArray<PGPUInt8>	blanks(kPGPdiskBlockSize);
		blanks.Wipe();

		diskFile.Write(blanks.Get(), offset + HeaderSize(), 
			kPGPdiskBlockSize - (HeaderSize()%kPGPdiskBlockSize));
	}
}

void 
CPGPdiskHeader::Assign(const CPGPdiskHeader *pHeader)
{
	pgpAssertAddrValid(pHeader, CPGPdiskHeader);

	mHeaderBytes.Resize(pHeader->HeaderSize());

	pgpCopyMemory(pHeader->mHeaderBytes.Get(), mHeaderBytes.Get(), 
		pHeader->HeaderSize());

	// Make sure everything is kosher.
	CanonicalizeHeader();
}

PGPUInt32 
CPGPdiskHeader::ComputeCRC()
{
	CCRC32	theCrc;

	// Always set CRC value to zero before computing.
	PGPUInt32	oldCRC = HeaderCRC();
	HeaderCRC() = 0;

	theCrc.Init();

	theCrc.Continue(reinterpret_cast<const PGPUInt32 *>(mHeaderBytes.Get()), 
		HeaderSize() / sizeof(PGPUInt32));

	HeaderCRC() = oldCRC;

	return theCrc.GetCRC();
}

void 
CPGPdiskHeader::ValidateHeaderSize()
{
	if ((HeaderSize() < sizeof(OnDiskHeaderInfo)) || 
		(HeaderSize() > OnDiskHeaderInfo::kMaxBlocksHeader * 
		kPGPdiskBlockSize))
	{
		THROW_PGPERROR(kPGPClientError_CorruptDiskHeader);
	}
}

void 
CPGPdiskHeader::CanonicalizeHeader()
{
	if (HeaderMagic() != OnDiskHeaderInfo::kHeaderMagic)
		THROW_PGPERROR(kPGPClientError_CorruptDiskHeader);

	if (HeaderCRC() != ComputeCRC())
		THROW_PGPERROR(kPGPClientError_CorruptDiskHeader);

	ValidateHeaderSize();
}


// Class CPGPdiskMainHeader member functions

CPGPdiskMainHeader::CPGPdiskMainHeader(const CPGPdiskContext *pContext) : 
	CPGPdiskHeader(pContext), mIsOldHeaderAvailable(FALSE), mOldHeader(NULL)
{
}

CPGPdiskMainHeader::~CPGPdiskMainHeader()
{
	if (IsOldHeaderAvailable())
	{
		delete mOldHeader;
		mOldHeader = NULL;
	}
}

void 
CPGPdiskMainHeader::Create(
	PGPUInt64						headerBlocks, 
	PGPUInt64						dataBlocks, 
	PGPdiskEncryptionAlgorithm		algorithm, 
	const Crypto::PassphraseSalt&	salt, 
	const char *					defaultRoot)
{
	CommonCreate(headerBlocks, dataBlocks, algorithm, salt, defaultRoot);
}

void 
CPGPdiskMainHeader::ConvertOldVersion()
{
	if (MajorVersion() >= OnDiskMainHeader::kPreSubKeyMajorVersion)
		return;

	// Store old header for retrieval by re-encrypters.
	if (IsOldHeaderAvailable())
	{
		delete mOldHeader;
		mOldHeader = NULL;
	}

	mOldHeader = new PGPdiskOldFormat::MainHeader;
	mIsOldHeaderAvailable = TRUE;

	pgpCopyMemory(mHeaderBytes.Get(), mOldHeader, 
		sizeof(PGPdiskOldFormat::MainHeader));

	// Fill in new header structure.
	Crypto::PassphraseSalt	salt;
	pgpClearMemory(&salt.bytes, sizeof(salt.bytes));

	pgpCopyMemory(mOldHeader->salt.bytes, salt.bytes, 
		sizeof(mOldHeader->salt.bytes));

	CString	root;
	root.Format("%c:\\", UString::NumberToLetter(mOldHeader->drive));

	CommonCreate(mOldHeader->numHeaderBlocks, mOldHeader->numDataBlocks, 
		mOldHeader->algorithm, salt, root);

	NextHeaderOffset() = mOldHeader->headerInfo.nextHeaderOffset;
}

void 
CPGPdiskMainHeader::CanonicalizeHeader()
{
	CPGPdiskHeader::CanonicalizeHeader();

	if (!IsMainType())
		THROW_PGPERROR(kPGPClientError_CorruptDiskHeader);

	OnDiskMainHeader	*pHeader	= GetAsMainHeader();

	if (MajorVersion() < OnDiskMainHeader::kCompatMajorVersion)
		THROW_PGPERROR(kPGPClientError_DiskTooOld);

	if (MajorVersion() < OnDiskMainHeader::kMajorVersion)
		ConvertOldVersion();

	if (MajorVersion() > OnDiskMainHeader::kMajorVersion)
		THROW_PGPERROR(kPGPClientError_DiskTooNew);

	if (NumDataBlocks() > NumFileBlocks())
		THROW_PGPERROR(kPGPClientError_CorruptDiskHeader);

	if (NumHeaderBlocks() + NumDataBlocks() > NumFileBlocks())
		THROW_PGPERROR(kPGPClientError_CorruptDiskHeader);
}

void 
CPGPdiskMainHeader::CommonCreate(
	PGPUInt64						headerBlocks, 
	PGPUInt64						dataBlocks, 
	PGPdiskEncryptionAlgorithm		algorithm, 
	const Crypto::PassphraseSalt&	salt, 
	const char						*defaultRoot)
{
	pgpAssertStrValid(defaultRoot);

	// Allocate header.
	mHeaderBytes.Resize(sizeof(OnDiskMainHeader));
	mHeaderBytes.Wipe();

	// Initialize fields.
	HeaderMagic()	= OnDiskHeaderInfo::kHeaderMagic;
	HeaderType()	= OnDiskHeaderInfo::kMainType;
	HeaderSize()	= sizeof(OnDiskMainHeader);

	MajorVersion() = OnDiskMainHeader::kMajorVersion;
	MinorVersion() = OnDiskMainHeader::kMinorVersion;

	Algorithm() = algorithm;

	NumHeaderBlocks()	= headerBlocks;
	NumDataBlocks()		= dataBlocks;
	NumFileBlocks()		= headerBlocks + dataBlocks;

	Salt() = salt;
	SetDefaultRoot(defaultRoot);
}


// Class CPGPdiskUserHeader member functions

CPGPdiskUserHeader::CPGPdiskUserHeader(const CPGPdiskContext *pContext) : 
	CPGPdiskHeader(pContext)
{
}

void 
CPGPdiskUserHeader::GetUser(CPGPdiskUser& user)
{
	user.Import(mHeaderBytes.Get() + ExportedUserOffset(), 
		ExportedUserSize());
}

void 
CPGPdiskUserHeader::Create(const CPGPdiskUser& user)
{
	pgpAssert(user.IsAssigned());

	// Expand header large enough to hold user.
	PGPUInt32	onDiskSize	= sizeof(OnDiskUserHeader) + 
		user.GetSizeForExport();

	mHeaderBytes.Resize(onDiskSize);
	mHeaderBytes.Wipe();

	// Initialize fields.
	HeaderMagic()	= OnDiskHeaderInfo::kHeaderMagic;
	HeaderType()	= OnDiskHeaderInfo::kUserType;
	HeaderSize()	= onDiskSize;

	// Get exported user.
	ExportedUserOffset() = sizeof(OnDiskUserHeader);
	ExportedUserSize() = user.GetSizeForExport();

	user.Export(mHeaderBytes.Get() + ExportedUserOffset(), 
		user.GetSizeForExport());
}

void 
CPGPdiskUserHeader::CreateFromOldKeyInfo(
	const PGPdiskOldCrypto::PassphraseKeyInfo	*pOldKeyInfo,  
	PGPBoolean									isAdmin)
{
	pgpAssertAddrValid(pOldKeyInfo, PGPdiskOldCrypto::PassphraseKeyInfo);

	// Re-encrypt crypto.
	Crypto::PassphraseKeyInfo	keyInfo;
	pgpClearMemory(&keyInfo, sizeof(keyInfo));

	pgpCopyMemory(pOldKeyInfo->encryptedKey.bytes, 
		keyInfo.encryptedKey.bytes, sizeof(pOldKeyInfo->encryptedKey));

	pgpCopyMemory(pOldKeyInfo->checkBytes.bytes, 
		keyInfo.checkBytes.bytes, sizeof(pOldKeyInfo->checkBytes));

	keyInfo.hashReps = pOldKeyInfo->hashReps;

	CPGPdiskUser	user(Context());

	user.ImportRawSymmetricKey(keyInfo, UCommonStrings::Get(
		UCommonStrings::kUnknownUserString), pOldKeyInfo->readOnly);

	if (isAdmin)
		user.SetIsAdminUser(TRUE);

	// Punt to Create();
	Create(user);
}

void 
CPGPdiskUserHeader::ConvertFromOldPubKeyHeader()
{
	PGPdiskOldFormat::PublicKeyHeader	*pOldHeader	= 
		reinterpret_cast<PGPdiskOldFormat::PublicKeyHeader *>(
		mHeaderBytes.Get());

	const PGPByte	*pExportedKeyID	= mHeaderBytes.Get() + 
		pOldHeader->keyIDOffset;
	const PGPByte	*pEncryptedKey	= mHeaderBytes.Get() + 
		pOldHeader->encryptedKeyOffset;

	Crypto::CheckBytes	checkBytes;
	pgpClearMemory(&checkBytes.bytes, sizeof(checkBytes.bytes));

	pgpCopyMemory(pOldHeader->checkBytes.bytes, checkBytes.bytes, 
		sizeof(pOldHeader->checkBytes.bytes));

	CPGPdiskUser	user(Context());

	user.ImportRawPublicKey(pExportedKeyID, pOldHeader->keyIDSize, 
		pEncryptedKey, pOldHeader->encryptedKeySize, checkBytes, 
		pOldHeader->algorithm, pOldHeader->readOnly, pOldHeader->locked);

	// Punt to Create();
	Create(user);
}

void 
CPGPdiskUserHeader::CanonicalizeHeader()
{
	CPGPdiskHeader::CanonicalizeHeader();

	if (IsOldPubKeyType())
		ConvertFromOldPubKeyHeader();

	if (!IsUserType())
		THROW_PGPERROR(kPGPClientError_CorruptDiskHeader);

	if (ExportedUserOffset() + ExportedUserSize() > HeaderSize())
		THROW_PGPERROR(kPGPClientError_CorruptDiskHeader);
}


// Class CPGPdiskHeaders member functions

CPGPdiskHeaders::~CPGPdiskHeaders()
{
	ClearHeaders();
}

PGPBoolean 
CPGPdiskHeaders::HasOldHeaders() const
{
	pgpAssert(HeadersRead());
	return MainHeader()->IsOldHeaderAvailable();
}

PGPUInt64 
CPGPdiskHeaders::NumFileBlocks() const
{
	pgpAssert(HeadersRead());
	return MainHeader()->NumFileBlocks();
}

PGPUInt64 
CPGPdiskHeaders::NumHeaderBlocks() const
{
	pgpAssert(HeadersRead());
	return MainHeader()->NumHeaderBlocks();
}

PGPUInt64 
CPGPdiskHeaders::NumDataBlocks() const
{
	pgpAssert(HeadersRead());
	return MainHeader()->NumDataBlocks();
}

PGPdiskEncryptionAlgorithm 
CPGPdiskHeaders::Algorithm() const
{
	pgpAssert(HeadersRead());
	return MainHeader()->Algorithm();
}

PGPBoolean 
CPGPdiskHeaders::IsBeingReEncrypted() const
{
	pgpAssert(HeadersRead());
	return MainHeader()->GetIsBeingReEncrypted();
}

PGPBoolean 
CPGPdiskHeaders::AreUsersWiped() const
{
	pgpAssert(HeadersRead());
	return MainHeader()->GetIsWiped();
}

PGPBoolean 
CPGPdiskHeaders::UsesCustomTimeout() const
{
	// Could happen if the network cable is unplugged--headers couldn't be read
	// This was reported to happen when PGPdisk is mounted from a network
	// drive, PGPkeys is opened to change passphrase, and then network is
	// disconnected before closing PGPkeys. 
	if(!HeadersRead())
	{
		return FALSE;
	}

	pgpAssert(HeadersRead());

	return MainHeader()->GetUsesCustomTimeout();
}

const char * 
CPGPdiskHeaders::DefaultRoot() const
{
	pgpAssert(HeadersRead());
	return MainHeader()->DefaultRoot();
}

PGPUInt32 
CPGPdiskHeaders::CustomTimeout() const
{
	pgpAssert(HeadersRead());
	return MainHeader()->CustomTimeout();
}

PGPUInt64 
CPGPdiskHeaders::BlocksDisk() const
{
	pgpAssert(HeadersRead());
	return MainHeader()->NumDataBlocks();
}

void 
CPGPdiskHeaders::SetUsesCustomTimeout(PGPBoolean usesCustom)
{
	pgpAssert(HeadersRead());
	MainHeader()->SetUsesCustomTimeout(usesCustom);
}

void 
CPGPdiskHeaders::SetCustomTimeout(PGPUInt32 customTimeout)
{
	pgpAssert(HeadersRead());
	MainHeader()->CustomTimeout() = customTimeout;
}

void 
CPGPdiskHeaders::SetDefaultRoot(const char * root)
{
	pgpAssert(HeadersRead());
	MainHeader()->SetDefaultRoot(root);
}

void 
CPGPdiskHeaders::ExportUserSet(CPGPdiskUserSet& userSet) const
{
	CPGPdiskHeader	*pHeader	= MainHeader();
	userSet.Empty();

	while (IsntNull(pHeader))
	{
		if (pHeader->IsUserType())
		{
			CPGPdiskUserHeader	*pUserHeader	= 
				reinterpret_cast<CPGPdiskUserHeader *>(pHeader);

			if (!pUserHeader->GetIsNotExported())
			{
				CPGPdiskUser	user(Context());
				pUserHeader->GetUser(user);

				userSet.Add(user);
			}
		}

		pHeader = mHeaders.Next(pHeader);
	}
}

void 
CPGPdiskHeaders::ImportUserSet(const CPGPdiskUserSet& userSet)
{
	CPGPdiskUser		*pUser;
	CPGPdiskUserIter	userIter(&userSet);

	if (userSet.Count() == 0)
		THROW_PGPERROR(kPGPError_BadParams);

	NukeAllUserHeaders();

	// First pass: ensure only one user is the admin.
	PGPBoolean	sawAdmin	= FALSE;

	while (TRUE)
	{
		pUser = userIter.Next();

		if (IsNull(pUser))
			break;

		if (pUser->IsAdminUser())
		{
			if (sawAdmin)
				pUser->SetIsAdminUser(FALSE);

			sawAdmin = TRUE;
		}
	}

	userIter.Rewind();

	// Second pass: create user headers.
	while (TRUE)
	{
		pUser = userIter.Next();

		if (IsNull(pUser))
			break;

		// If no admin was seen on first pass, first user becomes admin.
		if (!sawAdmin)
		{
			pUser->SetIsAdminUser(TRUE);
			sawAdmin = TRUE;
		}

		// Create and add new user header.
		auto_ptr<CPGPdiskUserHeader>	pUserHeader(
			new CPGPdiskUserHeader(Context()));

		pUserHeader->Create(*pUser);
		mHeaders.AddTail(pUserHeader.release());
	}
}

void 
CPGPdiskHeaders::EnableCipheringOnUser(
	CPGPdiskUser&	user, 
	const char		*passphrase,
	int				passkeylen)
{
	pgpAssert(HeadersRead());
	pgpAssert(user.IsAssigned());

	user.EnableCiphering(passphrase, passkeylen, Algorithm(), MainHeader()->Salt());
}

void 
CPGPdiskHeaders::WipeUsers(CFile& diskFile)
{
	pgpAssert(HeadersRead());
	pgpAssert(diskFile.IsOpened());

	// Remove all symmetric user headers.
	CPGPdiskUserSet	userSet, usersToRemove;
	ExportUserSet(userSet);

	CPGPdiskUserIter	userIter(&userSet);
	CPGPdiskUser		*pUser;
	PGPBoolean			sawPubKey	= FALSE;

	while (TRUE)
	{
		pUser = userIter.Next();

		if (IsNull(pUser))
			break;

		if (pUser->HasPublicKey())
			sawPubKey = TRUE;

		if (pUser->HasSymmetricKey())
			usersToRemove.Add(*pUser);
	}

	// Only proceed if at least one public key was seen.
	if (!sawPubKey)
		THROW_PGPERROR(kPGPClientError_NeedPubKeyToWipeDisk);

	// Remove and update users.
	userSet.Remove(usersToRemove);
	ImportUserSet(userSet);

	// Mark disk as wiped.
	MainHeader()->SetIsWiped(TRUE);

	// Make sure we write out now and flush.
	WriteHeaders(diskFile);
	diskFile.Flush();
}

void 
CPGPdiskHeaders::BeginReEncryption(
	CFile&						diskFile, 
	const char					*adminPassphrase, 
	PGPdiskEncryptionAlgorithm	newAlgorithm)
{
	pgpAssert(HeadersRead());
	pgpAssert(!IsBeingReEncrypted());
	pgpAssert(diskFile.IsOpened());
	pgpAssertStrValid(adminPassphrase);

	// Get set of users on disk.
	CPGPdiskUserSet	userSet;
	ExportUserSet(userSet);

	// Find current admin user and copy it.
	CPGPdiskUserIter	userIter(&userSet);
	CPGPdiskUser		oldAdmin(*userIter.FindAdminUser());

	// Create new session user and salt from random data.
	CCipherContext	newCipher(newAlgorithm);
	GenerateNewKeyAndSalt(newCipher);

	Crypto::PassphraseSalt	newSalt;
	newCipher.GetSalt(newSalt);

	// Convert all users to the new session key.
	userIter.Rewind();

	while (TRUE)
	{
		CPGPdiskUser	*pUser	= userIter.Next();

		if (IsNull(pUser))
			break;

		PGPBoolean	removeUser	= FALSE;

		if (pUser->IsAdminUser())
		{
			// New admin gets converted whether symmetric or not.
			pUser->ChangeCipher(newCipher, adminPassphrase);
		}
		else if (pUser->HasSymmetricKey())
		{
			// Non-admin symmetrics are hopeless, no passphrase.
			removeUser = TRUE;
		}
		else
		{
			// We have a non-admin public key user, attempt a conversion.
			try
			{
				pUser->ChangeCipher(newCipher, adminPassphrase);
			}
			catch (CComboError&)
			{
				if (pUser->IsLocked())
					THROW_PGPERROR(kPGPClientError_DiskReEncryptWithBadADK);

				removeUser = TRUE;
			}
		}

		if (removeUser)
			userSet.Remove(*pUser);
	}

	// Import the converted keys.
	ImportUserSet(userSet);

	// Re-create old admin user header.
	auto_ptr<CPGPdiskUserHeader>	pOldAdmin(
		new CPGPdiskUserHeader(Context()));

	pOldAdmin->Create(oldAdmin);
	pOldAdmin->SetIsNotExported(TRUE);

	// Create new main header.
	auto_ptr<CPGPdiskMainHeader>	pNewMain(
		new CPGPdiskMainHeader(Context()));

	pNewMain->Create(MainHeader()->NumHeaderBlocks(), 
		MainHeader()->NumDataBlocks(), newAlgorithm, newSalt, 
		MainHeader()->DefaultRoot());

	// Add old admin user header and new main header.
	mHeaders.InsertAfter(MainHeader(), pOldAdmin.release());
	mHeaders.AddHead(pNewMain.release());

	// Mark disk as being reencrypted.
	MainHeader()->SetIsBeingReEncrypted(TRUE);

	WriteHeaders(diskFile);
	diskFile.Flush();
}

void 
CPGPdiskHeaders::GetReEncryptionInfo(
	const char		*adminPassphrase, 
	CCipherContext&	oldCipher, 
	CCipherContext&	newCipher, 
	PGPUInt64&		numBlocksReEncrypted) const
{
	pgpAssert(HeadersRead());
	pgpAssert(IsBeingReEncrypted());
	pgpAssertStrValid(adminPassphrase);
	pgpAssert(!oldCipher.IsInitialized());
	pgpAssert(!newCipher.IsInitialized());

	CPGPdiskMainHeader	*pOldMain	= reinterpret_cast<CPGPdiskMainHeader *>(
		mHeaders.Next(MainHeader()));
	CPGPdiskUserHeader	*pOldAdmin	= reinterpret_cast<CPGPdiskUserHeader *>(
		mHeaders.Next(pOldMain));

	// Get old admin user.
	CPGPdiskUser	oldAdminUser(Context());
	pOldAdmin->GetUser(oldAdminUser);

	// Get old cipher context.
	// This is just conventional and the zero ok. -wjb
	oldAdminUser.EnableCiphering(adminPassphrase, 0, pOldMain->Algorithm(), 
		pOldMain->Salt());
	oldAdminUser.GetCipherContext(oldCipher);

	// Get new admin user.
	CPGPdiskUserSet	userSet;
	ExportUserSet(userSet);

	CPGPdiskUserIter	userIter(&userSet);
	CPGPdiskUser		*pNewAdminUser	= userIter.FindAdminUser();

	pgpAssertAddrValid(pNewAdminUser, CPGPdiskUser);

	// Get new cipher context.
	// This is just conventional and the zero ok. -wjb
	pNewAdminUser->EnableCiphering(adminPassphrase, 0,
		MainHeader()->Algorithm(), MainHeader()->Salt());
	pNewAdminUser->GetCipherContext(newCipher);

	// Get next block to reencrypt.
	numBlocksReEncrypted = MainHeader()->NumBlocksReEncrypted();
}

void 
CPGPdiskHeaders::WriteReEncryptionProgress(
	CFile&		diskFile, 
	PGPUInt64	numBlocksReEncrypted)
{
	pgpAssert(HeadersRead());
	pgpAssert(IsBeingReEncrypted());
	pgpAssert(diskFile.IsOpened());

	MainHeader()->NumBlocksReEncrypted() = numBlocksReEncrypted;

	WriteHeaders(diskFile);
	diskFile.Flush();
}

void 
CPGPdiskHeaders::FinishReEncryption(CFile& diskFile)
{
	pgpAssert(HeadersRead());
	pgpAssert(IsBeingReEncrypted());
	pgpAssert(diskFile.IsOpened());

	// Remove old main and admin user headers.
	CPGPdiskMainHeader	*pOldMain	= reinterpret_cast<CPGPdiskMainHeader *>(
		mHeaders.Next(MainHeader()));
	CPGPdiskUserHeader	*pOldAdmin	= reinterpret_cast<CPGPdiskUserHeader *>(
		mHeaders.Next(pOldMain));

	mHeaders.Remove(pOldMain);
	mHeaders.Remove(pOldAdmin);

	delete pOldMain;
	delete pOldAdmin;

	// Update new main header.
	MainHeader()->SetIsBeingReEncrypted(FALSE);

	WriteHeaders(diskFile);
	diskFile.Flush();
}

void 
CPGPdiskHeaders::ReadHeaders(
	const CPGPdiskContext	*pContext, 
	const CFile&			diskFile)
{
	pgpAssertAddrValid(pContext, CPGPdiskContext);
	pgpAssert(diskFile.IsOpened());

	mPContext = pContext;
	ClearHeaders();

	// Read in main header.
	auto_ptr<CPGPdiskMainHeader>	pMainHeader(
		new CPGPdiskMainHeader(pContext));

	pMainHeader->ReadAtOffset(diskFile, 0);
	mHeaders.AddTail(pMainHeader.release());

	// If this is an older PGPdisk, create symmetric user headers from main.
	if (pMainHeader->IsOldHeaderAvailable())
		CreateHeadersFromOldMainHeader(pMainHeader->GetOldHeader());

	// Read in the remaining PGPdisk headers.
	try
	{
		PGPUInt64	curOffset	= MainHeader()->NextHeaderOffset();

		while (curOffset != 0)
		{
			CPGPdiskHeader	genericHeader(pContext);

			// Read in current header.
			genericHeader.ReadAtOffset(diskFile, curOffset);

			// Generate and add header object of appropriate type.
			CopyAndAddTypedHeader(&genericHeader);

			// Determine offset of next header.
			curOffset = genericHeader.NextHeaderOffset();
		}
	}
	catch (CComboError&) { }

	mHeadersRead = TRUE;
}

void 
CPGPdiskHeaders::WriteHeaders(CFile& diskFile)
{
	pgpAssert(HeadersRead());
	pgpAssert(diskFile.IsOpened());

	PGPUInt64	blockAfterData	= MainHeader()->NumHeaderBlocks() + 
		 MainHeader()->NumDataBlocks();
	PGPUInt64	byteAfterData	= blockAfterData * kPGPdiskBlockSize;

	// Set file size to max we will need BEFORE beginning the writes. Headers
	// after the first go at the END of the file.

	CPGPdiskHeader	*pHeader	= mHeaders.Next(MainHeader());
	PGPUInt32		blocksTail	= 0;

	while (IsntNull(pHeader))
	{
		blocksTail += UMath::CeilDiv<PGPUInt32>(pHeader->HeaderSize(), 
			kPGPdiskBlockSize);
		pHeader = mHeaders.Next(pHeader);
	}

	diskFile.SetLength(byteAfterData + blocksTail * kPGPdiskBlockSize);

	// Update file size in main header.
	MainHeader()->NumFileBlocks() = MainHeader()->NumHeaderBlocks() + 
		MainHeader()->NumDataBlocks() + blocksTail;

	// Wipe reserved header blocks.
	CArray<PGPUInt8>	blanks(kPGPdiskBlockSize);
	blanks.Wipe();

	for (PGPUInt32 i = 0; i <  MainHeader()->NumHeaderBlocks(); i++)
	{
		diskFile.Write(blanks.Get(), i * kPGPdiskBlockSize, 
			kPGPdiskBlockSize);
	}

	// Calculate new offsets and write out all headers.
	MainHeader()->NextHeaderOffset() = byteAfterData;
	MainHeader()->WriteAtOffset(diskFile, 0);

	pHeader = mHeaders.Next(MainHeader());

	while (IsntNull(pHeader))
	{
		PGPUInt64	offset		= mHeaders.Prev(pHeader)->NextHeaderOffset();
		PGPUInt64	nextOffset	= offset + UMath::RoundUpToMultiple<
			PGPUInt32>(pHeader->HeaderSize(), kPGPdiskBlockSize);
			
		if (IsNull(mHeaders.Next(pHeader)))
			pHeader->NextHeaderOffset() = 0;
		else
			pHeader->NextHeaderOffset() = nextOffset;

		pHeader->WriteAtOffset(diskFile, offset);
		pHeader = mHeaders.Next(pHeader);
	}
}

void 
CPGPdiskHeaders::ClearHeaders()
{
	mHeaders.EmptyWithDelete();
	mHeadersRead = FALSE;
}

void 
CPGPdiskHeaders::CreateHeaders(
	const CPGPdiskContext		*pContext, 
	const CFile&				diskFile, 
	PGPUInt64					blocksDisk, 
	PGPdiskEncryptionAlgorithm	algorithm, 
	const char					*adminPassphrase, 
	const char					*userName, 
	const char					*defaultRoot, 
	CCipherContext&				cipher)
{
	ClearHeaders();

	CreateHeadersCommon(pContext, diskFile, blocksDisk, algorithm, 
		adminPassphrase, userName, NULL, defaultRoot, cipher);
}

void 
CPGPdiskHeaders::CreateHeaders(
	const CPGPdiskContext		*pContext, 
	const CFile&				diskFile, 
	PGPUInt64					blocksDisk, 
	PGPdiskEncryptionAlgorithm	algorithm, 
	const CPGPKey&				pubKey, 
	const char					*defaultRoot, 
	CCipherContext&				cipher)
{
	ClearHeaders();

	CreateHeadersCommon(pContext, diskFile, blocksDisk, algorithm, NULL, 
		NULL, &pubKey, defaultRoot, cipher);
}

void 
CPGPdiskHeaders::NukeAllUserHeaders()
{
	CPGPdiskHeader	*pHeader	= MainHeader();

	while (IsntNull(pHeader))
	{
		CPGPdiskHeader	*pNextHeader	= mHeaders.Next(pHeader);

		if (pHeader->IsUserType())
		{
			mHeaders.Remove(pHeader);
			delete pHeader;
		}

		pHeader = pNextHeader;
	}
}

void 
CPGPdiskHeaders::CopyAndAddTypedHeader(const CPGPdiskHeader *pHeader)
{
	pgpAssertAddrValid(pHeader, CPGPdiskHeader);

	// Allocate new header object according to type.
	auto_ptr<CPGPdiskHeader>	pNewHeader;

	if (pHeader->IsMainType())
	{
		pNewHeader = auto_ptr<CPGPdiskHeader>(
			new CPGPdiskMainHeader(Context()));
	}
	else if ((pHeader->IsUserType() || pHeader->IsOldPubKeyType()))
	{
		pNewHeader = auto_ptr<CPGPdiskHeader>(
			new CPGPdiskUserHeader(Context()));
	}
	else
	{
		THROW_PGPERROR(kPGPClientError_CorruptDiskHeader);
	}

	pNewHeader->Assign(pHeader);

	// Add the new header to the header list.
	mHeaders.AddTail(pNewHeader.release());
}

void 
CPGPdiskHeaders::CreateHeadersFromOldMainHeader(
	const PGPdiskOldFormat::MainHeader	*pOldHeader)
{
	pgpAssertAddrValid(pOldHeader, PGPdiskOldFormat::MainHeader);

	// For each passphrase, add a new user header.
	for (PGPUInt32 i = 0; 
		i <= PGPdiskOldFormat::MainHeader::kMaxAlternatePassphrases;
		i++)
	{
		const PGPdiskOldCrypto::PassphraseKeyInfo	*pUserInfo	= 
			&pOldHeader->passphrases[i];

		if (pUserInfo->inUse)
		{
			auto_ptr<CPGPdiskUserHeader>	pNewHeader(
				new CPGPdiskUserHeader(Context()));

			PGPBoolean	isAdmin	= (i == 0);
			pNewHeader->CreateFromOldKeyInfo(pUserInfo, isAdmin);

			mHeaders.AddTail(pNewHeader.release());
		}
	}
}

void 
CPGPdiskHeaders::GenerateNewKeyAndSalt(CCipherContext& cipher) const
{
	pgpAssert(cipher.IsInitialized());

	CPGPGlobalRandomPool	randomPool(Context()->PGPContext());
	CSecureArray<PGPUInt8>	keyData(cipher.GetNeededRandomDataSize());
	CSecureArray<PGPUInt8>	mungeData(cipher.GetNeededRandomDataSize());
	Crypto::PassphraseSalt	salt;

	randomPool.GetRandomBytes(keyData.Get(), keyData.Size());
	randomPool.GetRandomBytes(salt.bytes, sizeof(salt.bytes));
	randomPool.GetRandomBytes(mungeData.Get(), mungeData.Size());

	cipher.GenerateNewSymmetricKey(keyData.Get(), mungeData.Get(), 
		keyData.Size(), salt);
}

void 
CPGPdiskHeaders::CreateHeadersCommon(
	const CPGPdiskContext		*pContext, 
	const CFile&				diskFile, 
	PGPUInt64					blocksDisk, 
	PGPdiskEncryptionAlgorithm	algorithm, 
	const char					*adminPass, 
	const char					*userName, 
	const CPGPKey				*pPubKey, 
	const char					*defaultRoot, 
	CCipherContext&				cipher)
{
	pgpAssertAddrValid(pContext, CPGPdiskContext);
	pgpAssert(diskFile.IsOpened());
	pgpAssertStrValid(defaultRoot);

	mPContext = pContext;

	// Create new session user and salt from random data.
	if (!cipher.IsInitialized())
		cipher.Init(algorithm);

	Crypto::PassphraseSalt	salt;

	GenerateNewKeyAndSalt(cipher);
	cipher.GetSalt(salt);

	// Create main header.
	auto_ptr<CPGPdiskMainHeader>	pMainHeader(
		new CPGPdiskMainHeader(Context()));

	pMainHeader->Create(OnDiskHeaderInfo::kReservedHeaderBlocks, blocksDisk, 
		algorithm, salt, defaultRoot);

	// Create admin user header.
	CPGPdiskUser	user(Context());

	if (IsntNull(pPubKey))
		user.Create(cipher, *pPubKey);
	else
		user.Create(cipher, userName, adminPass);

	user.SetIsAdminUser(TRUE);

	auto_ptr<CPGPdiskUserHeader>	pAdminUserHeader(
		new CPGPdiskUserHeader(Context()));

	pAdminUserHeader->Create(user);

	// Add the headers to the disk
	mHeaders.AddHead(pAdminUserHeader.release());
	mHeaders.AddHead(pMainHeader.release());

	if(PGPclEnterprise())
	{
		// Create ADK header if we must.
		if (Context()->IsTherePGPdiskADK())
		{
			CPGPKey	adkKey;
			Context()->GetPGPdiskADKPubKey(adkKey);

			user.Create(cipher, adkKey, FALSE, TRUE);

			auto_ptr<CPGPdiskUserHeader>	pADKeyHeader(
				new CPGPdiskUserHeader(Context()));

			pADKeyHeader->Create(user);
			mHeaders.AddTail(pADKeyHeader.release());
		}
	}

	mHeadersRead = TRUE;
}
