/*____________________________________________________________________________
		Copyright (C) 2002 PGP Corporation
        All rights reserved.

        $Id: UNativeCryptoModule.cpp,v 1.11 2002/08/06 20:10:19 dallen Exp $
____________________________________________________________________________*/

#include "pgpClassesConfig.h"

#include "CSecureObject.h"
#include "UMath.h"
#include "UString.h"
#include "UTime.h"

#include "CCast5Cipher.h"
#include "CCipher.h"
#include "CSHA.h"
#include "CTwofishCipher.h"
#include "UCommonStrings.h"
#include "UNativeCryptoModule.h"
#include "pgpClientErrors.h"

_USING_PGP

_UNNAMED_BEGIN

// Constants

enum
{
	kBlockSizeIn32		= Crypto::kCipherBlockSize / sizeof(PGPUInt32), 
	kMaxHashReps		= 16000, 
	kMaxMsTimeForHash	= 500
};


// Macros

#define EXTRACT_CIPHER(keyHandle)										\
	CCipher	*pCipher	= static_cast<CCipher *>(keyHandle);			\
	pgpAssertAddrValid(pCipher, CCipher);


// Class CCipherWrapper

class CCipherWrapper SMART_ERROR_INHERIT
{
public:
	CCipherWrapper(CCipher *pCipher) : mPCipher(pCipher)
	{
		pgpAssertAddrValid(pCipher, CCipher);
	}

	CCipherWrapper(PGPdiskEncryptionAlgorithm algorithm) : 
		mWeCreated(FALSE), mPCipher(NULL)
	{
	#if PGP_EXCEPTIONS
		Make(algorithm);
		mWeCreated = TRUE;
	#else	// !PGP_EXCEPTIONS
		Status() = Make(algorithm);

		if (Status().IsntError())
			mWeCreated = TRUE;
	#endif	// PGP_EXCEPTIONS
	}

	~CCipherWrapper()
	{
		if (WeCreated())
		{
			delete mPCipher;
			mPCipher = NULL;
		}
	}

	operator CCipher *() {return mPCipher;}
	operator const CCipher *() const {return mPCipher;}

	CCipher *		operator->() {return mPCipher;}
	const CCipher *	operator->() const {return mPCipher;}

	CCipher&		operator*() {return *mPCipher;}
	const CCipher&	operator*() const {return *mPCipher;}

	CCipher *	Get() {return mPCipher;}
	PGPBoolean	WeCreated() const {return mWeCreated;}

	CCipher *	Release()
	{
		CCipher	*pCipher	= mPCipher;

		mWeCreated = FALSE;
		mPCipher = NULL;

		return pCipher;
	}

private:
	PGPBoolean	mWeCreated;
	CCipher		*mPCipher;

	SMART_ERROR	Make(PGPdiskEncryptionAlgorithm algorithm);
};


// Function declarations

void		GetAlgorithmIdString(PGPdiskEncryptionAlgorithm algorithm, 
	char *idString, PGPUInt32 availSize);

PGPUInt32	GetEncryptSize(Crypto::KeyHandle keyHandle);
PGPUInt32	GetDecryptSize(Crypto::KeyHandle keyHandle);
PGPUInt32	GetKeySize(Crypto::KeyHandle keyHandle);
PGPUInt32	GetNeededRandomDataSize(Crypto::KeyHandle keyHandle);
PGPUInt32	GetSizeForExport(Crypto::KeyHandle keyHandle);

CComboError	Validate(Crypto::KeyHandle keyHandle);

CComboError	Init(Crypto::KeyHandle *pKeyHandle, 
	PGPdiskEncryptionAlgorithm algorithm);
void		Cleanup(Crypto::KeyHandle keyHandle);

CComboError	AttachSymmetricKey(Crypto::KeyHandle keyHandle, 
	const Crypto::SymmetricKey *pKey, 
	const Crypto::PassphraseSalt *pSalt);
CComboError	GetSymmetricKey(Crypto::KeyHandle keyHandle, 
	Crypto::SymmetricKey *pKey);
CComboError	GetSalt(Crypto::KeyHandle keyHandle, 
	Crypto::PassphraseSalt *pSalt);

CComboError	FlipKeyBytes(Crypto::KeyHandle keyHandle);

CComboError	Encrypt(Crypto::KeyHandle keyHandle, const void *in, 
	void *out, PGPUInt32 numBlocks = 1);
CComboError	Decrypt(Crypto::KeyHandle keyHandle, const void *in, 
	void *out, PGPUInt32 numBlocks = 1);

CComboError	EncryptCFB(Crypto::KeyHandle keyHandle, 
	PGPUInt64 startBlockIndex, PGPUInt32 numBlocks, 
	const void *inBlocks, void *outBlocks);
CComboError	DecryptCFB(Crypto::KeyHandle keyHandle, 
	PGPUInt64 startBlockIndex, PGPUInt32 numBlocks, 
	const void *inBlocks, void *outBlocks);

CComboError	EncryptPassphraseKey(Crypto::KeyHandle keyHandle, 
	const char *passphrase, Crypto::PassphraseKeyInfo *pKeyInfo);
CComboError	DecryptPassphraseKey(Crypto::KeyHandle keyHandle, 
	const char *passphrase, const Crypto::PassphraseSalt *pSalt, 
	const Crypto::PassphraseKeyInfo *pKeyInfo);

CComboError	GenerateNewSymmetricKey(Crypto::KeyHandle keyHandle, 
	PGPUInt8 *randomData1, PGPUInt8 *randomData2, PGPUInt32 bytesData, 
	const Crypto::PassphraseSalt *pSalt);

CComboError	Export(Crypto::KeyHandle keyHandle, void *buffer, 
	PGPUInt32 sizeBuffer);
CComboError	Import(Crypto::KeyHandle keyHandle, const void *buffer, 
	PGPUInt32 sizeBuffer);

void		EncryptBlock(const CCipher *pCipher, PGPUInt64 blockNumber, 
	const PGPUInt32 *inBlock, PGPUInt32 *outBlock);
void		DecryptBlock(const CCipher *pCipher, PGPUInt64 blockNumber, 
	const PGPUInt32 *inBlock, PGPUInt32 *outBlock);

SMART_ERROR	HashSaltSchedulePassphrase(const char *passphrase, 
	const Crypto::PassphraseSalt *pSalt, CCipher *pCipher, 
	PGPUInt16& hashReps, PGPBoolean generateNewHashReps);


// Class CCipherWrapper member functions

SMART_ERROR 
CCipherWrapper::Make(PGPdiskEncryptionAlgorithm	algorithm)
{
	pgpAssert(!mWeCreated);

	SMART_ERROR_DECLARE

	if ((algorithm == kPGPdiskOldCAST5Algorithm) || 
		(algorithm == kPGPdiskCAST5Algorithm))
	{
		mPCipher = new CCast5Cipher(algorithm);
	}
	else if (algorithm == kPGPdiskTwoFishAlgorithm)
	{
		mPCipher = new CTwofishCipher(algorithm);
	}
	else
	{
	#if PGP_EXCEPTIONS
		THROW_PGPERROR(kPGPError_BadParams);
	#else	// !PGP_EXCEPTIONS
		error.pgpErr = kPGPError_BadParams;
	#endif	// PGP_EXCEPTIONS
	}
	
#if !PGP_EXCEPTIONS
	if (error.IsntError())
	{
		if (IsNull(mPCipher))
			error.pgpErr = kPGPError_OutOfMemory;
	}

	if (error.IsntError())
	{
		error = mPCipher->Status();

		if (error.IsError())
		{
			delete mPCipher;
			mPCipher = NULL;
		}
	}
#endif	// !PGP_EXCEPTIONS

	SMART_ERROR_RETURN
}


// Local function definitions.

void 
GetAlgorithmIdString(
	PGPdiskEncryptionAlgorithm	algorithm, 
	char						*idString, 
	PGPUInt32					availSize)
{
	using namespace UCommonStrings;
	pgpAssertAddrValid(idString, char);

	switch (algorithm)
	{
	case kPGPdiskCAST5Algorithm:
		UString::SmartStringCopy(idString, Get(kAlgorithmCast5IDString), 
			availSize);
		break;
	
	case kPGPdiskOldCAST5Algorithm:
		UString::SmartStringCopy(idString, Get(kAlgorithmOldCast5IDString), 
			availSize);
		break;

	case kPGPdiskTwoFishAlgorithm:
		UString::SmartStringCopy(idString, Get(kAlgorithmTwoFishIDString), 
			availSize);
		break;

	default:
		pgpAssert(FALSE);
		idString[0] = '\0';
		break;
	}
}

PGPUInt32 
GetEncryptSize(Crypto::KeyHandle keyHandle)
{
	EXTRACT_CIPHER(keyHandle)
	return pCipher->NumEncryptBytes();
}

PGPUInt32 
GetDecryptSize(Crypto::KeyHandle keyHandle)
{
	EXTRACT_CIPHER(keyHandle)
	return pCipher->NumDecryptBytes();
}

PGPUInt32 
GetKeySize(Crypto::KeyHandle keyHandle)
{
	EXTRACT_CIPHER(keyHandle)
	return pCipher->NumKeyBytes();
}

PGPUInt32 
GetNeededRandomDataSize(Crypto::KeyHandle keyHandle)
{
	EXTRACT_CIPHER(keyHandle)
	return pCipher->NumKeyBytes();
}

PGPUInt32 
GetSizeForExport(Crypto::KeyHandle keyHandle)
{
	EXTRACT_CIPHER(keyHandle)
	return pCipher->SizeExportedCipher();
}

CComboError 
Validate(Crypto::KeyHandle keyHandle)
{
	CComboError	error;
	EXTRACT_CIPHER(keyHandle)

#if PGP_EXCEPTIONS
	try
	{
		pCipher->Validate();
	}
	catch (CComboError& caughtErr)
	{
		error = caughtErr;
	}
#else	// !PGP_EXCEPTIONS
	error = pCipher->Validate();
#endif	// PGP_EXCEPTIONS

	return error;
}

CComboError 
Init(
	Crypto::KeyHandle			*pKeyHandle, 
	PGPdiskEncryptionAlgorithm	algorithm)
{
	CComboError	error;

	CCipher	**ppCipher	= reinterpret_cast<CCipher **>(pKeyHandle);
	pgpAssertAddrValid(ppCipher, CCipher *);

#if PGP_EXCEPTIONS
	try
	{
		CCipherWrapper	pCipher(algorithm);
		*ppCipher = pCipher.Release();
	}
	catch (CComboError& caughtErr)
	{
		error = caughtErr;
	}
#else	// !PGP_EXCEPTIONS
	CCipherWrapper	pCipher(algorithm);
	error = pCipher.Status();

	if (error.IsntError())
		*ppCipher = pCipher.Release();
#endif	// PGP_EXCEPTIONS

	return error;
}

void 
Cleanup(Crypto::KeyHandle keyHandle)
{
	EXTRACT_CIPHER(keyHandle)
	delete pCipher;
}

CComboError 
AttachSymmetricKey(
	Crypto::KeyHandle				keyHandle, 
	const Crypto::SymmetricKey		*pKey, 
	const Crypto::PassphraseSalt	*pSalt)
{
	CComboError	error;
	EXTRACT_CIPHER(keyHandle)

	pgpAssertAddrValid(pKey, Crypto::SymmetricKey);
	pgpAssertAddrValid(pSalt, Crypto::PassphraseSalt);

	pCipher->AttachKey(*pKey, *pSalt);

	return error;
}

CComboError 
GetSymmetricKey(Crypto::KeyHandle keyHandle, Crypto::SymmetricKey *pKey)
{
	pgpAssertAddrValid(pKey, Crypto::SymmetricKey);

	CComboError	error;
	EXTRACT_CIPHER(keyHandle)

	*pKey = pCipher->Key();

	return error;
}

CComboError 
GetSalt(Crypto::KeyHandle keyHandle, Crypto::PassphraseSalt *pSalt)
{
	pgpAssertAddrValid(pSalt, Crypto::PassphraseSalt);

	CComboError	error;
	EXTRACT_CIPHER(keyHandle)

	*pSalt = pCipher->Salt();

	return error;
}

CComboError 
FlipKeyBytes(Crypto::KeyHandle keyHandle)
{
	CComboError	error;
	EXTRACT_CIPHER(keyHandle)

	pCipher->FlipKeyBytes();

	return error;
}

CComboError 
Encrypt(
	Crypto::KeyHandle	keyHandle, 
	const void			*in, 
	void				*out, 
	PGPUInt32			numBlocks)
{
	pgpAssertAddrValid(in, VoidAlign);
	pgpAssertAddrValid(out, VoidAlign);

	CComboError	error;
	EXTRACT_CIPHER(keyHandle)

	pCipher->Encrypt(in, out, numBlocks);

	return error;
}

CComboError 
Decrypt(
	Crypto::KeyHandle	keyHandle, 
	const void			*in, 
	void				*out, 
	PGPUInt32			numBlocks)
{
	pgpAssertAddrValid(in, VoidAlign);
	pgpAssertAddrValid(out, VoidAlign);

	CComboError	error;
	EXTRACT_CIPHER(keyHandle)

	pCipher->Decrypt(in, out, numBlocks);

	return error;
}

CComboError 
EncryptCFB(
	Crypto::KeyHandle	keyHandle, 
	PGPUInt64			startBlockIndex, 
	PGPUInt32			numBlocks, 
	const void			*inBlocks, 
	void				*outBlocks)
{
	pgpAssertAddrValid(inBlocks, VoidAlign);
	pgpAssertAddrValid(outBlocks, VoidAlign);

	CComboError	error;
	EXTRACT_CIPHER(keyHandle)

	const PGPUInt8	*curInBlock		= static_cast<const PGPUInt8 *>(inBlocks);
	PGPUInt8		*curOutBlock	= static_cast<PGPUInt8 *>(outBlocks);

	for (PGPUInt32 blockIndex = 0; blockIndex < numBlocks; blockIndex++)
	{
		PGPUInt64 seedValue = startBlockIndex + blockIndex;

		EncryptBlock(pCipher, seedValue, 
			reinterpret_cast<const PGPUInt32 *>(curInBlock), 
			reinterpret_cast<PGPUInt32 *>(curOutBlock));

		curInBlock	+= Crypto::kCipherBlockSize;
		curOutBlock	+= Crypto::kCipherBlockSize;
	}

	return error;
}

CComboError 
DecryptCFB(
	Crypto::KeyHandle	keyHandle, 
	PGPUInt64			startBlockIndex, 
	PGPUInt32			numBlocks, 
	const void			*inBlocks, 
	void				*outBlocks)
{
	pgpAssertAddrValid(inBlocks, VoidAlign);
	pgpAssertAddrValid(outBlocks, VoidAlign);

	CComboError	error;
	EXTRACT_CIPHER(keyHandle)

	const PGPUInt8	*curInBlock		= static_cast<const PGPUInt8 *>(inBlocks);
	PGPUInt8		*curOutBlock	= static_cast<PGPUInt8 *>(outBlocks);

	for (PGPUInt32 blockIndex = 0; blockIndex < numBlocks; blockIndex++)
	{
		PGPUInt64 seedValue = startBlockIndex + blockIndex;

		DecryptBlock(pCipher, seedValue, 
			reinterpret_cast<const PGPUInt32 *>(curInBlock), 
			reinterpret_cast<PGPUInt32 *>(curOutBlock));

		curInBlock	+= Crypto::kCipherBlockSize;
		curOutBlock	+= Crypto::kCipherBlockSize;
	}

	return error;
}

CComboError 
EncryptPassphraseKey(
	Crypto::KeyHandle			keyHandle, 
	const char					*passphrase, 
	Crypto::PassphraseKeyInfo	*pKeyInfo)
{
	pgpAssertStrValid(passphrase);
	pgpAssertAddrValid(pKeyInfo, Crypto::PassphraseKeyInfo);

	CComboError	error;
	EXTRACT_CIPHER(keyHandle)

#if PGP_EXCEPTIONS
	try
#endif	// PGP_EXCEPTIONS
	{
		// Create new cipher out of passphrase.
		CCipherWrapper	pPassCipher(pCipher->Algorithm());

	#if !PGP_EXCEPTIONS
		error = pPassCipher.Status();

		if (error.IsntError())
	#endif	// !PGP_EXCEPTIONS
		{
			pgpClearMemory(pKeyInfo, sizeof(Crypto::PassphraseKeyInfo));

		#if PGP_EXCEPTIONS
			HashSaltSchedulePassphrase(passphrase, &pCipher->Salt(), 
				pPassCipher, pKeyInfo->hashReps, TRUE);
		#else	// !PGP_EXCEPTIONS
			error = HashSaltSchedulePassphrase(passphrase, &pCipher->Salt(), 
				pPassCipher, pKeyInfo->hashReps, TRUE);
		#endif	// PGP_EXCEPTIONS
		}

	#if !PGP_EXCEPTIONS
		if (error.IsntError())
	#endif	// !PGP_EXCEPTIONS
		{
			// Encrypt the session key with the key derived from passphrase.
			pgpClearMemory(pKeyInfo->encryptedKey.bytes, 
				sizeof(pKeyInfo->encryptedKey.bytes));

			pPassCipher->Encrypt(pCipher->Key().bytes, 
				pKeyInfo->encryptedKey.bytes, 
				pCipher->NumKeyBytes() / pCipher->NumEncryptBytes());

			// Construct the checkbytes.
			pgpClearMemory(pKeyInfo->checkBytes.bytes, 
				sizeof(pKeyInfo->checkBytes.bytes));

			pPassCipher->Encrypt(pPassCipher->Key().bytes, 
				pKeyInfo->checkBytes.bytes);
		}
	}
#if PGP_EXCEPTIONS
	catch (CComboError& caughtErr)
	{
		error = caughtErr;
	}
#endif	// PGP_EXCEPTIONS

	return error;
}

CComboError 
DecryptPassphraseKey(
	Crypto::KeyHandle				keyHandle, 
	const char						*passphrase, 
	const Crypto::PassphraseSalt	*pSalt, 
	const Crypto::PassphraseKeyInfo	*pKeyInfo)
{
	pgpAssertStrValid(passphrase);
	pgpAssertAddrValid(pSalt, Crypto::PassphraseSalt);
	pgpAssertAddrValid(pKeyInfo, Crypto::PassphraseKeyInfo);

	CComboError	error;
	EXTRACT_CIPHER(keyHandle)

#if PGP_EXCEPTIONS
	try
#endif	// PGP_EXCEPTIONS
	{
		// Create new cipher out of passphrase.
		CCipherWrapper	pPassCipher(pCipher->Algorithm());

	#if !PGP_EXCEPTIONS
		error = pPassCipher.Status();

		if (error.IsntError())
	#endif	// !PGP_EXCEPTIONS
		{
			PGPUInt16	hashReps	= pKeyInfo->hashReps;

		#if PGP_EXCEPTIONS
			HashSaltSchedulePassphrase(passphrase, pSalt, pPassCipher, 
				hashReps, FALSE);
		#else	// !PGP_EXCEPTIONS
			error = HashSaltSchedulePassphrase(passphrase, pSalt, 
				pPassCipher, hashReps, FALSE);
		#endif	// PGP_EXCEPTIONS
		}

	#if !PGP_EXCEPTIONS
		if (error.IsntError())
	#endif	// !PGP_EXCEPTIONS
		{
			// Verify that the passphrase is valid by examining check bytes.
			Crypto::CheckBytes	checkBytes;
			pgpClearMemory(checkBytes.bytes, sizeof(checkBytes.bytes));

			pPassCipher->Encrypt(pPassCipher->Key().bytes, checkBytes.bytes);

			if (!pgpMemoryEqual(checkBytes.bytes, 
				pKeyInfo->checkBytes.bytes, pCipher->NumEncryptBytes()))
			{
			#if PGP_EXCEPTIONS
				THROW_PGPERROR(kPGPClientError_BadDiskPassphrase);
			#else	// !PGP_EXCEPTIONS
				error.pgpErr = kPGPClientError_BadDiskPassphrase;
			#endif	// PGP_EXCEPTIONS
			}
		}

	#if !PGP_EXCEPTIONS
		if (error.IsntError())
	#endif	// !PGP_EXCEPTIONS
		{
			// Decrypt the encrypted session key.
			CSecureObject<Crypto::SymmetricKey>	pSessionKey;

		#if !PGP_EXCEPTIONS
			error = pSessionKey.Status();

			if (error.IsntError())
		#endif	// !PGP_EXCEPTIONS
			{
				pPassCipher->Decrypt(pKeyInfo->encryptedKey.bytes, 
					pSessionKey->bytes, pCipher->NumKeyBytes() / 
					pCipher->NumDecryptBytes());

				// Assign the final session key and salt to the keyholder.
				pCipher->AttachKey(*pSessionKey, *pSalt);
			}
		}
	}
#if PGP_EXCEPTIONS
	catch (CComboError& caughtErr)
	{
		error = caughtErr;
	}
#endif	// PGP_EXCEPTIONS

	return error;
}

CComboError 
GenerateNewSymmetricKey(
	Crypto::KeyHandle				keyHandle, 
	PGPUInt8						*randomData1, 
	PGPUInt8						*randomData2, 
	PGPUInt32						bytesRandomData, 
	const Crypto::PassphraseSalt	*pSalt)
{
	CComboError	error;
	EXTRACT_CIPHER(keyHandle)

	pgpAssertAddrValid(randomData1, PGPUInt8);
	pgpAssertAddrValid(randomData2, PGPUInt8);
	pgpAssertAddrValid(pSalt, Crypto::PassphraseSalt);

#if PGP_EXCEPTIONS
	try
#endif	// PGP_EXCEPTIONS
	{
		// Make sure we have enough random data.
		if (bytesRandomData < pCipher->NumKeyBytes())
		{
		#if PGP_EXCEPTIONS
			THROW_PGPERROR(kPGPError_BufferTooSmall);
		#else	// !PGP_EXCEPTIONS
			error.pgpErr = kPGPError_BufferTooSmall;
			return error;
		#endif	// PGP_EXCEPTIONS
		}

	#if !PGP_EXCEPTIONS
		if (error.IsntError())
	#endif	// !PGP_EXCEPTIONS
		{
			// Get buffer for keys.
			CSecureObject<Crypto::SymmetricKey>	pSymKey;

		#if !PGP_EXCEPTIONS
			error = pSymKey.Status();

			if (error.IsntError())
		#endif	// !PGP_EXCEPTIONS
			{
				// Create new cipher out of randomData2 to munge randomData1
				CCipherWrapper	pMungeCipher(pCipher->Algorithm());

			#if !PGP_EXCEPTIONS
				error = pMungeCipher.Status();

				if (error.IsntError())
			#endif	// !PGP_EXCEPTIONS
				{
					// Copy data for munge key.
					pgpCopyMemory(randomData2, pSymKey->bytes, 
						pMungeCipher->NumKeyBytes());
					pMungeCipher->AttachKey(*pSymKey, *pSalt);

					// Generate new session key using randomData1.
					pgpCopyMemory(randomData1, pSymKey->bytes, 
						pCipher->NumKeyBytes());

					// Munge the new session key.
					for (PGPUInt32 i = 0; i < 2; i++)
					{
						pMungeCipher->Encrypt(pSymKey->bytes, 
							pSymKey->bytes, pCipher->NumKeyBytes() / 
							pCipher->NumEncryptBytes());
					}

					// Assign the final session key and salt.
					pCipher->AttachKey(*pSymKey, *pSalt);
				}
			}
		}
	}
#if PGP_EXCEPTIONS
	catch (CComboError& caughtErr)
	{
		error = caughtErr;
	}
#endif	// PGP_EXCEPTIONS

	return error;
}

CComboError 
Export(
	Crypto::KeyHandle	keyHandle, 
	void				*buffer, 
	PGPUInt32			sizeBuffer)
{
	CComboError	error;
	EXTRACT_CIPHER(keyHandle);

#if PGP_EXCEPTIONS
	try
	{
		pCipher->ExportCipher(buffer, sizeBuffer);
	}
	catch (CComboError& caughtErr)
	{
		error = caughtErr;
	}
#else	// !PGP_EXCEPTIONS
	error = pCipher->ExportCipher(buffer, sizeBuffer);
#endif	// PGP_EXCEPTIONS

	return error;
}

CComboError 
Import(
	Crypto::KeyHandle	keyHandle, 
	const void			*buffer, 
	PGPUInt32			sizeBuffer)
{
	CComboError	error;
	EXTRACT_CIPHER(keyHandle);

#if PGP_EXCEPTIONS
	try
	{
		pCipher->ImportCipher(buffer, sizeBuffer);
	}
	catch (CComboError& caughtErr)
	{
		error = caughtErr;
	}
#else	// !PGP_EXCEPTIONS
	error = pCipher->ImportCipher(buffer, sizeBuffer);
#endif	// PGP_EXCEPTIONS

	return error;
}

void 
EncryptBlock(
	const CCipher	*pCipher, 
	PGPUInt64		blockNumber, 
	const PGPUInt32	*inBlock, 
	PGPUInt32		*outBlock)
{
	pgpAssertAddrValid(pCipher, CCipher);
	pgpAssert(pCipher->IsKeyAttached());
	pgpAssertAddrValid(inBlock, PGPUInt32);
	pgpAssertAddrValid(outBlock, PGPUInt32);
	pgpAssert(pCipher->NumIvsNeeded()%2 == 0);

	PGPInt32	i, j;
	PGPInt32	halfIvs	= pCipher->NumIvsNeeded() / 2;
	PGPUInt32	ivs[8], sums[4];

	pgpAssert(halfIvs <= 4);	// if not so, bump up size of above arrays

	for (i = 0; i < halfIvs; i++)
	{
		sums[i] = pCipher->Salt().dwords[i] + 
			UMath::GetHighDWord(blockNumber) + 
			UMath::GetLowDWord(blockNumber);
	}

	for (i = 0; i < kBlockSizeIn32 - halfIvs; i++)
	{
		// simple Fletcher checksum of block
		sums[halfIvs - 1] += inBlock[i];

		for (j = halfIvs - 2; j >= 0; j--)
		{
			sums[j] += sums[j+1];
		}
	}

	for (i = 0; i < halfIvs; i++)
	{
		ivs[i] = sums[i] ^ inBlock[kBlockSizeIn32 - halfIvs + i];
	}

	for (i = 0; i < halfIvs; i++)
	{
		ivs[i + halfIvs] = ~ivs[i];
	}

	pCipher->EncryptCFBdbl(ivs, inBlock, outBlock, 
		Crypto::kCipherBlockSize/(pCipher->NumEncryptBytes()*2));

	for (i = 0; i < halfIvs; i++)
	{
		outBlock[kBlockSizeIn32 - halfIvs + i] ^= sums[i];
	}
}

void 
DecryptBlock(
	const CCipher	*pCipher, 
	PGPUInt64		blockNumber, 
	const PGPUInt32	*inBlock, 
	PGPUInt32		*outBlock)
{
	pgpAssertAddrValid(pCipher, CCipher);
	pgpAssert(pCipher->IsKeyAttached());
	pgpAssertAddrValid(inBlock, PGPUInt32);
	pgpAssertAddrValid(outBlock, PGPUInt32);
	pgpAssert(pCipher->NumIvsNeeded()%2 == 0);

	PGPInt32	i, j;
	PGPInt32	halfIvs	= pCipher->NumIvsNeeded() / 2;
	PGPUInt32	ivs[8];

	pgpAssert(halfIvs <= 4);	// if not so, bump up size of above arrays

	PGPUInt32	kSplitPointIn32;
	size_t		kBytesBeforeSplit;
	size_t		kBytesAfterSplit;

	kSplitPointIn32		= kBlockSizeIn32 / 2;
	kBytesBeforeSplit	= kSplitPointIn32 * sizeof(PGPUInt32);
	kBytesAfterSplit	= (kBlockSizeIn32 - kSplitPointIn32) * 
		sizeof(PGPUInt32);

	// This must be done in two halves to recover the IV.
	for (i = 0; i < halfIvs*2; i++)
	{
		ivs[i] = inBlock[kSplitPointIn32 - halfIvs*2 + i];
	}

	pCipher->DecryptCFBdbl(ivs, inBlock + kSplitPointIn32, 
		outBlock + kSplitPointIn32, 
		kBytesAfterSplit/(pCipher->NumDecryptBytes()*2));

	for (i = 0; i < halfIvs; i++)
	{
		ivs[i] = outBlock[kBlockSizeIn32 - halfIvs + i];
	}

	for (i = 0; i < halfIvs; i++)
	{
		ivs[i + halfIvs] = ~ivs[i];
	}

	pCipher->DecryptCFBdbl(ivs, inBlock, outBlock, 
		kBytesBeforeSplit/(pCipher->NumDecryptBytes()*2));

	// Undo the checksum.
	for (i = 0; i < halfIvs; i++)
	{
		ivs[i] = pCipher->Salt().dwords[i] + 
			UMath::GetHighDWord(blockNumber) + 
			UMath::GetLowDWord(blockNumber);
	}

	for (i = 0; i < kBlockSizeIn32 - halfIvs; i++)
	{
		// simple Fletcher checksum of block
		ivs[halfIvs - 1] += outBlock[i];

		for (j = halfIvs - 2; j >= 0; j--)
		{
			ivs[j] += ivs[j+1];
		}
	}
	
	for (i = 0; i < halfIvs; i++)
	{
		outBlock[kBlockSizeIn32 - halfIvs + i] ^= ivs[i];
	}
}

SMART_ERROR 
HashSaltSchedulePassphrase(
	const char						*passphrase, 
	const Crypto::PassphraseSalt	*pSalt, 
	CCipher							*pCipher, 
	PGPUInt16&						hashReps, 
	PGPBoolean						generateNewHashReps)
{
	pgpAssertStrValid(passphrase);
	pgpAssertAddrValid(pSalt, Crypto::PassphraseSalt);
	pgpAssertAddrValid(pCipher, CCipher);

	SMART_ERROR_DECLARE

	// Get secure objects.
	CSecureObject<CSHA>					pHasher;
	CSecureObject<Crypto::SymmetricKey>	pNewKey;
	CSecureObject<CSHA::Digest>			pHashResult;

#if !PGP_EXCEPTIONS
	error = pHasher.Status();

	if (error.IsntError())
		error = pNewKey.Status();

	if (error.IsntError())
		error = pHashResult.Status();

	if (error.IsntError())
#endif	// !PGP_EXCEPTIONS
	{
		// Generate key material.
		PGPUInt32	bytesNeeded	= pCipher->NumKeyBytes();
		PGPUInt32	offset		= 0;		// offset into output buffer

		while (bytesNeeded > 0)
		{
			// # bytes key material we need this iteration.
			PGPUInt32	bytesThisTime	= pgpMin(
				sizeof(pHashResult->bytes), bytesNeeded);

			// Init hash for this new round.
			pHasher->Init();

			// Update with previous hash result if not on first hash.
			if (offset > 0)
			{
				pHasher->Update(pHashResult->bytes, 
					sizeof(pHashResult->bytes));
			}

			// Update with the bare passphrase.
			pHasher->Update(reinterpret_cast<const PGPUInt8 *>(passphrase), 
				strlen(passphrase));

			// Save the hashed passphrase for use below.
			pHasher->Final(pHashResult.Get());

			// Re-initialize the hash and update with the salt.
			pHasher->Init();
			pHasher->Update(pSalt->bytes, pCipher->NumSaltBytes());

			// Hash in the hashed passphrase a number of times.
			PGPUInt8	j;
			PGPUInt16	i;

			if (generateNewHashReps && (offset == 0))
			{
				// We need to calculate how many times.
				PGPUInt64 endTicks	= UTime::GetSystemTicks() + 
					kMaxMsTimeForHash;

				for (i = 0, j = 0; (UTime::GetSystemTicks() < 
					endTicks) && (i < kMaxHashReps); i++, j++)
				{
					pHasher->Update(pHashResult->bytes, bytesThisTime);
					pHasher->Update(&j, 1);
				}

				hashReps = i;
			}
			else
			{
				// We already know how many times.
				for (i = 0, j = 0; i < hashReps; i++, j++)
				{
					pHasher->Update(pHashResult->bytes, bytesThisTime);
					pHasher->Update(&j, 1);
				}
			}

			// Finalize the hash.
			pHasher->Final(pHashResult.Get());

			// Copy the newly generated key bytes to the output buffer.
			pgpCopyMemory(pHashResult->bytes, pNewKey->bytes + offset, 
				bytesThisTime);

			bytesNeeded -= bytesThisTime;
			offset += bytesThisTime;
		}

		// Initialize the cipher with the final key.
		pCipher->AttachKey(*pNewKey, *pSalt);
	}

	SMART_ERROR_RETURN
}

_UNNAMED_END


// UNativeCryptoModule functions

void 
UNativeCryptoModule::GetVTable(Crypto::CipherVTable& vTable)
{
	vTable.pfnGetAlgorithmIdString	= GetAlgorithmIdString;

	vTable.pfnGetEncryptSize			= GetEncryptSize;
	vTable.pfnGetDecryptSize			= GetDecryptSize;
	vTable.pfnGetKeySize				= GetKeySize;
	vTable.pfnGetNeededRandomDataSize	= GetNeededRandomDataSize;
	vTable.pfnGetSizeForExport			= GetSizeForExport;

	vTable.pfnValidate	= Validate;

	vTable.pfnInit		= Init;
	vTable.pfnCleanup	= Cleanup;

	vTable.pfnAttachSymmetricKey	= AttachSymmetricKey;
	vTable.pfnGetSymmetricKey		= GetSymmetricKey;
	vTable.pfnGetSalt				= GetSalt;

	vTable.pfnFlipKeyBytes	= FlipKeyBytes;

	vTable.pfnEncrypt		= Encrypt;
	vTable.pfnDecrypt		= Decrypt;
	vTable.pfnEncryptCFB	= EncryptCFB;
	vTable.pfnDecryptCFB	= DecryptCFB;

	vTable.pfnEncryptPassphraseKey		= EncryptPassphraseKey;
	vTable.pfnGenerateNewSymmetricKey	= GenerateNewSymmetricKey;
	vTable.pfnDecryptPassphraseKey		= DecryptPassphraseKey;
	vTable.pfnExport					= Export;
	vTable.pfnImport					= Import;
}
