/*____________________________________________________________________________
		Copyright (C) 2002 PGP Corporation
		All rights reserved.

		$Id: pgpDiskEngine.c,v 1.47 2002/11/18 02:21:06 jeffc Exp $
____________________________________________________________________________*/

#include <sys/types.h>
#include <pwd.h>
#include <sys/socket.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/un.h>
#include <errno.h>

#include "pflPrefs.h"
#include "pgpPFLPriv.h"
#include "pgpPubTypes.h"
#include "pgpMem.h"
#include "pgpErrors.h"
#include "pgpUtilities.h"
#include "pgpKeys.h"
#include "pgpDiskBase.h"

#include "pgpDiskEngine.h"
#include "pgpRPCPacker.h"

#define CKERR		if( IsPGPError( err ) ) goto done
#define CKNULL(x)	do { if( IsNull( x ) ) { err = kPGPError_OutOfMemory; goto done; } } while(0)
#define RETERR(x)	do { err = x; goto done; } while(0)

#define CKENGERR	if( IsPGPError( engerr ) ) RETERR( engerr )

/*
 * PGPdisk data structures
 */

typedef struct PGPdiskContextPriv
{
	PGPContextRef		context;
	PGPInt32			sock;
	char				sockPath[PATH_MAX];
} PGPdiskContextPriv;

typedef struct PGPdiskDiskPriv
{
	PGPdiskContextRef	diskcon;
	PGPUInt32			beDiskRef;
} PGPdiskDiskPriv;

typedef struct PGPdiskUserPriv
{
	PGPdiskContextRef	diskcon;
	PGPUInt32			beUserRef;
} PGPdiskUserPriv;

typedef struct PGPdiskUserSetPriv
{
	PGPdiskContextRef	diskcon;
	PGPdiskUserRef *	users;		/* Array of PGPdiskUserRefs */
	PGPUInt32			numUsers;
} PGPdiskUserSetPriv;

typedef struct PGPdiskUserIterPriv
{
	PGPdiskContextRef	diskcon;
	PGPdiskUserSetRef	userset;
	PGPInt32			index;
} PGPdiskUserIterPriv;

/*
 * Error handling constants
 */

typedef struct ErrorEntry
{
    PGPError const      number;
    const char * const  string;
} ErrorEntry;

#define EN(x, y)      { x, y }

static const ErrorEntry sPGPdiskErrors[] = 
{
EN( kPGPdiskError_NotInstalledCorrectly, "PGPdisk is not installed correctly" ),
EN( kPGPdiskError_DiskAlreadyMounted, "This PGPdisk has already been mounted" ),
EN( kPGPdiskError_DiskNotMounted, "This PGPdisk is not mounted" ),
EN( kPGPdiskError_CantUnmountDisk, "This PGPdisk cannot be unmounted" ),
EN( kPGPdiskError_DiskNeedsReEncryption, "This PGPdisk is only partially reencrypted" ),
EN( kPGPdiskError_OS9DiskNeedsConversion, "This Mac OS 9 PGPdisk needs to be converted to the new OS X PGPdisk format" ),
EN( kPGPdiskError_DiskIsReadOnly, "This read-only PGPdisk cannot be modified" ),
EN( kPGPdiskError_CantFormatDisk, "This PGPdisk cannot be formatted" ),
EN( kPGPdiskError_IsAdminUser, "User is adminstrator" ),
EN( kPGPdiskError_IsLocked, "User is locked" ),
EN( kPGPdiskError_IsDisabled, "User is disabled" )
};

#define kPGPdiskErrors_NumErrorTableEntries     \
    ( sizeof( sPGPdiskErrors ) / sizeof( sPGPdiskErrors[ 0 ] ) )

#undef EN

/*____________________________________________________________
	DLL init/exit functions 
  ____________________________________________________________*/

PGPError 
PGPdiskEngineInitLibrary()
{
	return kPGPError_NoErr;
}

PGPError 
PGPdiskEngineCleanupLibrary()
{
	return kPGPError_NoErr;
}

/*____________________________________________________________
	PGPdiskContextRef functions
  ____________________________________________________________*/
  
static PGPError
sConnectToEngine(
	PGPdiskContextRef diskCon )
{
	char				sockPath[PATH_MAX];
	PGPInt32			sock		= 0;
	struct sockaddr_un	sun;
	PGPInt32			retval		= 0;
		
	(void) PGPdiskGetSockPath( sockPath );
	
	sock = socket( AF_LOCAL, SOCK_STREAM, 0 );
	if( sock < 0 )
		goto done;
		
	sun.sun_family = AF_LOCAL;
	memcpy( sun.sun_path, sockPath, strlen( sockPath ) + 1 );
	
	retval = connect( sock, (struct sockaddr *) &sun, sizeof( sun ) );
	if( retval < 0 )
		goto done;
		
	diskCon->sock = sock;
	
done:
	if( retval < 0 )
	{
		if( sock > 0 )
			(void) close( sock );
		return kPGPError_UnknownError;
	}
	return kPGPError_NoErr;
}

static PGPError
sCreateIPCSocket(
	char *				sockPath,
	PGPInt32 *			outSock )
{
	PGPError			err			= kPGPError_NoErr;
	struct sockaddr_un	sun;
	PGPInt32			sock		= 0;
	PGPInt32			retval		= 0;
	
	sock = socket( AF_LOCAL, SOCK_STREAM, 0 );
	if( sock < 0 )
		return kPGPError_UnknownError;
		
	/* In case the sock is sitting around from a previous run */
	(void) unlink( sockPath );
	
	sun.sun_family = AF_LOCAL;
	pgpCopyMemory( sockPath, sun.sun_path, strlen( sockPath ) + 1 );
	
	retval = bind( sock, (struct sockaddr *) &sun, SUN_LEN( &sun ) );
	if( retval < 0 )
		return kPGPError_UnknownError;
		
	retval = listen( sock, 10 );
	if( retval < 0 )
		return kPGPError_UnknownError;
		
	*outSock = sock;

	return err;
}

PGPError 
PGPdiskNewContext(
	PGPContextRef context,
	PGPUInt32 diskAPIVersion,
   	PGPdiskContextRef *pContext )
{
	PGPError				err			= kPGPError_NoErr;
	PGPError				engerr		= kPGPError_NoErr;
	PGPdiskContextPriv *	diskCon		= NULL;
	
	PGPValidatePtr( pContext );

	if ((PGPdiskMajorVersion(diskAPIVersion) > 
		PGPdiskMajorVersion(kPGPdiskAPIVersion)) ||
		(PGPdiskMajorVersion(diskAPIVersion) == 0))
	{
		return kPGPError_IncompatibleAPI;
	}

	diskCon = PGPNewData( PGPPeekContextMemoryMgr( context ),
		sizeof( PGPdiskContextPriv ),
		kPGPMemoryMgrFlags_Clear );
	if( IsNull( diskCon ) )
		return kPGPError_OutOfMemory;

	pgpClearMemory( diskCon, sizeof( PGPdiskContextPriv ) );
	diskCon->context = context;
	
	/* Create communication socket */
	err = sConnectToEngine( diskCon ); CKERR;
		
	err = PGPrpcPackAndWrite( context, diskCon->sock,
		kPGPdiskEngineMsg_NewContext, "" ); CKERR;
	
	err = PGPrpcReadAndUnpack( context, diskCon->sock,
		kPGPdiskEngineMsg_NewContext, "u", &engerr );
	CKERR; CKENGERR;

	*pContext = diskCon;

done:
	return err;
}

PGPError 
PGPdiskFreeContext(
	PGPdiskContextRef context )
{
	/* Send message to engine so it can clean things up */
	(void) PGPrpcPackAndWrite( context->context, context->sock,
		kPGPdiskEngineMsg_FreeContext, "" );
	
	PGPValidatePtr( context );
	(void) PGPFreeData( context );

	return kPGPError_NoErr;
}

PGPError
PGPdiskLoadKeyrings(
	PGPdiskContextRef context,
	char *publicFilename,
	char *privateFilename )
{
	PGPError		err			= kPGPError_NoErr;
	PGPError		engerr		= kPGPError_NoErr;
	
	PGPdiskValidateContextRef( context );
	PGPValidatePtr( publicFilename );
	PGPValidatePtr( privateFilename );
	
	err = PGPrpcPackAndWrite( context->context, context->sock,
		kPGPdiskEngineMsg_LoadKeyrings, "ss", publicFilename, privateFilename ); CKERR;
	
	err = PGPrpcReadAndUnpack( context->context, context->sock,
		kPGPdiskEngineMsg_LoadKeyrings, "u", &engerr );
	CKERR; CKENGERR;
	
done:
	return err;
}

PGPContextRef 
PGPdiskPeekContextPGPContext(
	PGPdiskContextRef context )
{
	return context->context;
}

PGPError 
PGPdiskContextSetADK(
	PGPdiskContextRef context,
   	PGPKeyID * adkKeyID )
{
	PGPError		err			= kPGPError_NoErr;
	PGPError		engerr		= kPGPError_NoErr;
	
	PGPdiskValidateContextRef( context );
	PGPValidatePtr( adkKeyID );
	
	err = PGPrpcPackAndWrite( context->context, context->sock,
		kPGPdiskEngineMsg_SetADK, "B", adkKeyID, sizeof( PGPKeyID ) ); CKERR;
	
	err = PGPrpcReadAndUnpack( context->context, context->sock,
		kPGPdiskEngineMsg_SetADK, "u", &engerr );
	CKERR; CKENGERR;
	
done:
	return err;
}

PGPError
PGPdiskContextSetGlobalDiskTimeout(
	PGPdiskContextRef context,
	PGPUInt32 timeout )
{
	PGPError		err			= kPGPError_NoErr;
	PGPError		engerr		= kPGPError_NoErr;
	
	PGPdiskValidateContextRef( context );
	
	err = PGPrpcPackAndWrite( context->context, context->sock,
		kPGPdiskEngineMsg_SetGlobalDiskTimeout, "u", timeout ); CKERR;
	
	err = PGPrpcReadAndUnpack( context->context, context->sock,
		kPGPdiskEngineMsg_SetGlobalDiskTimeout, "u", &engerr );
	CKERR; CKENGERR;
	
done:
	return err;
}

PGPError
PGPdiskContextSetGlobalEjectOnLogout(
	PGPdiskContextRef context,
	PGPBoolean ejectOnLogout )
{
	PGPError		err			= kPGPError_NoErr;
	PGPError		engerr		= kPGPError_NoErr;
	
	PGPdiskValidateContextRef( context );
	
	err = PGPrpcPackAndWrite( context->context, context->sock,
		kPGPdiskEngineMsg_SetGlobalEjectOnLogout, "b", ejectOnLogout ); CKERR;
	
	err = PGPrpcReadAndUnpack( context->context, context->sock,
		kPGPdiskEngineMsg_SetGlobalEjectOnLogout, "u", &engerr );
	CKERR; CKENGERR;
	
done:
	return err;
}

PGPError
PGPdiskContextSetPassphraseCaching(
	PGPdiskContextRef context,
	PGPBoolean cache,
	PGPUInt32 howlong,
	PGPBoolean shareCache )
{
	PGPError		err			= kPGPError_NoErr;
	PGPError		engerr		= kPGPError_NoErr;
	
	PGPdiskValidateContextRef( context );
	
	err = PGPrpcPackAndWrite( context->context, context->sock,
		kPGPdiskEngineMsg_SetPassphraseCaching, "bub", cache, howlong, shareCache ); CKERR;
	
	err = PGPrpcReadAndUnpack( context->context, context->sock,
		kPGPdiskEngineMsg_SetPassphraseCaching, "u", &engerr );
	CKERR; CKENGERR;
	
done:
	return err;
}

PGPError
PGPdiskContextSetUnmountOnSleep(
	PGPdiskContextRef context,
	PGPBoolean unmountOnSleep )
{
	PGPError		err			= kPGPError_NoErr;
	PGPError		engerr		= kPGPError_NoErr;
	
	PGPdiskValidateContextRef( context );
	
	err = PGPrpcPackAndWrite( context->context, context->sock,
		kPGPdiskEngineMsg_SetUnmountOnSleep, "b", unmountOnSleep ); CKERR;
	
	err = PGPrpcReadAndUnpack( context->context, context->sock,
		kPGPdiskEngineMsg_SetUnmountOnSleep, "u", &engerr );
	CKERR; CKENGERR;
	
done:
	return err;
}

PGPError
PGPdiskContextSetForcedAutoDiskUnmount(
	PGPdiskContextRef context,
	PGPBoolean forced )
{
	PGPError		err			= kPGPError_NoErr;
	PGPError		engerr		= kPGPError_NoErr;
	
	PGPdiskValidateContextRef( context );
	
	err = PGPrpcPackAndWrite( context->context, context->sock,
		kPGPdiskEngineMsg_SetForcedAutoDiskUnmount, "b", forced ); CKERR;
	
	err = PGPrpcReadAndUnpack( context->context, context->sock,
		kPGPdiskEngineMsg_SetForcedAutoDiskUnmount, "u", &engerr );
	CKERR; CKENGERR;
	
done:
	return err;
}

PGPError
PGPdiskContextSetPreventSleep(
	PGPdiskContextRef context,
	PGPBoolean preventSleep )
{
	PGPError		err			= kPGPError_NoErr;
	PGPError		engerr		= kPGPError_NoErr;
	
	PGPdiskValidateContextRef( context );
	
	err = PGPrpcPackAndWrite( context->context, context->sock,
		kPGPdiskEngineMsg_SetPreventSleep, "b", preventSleep ); CKERR;
	
	err = PGPrpcReadAndUnpack( context->context, context->sock,
		kPGPdiskEngineMsg_SetPreventSleep, "u", &engerr );
	CKERR; CKENGERR;
	
done:
	return err;
}


/*____________________________________________________________
	PGPdiskUserRef functions
  ____________________________________________________________*/
  
static PGPError
sNewUser(
	PGPdiskContextRef	diskcon,
	PGPUInt32			beUserRef,
	PGPdiskUserRef *	pNewUser )
{
	PGPError			err			= kPGPError_NoErr;
	PGPdiskUserRef		newUser		= kInvalidPGPdiskUserRef;
	
	newUser = PGPNewData( PGPPeekContextMemoryMgr( diskcon->context ),
		sizeof( PGPdiskUserPriv ),
		kPGPMemoryMgrFlags_Clear );
	CKNULL( newUser );
	
	pgpClearMemory( newUser, sizeof( PGPdiskUserPriv ) );
	
	newUser->diskcon = diskcon;
	newUser->beUserRef = beUserRef;
	
	*pNewUser = newUser;
	
done:
	return err;
}

static PGPError 
sFreeUser(
	PGPdiskUserRef user )
{
	PGPdiskValidateUserRef( user );
	(void) PGPFreeData( user );
	return kPGPError_NoErr;
}

static PGPError 
sAddUserToUserSet(
	PGPdiskUserSetRef userSet, 
	PGPdiskUserRef userToAdd )
{
	PGPError		err			= kPGPError_NoErr;
	
	PGPdiskValidateUserSetRef( userSet );
	PGPdiskValidateUserRef( userToAdd );
	
	err = PGPReallocData( PGPPeekContextMemoryMgr( userSet->diskcon->context ),
		(void **) &userSet->users,
		( userSet->numUsers + 1 ) * sizeof( PGPdiskUserRef ),
		kPGPMemoryMgrFlags_Clear ); CKERR;
	
	userSet->users[userSet->numUsers] = userToAdd;
	userSet->numUsers++;
	
done:
	return err;
}


PGPError 
PGPdiskNewUserWithSymmetricKey(
	PGPdiskUserSetRef userSet, 
	const char *userName,
   	const char *passphrase,
   	PGPBoolean isReadOnly, 
	PGPdiskUserRef *pNewUser )
{
	PGPError			err			= kPGPError_NoErr;
	PGPError			engerr		= kPGPError_NoErr;
	PGPdiskUserRef		newUser		= kInvalidPGPdiskUserRef;
	PGPUInt32			beUserRef	= 0;
	PGPdiskContextRef	diskcon		= userSet->diskcon;
	
	PGPdiskValidateUserSetRef( userSet );
	PGPValidatePtr( userName );
	PGPValidatePtr( passphrase );
	PGPValidatePtr( pNewUser );
	
	err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock,
		kPGPdiskEngineMsg_NewUserWithSymmetricKey, "ssb", userName, passphrase, isReadOnly ); CKERR;
	
	err = PGPrpcReadAndUnpack( diskcon->context, diskcon->sock,
		kPGPdiskEngineMsg_NewUserWithSymmetricKey, "uu", &beUserRef, &engerr );
	CKERR; CKENGERR;
	
	err = sNewUser( diskcon, beUserRef, &newUser ); CKERR;
	
	err = sAddUserToUserSet( userSet, newUser ); CKERR;
	
	*pNewUser = newUser;

done:
	if( IsPGPError( err ) )
		(void) PGPFreeData( newUser );

	return err;
}

PGPError 
PGPdiskNewUserWithPublicKey(
	PGPdiskUserSetRef userSet, 
	PGPKeyID *pubKeyID,
   	PGPBoolean isReadOnly,
   	PGPBoolean isLocked, 
	PGPdiskUserRef *pNewUser )
{
	PGPError			err			= kPGPError_NoErr;
	PGPError			engerr		= kPGPError_NoErr;
	PGPdiskUserRef		newUser		= kInvalidPGPdiskUserRef;
	PGPUInt32			beUserRef	= 0;
	PGPdiskContextRef	diskcon		= userSet->diskcon;
	
	PGPdiskValidateUserSetRef( userSet );
	PGPValidatePtr( pNewUser );
	PGPValidatePtr( pubKeyID );
	
	err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock,
		kPGPdiskEngineMsg_NewUserWithPublicKey, "Bbb", pubKeyID, sizeof( PGPKeyID ), isReadOnly, isLocked ); CKERR;
	
	err = PGPrpcReadAndUnpack( diskcon->context, diskcon->sock,
		kPGPdiskEngineMsg_NewUserWithPublicKey, "uu", &beUserRef, &engerr );
	CKERR; CKENGERR;
	
	err = sNewUser( diskcon, beUserRef, &newUser ); CKERR;
	
	err = sAddUserToUserSet( userSet, newUser ); CKERR;
	
	*pNewUser = newUser;

done:
	if( IsPGPError( err ) )
		(void) PGPFreeData( newUser );

	return err;
}

PGPError 
PGPdiskDuplicateUser(
	PGPdiskUserRef user,
   	PGPdiskUserRef *pNewUser )
{
	PGPError		err			= kPGPError_NoErr;
	PGPdiskUserRef	newUser		= kInvalidPGPdiskUserRef;
	
	PGPdiskValidateUserRef( user );
	PGPValidatePtr( pNewUser );
	
	err = sNewUser( user->diskcon, user->beUserRef, &newUser ); CKERR;
	*pNewUser = newUser;

done:
	return err;
}

PGPdiskContextRef 
PGPdiskPeekUserContext(
	PGPdiskUserRef user )
{
	return user->diskcon;
}

PGPError 
PGPdiskUserGetPGPKeyID(
	PGPdiskUserRef user,
   	PGPKeyID *pKeyID )
{
	PGPError		err			= kPGPError_NoErr;
	PGPError		engerr		= kPGPError_NoErr;
	PGPUInt32		len			= 0;
	PGPByte *		buf			= NULL;

	PGPdiskValidateUserRef( user );
	PGPValidatePtr( pKeyID );

	err = PGPrpcPackAndWrite( user->diskcon->context, user->diskcon->sock,
		kPGPdiskEngineMsg_UserGetPGPKeyID, "u", user->beUserRef ); CKERR;
	err = PGPrpcReadAndUnpack( user->diskcon->context, user->diskcon->sock,
		kPGPdiskEngineMsg_UserGetPGPKeyID, "Bu", &buf, &len, &engerr );
	CKERR; CKENGERR;
	
	pgpCopyMemory( buf, pKeyID, pgpMin( sizeof( PGPKeyID ), len ) );

done:
	return err;
}

PGPError
PGPdiskIsUserPassphraseCorrect(
	PGPdiskUserRef user,
	const char *passphrase,
	PGPBoolean * isCorrect )
{
	PGPError		err			= kPGPError_NoErr;
	PGPError		engerr		= kPGPError_NoErr;
	PGPBoolean		bProp		= FALSE;
	
	PGPdiskValidateUserRef( user );
	PGPValidatePtr( passphrase );
	PGPValidatePtr( isCorrect );
	
	err = PGPrpcPackAndWrite( user->diskcon->context, user->diskcon->sock,
		kPGPdiskEngineMsg_IsUserPassphraseCorrect, "us", user->beUserRef, passphrase ); CKERR;
	err = PGPrpcReadAndUnpack( user->diskcon->context, user->diskcon->sock,
		kPGPdiskEngineMsg_IsUserPassphraseCorrect, "ub", &engerr, &bProp );
	CKERR; CKENGERR;
	
	*isCorrect = bProp;

done:
	return err;
}

PGPError 
PGPdiskUserChangePassphrase(
	PGPdiskUserRef user, 
	const char *oldPassphrase,
	const char *newPassphrase )
{
	PGPError		err			= kPGPError_NoErr;
	PGPError		engerr		= kPGPError_NoErr;
	
	PGPdiskValidateUserRef( user );
	PGPValidatePtr( oldPassphrase );
	PGPValidatePtr( newPassphrase );
	
	err = PGPrpcPackAndWrite( user->diskcon->context, user->diskcon->sock,
		kPGPdiskEngineMsg_UserChangePassphrase, "uss", user->beUserRef, oldPassphrase, newPassphrase ); CKERR;
	err = PGPrpcReadAndUnpack( user->diskcon->context, user->diskcon->sock,
		kPGPdiskEngineMsg_UserChangePassphrase, "u", &engerr );
	CKERR; CKENGERR;

done:
	return err;
}

PGPBoolean 
PGPdiskUserAreUsersEqual(
	PGPdiskUserRef user1,
   	PGPdiskUserRef user2 )
{
	return ( user1->beUserRef == user2->beUserRef );
}

PGPError 
PGPdiskGetUserBooleanProp(
	PGPdiskUserRef user,
   	PGPdiskUserPropName prop, 
	PGPBoolean *pBool )
{
	PGPError		err			= kPGPError_NoErr;
	PGPError		engerr		= kPGPError_NoErr;
	PGPBoolean		bProperty	= FALSE;
	
	PGPdiskValidateUserRef( user );
	PGPValidatePtr( pBool );
	
	err = PGPrpcPackAndWrite( user->diskcon->context, user->diskcon->sock,
		kPGPdiskEngineMsg_GetUserBooleanProp, "uu", user->beUserRef, prop ); CKERR;
	err = PGPrpcReadAndUnpack( user->diskcon->context, user->diskcon->sock,
		kPGPdiskEngineMsg_GetUserBooleanProp, "bu", &bProperty, &engerr );
	CKERR; CKENGERR;

	*pBool = bProperty;

done:
	return err;
}

PGPError 
PGPdiskGetUserNumericProp(
	PGPdiskUserRef user,
   	PGPdiskUserPropName prop, 
	PGPInt32 *pNumber )
{
	/* There are no numeric user properties at this time */
	return kPGPError_FeatureNotAvailable;
}

PGPError 
PGPdiskGetUserDataProp(
	PGPdiskUserRef user,
   	PGPdiskUserPropName prop, 
	void *pData,
   	PGPSize availLength,
   	PGPSize *pUsedLength )
{
	PGPError		err			= kPGPError_NoErr;
	PGPError		engerr		= kPGPError_NoErr;
	char *			pszProperty	= NULL;
	
	PGPdiskValidateUserRef( user );
	
	err = PGPrpcPackAndWrite( user->diskcon->context, user->diskcon->sock,
		kPGPdiskEngineMsg_GetUserDataProp, "uu", user->beUserRef, prop ); CKERR;
	err = PGPrpcReadAndUnpack( user->diskcon->context, user->diskcon->sock,
		kPGPdiskEngineMsg_GetUserDataProp, "su", &pszProperty, &engerr );
	CKERR; CKENGERR;
	
	if( IsntNull( pData ) )
		strncpy( pData, pszProperty, availLength );
	if( IsntNull( pUsedLength ) )
		*pUsedLength = pgpMin( strlen( pszProperty ), availLength );
		
	if( availLength < strlen( pszProperty ) + 1 )
		RETERR( kPGPError_BufferTooSmall );

done:
	if( IsntNull( pszProperty ) )
		(void) PGPFreeData( pszProperty );
	return err;
}

PGPError 
PGPdiskSetUserBooleanProp(
	PGPdiskUserRef user,
   	PGPdiskUserPropName prop, 
	PGPBoolean boolVal )
{
	PGPError		err			= kPGPError_NoErr;
	PGPError		engerr		= kPGPError_NoErr;
	
	PGPdiskValidateUserRef( user );
	
	err = PGPrpcPackAndWrite( user->diskcon->context, user->diskcon->sock,
		kPGPdiskEngineMsg_SetUserBooleanProp, "uub", user->beUserRef, prop, boolVal ); CKERR;
	err = PGPrpcReadAndUnpack( user->diskcon->context, user->diskcon->sock,
		kPGPdiskEngineMsg_SetUserBooleanProp, "u", &engerr );
	CKERR; CKENGERR;

done:
	return err;
}

PGPError 
PGPdiskSetUserNumericProp(
	PGPdiskUserRef user,
   	PGPdiskUserPropName prop, 
	PGPInt32 number )
{
	/* There are no numeric user properties at this time */
	return kPGPError_FeatureNotAvailable;
}

PGPError 
PGPdiskSetUserDataProp(
	PGPdiskUserRef user,
   	PGPdiskUserPropName prop, 
	const void *pData,
   	PGPSize length )
{
	PGPError		err			= kPGPError_NoErr;
	PGPError		engerr		= kPGPError_NoErr;
	
	PGPdiskValidateUserRef( user );
	
	err = PGPrpcPackAndWrite( user->diskcon->context, user->diskcon->sock,
		kPGPdiskEngineMsg_SetUserDataProp, "uus", user->beUserRef, prop, pData ); CKERR;
	err = PGPrpcReadAndUnpack( user->diskcon->context, user->diskcon->sock,
		kPGPdiskEngineMsg_SetUserDataProp, "u", &engerr );
	CKERR; CKENGERR;
	
done:
	return err;
}

/*____________________________________________________________
	PGPdiskUserSetRef functions
  ____________________________________________________________*/

PGPError 
PGPdiskNewUserSet(
	PGPdiskContextRef context,
	PGPdiskUserSetRef *pUserSet )
{
	PGPError				err			= kPGPError_NoErr;
	PGPdiskUserSetPriv *	userset		= NULL;
	
	PGPdiskValidateContextRef( context );
	PGPValidatePtr( pUserSet );
	
	userset = PGPNewData( PGPPeekContextMemoryMgr( context->context ),
		sizeof( PGPdiskUserSetPriv ),
		kPGPMemoryMgrFlags_Clear );
	CKNULL( userset );
	
	pgpClearMemory( userset, sizeof( PGPdiskUserSetPriv ) );
	
	userset->diskcon = context;
	
	*pUserSet = userset;
	
done:
	return err;
}

PGPError
PGPdiskFreeUserSet(
	PGPdiskUserSetRef userSet )
{
	PGPUInt32		i		= 0;
	
	PGPdiskValidateUserSetRef( userSet );
	
	if( IsntNull( userSet->users ) )
	{
		for( i = 0; i < userSet->numUsers; i++ )
			(void) sFreeUser( userSet->users[i] );
		(void) PGPFreeData( userSet->users );
	}
	(void) PGPFreeData( userSet );

	return kPGPError_NoErr;
}

PGPUInt32 
PGPdiskUserSetCount(
	PGPdiskUserSetRef userSet )
{
	return userSet->numUsers;
}

PGPError 
PGPdiskUserSetRemoveUser(
	PGPdiskUserSetRef userSet, 
	PGPdiskUserRef userToRemove )
{
	PGPError			err			= kPGPError_NoErr;
	PGPBoolean			bAdminUser	= FALSE;
	PGPBoolean			bLocked		= FALSE;
	PGPUInt32			i			= 0;
	PGPUInt32			j			= 0;
	PGPUInt32			numUsers	= 0;
	PGPdiskUserRef *	users		= NULL;
	
	/* Don't remove user if admin user or locked */
	err = PGPdiskGetUserBooleanProp( userToRemove, kPGPdiskUserProp_IsAdminUser, &bAdminUser ); CKERR;
	if( bAdminUser )
		RETERR( kPGPdiskError_IsAdminUser );
	err = PGPdiskGetUserBooleanProp( userToRemove, kPGPdiskUserProp_IsLocked, &bLocked ); CKERR;
	if( bLocked )
		RETERR( kPGPdiskError_IsLocked );
	
	numUsers = PGPdiskUserSetCount( userSet ) - 1;
	users = PGPNewData( PGPPeekContextMemoryMgr( userSet->diskcon->context ),
		numUsers * sizeof( PGPdiskUserRef ),
		kPGPMemoryMgrFlags_Clear );
	CKNULL( users );
	
	for( i = 0, j = 0; i < PGPdiskUserSetCount( userSet ); i++ )
	{
		if( PGPdiskUserAreUsersEqual( userSet->users[i], userToRemove ) )
			(void) sFreeUser( userSet->users[i] );
		else
			users[j++] = userSet->users[i];
	}
	
	(void) PGPFreeData( userSet->users );
	userSet->users = users;
	userSet->numUsers = numUsers;
	
done:
	return err;
}

/*____________________________________________________________
	PGPdiskUserIterRef functions
  ____________________________________________________________*/

PGPError 
PGPdiskNewUserIter(
	PGPdiskUserSetRef userSet,
   	PGPdiskUserIterRef *pUserIter )
{
	PGPError				err			= kPGPError_NoErr;
	PGPdiskUserIterPriv *	useriter	= NULL;
	
	PGPdiskValidateUserSetRef( userSet );
	PGPValidatePtr( pUserIter );
	
	useriter = PGPNewData( PGPPeekContextMemoryMgr( userSet->diskcon->context ),
		sizeof( PGPdiskUserIterPriv ),
		kPGPMemoryMgrFlags_Clear );
	CKNULL( useriter );
	
	pgpClearMemory( useriter, sizeof( PGPdiskUserIterPriv ) );
	
	useriter->diskcon = userSet->diskcon;
	useriter->userset = userSet;
	useriter->index = -1;	/* So that PGPdiskUserIterNext gives us index 0 */
	
	*pUserIter = useriter;
	
done:
	return err;
}

PGPError 
PGPdiskFreeUserIter(
	PGPdiskUserIterRef userIter )
{
	PGPdiskValidateUserIterRef( userIter );
	
	(void) PGPFreeData( userIter );

	return kPGPError_NoErr;
}

PGPdiskUserSetRef 
PGPdiskPeekUserIterUserSet(
	PGPdiskUserIterRef userIter )
{
	return userIter->userset;
}

PGPInt32 
PGPdiskUserIterIndex(
	PGPdiskUserIterRef userIter )
{
	PGPdiskValidateUserIterRef( userIter );
	return userIter->index;
}

PGPError 
PGPdiskUserIterMove(
	PGPdiskUserIterRef userIter,
   	PGPInt32 relOffset, 
	PGPBoolean fromStart,
   	PGPdiskUserRef *pUser )
{
	PGPError		err			= kPGPError_NoErr;
	
	PGPdiskValidateUserIterRef( userIter );
	PGPValidatePtr( pUser );
	
	if( fromStart )
	{
		if( relOffset < 0 )
			RETERR( kPGPError_BadParams );
		userIter->index = relOffset;
	}
	else
	{
		userIter->index += relOffset;
	}
	
	if( userIter->index >= userIter->userset->numUsers )
	{
		userIter->index = userIter->userset->numUsers;
		RETERR( kPGPError_EndOfIteration );
	}
	
	if( userIter->index < 0 )
	{
		userIter->index = 0;
		RETERR( kPGPError_EndOfIteration );
	}

	*pUser = userIter->userset->users[ userIter->index ];

done:
	return err;
}

PGPInt32 
PGPdiskUserIterSeek(
	PGPdiskUserIterRef userIter,
   	PGPdiskUserRef user )
{
	PGPError			err			= kPGPError_NoErr;
	PGPdiskUserRef		walk		= kInvalidPGPdiskUserRef;
	PGPInt32			index		= -1;
	
	(void) PGPdiskUserIterRewind( userIter );
	do
	{
		index++;
		err = PGPdiskUserIterNext( userIter, &walk );
	} while( IsntPGPError( err ) && !PGPdiskUserAreUsersEqual( user, walk ) );

	return IsPGPError( err ) ? -1 : index;
}

PGPError 
PGPdiskUserIterNext(
	PGPdiskUserIterRef userIter,
   	PGPdiskUserRef *pUser )
{
	return PGPdiskUserIterMove( userIter, 1, FALSE, pUser );
}

PGPError 
PGPdiskUserIterPrev(
	PGPdiskUserIterRef userIter,
   	PGPdiskUserRef *pUser )
{
	return PGPdiskUserIterMove( userIter, -1, FALSE, pUser );
}

PGPError 
PGPdiskUserIterRewind(
	PGPdiskUserIterRef userIter )
{
	PGPdiskValidateUserIterRef( userIter );

	userIter->index = -1;	/* So PGPdiskUserIterNext returns user at index 0 */
	return kPGPError_NoErr;
}

/*____________________________________________________________
	PGPdiskDiskRef functions
  ____________________________________________________________*/

PGPError 
PGPdiskOpenDisk(
	PGPdiskContextRef context,
   	const char *path, 
	PGPBoolean readOnly,
   	PGPdiskDiskRef *pDisk )
{
	PGPError		err			= kPGPError_NoErr;
	PGPError		engerr		= kPGPError_NoErr;
	PGPdiskDiskRef	disk		= kInvalidPGPdiskDiskRef;
	PGPUInt32		beDiskRef	= 0;
	
	PGPdiskValidateContextRef( context );
	PGPValidatePtr( path );
	PGPValidatePtr( pDisk );
	
	disk = PGPNewData( PGPPeekContextMemoryMgr( context->context ),
		sizeof( PGPdiskDiskPriv ),
		kPGPMemoryMgrFlags_Clear );
	CKNULL( disk );
	
	disk->diskcon = context;
	
	err = PGPrpcPackAndWrite( context->context, context->sock,
		kPGPdiskEngineMsg_OpenDisk, "su", path, readOnly ); CKERR;
	
	err = PGPrpcReadAndUnpack( context->context, context->sock,
		kPGPdiskEngineMsg_OpenDisk, "uu", &beDiskRef, &engerr );
	CKERR; CKENGERR;
	
	disk->beDiskRef = beDiskRef;
	
	*pDisk = disk;

done:
	if( IsPGPError( err ) )
		(void) PGPFreeData( disk );

	return err;
}

PGPError
PGPdiskCreateDisk(
	PGPdiskContextRef	context,
	const char *		path,
	PGPUInt64 			sizeBytes,
	const char *		defaultMountPoint,
	PGPdiskEncryptionAlgorithm	alg,
	PGPBoolean 			bPubKeyUser,
	PGPKeyID *	 		keyid,
	const char *		username,
	const char *		passphrase,
	PGPdiskNewDiskStatusFunc	callback,
   	void *				userValue,
	PGPdiskDiskRef *	pDisk )
{
	PGPError		err				= kPGPError_NoErr;
	PGPError		engerr			= kPGPError_NoErr;
	PGPError		cberr			= kPGPError_NoErr;
	PGPdiskDiskRef	disk			= kInvalidPGPdiskDiskRef;
	PGPSize			len				= 0;
	PGPUInt32		percent			= 0;
	PGPUInt32		beDiskRef		= 0;
	char			sockPath[PATH_MAX];
	struct passwd *	pw				= NULL;
	PGPInt32		callbackSock	= 0;
	PGPInt32		asock			= 0;
	struct sockaddr_un	sun;
	
	PGPdiskValidateContextRef( context );
	PGPValidateParam( !bPubKeyUser || IsntNull( keyid ) );
	PGPValidatePtr( path );
	PGPValidatePtr( pDisk );
	
	disk = PGPNewData( PGPPeekContextMemoryMgr( context->context ),
		sizeof( PGPdiskDiskPriv ),
		kPGPMemoryMgrFlags_Clear );
	CKNULL( disk );
	
	disk->diskcon = context;
	
	pw = getpwuid( getuid() );
	sprintf( sockPath, "/tmp/.pgpdisk-%s-%i-sock-XXXXXX", pw->pw_name, getuid() );
	(void) mktemp( sockPath );
	
	err = sCreateIPCSocket( sockPath, &callbackSock ); CKERR;
	
	err = PGPrpcPackAndWrite( context->context, context->sock,
		kPGPdiskEngineMsg_CreateDisk, "sUsubBsss",
		path,
		sizeBytes,
		defaultMountPoint,
		alg,
		bPubKeyUser,
		keyid, sizeof( PGPKeyID ),
		username,
		passphrase,
		sockPath ); CKERR;

	len = sizeof( sun );
	asock = accept( callbackSock, (struct sockaddr *) &sun, (int *) &len );
	if( asock < 0 )
		RETERR( kPGPError_UnknownError );

	do
	{
		err = PGPrpcReadAndUnpack( context->context, asock,
			kPGPdiskEngineMsg_CreateDiskCallback, "u", &percent ); CKERR;
		if( IsntNull( callback ) )
		{
			cberr = callback( FALSE, percent, userValue );
			
			/* Send the error back to the engine */
			err = PGPrpcPackAndWrite( context->context, asock,
				kPGPdiskEngineMsg_CreateDiskCallbackFinished, "u", cberr ); CKERR;
				
			if( cberr == kPGPError_UserAbort )
				break;
		}
	} while( percent < 100 );

	err = PGPrpcReadAndUnpack( context->context, context->sock,
		kPGPdiskEngineMsg_CreateDisk, "uu", &beDiskRef, &engerr );
	CKERR; CKENGERR;
	
	disk->beDiskRef = beDiskRef;
	
	*pDisk = disk;

done:
	if( asock > 0 )
		(void) close( asock );
	if( callbackSock > 0 )
		(void) close( callbackSock );
	(void) unlink( sockPath );

	if( IsPGPError( err ) )
		(void) PGPFreeData( disk );

	return err;
}

PGPError 
PGPdiskCloseDisk(
	PGPdiskDiskRef disk )
{
	PGPError			err			= kPGPError_NoErr;
	PGPError			engerr		= kPGPError_NoErr;
	PGPdiskContextRef	con			= kInvalidPGPdiskContextRef;
	
	PGPdiskValidateDiskRef( disk );

	con = PGPdiskPeekDiskContext( disk );
	
	err = PGPrpcPackAndWrite( con->context, con->sock,
		kPGPdiskEngineMsg_CloseDisk, "u", disk->beDiskRef ); CKERR;
		
	err = PGPrpcReadAndUnpack( con->context, con->sock,
		kPGPdiskEngineMsg_CloseDisk, "u", &engerr );
	CKERR; CKENGERR;
	
	if( PGPdiskDiskRefIsValid( disk ) )
		(void) PGPFreeData( disk );

done:
	return err;
}

PGPError
PGPdiskFormatDisk(
	PGPdiskDiskRef disk,
	const char * adminPassphrase,
	const char * volumeName,
	PGPdiskFilesystemType type )
{
	PGPError			err			= kPGPError_NoErr;
	PGPError			engerr		= kPGPError_NoErr;
	PGPdiskContextRef	con			= kInvalidPGPdiskContextRef;
	
	PGPdiskValidateDiskRef( disk );

	con = PGPdiskPeekDiskContext( disk );
	
	err = PGPrpcPackAndWrite( con->context, con->sock,
		kPGPdiskEngineMsg_FormatDisk, "ussu", disk->beDiskRef, adminPassphrase, volumeName, type ); CKERR;
		
	err = PGPrpcReadAndUnpack( con->context, con->sock,
		kPGPdiskEngineMsg_FormatDisk, "u", &engerr );
	CKERR; CKENGERR;
	
done:
	return err;
}
	
PGPdiskContextRef 
PGPdiskPeekDiskContext(
	PGPdiskDiskRef disk )
{
	return disk->diskcon;
}

PGPError
PGPdiskCanPassphraseMountDisk(
	PGPdiskDiskRef disk,
	const char *passphrase,
	PGPBoolean *canMount )
{
	PGPError			err			= kPGPError_NoErr;
	PGPError			engerr		= kPGPError_NoErr;
	PGPdiskContextRef	con			= kInvalidPGPdiskContextRef;
	PGPBoolean			bProp		= FALSE;

	PGPValidatePtr( canMount );
	
	con = PGPdiskPeekDiskContext( disk );
	
	err = PGPrpcPackAndWrite( con->context, con->sock,
		kPGPdiskEngineMsg_CanPassphraseMountDisk, "us",
		disk->beDiskRef, passphrase ); CKERR;
		
	err = PGPrpcReadAndUnpack( con->context, con->sock,
		kPGPdiskEngineMsg_CanPassphraseMountDisk, "bu", &bProp, &engerr );
	CKERR; CKENGERR;
	
	*canMount = bProp;
	
done:
	return err;
}

PGPError 
PGPdiskMountDiskAsUser(
	PGPdiskDiskRef disk,
   	const char *root,
	PGPdiskUserRef user,
	const char *passphrase,
   	PGPBoolean readOnly )
{
	PGPError			err			= kPGPError_NoErr;
	PGPError			engerr		= kPGPError_NoErr;
	PGPdiskContextRef	con			= kInvalidPGPdiskContextRef;

	con = PGPdiskPeekDiskContext( disk );
	
	err = PGPrpcPackAndWrite( con->context, con->sock,
		kPGPdiskEngineMsg_MountDisk, "ususb",
		disk->beDiskRef, root, user->beUserRef, passphrase, readOnly ); CKERR;
		
	err = PGPrpcReadAndUnpack( con->context, con->sock,
		kPGPdiskEngineMsg_MountDisk, "u", &engerr );
	CKERR; CKENGERR;
	
done:
	return err;
}

PGPError 
PGPdiskMountDisk(
	PGPdiskDiskRef disk,
   	const char *root, 
	const char *passphrase,
   	PGPBoolean readOnly )
{
	char				username[256];
	PGPSize				actualSize;
	PGPError			err			= kPGPError_NoErr;
	PGPdiskUserSetRef	userset		= kInvalidPGPdiskUserSetRef;
	PGPdiskUserIterRef	useriter	= kInvalidPGPdiskUserIterRef;
	PGPdiskUserRef		user		= kInvalidPGPdiskUserRef;
	
	PGPdiskValidateDiskRef( disk );
	
	err = PGPdiskNewUserSet( PGPdiskPeekDiskContext( disk ), &userset ); CKERR;
	err = PGPdiskDiskExportUserSet( disk, userset ); CKERR;
	err = PGPdiskNewUserIter( userset, &useriter ); CKERR;
	
	while( IsntPGPError( PGPdiskUserIterNext( useriter, &user ) ) )
	{
		err = PGPdiskGetUserDataProp(user, kPGPdiskUserProp_UserName, username, 256, &actualSize);
		err = PGPdiskMountDiskAsUser( disk,
			root, user, passphrase, readOnly );
		if( IsntPGPError( err ) )
		{
			/* Mounted successfully */
			break;
		}
	}
	CKERR;
	
done:
	if( PGPdiskUserSetRefIsValid( userset ) )
		(void) PGPdiskFreeUserSet( userset );
	if( PGPdiskUserIterRefIsValid( useriter ) )
		(void) PGPdiskFreeUserIter( useriter );
	return err;
}

PGPError 
PGPdiskUnmountDisk(
	PGPdiskDiskRef disk, 
	PGPBoolean isForced )
{
	PGPError		err			= kPGPError_NoErr;
	PGPError		engerr		= kPGPError_NoErr;
	
	PGPdiskValidateDiskRef( disk );
	
	err = PGPrpcPackAndWrite( disk->diskcon->context, disk->diskcon->sock,
		kPGPdiskEngineMsg_UnmountDisk, "ub", disk->beDiskRef, isForced ); CKERR;
	err = PGPrpcReadAndUnpack( disk->diskcon->context, disk->diskcon->sock,
		kPGPdiskEngineMsg_UnmountDisk, "u", &engerr );
	CKERR; CKENGERR;

done:
	return err;
}

PGPError 
PGPdiskDiskExportUserSet(
	PGPdiskDiskRef disk,
   	PGPdiskUserSetRef userSet )
{
	PGPError			err			= kPGPError_NoErr;
	PGPError			engerr		= kPGPError_NoErr;
	PGPdiskContextRef	con			= kInvalidPGPdiskContextRef;
	PGPByte *			buf			= NULL;
	PGPUInt32			size		= 0;
	PGPUInt32			i			= 0;
	PGPUInt32 *			beUserRefs	= NULL;
	
	PGPdiskValidateUserSetRef( userSet );

	con = PGPdiskPeekDiskContext( disk );
	
	err = PGPrpcPackAndWrite( con->context, con->sock,
		kPGPdiskEngineMsg_DiskExportUserSet, "u", disk->beDiskRef ); CKERR;
		
	err = PGPrpcReadAndUnpack( con->context, con->sock,
		kPGPdiskEngineMsg_DiskExportUserSet, "uB", &engerr, &buf, &size );
	CKERR; CKENGERR;
	
	userSet->numUsers = size / sizeof( PGPUInt32 );

	userSet->users = PGPNewData( PGPPeekContextMemoryMgr( disk->diskcon->context ),
		userSet->numUsers * sizeof( PGPdiskUserRef ),
		kPGPMemoryMgrFlags_Clear );
	CKNULL( userSet->users );
	
	beUserRefs = (PGPUInt32 *) buf;
	for( i = 0; i < userSet->numUsers; i++ )
	{
		err = sNewUser( disk->diskcon, beUserRefs[i], &userSet->users[i] );
		CKERR;
	}
	
done:
	if( IsntNull( buf ) )
		(void) PGPFreeData( buf );
	return err;
}

PGPError 
PGPdiskDiskImportUserSet(
	PGPdiskDiskRef disk,
	const char *adminPassphrase,
   	PGPdiskUserSetRef userSet, 
	PGPdiskUserRef newAdminUser )
{
	PGPError			err			= kPGPError_NoErr;
	PGPError			engerr		= kPGPError_NoErr;
	PGPdiskContextRef	con			= kInvalidPGPdiskContextRef;
	PGPUInt32 *			beUserRefs	= NULL;
	PGPUInt32			i			= 0;
	
	PGPdiskValidateUserSetRef( userSet );

	con = PGPdiskPeekDiskContext( disk );

	beUserRefs = PGPNewData( PGPPeekContextMemoryMgr( disk->diskcon->context ),
		userSet->numUsers * sizeof( PGPUInt32 ),
		kPGPMemoryMgrFlags_Clear );
	CKNULL( beUserRefs );
	
	for( i = 0; i < userSet->numUsers; i++ )
		beUserRefs[i] = userSet->users[i]->beUserRef;
	
	err = PGPrpcPackAndWrite( con->context, con->sock,
		kPGPdiskEngineMsg_DiskImportUserSet, "usBu",
		disk->beDiskRef,
		adminPassphrase,
		beUserRefs, userSet->numUsers * sizeof( PGPUInt32 ),
		newAdminUser->beUserRef ); CKERR;
		
	err = PGPrpcReadAndUnpack( con->context, con->sock,
		kPGPdiskEngineMsg_DiskImportUserSet, "u", &engerr );
	CKERR; CKENGERR;
	
done:
	if( IsntNull( beUserRefs ) )
		(void) PGPFreeData( beUserRefs );

	return err;
}

PGPError 
PGPdiskReEncryptDisk(
	PGPdiskDiskRef disk,
   	const char *adminPassphrase, 
	PGPdiskEncryptionAlgorithm newAlgorithm, 
	PGPdiskReEncryptDiskStatusFunc statusFunc,
   	void *userValue )
{
	PGPError			err				= kPGPError_NoErr;
	PGPError			engerr			= kPGPError_NoErr;
	PGPError			cberr			= kPGPError_NoErr;
	PGPdiskContextRef	diskcon			= kInvalidPGPdiskContextRef;
	PGPSize				len				= 0;
	PGPUInt32			percent			= 0;
	char				sockPath[PATH_MAX];
	struct passwd *		pw				= NULL;
	PGPInt32			callbackSock	= 0;
	PGPInt32			asock			= 0;
	struct sockaddr_un	sun;
	
	PGPdiskValidateDiskRef( disk );
	
	diskcon = disk->diskcon;

	pw = getpwuid( getuid() );
	sprintf( sockPath, "/tmp/.pgpdisk-%s-%i-sock-XXXXXX", pw->pw_name, getuid() );
	(void) mktemp( sockPath );
	
	err = sCreateIPCSocket( sockPath, &callbackSock ); CKERR;

	err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock,
		kPGPdiskEngineMsg_ReEncryptDisk, "usus",
		disk->beDiskRef, adminPassphrase, newAlgorithm, sockPath ); CKERR;
		
	len = sizeof( sun );
	asock = accept( callbackSock, (struct sockaddr *) &sun, (int *) &len );
	if( asock < 0 )
		RETERR( kPGPError_UnknownError );

	do
	{
		err = PGPrpcReadAndUnpack( diskcon->context, asock,
			kPGPdiskEngineMsg_ReEncryptDiskCallback, "u", &percent ); CKERR;
		if( IsntNull( statusFunc ) && ( percent >= 0 ) )
		{
			cberr = statusFunc( percent, userValue );
			
			/* Send the error back to the engine */
			err = PGPrpcPackAndWrite( diskcon->context, asock,
				kPGPdiskEngineMsg_ReEncryptDiskCallbackFinished, "u", cberr ); CKERR;
				
			if( cberr == kPGPError_UserAbort )
				break;
		}
	} while( ( percent >= 0 ) && ( percent < 100 ) );

	err = PGPrpcReadAndUnpack( diskcon->context, diskcon->sock,
		kPGPdiskEngineMsg_ReEncryptDisk, "u", &engerr );
	CKERR; CKENGERR;
	
done:
	if( asock > 0 )
		(void) close( asock );
	if( callbackSock > 0 )
		(void) close( callbackSock );
	(void) unlink( sockPath );

	return err;
}

PGPError
PGPdiskConvertDisk(
	PGPdiskContextRef context,
	PGPdiskDiskFormat format,
	const char * originalFilename,
	const char * adminPassphrase,
	const char * newFilename,
	PGPdiskConvertDiskStatusFunc statusFunc,
	void *userValue )
{
	PGPError			err				= kPGPError_NoErr;
	PGPError			engerr			= kPGPError_NoErr;
	PGPError			cberr			= kPGPError_NoErr;
	PGPSize				len				= 0;
	PGPUInt32			percent			= 0;
	char				sockPath[PATH_MAX];
	struct passwd *		pw				= NULL;
	PGPInt32			callbackSock	= 0;
	PGPInt32			asock			= 0;
	struct sockaddr_un	sun;
	
	PGPValidatePtr( originalFilename );
	PGPValidatePtr( newFilename );
	
	pw = getpwuid( getuid() );
	sprintf( sockPath, "/tmp/.pgpdisk-%s-%i-sock-XXXXXX", pw->pw_name, getuid() );
	(void) mktemp( sockPath );
	
	err = sCreateIPCSocket( sockPath, &callbackSock ); CKERR;

	err = PGPrpcPackAndWrite( context->context, context->sock,
		kPGPdiskEngineMsg_ConvertDisk, "ussss",
		format, originalFilename, adminPassphrase, newFilename, sockPath ); CKERR;
		
	len = sizeof( sun );
	asock = accept( callbackSock, (struct sockaddr *) &sun, (int *) &len );
	if( asock < 0 )
		RETERR( kPGPError_UnknownError );

	do
	{
		err = PGPrpcReadAndUnpack( context->context, asock,
			kPGPdiskEngineMsg_ConvertDiskCallback, "u", &percent ); CKERR;
		if( IsntNull( statusFunc ) && ( percent >= 0 ) )
		{
			cberr = statusFunc( percent, userValue );
			
			/* Send the error back to the engine */
			err = PGPrpcPackAndWrite( context->context, asock,
				kPGPdiskEngineMsg_ConvertDiskCallbackFinished, "u", cberr ); CKERR;
				
			if( cberr == kPGPError_UserAbort )
				break;
		}
	} while( ( percent >= 0 ) && ( percent < 100 ) );

	err = PGPrpcReadAndUnpack( context->context, context->sock,
		kPGPdiskEngineMsg_ConvertDisk, "u", &engerr );
	CKERR; CKENGERR;
	
done:
	if( asock > 0 )
		(void) close( asock );
	if( callbackSock > 0 )
		(void) close( callbackSock );
	(void) unlink( sockPath );

	return err;
}

PGPError 
PGPdiskGetDiskBooleanProp(
	PGPdiskDiskRef disk,
   	PGPdiskDiskPropName prop, 
	PGPBoolean *pBool )
{
	PGPError		err			= kPGPError_NoErr;
	PGPError		engerr		= kPGPError_NoErr;
	PGPBoolean		bProperty	= FALSE;
	
	PGPdiskValidateDiskRef( disk );
	PGPValidatePtr( pBool );
	
	err = PGPrpcPackAndWrite( disk->diskcon->context, disk->diskcon->sock,
		kPGPdiskEngineMsg_GetDiskBooleanProp, "uu", disk->beDiskRef, prop ); CKERR;
	err = PGPrpcReadAndUnpack( disk->diskcon->context, disk->diskcon->sock,
		kPGPdiskEngineMsg_GetDiskBooleanProp, "bu", &bProperty, &engerr );
	CKERR; CKENGERR;

	*pBool = bProperty;

done:
	return err;
}

PGPError 
PGPdiskGetDiskNumericProp(
	PGPdiskDiskRef disk,
   	PGPdiskDiskPropName prop, 
	PGPInt32 *pNumber )
{
	PGPError		err			= kPGPError_NoErr;
	PGPError		engerr		= kPGPError_NoErr;
	PGPInt32		iProperty	= FALSE;
	
	PGPdiskValidateDiskRef( disk );
	PGPValidatePtr( pNumber );
	
	err = PGPrpcPackAndWrite( disk->diskcon->context, disk->diskcon->sock,
		kPGPdiskEngineMsg_GetDiskNumericProp, "uu", disk->beDiskRef, prop ); CKERR;
	err = PGPrpcReadAndUnpack( disk->diskcon->context, disk->diskcon->sock,
		kPGPdiskEngineMsg_GetDiskNumericProp, "iu", &iProperty, &engerr );
	CKERR; CKENGERR;

	*pNumber = iProperty;

done:
	return err;
}

PGPError 
PGPdiskGetDiskDataProp(
	PGPdiskDiskRef disk,
   	PGPdiskDiskPropName prop, 
	void *pData,
   	PGPSize availLength,
   	PGPSize *pUsedLength )
{
	PGPError		err			= kPGPError_NoErr;
	PGPError		engerr		= kPGPError_NoErr;
	char *			pszProperty	= NULL;
	
	PGPdiskValidateDiskRef( disk );
	
	err = PGPrpcPackAndWrite( disk->diskcon->context, disk->diskcon->sock,
		kPGPdiskEngineMsg_GetDiskDataProp, "uu", disk->beDiskRef, prop ); CKERR;
	err = PGPrpcReadAndUnpack( disk->diskcon->context, disk->diskcon->sock,
		kPGPdiskEngineMsg_GetDiskDataProp, "su", &pszProperty, &engerr );
	CKERR; CKENGERR;
	
	if( IsntNull( pData ) )
		strncpy( pData, pszProperty, availLength );
	if( IsntNull( pUsedLength ) )
		*pUsedLength = pgpMin( strlen( pszProperty ), availLength );
		
	if( availLength < strlen( pszProperty ) + 1 )
		RETERR( kPGPError_BufferTooSmall );

done:
	if( IsntNull( pszProperty ) )
		(void) PGPFreeData( pszProperty );
	return err;
}
	
PGPError 
PGPdiskSetDiskBooleanProp(
	PGPdiskDiskRef disk,
   	PGPdiskDiskPropName prop, 
	PGPBoolean boolVal )
{
	PGPError		err			= kPGPError_NoErr;
	PGPError		engerr		= kPGPError_NoErr;
	
	PGPdiskValidateDiskRef( disk );
	
	err = PGPrpcPackAndWrite( disk->diskcon->context, disk->diskcon->sock,
		kPGPdiskEngineMsg_SetDiskBooleanProp, "uub", disk->beDiskRef, prop, boolVal ); CKERR;
	err = PGPrpcReadAndUnpack( disk->diskcon->context, disk->diskcon->sock,
		kPGPdiskEngineMsg_SetDiskBooleanProp, "u", &engerr );
	CKERR; CKENGERR;

done:
	return err;
}

PGPError 
PGPdiskSetDiskNumericProp(
	PGPdiskDiskRef disk,
   	PGPdiskDiskPropName prop, 
	PGPInt32 number )
{
	PGPError		err			= kPGPError_NoErr;
	PGPError		engerr		= kPGPError_NoErr;
	
	PGPdiskValidateDiskRef( disk );
	
	err = PGPrpcPackAndWrite( disk->diskcon->context, disk->diskcon->sock,
		kPGPdiskEngineMsg_SetDiskNumericProp, "uui", disk->beDiskRef, prop, number ); CKERR;
	err = PGPrpcReadAndUnpack( disk->diskcon->context, disk->diskcon->sock,
		kPGPdiskEngineMsg_SetDiskNumericProp, "u", &engerr );
	CKERR; CKENGERR;

done:
	return err;
}

PGPError 
PGPdiskSetDiskDataProp(
	PGPdiskDiskRef disk,
   	PGPdiskDiskPropName prop, 
	const void *pData,
   	PGPSize length )
{
	PGPError		err			= kPGPError_NoErr;
	PGPError		engerr		= kPGPError_NoErr;
	
	PGPdiskValidateDiskRef( disk );
	
	err = PGPrpcPackAndWrite( disk->diskcon->context, disk->diskcon->sock,
		kPGPdiskEngineMsg_SetDiskDataProp, "uus", disk->beDiskRef, prop, pData ); CKERR;
	err = PGPrpcReadAndUnpack( disk->diskcon->context, disk->diskcon->sock,
		kPGPdiskEngineMsg_SetDiskDataProp, "u", &engerr );
	CKERR; CKENGERR;
	
done:
	return err;
}

/*____________________________________________________________
	Miscellaneous functions
  ____________________________________________________________*/

PGPError
PGPdiskGetSockPath( char * sockPath )
{
	struct passwd *		pw	= NULL;
	
	pw = getpwuid( getuid() );
	sprintf( sockPath, "/tmp/.pgpdisk-%s-%i-sock", pw->pw_name, getuid() );
	
	return kPGPError_NoErr;
}

PGPBoolean
PGPdiskIsEngineRunning( void )
{
	char				sockPath[PATH_MAX];
	PGPInt32			sock		= 0;
	struct sockaddr_un	sun;
	PGPInt32			retval		= 0;
		
	(void) PGPdiskGetSockPath( sockPath );
	
	sock = socket( AF_LOCAL, SOCK_STREAM, 0 );
	if( sock < 0 )
		return FALSE;
		
	sun.sun_family = AF_LOCAL;
	memcpy( sun.sun_path, sockPath, strlen( sockPath ) + 1 );
	
	retval = connect( sock, (struct sockaddr *) &sun, sizeof( sun ) );

	(void) close( sock );
	if( retval < 0 )
		return FALSE;
	else
		return TRUE;
}

PGPError
PGPdiskGetEnginePID(
	PGPdiskContextRef context,
	pid_t *pid )
{
	PGPError			err			= kPGPError_NoErr;
	PGPError			engerr		= kPGPError_NoErr;
	PGPInt32			iPID		= 0;

	PGPdiskValidateContextRef( context );
	PGPValidatePtr( pid );
	
	err = PGPrpcPackAndWrite( context->context, context->sock,
		kPGPdiskEngineMsg_GetEnginePID, "" ); CKERR;
		
	err = PGPrpcReadAndUnpack( context->context, context->sock,
		kPGPdiskEngineMsg_GetEnginePID, "iu", &iPID, &engerr );
	CKERR; CKENGERR;
	
	*pid = (pid_t) iPID;
	
done:
	return err;
}

static PGPError
sGetPGPdiskErrorString( PGPError theError,
	PGPSize bufferSize,
	char * theString )
{
    PGPUInt32       idx;
    char            temp[ 256 ];
    char const *    errStr  = NULL;
    PGPSize         len = 0;
    PGPBoolean      bufferBigEnough = FALSE;

    PGPValidateParam( bufferSize >= 1 );
    PGPValidatePtr( theString );
    *theString  = '\0';

    for( idx = 0; idx < kPGPdiskErrors_NumErrorTableEntries; ++idx )
    {
        const ErrorEntry *  entry;

        entry   = &sPGPdiskErrors[ idx ];
        if ( entry->number == theError )
        {
            errStr  = entry->string;
            break;
        }
    }

    if ( IsNull( errStr ) )
    {
        /* Produce something for missing errors */
        sprintf( temp, "PGPdiskError #%ld", (long)theError );
        errStr  = temp;
    }

    len = strlen( errStr );
    bufferBigEnough = ( len + 1 <= bufferSize );
    if ( bufferBigEnough )
    {
        strcpy( theString, errStr );
    }
    else
    {
        pgpCopyMemory( errStr, theString, bufferSize - 1 );
        theString[ bufferSize - 1 ] = '\0';
    }

    return( bufferBigEnough ? kPGPError_NoErr : kPGPError_BufferTooSmall );
}

PGPError
PGPdiskGetErrorString( PGPError theError,
	PGPSize bufferSize,
	char * theString )
{
	PGPError		err			= kPGPError_NoErr;
	
	if( IsPGPdiskError( theError ) )
		err = sGetPGPdiskErrorString( theError, bufferSize, theString );
	else
		err = PGPGetErrorString( theError, bufferSize, theString );

	return err;
}
